
/*
  cc -g -D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE -o cd_backup_planer cd_backup_planer.c
*/

#define PROG_VERSION "0.8"
static int verbous=1;


#include <ctype.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/wait.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include <pwd.h>
#include <grp.h>


/* Macro for creation of arrays of objects (or single objects) */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


/* a constant dev_t to be used as placeholder */
static dev_t dev_dummy= 0;


/* global pacifier display time stamp */
time_t Cd_backup_planer_pacifier_timestamP= 0;
int Cd_backup_planer_pacifier_intervaL= 1;


/* usually Large File Support is gained by 
     cc ... -D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE ...
   better leave Cd_backup_planer_use_stat64 and Cd_backup_planer_define_stat64
   undefined unless you do not see any other way.
*/
#ifdef Cd_backup_planer_use_stat64

#define Cd_backup_planer_struct_staT stat64
#define Cd_backup_planer_call_staT stat64
#define Cd_backup_planer_call_lstaT lstat64

#else /* Cd_backup_planer_use_stat64 */

#define Cd_backup_planer_struct_staT stat
#define Cd_backup_planer_call_staT stat
#define Cd_backup_planer_call_lstaT lstat

#endif /* ! Cd_backup_planer_use_stat64 */


#ifdef Cd_backup_planer_define_stat64

/* if you got the functions stat64(), lstat64() but no header then try this */

struct stat64 {
        unsigned short  st_dev;
        unsigned char   __pad0[10];

        unsigned long   __st_ino;

        unsigned int    st_mode;
        unsigned int    st_nlink;

        unsigned long   st_uid;
        unsigned long   st_gid;

        unsigned short  st_rdev;
        unsigned char   __pad3[10];

        long long       st_size;
        unsigned long   st_blksize;

        unsigned long   st_blocks;      /* Number 512-byte blocks allocated. */
         unsigned long   __pad4;         /* future possible st_blocks high bits */
        unsigned long   st_atime;
        unsigned long   __pad5;

        unsigned long   st_mtime;
        unsigned long   __pad6;

        unsigned long   st_ctime;
        unsigned long   __pad7;         /* will be high 32 bits of ctime someday */

        unsigned long long      st_ino;
};

#else /* Cd_backup_planer_define_stat64 */

#ifdef Cd_backup_planer_use_stat64
/* >>> possible need to include the header with stat64.h */
#endif /* Cd_backup_planer_use_stat64 */

#endif /* ! Cd_backup_planer_define_stat64 */



/* --------------------------- nonportable --------------------------*/

/* Wrappers for code not usable on all target systems */



int Sfile_is_local_fs(name,flag)
char *name;
int flag;
{
 /* A00907_0060_dave_erikson , code and ideas contributed by Dave Erikson */
 /* Do not place more executable statements outside the #ifdef __xxx */
 int ret= 1;

#ifdef __sgi
 struct stat stbuf;
 if(Cd_backup_planer_call_lstaT(name,&stbuf)==-1) {
   ret= 0;
 } else {
   if(strncasecmp ("nfs", stbuf.st_fstype, strlen ("nfs")) == 0)
     ret= 0;
 }

#endif /* __sgi */


#ifdef __linux
#ifdef Cd_backup_planer_localdev_linuX

/* #include <sys/statfs.h> */ /* as alternative to <sys/vfs.h> */
#include <sys/vfs.h>

/* #include <linux/nfs_fs.h> */ /* this is a complete mess in SuSE 6.4 */
/* so i have to define it by hand */
#define NFS_SUPER_MAGIC       0x6969
#define SMB_SUPER_MAGIC       0x517B

 struct statfs fstbuf;

 if(statfs (name,&fstbuf)==-1) {
   ret= 0;
 } else {
   switch (fstbuf.f_type){
   case NFS_SUPER_MAGIC: case SMB_SUPER_MAGIC:
     ret= 0;
     break;
   default:
     break;
   }
 }

#endif /* Cd_backup_planer_localdev_linuX */
#endif /* __linux */


 return(ret);
}



/* ----------------- Misc --------------- */

char *Text_shellsafe(in_text,out_text,flag)
char *in_text,*out_text;
int flag;
{
 int l,i,w=0;

 /* enclose everything by hard quotes */
 l= strlen(in_text);
 out_text[w++]= '\'';
 for(i=0;i<l;i++){
   if(in_text[i]=='\''){
     /* escape hard quote within the text */
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
   } else {
     out_text[w++]= in_text[i];
   }
 }
 out_text[w++]= '\'';
 out_text[w++]= 0;
 return(out_text);
}


int Decode_date_input_format(erg,text,flag)
/* MMDDhhmm[[CC]YY][.ss]] */
struct tm *erg;
char *text;
int flag;
{
 int i,l,year;
 time_t current_time;
 struct tm *now;
 char line[1024];

 current_time= time(0);
 now= localtime(&current_time);
 for(i=0;i<sizeof(struct tm);i++)
   ((char *) erg)[i]= ((char *) now)[i];

 l= strlen(text);
 for(i=0;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=8 && i!=10 && i!=12)
   return(0);
 if(text[i]==0)
   goto decode;
 if(text[i]!='.' || l!=15)
   return(0);
 i++;
 if(text[i]<'0'||text[i]>'9')
   return(0);  
 i++;
 if(text[i]<'0'||text[i]>'9')
   return(0);

decode:; 
 /* MMDDhhmm[[CC]YY][.ss]] */
 i= 0;
 erg->tm_mon=  10*(text[0]-'0')+text[1]-'0'-1;
 erg->tm_mday= 10*(text[2]-'0')+text[3]-'0';
 erg->tm_hour= 10*(text[4]-'0')+text[5]-'0';
 erg->tm_min=  10*(text[6]-'0')+text[7]-'0';
 erg->tm_sec= 0;
 if(l==8)
   return(1);
 if(l>10){
   year= 1000*(text[8]-'0')+100*(text[9]-'0')+10*(text[10]-'0')+(text[11]-'0');
 }else{
   year= 1900+10*(text[8]-'0')+(text[9]-'0');
   if(year<1970)
     year+= 100;
 }
 erg->tm_year= year-1900;
 if(l<=12)
   return(1);
 erg->tm_sec=  10*(text[13]-'0')+text[14]-'0';
 return(1);
}


int Zeitdefinition(code,test_an,date,flag)
char *code;
int *test_an; /* out: bit0= usable for age test (else disable age_test) */ 
time_t *date;
int flag;
{
 char *cpt,fakchr;
 double wert,zeit;
 struct tm erg_tm;

 (*test_an)&= ~1;
 *date= 0;
 cpt= code;
 if(code[0]=='-' || code[0]=='+'){
   if(code[1]==0)
     return(1);
   if(code[1]>='0' && code[1]<='9') {
     zeit= time(0);
     wert= -1;
     sscanf(code,"%lf",&wert);
     fakchr= code[strlen(code)-1];
     if(isalpha(fakchr))
       fakchr= tolower(fakchr);
     if     (fakchr=='s') zeit+= 1.0*wert;
     else if(fakchr=='h') zeit+= 3600.0*wert;
     else if(fakchr=='d') zeit+= 86400.0*wert;
     else if(fakchr=='w') zeit+= 86400.0*7.0*wert;
     else if(fakchr=='m') zeit+= 86400.0*31.0*wert;
     else if(fakchr=='y') zeit+= 86400.0*(365.25*wert+1.0);
     else                 zeit+= 1.0*wert;
     goto fertig;
   }
   cpt= code+1;
 } else if(Sfile_decode_datestr(&erg_tm,code,0)) {
   /* YYMMDD[.hhmm[ss]] */
   erg_tm.tm_isdst= -1;
   zeit= mktime(&erg_tm);
   goto fertig;
 } else if(Decode_date_input_format(&erg_tm,code,0)) {
   /* MMDDhhmm[[CC]YY][.ss]] */
   erg_tm.tm_isdst= -1;
   zeit= mktime(&erg_tm);
   goto fertig;
 } else if(0) { /* ??? Mth dd hh:mm:ss Tzn YYYY */
   /* >>> */;
 }
 if(strcmp(cpt,"time")==0 || strcmp(cpt,"bigbang")==0 || strcmp(cpt,"off")==0)
   return(1);
 return(0);
fertig:;
 if(zeit<0 || zeit>=0x7FFFFFFF)
   return(0);
 *date= zeit;
 (*test_an)|= 1;
 return(1);
}


int Open_listfile(fppt,adr,mode,flag)
FILE **fppt;
char *adr;
char *mode;
int flag;
/*
 bit0= pipe through gzip/gunzip if adr indicates so (*.gz and *.GZ) 
 bit1= mode w : unlink but do not create new file nor return file pointer
return:
 <=0 = error
 1 = Sfile_fopen() successful
 2 = popen() successful
*/
{
 int ret= 0,adr_has_dotgz= 0,l;
 char cmd[4096+80],shell_adr[3*4096],dotgz[4];
 FILE *Sfile_fopen();
 struct stat stbuf;

 /* Remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 *fppt= NULL;
 l= strlen(adr);
 if(l>3)
   if(strcmp(adr+(l-3),".gz")==0 || strcmp(adr+(l-3),".GZ")==0)
     adr_has_dotgz= 1;
 if(strcmp(mode,"w")==0) {
   unlink(adr);
   if(Cd_backup_planer_call_staT(adr,&stbuf)!=-1) {
     fprintf(stderr,"--- Open_listfile( %s ,w) : obstinate file\n",adr);
     ret= 0; goto ex;
   }
   if(flag&2)
     return(1);
   if((flag&1) && adr_has_dotgz) {
     sprintf(cmd,"gzip >%s",Text_shellsafe(adr,shell_adr,0));
     *fppt= popen(cmd,"w");
     if(*fppt==NULL) {
       fprintf(stderr,"--- popen(\"%s\",\"w\") failed\n",cmd);
       fprintf(stderr,"--- errno= %ld , '%s'\n",errno,strerror(errno));
     }
     ret= 2;
   } else {
     *fppt= Sfile_fopen(adr,"w",0);
     ret= 1;
   }
 } else if(strcmp(mode,"r")==0) {
   dotgz[0]= 0;
   if((flag&1) && adr_has_dotgz)
     goto try_popen_gunzip;
   *fppt= Sfile_fopen(adr,"r",1);
   if(*fppt==NULL) {
     /* temporarily store the error message of fopen in case there is no .gz */
     sprintf(shell_adr,"--- errno= %ld , '%s'\n",errno,strerror(errno));
     /* try to find old_complete_list_adr.gz */
     strcpy(dotgz,".gz");
try_popen_gunzip:;
     sprintf(cmd,"%s%s",adr,dotgz);
     *fppt=  Sfile_fopen(cmd,"r",1);
     if(*fppt!=NULL) {
       Sfile_fclose(*fppt);
       sprintf(cmd,"gunzip <%s%s",Text_shellsafe(adr,shell_adr,0),dotgz);
       *fppt=  popen(cmd,"r");
       if(*fppt==NULL) {
         fprintf(stderr,"--- popen(\"%s\",\"r\") failed\n",cmd);
         fprintf(stderr,"--- errno= %ld , '%s'\n",errno,strerror(errno));
       }
       ret= 2;
     } else {
       fprintf(stderr,"failed to open %s in read mode\n",adr);
       fprintf(stderr,"%s",shell_adr);  /* the stored error message of fopen */
       ret= 0;
     }
   } else 
     ret= 1;
 } else {
   fprintf(stderr,"--- Open_listfile : mode %s not supported\n",mode);
 }
 if(*fppt==NULL)
   ret= 0;
ex:;
 return(ret);
}


int Is_forbidden_mountpoint(name,parent_dev,return_dev,flag)
char *name;
dev_t parent_dev;
dev_t *return_dev;
int flag;
/*
 bit0-3= mode
    0= anything is acceptable
    1= reject any mount point
    2= reject net mount points
    3= mode 1 (for compatibility reasons)
*/
{
 int mode;
 dev_t name_dev;
 struct stat stbuf;

 /* check for mountpoints */
 /* A00907_0060_dave_erikson , code and ideas contributed by Dave Erikson */
 mode= flag&15;
 if(mode==0)
   return(0);
 if(Cd_backup_planer_call_lstaT(name,&stbuf)==-1)
   return(-1);
 name_dev= stbuf.st_dev;
 if(return_dev!=NULL)
   *return_dev= name_dev;
 if(name_dev!=parent_dev) {
   if(mode==1 || mode==3) {
     return(1);
   } else if(mode==2) {
     if(!Sfile_is_local_fs(name,0))
       return(1);
   }
 }
 return(0);
}


int Split_fileaddress(line,target,source,flag)
char line[4096],target[4096],source[4096];
int flag;
{
 char *cpt;

 cpt= strchr(line,'=');
 if(cpt!=NULL) {
   strncpy(target,line,cpt-line);
   target[cpt-line]= 0;
   strcpy(source,cpt+1);
 } else {
   target[0]= 0;
   strcpy(source,line);
   if(strcmp(line,"-")==0 || strcmp(line,".")==0)
     source[0]= 0;
 }
 return(1);
}


int Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                        valid_flag,flag)
char *typetext,*lengthtext,*crctext;
int typ;
double size;
unsigned long crc;
unsigned char md5[16];
int valid_flag; /* bit1=typ , bit2=size , 
                   bit3=crc , bit4=md5 rathern than crc */
int flag;
/*
 bit9= prepend type
bit10= prepend length
bit11= preprend crc
bit12= prepend md5 rather than crc
*/
{
 int i;

 typetext[0]= lengthtext[0]= crctext[0]= 0;
 if(flag&512) {
   if(valid_flag&2)
     strcpy(typetext,typ==1?"- ":typ==2?"d ":typ==3?"l ":
                     typ==4?"p ":typ==5?"m ":"0 ");
   else
     strcpy(typetext,"0 ");
 }
 if(flag&1024) {
   if(valid_flag&4)
     sprintf(lengthtext,"%.lf ",size);
   else
     strcpy(lengthtext,"-1 ");
 }
 if(flag&2048) {
   if(valid_flag&8) {
     if(valid_flag&16) {
       for(i=0;i<16;i++) 
         sprintf(crctext+2*i,"%2.2x ",md5[i]);
     } else
       sprintf(crctext,"%X ",crc);
   } else
     strcpy(crctext,"-1 ");
 }
 return(1);
}


int Get_user_group_name(uid,gid,username,groupname,flag)
uid_t uid;
uid_t gid;
char username[256];
char groupname[256];
int flag;
{
 struct passwd *pwd;
 struct group *grp;

 strcpy(username,"?");
 strcpy(groupname,"?");
 pwd= getpwuid(uid);
 if(pwd==NULL)
   return(0);
 if(strlen(pwd->pw_name)>255)
   return(0);
 strcpy(username,pwd->pw_name);
 grp= getgrgid(gid);
 if(grp==NULL)
   return(0);
 if(strlen(grp->gr_name)>255)
   return(0);
 strcpy(groupname,grp->gr_name);
 return(1);
}


int Get_user_group_id(username,groupname,uid,gid,flag)
char username[256];
char groupname[256];
uid_t *uid;
uid_t *gid;
int flag;
{
 struct passwd *pwd;
 struct group *grp;

 pwd= getpwnam(username);
 if(pwd==NULL)
   return(0);
 *uid= pwd->pw_uid;
 grp= getgrnam(groupname);
 if(grp==NULL)
   return(0);
 *gid= grp->gr_gid;
 return(1);
}


int Record_dir_properties(fp,name,adr,lengthtext,crctext,flag)
FILE *fp;
char *name;                       /* address where to find file now */
char *adr;                        /* address text to be recorded */
char *lengthtext,*crctext;        /* to be used w/o bit0. Trailing blanks */
int flag;
/*
 bit0= report as shell commands which set the properties
       else as formated line with additionaly length and checksum
*/
{
 char shell_adr[3*4096],username[256],groupname[256],*npt;
 struct stat stbuf;
 int octomode= 0,i;
 char *Sfile_datestr();
 
 if(Cd_backup_planer_call_lstaT(name,&stbuf)==-1)
   return(0);
 /*  seems to be covered by  man 2 stat  and  man 1 chmod */
 octomode= stbuf.st_mode&0xfff;
 Get_user_group_name(stbuf.st_uid,stbuf.st_gid,username,groupname,0);
 if(flag&1) {
   /* report as some shell commands */;
   for(npt= adr;*npt=='/';npt++);
   fprintf(fp,"x=%s\n",Text_shellsafe(npt,shell_adr,0));
   fprintf(fp,"chmod %4.4o \"$x\"\n",octomode);
   fprintf(fp,"chown %s \"$x\"\n",Text_shellsafe(username,shell_adr,0));
   fprintf(fp,"chgrp %s \"$x\"\n",Text_shellsafe(groupname,shell_adr,0));
   fprintf(fp,"\n");
 } else {
   /* report as : */
   /* D permissionbits owner group time lengthtext crctext adr */
   for(i=strlen(username);i>=0;i--)
     if(isspace(username[i]))
       username[i]= '!';
   for(i=strlen(groupname);i>=0;i--)
     if(isspace(groupname[i]))
       groupname[i]= '!';
   /* capital D indicates presence of chmod,chown,chgrp,date */
   fprintf(fp,"D %4.4o %s %s %s %s%s%s\n",
              octomode,username,groupname,Sfile_datestr(stbuf.st_mtime,1|2),
              lengthtext,crctext,adr);
 }
 return(1);
}
 

int Type_and_size(filename,size,flag)
char *filename;
double *size;
int flag;
/*
 bit0= return -1 if file is missing 
 bit1= return a hardlink with siblings as type 5
 bit2= eventually return properties of link target
*/
/*
 return:  0=unknown  1=regular  2=directory  3=symbolic link  4=named pipe
          5=multiple hardlink (with bit1)
*/
{
 struct Cd_backup_planer_struct_staT stbuf;

 if(flag&4) {
   if(Cd_backup_planer_call_staT(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 } else {
   if(Cd_backup_planer_call_lstaT(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 }
 if(size!=NULL)
   *size= stbuf.st_size;
 if(S_ISREG(stbuf.st_mode)) {
   if(flag&2)
     if(stbuf.st_nlink>1)
       return(5);
   return(1);
 }
 if(S_ISDIR(stbuf.st_mode))
   return(2);
 if(S_ISLNK(stbuf.st_mode))
   return(3);
 if(S_ISFIFO(stbuf.st_mode))
   return(4);
 return(0);
}


int Is_older_file(candidate,reference,flag)
char *candidate;
char *reference;
int flag;
{
 struct stat stbuf_c,stbuf_r;
 time_t age_c,age_r;

 if(Cd_backup_planer_call_staT(candidate,&stbuf_c)==-1)
   return(-1);
 if(Cd_backup_planer_call_staT(reference,&stbuf_r)==-1)
   return(-1);
 if(stbuf_c.st_mtime>stbuf_c.st_ctime)
   age_c= stbuf_c.st_mtime;
 else
   age_c= stbuf_c.st_ctime;
 if(stbuf_r.st_mtime>stbuf_c.st_ctime)
   age_r= stbuf_c.st_mtime;
 else
   age_r= stbuf_c.st_ctime;
 if(age_r>age_c)
   return(1);
 return(0);
}


int Dir_is_empty(dir,flag)
char *dir;
int flag;
/*
 return:
  1= ok
  0= directory with content
 -1= not an accessible directory
*/
{
 FILE *fp;
 char buf[4*4096],shell_dir[3*4096];
 int ret;
 double size;
 char *Sfile_fgets();

 ret= Type_and_size(dir,&size,1);
 if(ret!=2)
   return(-1);
 sprintf(buf,"ls -a1 %s 2>/dev/null",Text_shellsafe(dir,shell_dir,0));
 wait3(NULL,WNOHANG,NULL);
 fp= popen(buf,"r");
 if(fp==NULL)
   return(-1);
 while(1) {
   if(Sfile_fgets(buf,sizeof(buf)-1,fp)==NULL)
 break;
   if(strcmp(buf,".")==0 || strcmp(buf,"..")==0)
 continue;
   ret= 0; goto ex;
 }
 ret= 1;
ex:;
 pclose(fp);
 return(ret);
}


int Remove_dir(dir,flag)
char *dir;
int flag;
/*
 bit0= report dir at stdout if it exists rather than deleting it
 bit1= with bit0==0 : 
       print rmdir command to stdout rather than execute rmdir()
 bit2= no error messages
*/
/*
 return:
  -1 = directory missing or not accessible
   0 = could not remove directory 
   1 = ok
*/
{
 int ret;
 char shell_adr[3*4096];

 ret= Dir_is_empty(dir,0);
 if(ret==1) {
   if(flag&1) {
     fprintf(stdout,"%s/\n",dir);
   } else {
     if(flag&2)
       fprintf(stdout,"rmdir %s\n",Text_shellsafe(dir,shell_adr,0));
     else {
       if(rmdir(dir)==-1) {
         if(!(flag&4))
           fprintf(stderr,"--- Cannot remove directory : %s \n    %s\n",
                          dir,strerror(errno));
         return(0);
       }
     }
   }
 } else if(ret==0) {
   if(flag&1) {
     fprintf(stdout,"%s/\n",dir);
   } else {
     if(!(flag&4))
       fprintf(stderr,"--- directory not empty : %s\n",dir);
     if(flag&2)
       fprintf(stdout,"# rmdir %s\n",Text_shellsafe(dir,shell_adr,0));
     return(0);
   }
 } else if(ret==-1) {
   if(!(flag&4))
     fprintf(stderr,"--- directory missing or not accessible : %s\n",dir);
   return(-1);
 }
 return(1);
}


int Remove_file(victim,remove_target,flag)
char *victim;
char *remove_target;
int flag;
{
 char line[2*4096],dirpath[4096],*cpt,*npt;
 int ret,errn,dtype;
 double size;
 
 if(remove_target[0]==0) {
   ret= unlink(victim);
   if(ret==-1) {
     fprintf(stderr,"Remove_file : unlink failed :\n");
     fprintf(stderr,"victim=%s\n",victim);
     fprintf(stderr,"%s\n",strerror(errno));
     {ret= 0; goto ex;}
   }
 } else {
   if(Sfile_type(remove_target,1)>=0) {
     fprintf(stderr,"Remove_file : remove_target already exists\n");
     fprintf(stderr,"victim=%s\n",victim);
     fprintf(stderr,"remove_target=%s\n",remove_target);
     {ret= 0; goto ex;}
   }

   /* >>> create necessary directories */
   cpt= remove_target;
   dirpath[0]= 0;
   if(*cpt=='/') {
     strcpy(dirpath,"/");
     cpt++;
   }
   while(1) {
     npt= strchr(cpt,'/');
     if(npt==NULL)
   break;
     if(npt-cpt>0) {
       strncat(dirpath,cpt,npt-cpt);
       dtype= Type_and_size(dirpath,&size,1);
       if(dtype==-1) {
         ret= mkdir(dirpath,0700);
         if(ret==-1) {
           fprintf(stderr,
                  "Remove_file : creation of directory failed :\n");
           fprintf(stderr,"dirpath=%s\n",dirpath);
           fprintf(stderr,"%s\n",strerror(errno));
           {ret= 0; goto ex;}
         }
         chmod(dirpath,0700);
       }
     } 
     strcat(dirpath,"/");
     cpt= npt+1;
   }

   ret= Sfile_move_file(victim,remove_target,&errn,line,0);
   if(ret<=0) {
     fprintf(stderr,"Remove_file : removal failed :\n");
     fprintf(stderr,"victim=%s\n",victim);
     fprintf(stderr,"remove_target=%s\n",remove_target);
     fprintf(stderr,"%s\n",line);
     goto ex;
   }
 }
 ret= 1;
ex:
 return(ret);
}


int Stdout_pipe_broker(data,datacount,flag)
char *data;
int datacount;
int flag;
/*
 bit0= set command (else write output)
 bit15= end output
*/
{
 static char cmd[4096];
 static FILE *fp= NULL;
 static int banned= 0;
 int ret;

 if(flag&1) {
   if(fp!=NULL && fp!=stdout)
     pclose(fp);
   fp= NULL;
   strcpy(cmd,data);
   banned= 0;
   return(1);
 }
 if(flag&(1<<15)) {
   if(fp!=NULL && fp!=stdout)
     pclose(fp);
   fp= NULL;
   strcpy(cmd,"");
   return(1);
 }
 if(banned)
   return(0);
 if(fp==NULL) {
   if(strlen(cmd)>0) {
     fp= popen(cmd,"w");
     if(fp==NULL){
       fprintf(stderr,"popen(\"%s\",\"w\") failed. errno=%ld",cmd,errno);
       banned= 1;
       return(0);
     }
   } else
     fp= stdout;
 }
 ret= fwrite(data,1,datacount,fp);
 if(ret<datacount) {
   if(strlen(cmd)>0) {
     fprintf(stderr,"write to pipe failed. errno=%ld",errno);
     fprintf(stderr,"pipe command was : %s\n",cmd);
     banned= 1;
     return(0);
   }
 }
 return(1);
}


int Determine_file_checksum(adr,fp_in,handle,crc,pre_crc,size,flag)
char *adr;
FILE *fp_in;
char **handle;
char *crc;        /* should be (void *) */
char *pre_crc;    /* should be (void *) */
double *size;
int flag;
/*
  bit0= assume datafile at fp_in 
  bit1= filtermode with bit0 : write data to stdout
  bit2= read only the first *size bytes of input
  bit3= with bit0 but not bit2: compute a pre_crc for the first *size bytes

  bit8= with bit0 : do not initialize but use handle
  bit9= with bit0 : do not finalize but return handle

 bit12= produce MD5 in crc.md5[] rather than CRC32 crc.crc32
*/
{
 FILE *fp;
 char buf[4*4096],dir[3*4096],*fret,*md5,*pre_md5;
 unsigned long *crc32,*pre_crc32;
 char *ctx= NULL, *pre_ctx= NULL;                     /* should be (void *) */
 int ret,first= 1,typ,make_pre_crc= 0,cut,i,read_bytes,bufsiz,extra_ret= 0;
 double size_limit;
 struct stat stbuf;
 char *Sfile_fgets();

 if(flag&(1<<12)) {
   md5= crc;
   pre_md5= pre_crc;
 } else {
   crc32= (unsigned long *) crc;
   pre_crc32= (unsigned long *) pre_crc;
 }
 size_limit= *size;
 make_pre_crc= !!(flag&8);
 if(flag&256) {
   if(flag&(1<<12))
     ctx= *handle;
   *handle= NULL;
 } else {
   if(flag&(1<<12)) {
     ret= Sregex_md5(&ctx,buf,0,md5,1);
     if(ret<=0)
       return(ret);
   } else
     *(crc32)= 0;
 }
 if(flag&1) {
   typ= 1;
 } else {
   typ= Type_and_size(adr,size,0);
   if(typ<=0)
     return(0);
 }
 if(typ==3) {                                              /* symbolic link */
#ifdef Stypes_on_ms_windowS
   return(0);
#else /* Stypes_on_ms_windowS */
   ret= readlink(adr,buf,sizeof(buf)-1);
   if(ret<0)
     return(0);
   buf[ret]= 0; 
   if(flag&(1<<12))
     Sregex_md5(&ctx,buf,ret,md5,0);
   else 
     Sregex_crc32_yenc(buf,ret,crc32,1);
   return(1);
#endif /* ! Stypes_on_ms_windowS */
 } if(typ==2) {                                                /* directory */
   sprintf(buf,"ls -a1 %s",Text_shellsafe(adr,dir,0));
   wait3(NULL,WNOHANG,NULL);
   fp= popen(buf,"r");
   if(fp==NULL)
     return(0);
   buf[0]= '\n'; /* line break at begin of each directory item */
   while(1) {
     fret= Sfile_fgets(buf+1,sizeof(buf)-2,fp);     /* saving the line break */
   if(fret==NULL) break;
     extra_ret= 1;
     if(flag&(1<<12))
       Sregex_md5(&ctx,buf,ret,md5,0);
     else 
       Sregex_crc32_yenc(buf,strlen(buf),crc32,first);
     first= 0;
   }
   pclose(fp);
 } else if(typ==1 || typ==5 || typ==6) {
                                  /* regular file , hardlink , block device */
   *size= 0;
   if(flag&1)
     fp= fp_in;
   else
     fp= fopen(adr,"rb");
   if(fp==NULL)
     return(0);
   bufsiz= sizeof(buf);
   while(1) {
     if((flag&4) || make_pre_crc) {
       if(*size+bufsiz>size_limit)
         bufsiz= size_limit - *size;
       if(bufsiz<=0)
   break;
     }
     ret= fread(buf,1,bufsiz,fp);
     if(ret<=0)
   break;


     if((flag&3)==3) {
       /* write to the pipe broker (idea : A30407_0208_marcus_rocha) */
       Stdout_pipe_broker(buf,ret,0);
     }

     (*size)+= ret;
     if(make_pre_crc && *size>=size_limit) {
       /* preview checksum for a fixed size start piece */
       read_bytes= ret;
       cut= ret-(*size-size_limit);
       if(flag&(1<<12)) {
         /* compute */
         Sregex_md5(&ctx,buf,cut,md5,0);
         /* obtain intermediate value */
         ret= Sregex_md5(&pre_ctx,ctx,0,pre_md5,1|4);
         if(ret<=0)
           return(ret);
         Sregex_md5(&pre_ctx,buf,0,pre_md5,2|(1<<15));
       } else {
         Sregex_crc32_yenc(buf,ret,crc32,first);
         /* obtain intermediate value */
         *(pre_crc32)= *(crc32);
       }
       /* shift remaining buffer butes */
       ret= read_bytes;
       for(i=cut;i<ret;i++)
         buf[i-cut]= buf[i];
       ret-= cut;
       make_pre_crc= 0;
       bufsiz= sizeof(buf);
   continue;
     }
     if(flag&(1<<12))
       Sregex_md5(&ctx,buf,ret,md5,0);
     else
       Sregex_crc32_yenc(buf,ret,crc32,first);
     first= 0;
   }
   if(!(flag&1))
     fclose(fp);
 } else if (typ==4) {                                         /* named pipe */
   /* >>> what to do rather than checksum 0 ? */;
 } else {                                                     /* unknown */
   /* >>> what to do rather than checksum 0 ? */;
 }
 if(flag&512) {
   /* do not destroy ctx but hand it over to caller */
   if(flag&(1<<12))
     *handle= ctx;
 } else {
   if(flag&(1<<12)) 
     Sregex_md5(&ctx,buf,0,md5,2|(1<<15));
 }
 return(1+extra_ret);
}


int File_is_changed(f_adr,f_flag,f_type,f_size,f_crc,f_md5,
                          o_flag,o_type,o_size,o_crc,o_md5,flag)
/* recent file properties */
char *f_adr;
int f_flag;     /* bit3= f_crc valid , bit4= use f_md5 rather than f_crc */
int f_type;
double f_size;
unsigned long f_crc;
unsigned char f_md5[16];
/* old file properties */
int o_flag;     /* bit3= o_crc valid , bit4= use o_md5 rather than o_crc */
int o_type;
double o_size;
unsigned long o_crc;
unsigned char o_md5[16];
int flag;
/*
 bit0= determine f_type and f_size from from f_adr, set f_flags=0
 bit1= count missing information as file change

 bit9= check type 
 bit10= check size
 bit11= check crc
return:
 <0 = error
  0 = no change detected
  1 = change detected
  2 = not found in filesystem (with bit0)
*/
{
 int ret,i;
 unsigned char md5[16];

 if(flag&1) {
   f_type= Type_and_size(f_adr,&f_size,1);
   if(f_type<0)
     return(2);
 } 
 if((flag&512)) {
   if(o_type<=0) {
     if(flag&2)
       return(1);
   } else if(o_type!=f_type)
     return(1);
 }
 if(flag&1024) {
   if(o_size<0 || f_size<0) {
     if(flag&2)
       return(1);
   } else if(f_size>=0 && o_size!=f_size)
     return(1);
 }
 if(flag&2048 && o_flag&8) {
   if(o_flag&16) {
     if((f_flag&(8|16))==(8|16)) {
       memcpy(md5,f_md5,16);
       ret= 1;
     } else 
       ret= Determine_file_checksum(f_adr,NULL,NULL,md5,md5,&f_size,1<<12);
     if(ret<=0 && (flag&2))
       return(1);
     if(ret>0) {
       for(i=0;i<16;i++)
         if(md5[i]!=o_md5[i])
       break;
       if(i<16)
         return(1);
     }
   } else {
     if((f_flag&(8|16))==8) {
       ret= 1;
     } else
       ret= Determine_file_checksum(f_adr,NULL,NULL,&f_crc,&f_crc,&f_size,0);
     if(ret<=0 && (flag&2))
       return(1);
     if(ret>0 && o_crc!=f_crc)
       return(1);
   }
 }
 return(0);
}


int Print_pacifier(count,text,flag)
int count;
char *text;
int flag;
/*
 bit0= print regardless of time (if not totally disabled)
 bit1= print count
 bit2= print text
 bit3= end of pacifying, print newline
 bit4= make line empty
 bit5= permission to truncate text if necessary
*/
{
 char endchar;

 if(Cd_backup_planer_pacifier_intervaL<0)
   return(2);
 if((!(flag&1)) && time(0)<Cd_backup_planer_pacifier_timestamP+
                           Cd_backup_planer_pacifier_intervaL)
   return(2);
 if(flag&16)
   fprintf(stderr,"\r%78.78s\r","");
 if(flag&8)
   endchar= '\n';
 else
   endchar= '\r';
 if((flag&6)==2)
   fprintf(stderr,"  %ld%c",count,endchar);
 else if((flag&6)==4) {
   if((flag&32) && strlen(text)>77)
     strcpy(text+74,"...");
   fprintf(stderr,"%s%c",text,endchar);
 } else if((flag&6)==6) {
   if((flag&32) && strlen(text)>65)
     strcpy(text+62,"...");
   fprintf(stderr,"  %-10ld %s%c",count,text,endchar);
 }
 Cd_backup_planer_pacifier_timestamP= time(0);
 return(1);
}


double Scanf_io_size(text,flag)
char *text;
int flag;
/*
 bit0= default value -1 rather than 0 
*/
{
 int c;
 double ret= 0.0;

 if(flag&1)
   ret= -1.0;
 if(text[0]==0)
   return(ret);
 sscanf(text,"%lf",&ret);
 if(ret>=1.0) {
   c= text[strlen(text)-1];
   if(c=='k' || c=='K') ret*= 1024.0;
   if(c=='m' || c=='M') ret*= 1024.0*1024.0;
   if(c=='g' || c=='G') ret*= 1024.0*1024.0*1024.0;
 }
 return(ret);
}


/* ----------------- Misc - end -------- */



/* ----------------- Sfile extract -------- */

#define Smem_malloC malloc
#define Smem_freE free

#define SfileadrL 4096


char *Sfile_fgets(line,maxl,fp)
char *line;
int maxl;
FILE *fp;
{
int l;
char *ret;
 ret= fgets(line,maxl,fp);
 if(ret==NULL) return(NULL);
 l= strlen(line);
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 if(l>0) if(line[l-1]=='\n') line[--l]= 0;
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 return(ret);
}


int Sfile_decode_datestr(reply,text,flag)
/* YYMMDD[.hhmm[ss]] */
struct tm *reply;
char *text;
int flag;
{
 int i,l,year;
 time_t current_time;
 struct tm *now;
 char line[1024];

 current_time= time(0);
 now= localtime(&current_time);
 for(i=0;i<sizeof(struct tm);i++)
   ((char *) reply)[i]= ((char *) now)[i];

 if(text[0]<'0'|| (text[0]>'9' && text[0]<'A') || text[0]>'Z')
   return(0);
 l= strlen(text);
 for(i=1;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=6)
   return(0);
 if(text[i]==0)
   goto decode;
 if(text[i]!='.' || (l!=11 && l!=13))
   return(0);
 for(i++;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=l)
   return(0);

decode:;
 reply->tm_hour= 0;
 reply->tm_min= 0;
 reply->tm_sec= 0;
 i= 0;
 if(text[0]>='A') 
   reply->tm_year= 100+(text[i]-'A')*10+text[1]-'0';
 else
   reply->tm_year= 10*(text[0]-'0')+text[1]-'0';
 reply->tm_mon=  10*(text[2]-'0')+text[3]-'0'-1;
 reply->tm_mday= 10*(text[4]-'0')+text[5]-'0';
 if(l==6)
   return(1);
 reply->tm_hour= 10*(text[7]-'0')+text[8]-'0';
 reply->tm_min=  10*(text[9]-'0')+text[10]-'0';
 if(l==11)
   return(1);
 reply->tm_sec=  10*(text[11]-'0')+text[12]-'0';
 return(1);
}



char *Sfile_datestr(tim,flag)
time_t tim;
short int flag;
/*
 bit0=with hours+minutes  
 bit1=with seconds

 bit8= local time rather than UTC
*/
{
 static char zeitcode[80]={"000000"};
 char puff[80];
 struct tm *azt;

 if(flag&256) 
   azt = localtime(&tim);
 else
   azt = gmtime(&tim);
 
 if(azt->tm_year>99)
   sprintf(zeitcode,"%c%01.1d%02.2d%02.2d",
           'A'+(azt->tm_year-100)/10,azt->tm_year%10,
           azt->tm_mon+1,azt->tm_mday);
 else
   sprintf(zeitcode,"%02.2d%02.2d%02.2d",
           azt->tm_year,azt->tm_mon+1,azt->tm_mday);
 if(flag&1){
   sprintf(puff,".%02.2d%02.2d",azt->tm_hour,azt->tm_min);
   strcat(zeitcode,puff);
 }
 if(flag&2){
   sprintf(puff,"%02.2d",azt->tm_sec);
   strcat(zeitcode,puff);
 }

 return(zeitcode);
}


FILE *Sfile_fopen(filename,mode,flag)
char *filename;
char *mode;
int flag;
/*
 bit0= do not print error message on failure
*/
{
 FILE *fp= NULL;
 if(strcmp(filename,"-")==0) {
   if(mode[0]=='a') fp= stdout;
   else             fp= stdin;
 } else if(strncmp(filename,"tcp:",4)==0){
   fprintf(stderr,"sorry - TCP/IP service isn't implemented yet.\n");
 } else if(strncmp(filename,"file:",5)==0){
   fp= fopen(filename+5,mode);
 } else {
   fp= fopen(filename,mode);
 }
 if(fp==NULL){
   if(!(flag&1))
     fprintf(stderr,"failed to open %s in %s mode\n",filename,mode);
   return(NULL);
 }
 return(fp);
}


int Sfile_fclose(fp)
FILE *fp;
{
 if(fp!=stdout && fp!=stdin && fp!=NULL)
   fclose(fp);
 return(1);
}


int Sfile_bin_to_file(filename,data,size,flag)
char *filename;
char *data;
int size;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= append to file
 bit2= add newline character
*/
{
 FILE *fp;
 struct stat stbuf;
 int ret;

 if(!(flag&3))
   if(Cd_backup_planer_call_staT(filename,&stbuf)!=-1)
     return(0);
 if(flag&2)
   fp= Sfile_fopen(filename,"ab",0);
 else
   fp= Sfile_fopen(filename,"wb",0);
 if(fp==NULL)
   return(-1);
 if(size>0)
   ret= fwrite(data,size,1,fp);
 else
   ret= 1;
 if(ret==1 && (flag&4))
   ret= (fputc('\n',fp)!=EOF);
 Sfile_fclose(fp);
 return(ret);
}


int Sfile_string_to_file(filename,text,flag)
char *filename;
char *text;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= append to file
 bit2= add newline character
*/
{
 int ret;

 ret= Sfile_bin_to_file(filename,text,strlen(text),flag&7);
 return(ret);
}


int Sfile_file_to_text(filename,text,textsize,flag)
char *filename;
char *text;
int textsize;
int flag;
/*
 bit0= return(2) if file is empty (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 FILE *fp;
 int ret;

 if(textsize<=0)
   return(-1);
 fp= Sfile_fopen(filename,"rb",(flag&2)>>1);
 if(fp==NULL)
   return(0);
 if(Sfile_fgets(text,textsize,fp)==NULL)
   ret= 2*(flag&1);
 else
   ret= 1;
ex:;
 Sfile_fclose(fp);
 return(ret);
}


int Sfile_file_to_double(filename,num,flag)
/*
 bit0= return(2) if file is empty (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
char *filename;
double *num;
int flag;
{
 char text[SfileadrL];
 double dv;
 int ret;

 if(Sfile_file_to_text(filename,text,sizeof(text),flag&2)<=0)
   return(0);
 ret= sscanf(text,"%lf",&dv);
 if(ret<=0)
   return(0);
 *num= dv;
 return(1);
}


int Sfile_file_to_int(filename,num,flag)
char *filename;
int *num;
int flag;
/*
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 double dv;
 int ret;

 if(Sfile_file_to_double(filename,&dv,flag&2)<=0)
   return(0);
 if(dv<-0x7FFFFFFF || dv>=0x7FFFFFFF)
   return(0);
 *num= dv;
 return(1);
}


int Sfile_identical(file1,file2,flag)
char *file1,*file2;
int flag;
/*
 bit0= check wether it is the same  i-node 
*/
{
 FILE *fp1,*fp2;
 struct stat stbuf;
 int ret,c1,c2;
 ino_t ino1,ino2;

 ret= 0;

 if(flag&1){
   if(Cd_backup_planer_call_staT(file1,&stbuf)==-1) return(0);
   ino1= stbuf.st_ino;
   if(Cd_backup_planer_call_staT(file2,&stbuf)==-1) return(0);
   ino2= stbuf.st_ino;
   return(ino1==ino2);
 }

 fp1= Sfile_fopen(file1,"rb",0);
 fp2= Sfile_fopen(file2,"rb",0);
 if(fp1==NULL||fp2==NULL) goto fertig;
 while(1){
   c1= fgetc(fp1);
   c2= fgetc(fp2);
   if(c1==EOF && c2==EOF){
     ret= 1;
 break;
   }
   if(c1!=c2)
 break;
 }
fertig:;
 if(fp1!=NULL) fclose(fp1);
 if(fp2!=NULL) fclose(fp2);
 return(ret);
}


int Sfile_copy_file(source,target,errn,errmsg,flag)
char *source;
char *target;
int *errn;
char errmsg[SfileadrL];
int flag;
{
 FILE *fp,*fw;
 char *line[1024];
 int ret;

 if(Sfile_identical(source,target,1)>0)
   return(2);
 fp= fopen(source,"r");
 if(fp==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening input file : %s\n%s",
                  errno,source,strerror(errno));
   return(0);
 }
 fw= fopen(target,"w");
 if(fw==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening output file : %s\n%s",
                  errno,target,strerror(errno));
   ret= 0;
   goto ex;
 }
 while((ret= fread(line,1,sizeof(line),fp))>0){
   fwrite(line,1,ret,fw);
   if(ret<sizeof(line))
 break;
 }
 ret= 1;
ex:;
 if(fp!=NULL) fclose(fp);
 if(fw!=NULL) fclose(fw);
 return(ret);
}


int Sfile_move_file(source,target,errn,errmsg,flag)
char *source;
char *target;
int *errn;
char *errmsg;
int flag;
/*
 bit0= do not unlink target if source cannot be unlinked
 bit1= touch ctime
 bit7= use overwrite+delete rather than rename
 bit8= like bit7 but do not unlink source (i.e. copy >>> with properties)
*/
{
 int ret,ovw_del=0;
 struct stat stbuf;

 if(flag&(128|256))
   ovw_del= 1;
 if(!ovw_del)
   if(rename(source,target)!=-1) {
     if(flag&2) 
       if(Cd_backup_planer_call_lstaT(target,&stbuf)!=-1) 
         chmod(target,stbuf.st_mode);
     return(1);
   }
 if(errno==EXDEV || ovw_del) {
   /* try copy and unlink */;
   ret= Sfile_copy_file(source,target,errn,errmsg,0);
   if(ret<=0)
     return(ret);
   if(flag&2) 
     if(Cd_backup_planer_call_lstaT(target,&stbuf)!=-1) 
       chmod(target,stbuf.st_mode);
   if(!(flag&128)) {
     /* transfer as many properties as possible */;
     Sfile_copy_properties(source,target,1|4);
   }
   if(flag&256)
     return(3);
   if(unlink(source)==-1) {
     *errn= errno;
     sprintf(errmsg,
             "Error %ld on removing file %s\nafter copying to %s\n\"%s\"",
             errno,source,target,strerror(errno));
     if(Cd_backup_planer_call_lstaT(source,&stbuf)!=-1 && !(flag&1)) {
       strcat(errmsg,"\n(Operation failed. Target file has been removed now)");
       unlink(target);
     }
     return(0);
   }
   return(2);
 }
 *errn= errno;
 sprintf(errmsg,"Error %ld on renaming\n   %s\nto %s\n\"%s\"",
                errno,source,target,strerror(errno));
 return(0);
}


#ifndef Stypes_on_ms_windowS

int Sfile_set_filetimes(fileadr,mtime,atime,flag)
char *fileadr;
time_t mtime;
time_t atime;
int flag;
{
 struct stat stbuf;
 struct utimbuf buf;
 int ret;
 
 if(Cd_backup_planer_call_staT(fileadr,&stbuf)==-1)
   return(flag&1);
 if(!S_ISREG(stbuf.st_mode))
   return(0);
 buf.actime= atime;
 buf.modtime= mtime;
 ret= utime(fileadr,&buf);
 if(ret<=0)
   return(0);
 return(1);
}

#endif /* ! Stypes_on_ms_windowS */


int Sfile_set_properties(target,ownership,uid,gid,atime,mtime,ctime,flag)
char *target;
int ownership; /* low 12 bit of stat.st_mode */
uid_t uid;
gid_t gid;
time_t atime;
time_t mtime;
time_t ctime;
int flag;
/*
 bit0= set owner access bits
 bit1= set group access bits
 bit2= set others access bits
 bit3= set special access bits (S_ISUID,S_ISGID,S_ISVTX)
 bit4= set uid
 bit5= set gid
 bit6= set atime
 bit7= set mtime
 reserved until i know how to set it: bit8= set ctime
 bit13= with bit0: make target objects writable by owner
*/
{
 int mode_mask= 0,ret;
 mode_t new_mode;
 struct stat stbuf;

 if(Cd_backup_planer_call_staT(target,&stbuf)==-1)
   return(0);
 if(flag&1) {
   mode_mask|= 0700;
   if(flag&(1<<13))
     ownership|= S_IWUSR;     
 }
 if(flag&2)
   mode_mask|= 070;
 if(flag&4)
   mode_mask|= 07;
 if(flag&8)
   mode_mask|= 07000;
 if(mode_mask!=0) {
   new_mode= stbuf.st_mode&~mode_mask;
   new_mode|= mode_mask&ownership;
   new_mode&= 07777;
   if(chmod(target,new_mode)==-1)
     return(0);
 }
 if(!(flag&16))
   uid= stbuf.st_uid;
 if(!(flag&32))
   gid= stbuf.st_gid;
 if(flag&(16|32)) {
   if(chown(target,uid,gid)==-1)
     return(0);
 }

#ifndef Stypes_on_ms_windowS
 if(!(flag&64))
   atime= stbuf.st_atime;
 if(!(flag&128))
   mtime= stbuf.st_mtime;
 if((flag&(64|128)) && S_ISREG(stbuf.st_mode)) {
   ret= Sfile_set_filetimes(target,mtime,atime,0);
   if(ret<=0)
     return(ret);
 }
#endif /* ! Stypes_on_ms_windowS */

 return(1);
}


int Sfile_copy_properties(source,target,flag)
char *source;
char *target;
int flag;
/*
 bit0= copy access control
 bit1= copy ownership
 bit2= copy times
 bit13= with bit0: make target objects writable by owner
*/
{
 struct stat stbuf;
 int ret,sflag= 0;

 if(Cd_backup_planer_call_staT(source,&stbuf)==-1)
   return(0);
 if(flag&1)
   sflag|= (1|2|4|8);
 if(flag&2)
   sflag|= (16|32);
 if(flag&4)
   sflag|= (64|128|256);
 sflag|= flag&(1<<13);
 ret= Sfile_set_properties(target,stbuf.st_mode&0xfff,
                                  stbuf.st_uid,stbuf.st_gid,
                                  stbuf.st_atime,stbuf.st_mtime,stbuf.st_ctime,
                                  sflag);
 return(ret);
}


int Sfile_read_argv(progname,filename,argc,argv,flag)
char *progname;
char *filename;
int *argc;
char ***argv;
int flag;
/*
 bit0= read progname as first argument from line
 bit1= just release argument list argv and return
*/
{
 int ret,i,pass,maxl=0,l,argzaehl=0,bufl;
 char *start;
 char buf[SfileadrL];
 FILE *fp= NULL;

 Sfile_destroy_argv(argc,argv,0);
 if(flag&2)
   return(1);

 for(pass=0;pass<2;pass++) {
   if(!(flag&1)){
     argzaehl= 1;
     if(pass==0)
       maxl= strlen(progname);
     else
       strcpy((*argv)[0],progname);
   } else {
     argzaehl= 0;
     if(pass==0) maxl= 0;
   }
   fp= Sfile_fopen(filename,"rb",0);
   if(fp==NULL) 
     {ret= 0; goto ex;}
   while(Sfile_fgets(buf,sizeof(buf)-1,fp)!=NULL) {
     l= strlen(buf);
     if(pass==0){
       if(l>maxl) 
         maxl= l;
     }else{
       strcpy((*argv)[argzaehl],buf);
     }
     argzaehl++;
   }
   if(pass==0){
     *argc= argzaehl;
     if(argzaehl>0) {
       *argv= (char **) Smem_malloC(argzaehl*sizeof(char *));
       if(*argv==NULL)
         {ret= -1; goto ex;}
     }
     for(i=0;i<*argc;i++) {
       (*argv)[i]= (char *) Smem_malloC((maxl+1));
       if((*argv)[i]==NULL)
         {ret= -1; goto ex;}
     }
   }     
   Sfile_fclose(fp,0); fp= NULL;
 }

 ret= 1;
ex:;
 if(fp!=NULL)
   Sfile_fclose(fp,0);
 return(ret);
}


int Sfile_destroy_argv(argc,argv,flag)
int *argc;
char ***argv;
int flag;
{
 int i;

 if(*argc>0 && *argv!=NULL){
   for(i=0;i<*argc;i++){
     if((*argv)[i]!=NULL) 
       Smem_freE((*argv)[i]);
   }
   Smem_freE((char *) *argv);
 }
 *argc= 0;
 *argv= NULL;
 return(1);
}



int Sfile_type(filename,flag)
char *filename;
int flag;
/*
 bit0= return -1 if file is missing
 bit1= return a hardlink with siblings as type 5
*/
/*
 return:  0=unknown  1=regular  2=directory  3=symbolic link  4=named pipe
          5=multiple hardlink (with bit1)
*/
{
 struct stat stbuf;
 if(Cd_backup_planer_call_lstaT(filename,&stbuf)==-1) {
   if(flag&1) return(-1);
   else       return(0);
 }
 if(S_ISREG(stbuf.st_mode)) {
   if(flag&2)
     if(stbuf.st_nlink>1)
       return(5);
   return(1);
 }
 if(S_ISDIR(stbuf.st_mode))
   return(2);
 if(S_ISLNK(stbuf.st_mode))
   return(3);
 if(S_ISFIFO(stbuf.st_mode))
   return(4);
 return(0);
}


int Sfile_readlink(filename,buff,buff_size,flag)
char *filename;
char *buff;
int buff_size;
int flag;
/*
 bit0= return -1 if file is missing
*/
{
 int l;

#ifdef Stypes_on_ms_windowS

 return(0);

#else /* Stypes_on_ms_windowS */

 if(Sfile_type(filename,0)!=3)
   return(0);
 l= readlink(filename,buff,buff_size-1);
 if(l<=0)
   return(0);
 buff[l]= 0;
 return(1);

#endif /* ! Stypes_on_ms_windowS */
}


int Sfile_scale(value,result,siz,thresh,flag)
double value;
char *result;
int siz;
double thresh;
int flag;
/*
 bit0= eventually ommit 'b'
 bit1= make text as short as possible
 bit2= no fraction (if it would fit at all)
*/
{
 char scale_c,scales[6],form[80];
 int i,dec_siz= 0,avail_siz;

 strcpy(scales,"bkmgt");
 scale_c= scales[0];
 for(i=1;scales[i]!=0;i++) {
   if(value<thresh) 
 break;
   value/= 1024;
   scale_c= scales[i];
 } 
 if(scale_c!='b' && !(flag&4)) { /* is there room for fractional part ? */
   avail_siz= siz-1;
   sprintf(form,"%%.lf");
   sprintf(result,"%.lf",value);
   if(strlen(result)<avail_siz-2) 
     dec_siz= 1;                                  /* we are very modest */
 }
 if(scale_c=='b' && (flag&1)) {
   if(flag&2)
     sprintf(form,"%%.lf");
   else
     sprintf(form,"%%%ld.lf",siz);
   sprintf(result,form,value);
 } else {
   if(flag&2)
     sprintf(form,"%%.lf%%c");
   else if(dec_siz>0) 
     sprintf(form,"%%%ld.%ldlf%%c",avail_siz,dec_siz);
   else
     sprintf(form,"%%%ld.lf%%c",siz-1);
   sprintf(result,form,value,scale_c);
 }
 return(1);
}


/* ----------------- Sfile extract end -------- */


/* ----------------- Sregex extract -------- */

int Sregex_string_cut(handle,text,len,flag)
char **handle;
char *text;
int len;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int l=0;
 char *old_handle;

 if((flag&1)&&*handle!=NULL)
   l+= strlen(*handle);
 old_handle= *handle;
 if(text!=NULL) {
   l+= len;
   *handle= TSOB_FELD(char,l+1);
   if(*handle==NULL) {
     *handle= old_handle;
     return(0);
   }
   if((flag&1) && old_handle!=NULL)
     strcpy(*handle,old_handle);
   else
     (*handle)[0]= 0;
   if(len>0)
     strncat(*handle,text,len);
 } else {
   *handle= NULL;
 }
 if(old_handle!=NULL)
   Smem_freE(old_handle);
 return(1);
}


int Sregex_string(handle,text,flag)
char **handle;
char *text;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int ret,l=0;

 if(text!=NULL)
   l= strlen(text);
 
/* #define Sregex_looking_for_contenT 1 */
#ifdef Sregex_looking_for_contenT
 /* a debugging point if a certain text content has to be caught */
 if(text!=NULL)
   if(strcmp(text,"clear")==0)
     ret= 0;
#endif

 ret= Sregex_string_cut(handle,text,l,flag&1);
 return(ret);
}


/*
  from Juergen Helbing's  yencode-c.txt 
  found at  www.yenc.org  , Feb 4,  2002
  license : "It is public domain - as well as the yEnc specification"
  some alterations made to adapt it to Sregex_crc32() 

  seems to be compatible with SFV/CSV files
*/                       

int Sregex_crc32_yenc(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
*/
{
 /* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
 static int crc_tab[256] = {
   0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 
   0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
   0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
   0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
   0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
   0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
   0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
   0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
   0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
   0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
   0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
   0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
   0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
   0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
   0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
   0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
   0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
   0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
   0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
   0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
   0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
   0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
   0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
   0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
   0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
   0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
   0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
   0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
   0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
   0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
   0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
   0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
   0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
   0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
   0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
   0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
   0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
   0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
   0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
   0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
   0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
   0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
   0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
   0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
   0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
   0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
   0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
   0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
   0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
   0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
   0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
   0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
   0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
   0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
   0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
   0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
   0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
   0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
   0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
   0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
   0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
   0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
   0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
   0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

 int k;
 unsigned long crc_val,ch1,ch2,cc;

 if(flag&1)
   crc_val = ~0;
 else
   crc_val = ~*crc;

 for(k=0;k<len;k++) {
   cc= data[k] & 0xffL;
   ch1= (crc_val ^ cc) & 0xffL;
   ch1= crc_tab[ch1];
   ch2= (crc_val>>8L) & 0xffffffL;
   crc_val= ch1 ^ ch2;
 }
 *crc= ~crc_val;
 return(1);
}


/* This code is derived from RFC 1321 and implements computation of the
    "RSA Data Security, Inc. MD5 Message-Digest Algorithm" */

#define Sregex_S11 7
#define Sregex_S12 12
#define Sregex_S13 17
#define Sregex_S14 22
#define Sregex_S21 5
#define Sregex_S22 9
#define Sregex_S23 14
#define Sregex_S24 20
#define Sregex_S31 4
#define Sregex_S32 11
#define Sregex_S33 16
#define Sregex_S34 23
#define Sregex_S41 6
#define Sregex_S42 10
#define Sregex_S43 15
#define Sregex_S44 21


/* F, G, H and I are basic MD5 functions.
 */
#define Sregex_F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define Sregex_G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define Sregex_H(x, y, z) ((x) ^ (y) ^ (z))
#define Sregex_I(x, y, z) ((y) ^ ((x) | (~z)))

/* ROTATE_LEFT rotates x left n bits.
 */
#define Sregex_ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */
#define Sregex_FF(a, b, c, d, x, s, ac) { \
 (a) += Sregex_F ((b), (c), (d)) + (x) + (unsigned long)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define Sregex_GG(a, b, c, d, x, s, ac) { \
 (a) += Sregex_G ((b), (c), (d)) + (x) + (unsigned long)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define Sregex_HH(a, b, c, d, x, s, ac) { \
 (a) += Sregex_H ((b), (c), (d)) + (x) + (unsigned long)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define Sregex_II(a, b, c, d, x, s, ac) { \
 (a) += Sregex_I ((b), (c), (d)) + (x) + (unsigned long)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }


/* MD5 context. */
typedef struct {
  unsigned long state[4];                                   /* state (ABCD) */
  unsigned long count[2];        /* number of bits, modulo 2^64 (lsb first) */
  unsigned char buffer[64];                         /* input buffer */
} Sregex_md5_ctX;


/* MD5 basic transformation. Transforms state based on block.
 */
int Sregex_md5__transform (state, block)
unsigned long state[4];
unsigned char block[64];
{
 unsigned long a = state[0], b = state[1], c = state[2], d = state[3], x[16];
 unsigned int i, j;

 for (i = 0, j = 0; j < 64; i++, j += 4)
   x[i] = ((unsigned long)block[j]) | (((unsigned long)block[j+1]) << 8) |
    (((unsigned long)block[j+2]) << 16) | (((unsigned long)block[j+3]) << 24);

  /* Round 1 */
  Sregex_FF (a, b, c, d, x[ 0], Sregex_S11, 0xd76aa478); /* 1 */
  Sregex_FF (d, a, b, c, x[ 1], Sregex_S12, 0xe8c7b756); /* 2 */
  Sregex_FF (c, d, a, b, x[ 2], Sregex_S13, 0x242070db); /* 3 */
  Sregex_FF (b, c, d, a, x[ 3], Sregex_S14, 0xc1bdceee); /* 4 */
  Sregex_FF (a, b, c, d, x[ 4], Sregex_S11, 0xf57c0faf); /* 5 */
  Sregex_FF (d, a, b, c, x[ 5], Sregex_S12, 0x4787c62a); /* 6 */
  Sregex_FF (c, d, a, b, x[ 6], Sregex_S13, 0xa8304613); /* 7 */
  Sregex_FF (b, c, d, a, x[ 7], Sregex_S14, 0xfd469501); /* 8 */
  Sregex_FF (a, b, c, d, x[ 8], Sregex_S11, 0x698098d8); /* 9 */
  Sregex_FF (d, a, b, c, x[ 9], Sregex_S12, 0x8b44f7af); /* 10 */
  Sregex_FF (c, d, a, b, x[10], Sregex_S13, 0xffff5bb1); /* 11 */
  Sregex_FF (b, c, d, a, x[11], Sregex_S14, 0x895cd7be); /* 12 */
  Sregex_FF (a, b, c, d, x[12], Sregex_S11, 0x6b901122); /* 13 */
  Sregex_FF (d, a, b, c, x[13], Sregex_S12, 0xfd987193); /* 14 */
  Sregex_FF (c, d, a, b, x[14], Sregex_S13, 0xa679438e); /* 15 */
  Sregex_FF (b, c, d, a, x[15], Sregex_S14, 0x49b40821); /* 16 */

  /* Round 2 */
  Sregex_GG (a, b, c, d, x[ 1], Sregex_S21, 0xf61e2562); /* 17 */
  Sregex_GG (d, a, b, c, x[ 6], Sregex_S22, 0xc040b340); /* 18 */
  Sregex_GG (c, d, a, b, x[11], Sregex_S23, 0x265e5a51); /* 19 */
  Sregex_GG (b, c, d, a, x[ 0], Sregex_S24, 0xe9b6c7aa); /* 20 */
  Sregex_GG (a, b, c, d, x[ 5], Sregex_S21, 0xd62f105d); /* 21 */
  Sregex_GG (d, a, b, c, x[10], Sregex_S22,  0x2441453); /* 22 */
  Sregex_GG (c, d, a, b, x[15], Sregex_S23, 0xd8a1e681); /* 23 */
  Sregex_GG (b, c, d, a, x[ 4], Sregex_S24, 0xe7d3fbc8); /* 24 */
  Sregex_GG (a, b, c, d, x[ 9], Sregex_S21, 0x21e1cde6); /* 25 */
  Sregex_GG (d, a, b, c, x[14], Sregex_S22, 0xc33707d6); /* 26 */
  Sregex_GG (c, d, a, b, x[ 3], Sregex_S23, 0xf4d50d87); /* 27 */
  Sregex_GG (b, c, d, a, x[ 8], Sregex_S24, 0x455a14ed); /* 28 */
  Sregex_GG (a, b, c, d, x[13], Sregex_S21, 0xa9e3e905); /* 29 */
  Sregex_GG (d, a, b, c, x[ 2], Sregex_S22, 0xfcefa3f8); /* 30 */
  Sregex_GG (c, d, a, b, x[ 7], Sregex_S23, 0x676f02d9); /* 31 */
  Sregex_GG (b, c, d, a, x[12], Sregex_S24, 0x8d2a4c8a); /* 32 */

  /* Round 3 */
  Sregex_HH (a, b, c, d, x[ 5], Sregex_S31, 0xfffa3942); /* 33 */
  Sregex_HH (d, a, b, c, x[ 8], Sregex_S32, 0x8771f681); /* 34 */
  Sregex_HH (c, d, a, b, x[11], Sregex_S33, 0x6d9d6122); /* 35 */
  Sregex_HH (b, c, d, a, x[14], Sregex_S34, 0xfde5380c); /* 36 */
  Sregex_HH (a, b, c, d, x[ 1], Sregex_S31, 0xa4beea44); /* 37 */
  Sregex_HH (d, a, b, c, x[ 4], Sregex_S32, 0x4bdecfa9); /* 38 */
  Sregex_HH (c, d, a, b, x[ 7], Sregex_S33, 0xf6bb4b60); /* 39 */
  Sregex_HH (b, c, d, a, x[10], Sregex_S34, 0xbebfbc70); /* 40 */
  Sregex_HH (a, b, c, d, x[13], Sregex_S31, 0x289b7ec6); /* 41 */
  Sregex_HH (d, a, b, c, x[ 0], Sregex_S32, 0xeaa127fa); /* 42 */
  Sregex_HH (c, d, a, b, x[ 3], Sregex_S33, 0xd4ef3085); /* 43 */
  Sregex_HH (b, c, d, a, x[ 6], Sregex_S34,  0x4881d05); /* 44 */
  Sregex_HH (a, b, c, d, x[ 9], Sregex_S31, 0xd9d4d039); /* 45 */
  Sregex_HH (d, a, b, c, x[12], Sregex_S32, 0xe6db99e5); /* 46 */
  Sregex_HH (c, d, a, b, x[15], Sregex_S33, 0x1fa27cf8); /* 47 */
  Sregex_HH (b, c, d, a, x[ 2], Sregex_S34, 0xc4ac5665); /* 48 */

  /* Round 4 */
  Sregex_II (a, b, c, d, x[ 0], Sregex_S41, 0xf4292244); /* 49 */
  Sregex_II (d, a, b, c, x[ 7], Sregex_S42, 0x432aff97); /* 50 */
  Sregex_II (c, d, a, b, x[14], Sregex_S43, 0xab9423a7); /* 51 */
  Sregex_II (b, c, d, a, x[ 5], Sregex_S44, 0xfc93a039); /* 52 */
  Sregex_II (a, b, c, d, x[12], Sregex_S41, 0x655b59c3); /* 53 */
  Sregex_II (d, a, b, c, x[ 3], Sregex_S42, 0x8f0ccc92); /* 54 */
  Sregex_II (c, d, a, b, x[10], Sregex_S43, 0xffeff47d); /* 55 */
  Sregex_II (b, c, d, a, x[ 1], Sregex_S44, 0x85845dd1); /* 56 */
  Sregex_II (a, b, c, d, x[ 8], Sregex_S41, 0x6fa87e4f); /* 57 */
  Sregex_II (d, a, b, c, x[15], Sregex_S42, 0xfe2ce6e0); /* 58 */
  Sregex_II (c, d, a, b, x[ 6], Sregex_S43, 0xa3014314); /* 59 */
  Sregex_II (b, c, d, a, x[13], Sregex_S44, 0x4e0811a1); /* 60 */
  Sregex_II (a, b, c, d, x[ 4], Sregex_S41, 0xf7537e82); /* 61 */
  Sregex_II (d, a, b, c, x[11], Sregex_S42, 0xbd3af235); /* 62 */
  Sregex_II (c, d, a, b, x[ 2], Sregex_S43, 0x2ad7d2bb); /* 63 */
  Sregex_II (b, c, d, a, x[ 9], Sregex_S44, 0xeb86d391); /* 64 */

  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;

  /* Zeroize sensitive information. */
  memset ((char *) x, 0, sizeof (x));
  return(1);
}


int Sregex_md5__encode(output, input, len)
unsigned char *output;
unsigned long *input;
unsigned int len;
{
  unsigned int i, j;

  for (i = 0, j = 0; j < len; i++, j += 4) {
    output[j] = (unsigned char)(input[i] & 0xff);
    output[j+1] = (unsigned char)((input[i] >> 8) & 0xff);
    output[j+2] = (unsigned char)((input[i] >> 16) & 0xff);
    output[j+3] = (unsigned char)((input[i] >> 24) & 0xff);
  }
  return(1);
}



int Sregex_md5_init(ctx,flag)
Sregex_md5_ctX *ctx;
int flag;
{
 ctx->count[0] = ctx->count[1] = 0;
 /* Load magic initialization constants. */
 ctx->state[0] = 0x67452301;
 ctx->state[1] = 0xefcdab89;
 ctx->state[2] = 0x98badcfe;
 ctx->state[3] = 0x10325476;
 return(1);
}


/* MD5 block update operation. Continues an MD5 message-digest
  operation, processing another message block, and updating the
  context.
 */
int Sregex_md5_update(ctx,data,datalen,flag)
Sregex_md5_ctX *ctx;
unsigned char *data;
int datalen;
int flag;
{
 unsigned int i, index, partlen;

 /* Compute number of bytes mod 64 */
 index = (unsigned int)((ctx->count[0] >> 3) & 0x3F);
 /* Update number of bits */
 if ((ctx->count[0] += ((unsigned long) datalen << 3)) < 
     ((unsigned long) datalen << 3))
   ctx->count[1]++;
 ctx->count[1] += ((unsigned long) datalen >> 29);
 partlen = 64 - index;

 /* Transform as many times as possible. */
 if (datalen >= partlen) {
   memcpy((char *) &ctx->buffer[index], (char *) data, partlen);
   Sregex_md5__transform(ctx->state, ctx->buffer);
   for (i = partlen; i + 63 < datalen; i += 64)
     Sregex_md5__transform(ctx->state, &data[i]);
   index = 0;
 } else
   i = 0;

 /* Buffer remaining data */
 memcpy((char *) &ctx->buffer[index], (char *) &data[i],datalen-i);

 return(1);
}


int Sregex_md5_final(ctx,result,flag)
Sregex_md5_ctX *ctx;
char result[16];
int flag;
{
 unsigned char bits[8];
 unsigned int index, padlen;
 static unsigned char PADDING[64] = {
   0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };

 /* Save number of bits */
 Sregex_md5__encode(bits, ctx->count, 8);
 /* Pad out to 56 mod 64.  */
 index = (unsigned int)((ctx->count[0] >> 3) & 0x3f);
 padlen = (index < 56) ? (56 - index) : (120 - index);
 Sregex_md5_update(ctx, PADDING, padlen,0);
 /* Append length (before padding) */
 Sregex_md5_update(ctx, bits, 8,0);
 /* Store state in result */
 Sregex_md5__encode(result, ctx->state, 16);
 /* Zeroize sensitive information.  */
 memset ((char *) ctx, 0, sizeof (*ctx));
 return(1);
}


int Sregex_md5(ctx,data,datalen,result,flag)
Sregex_md5_ctX **ctx;     /* *ctx has to be NULL or point to freeable memory */
unsigned char *data;
int datalen;
char result[16];
int flag;
/*
 bit0= allocate and init *ctx
 bit1= transfer ctx to result
 bit2= with bit 0 : clone new *ctx from data
 bit15= free *ctx
*/
{
 if(flag&1) {
   if(*ctx!=NULL)
     free((char *) *ctx);
   *ctx= TSOB_FELD(Sregex_md5_ctX,1);
   if(*ctx==NULL)
     return(-1);
   Sregex_md5_init(*ctx,0);
   if(flag&4)
     memcpy((char *) *ctx,data,sizeof(Sregex_md5_ctX));
 }
 if(*ctx==NULL)
   return(0);
 if(datalen>0)
   Sregex_md5_update(*ctx,data,datalen,0);
 if(flag&2) {
   Sregex_md5_final(*ctx,result,0);
 }
 if(flag&(1<<15)) {
   free((char *) *ctx);
   *ctx= NULL;
 }
 return(1);
}


/* ----------------- Sregex extract end -------- */


/* ----------------------- Lies_popen --------------------- */


struct Lies_popeN {
 char cmd[4096];
 int anz;          /* count */
 char **lines;
};


int Lies_popen_init(cmd,gelesen,flag)
char *cmd;
struct Lies_popeN *gelesen;
int flag;
{
 FILE *fp;
 struct ZeilE {
   struct ZeilE *prev;
   struct ZeilE *next;
   char *text;
 };
 struct ZeilE *zeile,*neuzeile,*startzeile= NULL;
 char line[4096],*fret;
 int i,zeilz=0,gesl,ret=0,l;

 /* remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 strcpy(gelesen->cmd,cmd);
 gelesen->anz= 0;
 gelesen->lines= NULL;

 fp= popen(cmd,"r");
 if(fp==NULL){
   fprintf(stderr,"popen(\"%s\",\"r\") failed. errno=%ld",cmd,errno);
   return(0);
 }

 /* create a list of lines  */
 startzeile=  TSOB_FELD(struct ZeilE,1);
 startzeile->prev= startzeile->next= NULL;
 startzeile->text= NULL;
 /* read in */
 zeile= startzeile;
 while(1) {
   fret= Sfile_fgets(line,sizeof(line)-1,fp);
 if(fret==NULL) break;
   zeile->text= malloc(strlen(line)+1);
 if(zeile->text == NULL) break;
   strcpy(zeile->text,line);
   neuzeile= TSOB_FELD(struct ZeilE,1);
 if(neuzeile==NULL) break;
   zeile->next= neuzeile;
   neuzeile->prev= zeile;
   neuzeile->next= NULL;
   neuzeile->text= NULL;
   zeile= neuzeile;
   zeilz++;
 } 
 if(fret!=NULL) {
   /* something went wrong */
speichermangel:;
   fprintf(stderr,"out of memory while reading output of: %s\n",cmd);
   ret= -1;
   goto tmp_freigeben;
 }

 pclose(fp);
 if(zeilz==0) goto tmp_freigeben;

 /* create return data */
 gelesen->anz= zeilz;
 gelesen->lines= TSOB_FELD(char *,gelesen->anz);
 if(gelesen->lines==NULL) goto speichermangel;
 for(i=0;i<gelesen->anz;i++)
   gelesen->lines[i]= NULL;
 gesl= 0;
 zeile= startzeile;
 for(i=0;i<zeilz;i++){
   l= strlen(zeile->text);
   gelesen->lines[i]= TSOB_FELD(char,l+1);
   if(gelesen->lines[i]==NULL) goto speichermangel;
   strcpy(gelesen->lines[i],zeile->text);
   gesl+= l+1;
   zeile= zeile->next;
 }

 ret= 1;
tmp_freigeben:;
 if(ret<=0)
    Lies_popen_cleanup(gelesen,0);
 zeile= startzeile;
 while(zeile!=NULL){
   if(zeile->text!=NULL) free(zeile->text);
   neuzeile= zeile->next;
   free((char *) zeile);
   zeile= neuzeile;
 } 
 return(ret);

}


int Lies_popen_cleanup(gelesen,flag)
struct Lies_popeN *gelesen;
int flag;
{
 int i;

 /* remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 if(gelesen==NULL)
   return(0);
 if(gelesen->lines!=NULL){
   for(i= 0;i<gelesen->anz;i++)
     if(gelesen->lines[i]!=NULL)
       free(gelesen->lines[i]);
   free((char *) gelesen->lines);
   gelesen->lines= NULL;
 }
 gelesen->anz= 0;
 return(1);
}


int Lies_popen_transplant(gelesen,target,index,flag)
struct Lies_popeN *gelesen;
char **target;
int index;
int flag;
{
 (*target)= gelesen->lines[index];
 gelesen->lines[index]= NULL;
 return(1);
}


/* ----------------------- Lies_popen - end -------------- */


/* ----------------- Checkrecord ----------------- */

struct CheckrecorD {

 char *name;
 time_t date;
 double size;
 int hashtype;   /* 0=CRC32 , 1=MD5 */
 union {
  unsigned long crc32;
  unsigned char md5[16];
 } checksum;

 struct CheckrecorD *prev,*next;

};


int Checkrecord_new(objpt,name,date,size,hashtype,crc32,md5,flag)
struct CheckrecorD **objpt;
char *name;
time_t date;
double size;
int hashtype;
unsigned long crc32;
unsigned char md5[16];
int flag;
/*
 bit0= set date automatically
*/
{
 struct CheckrecorD *o;
 int ret,i;

 *objpt= o= TSOB_FELD(struct CheckrecorD,1);
 if(o==NULL)
   return(-1);

 o->name= NULL;
 if(flag&1)
   o->date= time(0);
 else
   o->date= date;
 o->size= size;
 o->hashtype= hashtype;
 o->checksum.crc32= 0;
 for(i=0;i<16;i++)
   o->checksum.md5[i]= 0;
 if(hashtype==0)
   o->checksum.crc32= crc32;
 else if(md5!=NULL) 
   for(i=0;i<16;i++)
     o->checksum.md5[i]= md5[i];
 o->prev= NULL;
 o->next= NULL;

 if(Sregex_string(&(o->name),name,0)<=0)
   goto failed;

 return(1);
failed:;
 Checkrecord_destroy(objpt,0);
 return(-1);
}


int Checkrecord_destroy(objpt,flag)
struct CheckrecorD **objpt;
int flag;
{
 struct CheckrecorD *o;

 o= *objpt;
 if(o==NULL)
   return(0);

 Sregex_string(&(o->name),NULL,0);
 if(o->prev!=NULL)
   o->prev->next= o->next;
 if(o->next!=NULL)
   o->next->prev= o->prev;

 Smem_freE((char *) o);
 *objpt= NULL;
 return(1);
}


int Checkrecord_destroy_all(objpt,flag)
struct CheckrecorD **objpt;
int flag;
{
 struct CheckrecorD *o,*n;

 o= *objpt;
 if(o==NULL)
   return(0);
 for(;o->prev!=NULL;o= o->prev);
 for(;o!=NULL;o= n) {
   n= o->next;
   Checkrecord_destroy(&o,0);
 }
 *objpt= NULL;
 return(1);
}


int Checkrecord_get_name(o,pt,flag)
struct CheckrecorD *o;
char **pt;
int flag;
{
 *pt= o->name;
 return(1);
}


int Checkrecord_get_date(o,pt,flag)
struct CheckrecorD *o;
time_t *pt;
int flag;
{
 *pt= o->date;
 return(1);
}


int Checkrecord_get_size(o,pt,flag)
struct CheckrecorD *o;
double *pt;
int flag;
{
 *pt= o->size;
 return(1);
}


int Checkrecord_get_checksum(o,hashtype,crc32,md5,flag)
struct CheckrecorD *o;
int *hashtype;
unsigned long *crc32;
unsigned char md5[16];
int flag;
{
 int i;

 *hashtype= o->hashtype;
 if(o->hashtype==0)
   *crc32= o->checksum.crc32;
 else
   for(i=0;i<16;i++)
     md5[i]= o->checksum.md5[i];
 return(1);
}


int Checkrecord_get_prev(o,pt,flag)
struct CheckrecorD *o;
struct CheckrecorD **pt;
int flag;
{
 *pt= o->prev;
 return(1);
}


int Checkrecord_get_next(o,pt,flag)
struct CheckrecorD *o;
struct CheckrecorD **pt;
int flag;
{
 *pt= o->next;
 return(1);
}


int Checkrecord_link(o,link,flag)
struct CheckrecorD *o;
struct CheckrecorD *link;
int flag;
/*
  bit0= insert as link->prev rather than as link->next
*/
{
 if(o->prev!=NULL)
   o->prev->next= o->next;
 if(o->next!=NULL)
   o->next->prev= o->prev;
 o->prev= o->next= NULL;
 if(link==NULL)
   return(1);
 if(flag&1) {
   o->next= link;
   o->prev= link->prev;
   if(o->prev!=NULL)
     o->prev->next= o;
   link->prev= o;
 } else {
   o->prev= link;
   o->next= link->next;
   if(o->next!=NULL)
     o->next->prev= o;
   link->next= o;
 }
 return(1);
}


int Checkrecord_count(liststart,flag)
struct CheckrecorD *liststart;
int flag;
{
 struct CheckrecorD *o;
 int count= 0;

 for(o= liststart;o!=NULL;o= o->next) {
   if(o->hashtype<0)
 continue;
   count++;
 }
 return(count);
}


int Checkrecord_compare_checksum(f1,f2)
struct CheckrecorD **f1,**f2;
{
 int i;

 if((*f1)->hashtype<(*f2)->hashtype)
   return(-2);
 if((*f1)->hashtype>(*f1)->hashtype)
   return(2);
 if((*f1)->hashtype == 0) {
   if((*f1)->checksum.crc32<(*f2)->checksum.crc32)
     return(-1);
   if((*f1)->checksum.crc32>(*f2)->checksum.crc32)
     return(1);
 } else {
   for(i=0;i<16;i++) {
     if((*f1)->checksum.md5[i]<(*f2)->checksum.md5[i])
       return(-1);
     if((*f1)->checksum.md5[i]>(*f2)->checksum.md5[i])
       return(1);
   }
 }
 return(0);
}


int Checkrecord_compare_size(f1,f2)
struct CheckrecorD **f1,**f2;
{
 if((*f1)->size<(*f2)->size)
   return(-1);
 if((*f1)->size>(*f2)->size)
   return(1);
 return(0);
}


int Checkrecord_create_sorted_pt(liststart,pt_io,pt_z,Checkrecord_compare,flag)
struct CheckrecorD *liststart;
int (*Checkrecord_compare)();
struct CheckrecorD ***pt_io;
int *pt_z;
int flag;
{
 struct CheckrecorD *o;
 int i,ret,count= 0;

 *pt_io= NULL;
 count= Checkrecord_count(liststart,0);
 *pt_io= TSOB_FELD(struct CheckrecorD *,count+1);
 if(*pt_io==NULL)
   {ret= -1; goto ex;}
 i= 0;
 for(o= liststart;o!=NULL;o= o->next) {
   if(o->hashtype<0)
 continue;
   if(i>=count)
 break;
   (*pt_io)[i]= o;
   i++;
 }
 count= i;
 if(count>0)
   qsort(*pt_io,count,sizeof(struct CheckrecorD *),Checkrecord_compare);
 *pt_z= count;
 ret= 1;
ex:;
 if(ret<=0) {
   if(*pt_io!=NULL)
     Smem_freE((char *) *pt_io);
   *pt_io= NULL;
 }
 return(ret);
}


int Checkrecord_read_list(objpt,listadr,rec_count,flag)
struct CheckrecorD **objpt;
char *listadr;
int *rec_count;
int flag;
/*
 bit0= sorted by size 
*/
{
 int hashtype= 0,ret,valid,line_count= 0,i,sortcount= 0;
 char line[256],name[256],date_text[256],size_text[256],checksum_text[256];
 char buf[3];
 unsigned char md5[16];
 time_t date;
 double size;
 unsigned long crc32;
 FILE *fp= NULL;
 struct CheckrecorD *o,**sortpt= NULL;

 *rec_count= 0;
 fp= Sfile_fopen(listadr,"r",0);
 if(fp==NULL)
   {ret= 0; goto ex;}
 while(Sfile_fgets(line,sizeof(line)-1,fp)!=NULL) {
   (line_count)++;
   sscanf(line,"%s %s %s %s",name,date_text,size_text,checksum_text);
   ret= Zeitdefinition(date_text,&valid,&date,0);
   if(ret<=0)
 continue;
   size= -1;
   sscanf(size_text,"%lf",&size);
   if(size<=0)
 continue;
   /* determine hashtype and decode */
   if(strlen(checksum_text)==32) {
     hashtype= 1;
     for(i=0;i<16;i++) {
       buf[0]= checksum_text[2*i];
       buf[1]= checksum_text[2*i+1];
       buf[2]= 0;
       sscanf(buf,"%x",&ret);
       md5[i]= ret;
     }
   } else {
     hashtype= 0;
     crc32= 0;
     sscanf(checksum_text,"%X",&crc32);
   }
   /* create and link new record */
   o= *objpt;
   ret= Checkrecord_new(objpt,name,date,size,hashtype,crc32,md5,0);
   if(ret<=0)
     goto ex;
   Checkrecord_link(*objpt,o,1);
   (*rec_count)++;
 }
 if(*objpt==NULL)
   {ret= 2; goto ex;}

 if(flag&1) {
   /* sort to pointer array */
   ret= Checkrecord_create_sorted_pt(*objpt,&sortpt,&sortcount,
                                     Checkrecord_compare_size,0);
   if(ret<=0)
     goto ex;
   if(sortcount!=*rec_count) 
     {ret= -1; goto ex;}          /* that would be very strange */
   /* relink */
   for(i=0;i<sortcount;i++) {
     if(i==0)
       sortpt[i]->prev= NULL;
     else
       sortpt[i]->prev= sortpt[i-1];
     if(i<sortcount-1)
       sortpt[i]->next= sortpt[i+1];
     else
       sortpt[i]->next= NULL;
   }
   *objpt= sortpt[0];
 }
 ret= 1;
ex:;
 if(sortpt!=NULL)
   Smem_freE((char *) sortpt);
 if(fp!=NULL)
   Sfile_fclose(fp);
 return(ret);
}


/* ----------------- Checkrecord - ende -------- */


/* ----------------- Fileliste ------------------ */


/* extension record for checksum */
struct Filelist_crc_reC {
 unsigned long crc;
 unsigned char *md5;
};


int Filelist_crc_rec_new(o,flag)
struct Filelist_crc_reC **o;
int flag;
/*
 bit0= allocate storage for MD5
*/
{
 *o= TSOB_FELD(struct Filelist_crc_reC,1);
 if(o==NULL)
   goto ex;
 (*o)->crc= 0;
 (*o)->md5= NULL;
 if(flag&1) {
   (*o)->md5= TSOB_FELD(unsigned char,16);
   if((*o)->md5==NULL)
     goto ex;
 }
 return(1);
ex:;
 Filelist_crc_rec_destroy(o,0);
 return(-1);
}


int Filelist_crc_rec_destroy(o,flag)
struct Filelist_crc_reC **o;
int flag;
{
 if(*o==NULL)
   return(0);
 if((*o)->md5!=NULL)
   free((char *) (*o)->md5);
 free((char *) *o);
 *o= NULL;
 return(1);
}


/* extension record for file properties */
struct Filelist_xprop_reC {
 int mod;
 char *own; 
 char *grp;
 time_t date;
};


int Filelist_xprop_rec_new(o,flag)
struct Filelist_xprop_reC **o;
int flag;
{
 *o= TSOB_FELD(struct Filelist_xprop_reC,1);
 if(o==NULL)
   return(-1);
 (*o)->mod= -1;
 (*o)->own= NULL;
 (*o)->grp= NULL;
 (*o)->date= 0;
 return(1);
}


int Filelist_xprop_rec_destroy(o,flag)
struct Filelist_xprop_reC **o;
int flag;
{
 if(*o==NULL)
   return(0);
 if((*o)->own!=NULL)
   free((char *) (*o)->own);
 if((*o)->grp!=NULL)
   free((char *) (*o)->grp);
 free((char *) *o);
 *o= NULL;
 return(1);
}


/* This factor speeds up linear search in a list tree for the price of
   sizeof(int)*Fileliste_hash_sizE additional bytes in every directory node */
#define Fileliste_hash_sizE 20


struct FilelistE {
 int ref_count;
 struct FilelistE *boss;
 int boss_idx;

 char *dir;
 char *dir_size_cmd;
 int dir_size_internal;
 char *target_dir;
 char *main_dir;

 struct FilelistE *sperrliste;

 char target_inp_trenner;  /* english : target_inp_separator */
 char *impl_igno;          /* filename characters which cause the backup
                              command to ignore files if they are not given
                              explicitely. (For mkisofs-1.12b3 : ~# )
                            */ 
 int maxanz;               /* english : maxanzcount */
 int anz;                  /* english : count */
 int enlarge_step;

 int use_hashing;
 int hashstart[Fileliste_hash_sizE];

 char **adressen;          /* english : addresses */
 char **targets;
 int *typen; /* 1=Data file, 2=Directory, 3=Link */ /* english : types */
 int *flags;   /* bit0= map to root 
                  bit1= contains young files / is young file
                  bit2= containsl old file / is old file 
                  bit3= crcs[idx] valid
                  bit4= crcs[idx] is md5 rather than crc32
                */
 double *groessen;         /* english : sizes */

#ifdef NIX
 unsigned long *crcs;
 unsigned char **md5s;
#else
 struct Filelist_crc_reC **crcs;
#endif /* ! NIX */
 struct Filelist_xprop_reC **xprops;

 int *hashlinks;
 struct FilelistE **subdirs;
};

static char Fileliste_dir_size_cmd[4096]={"-internal"};  /* {"du -sk"}; */
static int Fileliste_dir_size_internal= 1;
/* estimation of management information in a backup archives directory per
   file entry . This is a generous estimation to be on the safe side.
   The number may be changed at run time by  -dir_size_cmd -internal+number  */
static double Fileliste_node_overhead= 256.0;
static int Fileliste_node_addrl= 1;

static char Fileliste_target_inp_trenner= '=';
static char Fileliste_impl_igno[256]= {""};

double Fileliste_size();
double Fileliste_dir_size();


int Fileliste_new(liste,maxanz,main_dir,flag)
struct FilelistE **liste;
int maxanz;
char *main_dir;
int flag;
/*
 bit0= list shall use hashing
*/
{
 int i;
 *liste= TSOB_FELD(struct FilelistE,1);
 if(*liste==NULL) return(0);
 (*liste)->ref_count= 0;
 (*liste)->boss= NULL;
 (*liste)->boss_idx= 0;
 (*liste)->dir= NULL;
 (*liste)->dir_size_cmd= NULL;
 (*liste)->dir_size_internal= 0;
 (*liste)->target_dir= NULL;
 (*liste)->main_dir= NULL;
 (*liste)->sperrliste= NULL;
 (*liste)->target_inp_trenner= Fileliste_target_inp_trenner;
 (*liste)->impl_igno= NULL;
 (*liste)->use_hashing= (flag&1);
 for(i=0;i<Fileliste_hash_sizE;i++)
   (*liste)->hashstart[i]= -1;
 (*liste)->adressen= NULL;
 (*liste)->targets= NULL;
 (*liste)->typen= NULL;
 (*liste)->flags= NULL;
 (*liste)->groessen= NULL;
 (*liste)->crcs= NULL;
 (*liste)->xprops= NULL;
 (*liste)->hashlinks= NULL;
 (*liste)->subdirs= NULL;
 (*liste)->maxanz= 0;

 if(Sregex_string(&((*liste)->dir),"",0)<=0)
   goto wiederweg;
 if(Sregex_string(&((*liste)->dir_size_cmd),Fileliste_dir_size_cmd,0)<=0)
   goto wiederweg;
 (*liste)->dir_size_internal= Fileliste_dir_size_internal;
 if(Sregex_string(&((*liste)->target_dir),"",0)<=0)
   goto wiederweg;
 if(Sregex_string(&((*liste)->main_dir),main_dir,0)<=0)
   goto wiederweg;
 (*liste)->impl_igno= TSOB_FELD(char,strlen(Fileliste_impl_igno)+1);
 strcpy((*liste)->impl_igno,Fileliste_impl_igno);
 (*liste)->maxanz= maxanz;
 (*liste)->anz= 0;
 (*liste)->enlarge_step= 10;
 if(maxanz>0) {
   if(Fileliste_new_vectors(*liste,maxanz,0)<=0)
     goto wiederweg;
 }
 return(1); 
wiederweg:;
 Fileliste_destroy(liste,1);
 return(0);
}


int Fileliste_init_vectors(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;

 for(i=0;i<liste->maxanz;i++) {
   liste->adressen[i]= NULL;
   liste->targets[i]= NULL;
   liste->typen[i]= 0;
   liste->flags[i]= 0;
   liste->groessen[i]= -1;
   liste->crcs[i]= NULL;
   liste->xprops[i]= NULL;
   liste->hashlinks[i] -1;
   liste->subdirs[i]= NULL;
 }
 return(1);
}


int Fileliste_new_vectors(liste,maxanz,flag)
struct FilelistE *liste;
int maxanz;
int flag;
{

/*  maxanz= liste->maxanz; */
 if(maxanz<=0) return(2);

 Fileliste_destroy_vectors(liste,0);
 liste->adressen= TSOB_FELD(char *,maxanz);
 if(liste->adressen==NULL)
   return(-1);
 liste->targets= TSOB_FELD(char *,maxanz);
 if(liste->targets==NULL)  
   return(-1);
 liste->typen= TSOB_FELD(int,maxanz);
 if(liste->typen==NULL)  
   return(-1);
 liste->flags= TSOB_FELD(int,maxanz);
 if(liste->flags==NULL)  
   return(-1);
 liste->groessen= TSOB_FELD(double,maxanz);
 if(liste->groessen==NULL)  
   return(-1);
 liste->crcs= TSOB_FELD(struct Filelist_crc_reC *,maxanz);
 if(liste->crcs==NULL)  
   return(-1);
 liste->xprops= TSOB_FELD(struct Filelist_xprop_reC *,maxanz);
 if(liste->xprops==NULL)  
   return(-1);
 liste->hashlinks= TSOB_FELD(int,maxanz);
 if(liste->hashlinks==NULL)  
   return(-1);
 liste->subdirs= TSOB_FELD(struct FilelistE *,maxanz);
 if(liste->subdirs==NULL)  
   return(-1);
 liste->maxanz= maxanz;
 Fileliste_init_vectors(liste,0);
 return(1);
}


int Fileliste_destroy(liste,flag)
struct FilelistE **liste;
int flag;
/*
 bit0= the list is not finished (this bit does not do anything yet)
 bit1= deletion of a dependend list : lower ref_count
*/
{
 int i;

 if(*liste==NULL)
   return(0);
 if(flag&2)
   if((*liste)->ref_count>0)
     ((*liste)->ref_count)-- ;
 if((*liste)->ref_count>0){
   /* This list is still used by another list */
   return(1);
 }

 Fileliste_destroy_vectors(*liste,0);
 if((*liste)->impl_igno!=NULL)
    free((char *) (*liste)->impl_igno);
 /* (*liste)->sperrliste was not created by Fileliste_new() */
 if((*liste)->target_dir!=NULL)
   free((char *) (*liste)->target_dir);
 if((*liste)->main_dir!=NULL)
   free((char *) (*liste)->main_dir);
 if((*liste)->dir_size_cmd!=NULL)
   free((char *) (*liste)->dir_size_cmd);
 if((*liste)->dir!=NULL)
   free((char *) (*liste)->dir);
 if((*liste)!=NULL) 
   free((char *) (*liste));
 *liste= NULL;
 return(1);
}


int Fileliste_destroy_vectors(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;

 if(liste->subdirs!=NULL) {
   for(i=0;i<liste->anz;i++)
     if(liste->subdirs[i]!=NULL) 
       Fileliste_destroy(&(liste->subdirs[i]),2);
   free((char *) liste->subdirs);
 }
 if(liste->hashlinks!=NULL)
   free((char *) liste->hashlinks);
 if(liste->crcs!=NULL) {
   for(i=0;i<liste->anz;i++)
     Filelist_crc_rec_destroy(&(liste->crcs[i]),0);
   free((char *) liste->crcs);
 }
 if(liste->xprops!=NULL) {
   for(i=0;i<liste->anz;i++)
     Filelist_xprop_rec_destroy(&(liste->xprops[i]),0);
   free((char *) liste->xprops);
 }
 if(liste->groessen!=NULL)
   free((char *) liste->groessen);
 if(liste->flags!=NULL)
   free((char *) liste->flags);
 if(liste->typen!=NULL)
   free((char *) liste->typen);
 if(liste->targets!=NULL){
   for(i=0;i<liste->anz;i++)
     if(liste->targets[i]!=NULL)
       free(liste->targets[i]);
   free((char *) liste->targets);
 }
 if(liste->adressen!=NULL) {
   for(i=0;i<liste->anz;i++)
     if(liste->adressen[i]!=NULL)
       free(liste->adressen[i]);
   free((char *) liste->adressen);
 }
 liste->anz= 0;
 liste->maxanz= 0;
 return(1);
}


int Fileliste_destroy_item(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
{
 if(liste->adressen[idx]!=NULL)
   free(liste->adressen[idx]);
 liste->adressen[idx]= NULL;
 if(liste->targets[idx]!=NULL)
   free(liste->targets[idx]);
 liste->targets[idx]= NULL;
 if(liste->subdirs!=NULL) 
   if(liste->subdirs[idx]!=NULL)
     Fileliste_destroy(&(liste->subdirs[idx]),2);
 Filelist_crc_rec_destroy(&(liste->crcs[idx]),0);
 Filelist_xprop_rec_destroy(&(liste->xprops[idx]),0);
 liste->typen[idx]= 0;
 return(1);
}


int Fileliste_enlarge(liste,maxanz,flag)
struct FilelistE *liste;
int maxanz;
int flag;
{
 int i,ti;
 struct FilelistE *neuliste;
 char *tr; /* should be (void *) */

 if(liste->maxanz/5>liste->enlarge_step)
   liste->enlarge_step= liste->maxanz/5;
 i= (liste->enlarge_step)*(maxanz/(liste->enlarge_step));
 if(i<maxanz) i+= liste->enlarge_step;
 maxanz= i;

 if(liste->maxanz>=maxanz || maxanz<=0) return(1);

 /* create new list  */
 if(Fileliste_new(&neuliste,maxanz,liste->main_dir,liste->use_hashing)<=0)
   return(-1);

 /* copy old content */
 for(i=0;i<liste->maxanz;i++)
   Fileliste_copy_item(liste,i,neuliste,i,2);

 /* swap data handles */
 tr= (char *) liste->adressen;
 liste->adressen= neuliste->adressen;
 neuliste->adressen= (char **) tr;
 tr= (char *) liste->targets;
 liste->targets= neuliste->targets;
 neuliste->targets= (char **) tr;
 tr= (char *) liste->typen;
 liste->typen= neuliste->typen;
 neuliste->typen= (int *) tr;
 tr= (char *) liste->flags;
 liste->flags= neuliste->flags;
 neuliste->flags= (int *) tr;
 tr= (char *) liste->groessen;
 liste->groessen= neuliste->groessen;
 neuliste->groessen= (double *) tr;
 tr= (char *) liste->crcs;
 liste->crcs= neuliste->crcs;
 neuliste->crcs= (struct Filelist_crc_reC **) tr;
 tr= (char *) liste->xprops;
 liste->xprops= neuliste->xprops;
 neuliste->xprops= (struct Filelist_xprop_reC **) tr;
 tr= (char *) liste->hashlinks;
 liste->hashlinks= neuliste->hashlinks;
 neuliste->hashlinks= (int *) tr;
 for(i=0;i<Fileliste_hash_sizE;i++) {
   ti= liste->hashstart[i];
   liste->hashstart[i]= neuliste->hashstart[i];
   neuliste->hashstart[i]= ti;
 }
 tr= (char *) liste->subdirs;
 liste->subdirs= neuliste->subdirs;
 neuliste->subdirs= (struct FilelistE **) tr;
 /* update back pointers of sub lists */
 for(i=0;i<liste->maxanz;i++)
   if(liste->subdirs[i]!=NULL) {
     liste->subdirs[i]->boss= liste;
     liste->subdirs[i]->boss_idx= i;
   }
 neuliste->maxanz= liste->maxanz;
 liste->maxanz= maxanz;  

 /* delete temporary list object */
 Fileliste_destroy(&neuliste,0);

 return(1);
}


int Fileliste_resize(liste,maxanz,flag)
struct FilelistE *liste;
int maxanz;
int flag;
{
 int i;

 liste->anz= 0;

 if(maxanz>liste->maxanz || liste->adressen==NULL ||
      liste->targets==NULL || liste->flags==NULL ||
      liste->typen==NULL || liste->groessen==NULL ||
      liste->subdirs==NULL || liste->crcs==NULL ||
      liste->xprops==NULL || liste->hashlinks==NULL) {
   /* liste->maxanz= maxanz; */
   if(Fileliste_new_vectors(liste,maxanz,0)<=0) 
    return(-1);
 } else {
   Fileliste_init_vectors(liste,0);
 }
 return(1);
}


int Fileliste__compute_hashvalue(adr,flag)
char *adr;
int flag;
{
 unsigned long crc;

 Sregex_crc32_yenc(adr,strlen(adr),&crc,1);
 crc= crc%Fileliste_hash_sizE;
 return((int) crc);
}


int Fileliste_hashlink(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
/*
 bit0= unlink
*/
{
 int start_idx,i,pred_idx,*pred_pt;

 if(liste->use_hashing==0)
   return(2);
 if(liste->adressen[idx]==NULL)
   return(2);
 start_idx= Fileliste__compute_hashvalue(liste->adressen[idx],0);
 if(flag&1) {
   /* search predecessor in hash list */
   pred_pt= &liste->hashstart[start_idx];
   pred_idx= *pred_pt;
   while(pred_idx>=0) { 
     if(liste->hashlinks[pred_idx]<0) {
       pred_idx= -1;
   break;
     }
     if(liste->hashlinks[pred_idx]==idx)
   break;
     pred_pt= &(liste->hashlinks[pred_idx]);
     pred_idx= *pred_pt;
   }
   /* set predecessor's hashlink to successor of idx */
   if(pred_idx<0)
     return(0);  
   *pred_pt= liste->hashlinks[idx];
 } else {
   liste->hashlinks[idx]= liste->hashstart[start_idx];
   liste->hashstart[start_idx]= idx;
 }
 return(1);
}


int Fileliste_set_address(liste,idx,address,flag)
struct FilelistE *liste;
int idx;
char *address;
int flag;
{
 if(liste->anz<=idx)
   return(0);
 if(liste->adressen[idx]!=NULL)
   free((char *) liste->adressen[idx]);
 liste->adressen[idx]= NULL;
 if(address!=NULL){
   liste->adressen[idx]=  TSOB_FELD(char,strlen(address)+1);
   if(liste->adressen[idx]==NULL)
     return(-1);
   strcpy(liste->adressen[idx],address);
 }
 return(1);
}


int Fileliste_dir(liste,dir,flag)
struct FilelistE *liste;
char *dir;
int flag;
/*
 bit0= forwarded to Fileliste_dir_size() (see there)
 bit1= no need to determine sizes
*/
{
 int i,w,l,dirl,ret,zaehl;
 char line[4096],wdmerk[4096];
 struct Lies_popeN *gelesen,gelesen_data;
 struct stat stbuf;

 gelesen= &gelesen_data;

 liste->anz= 0;
 if(Sregex_string(&(liste->dir),dir,0)<=0)
   return(-1);

 dirl= strlen(dir);
 
 getcwd(wdmerk,sizeof(wdmerk)-1);

 strcpy(line,"ls -a1"); 
 chdir(dir);
 if(Lies_popen_init(line,gelesen,0)<=0)
   {ret= 0; goto ex;}

 /* count items and create pointer list for file addresses */
 zaehl= 0;
 for(i=0;i<gelesen->anz;i++) {
   /* addresses that are not welcome */
   if(strcmp(gelesen->lines[i],".")==0 || strcmp(gelesen->lines[i],"..")==0)
 continue;
   if(Cd_backup_planer_call_lstaT(gelesen->lines[i],&stbuf)==-1)
 continue;
   zaehl++;
 }
 /* eventually attach a new line vector */
 if(Fileliste_resize(liste,zaehl+1,0)<=0)
   {ret= -1; goto ex;}

 /* read file addresses from popen-result */
 liste->anz= 0;
 w= 0;
 for(i=0;i<gelesen->anz;i++) {
   /* just in case there are suddenly more files than announced */
   if(w>=liste->maxanz)
 break;
   /* addresses that are not welcome */
   if(strcmp(gelesen->lines[i],".")==0 || strcmp(gelesen->lines[i],"..")==0)
 continue;
   if(Cd_backup_planer_call_lstaT(gelesen->lines[i],&stbuf)==-1)
 continue;
   /* prepend dir and add to list  */
   l= dirl+1+strlen(gelesen->lines[i]);
   liste->adressen[w]= TSOB_FELD(char,l+1);
   if(liste->adressen[w]==NULL)
     {ret= -1; goto ex;}
   if(dir[0] && dir[strlen(dir)-1]!='/') 
     sprintf(liste->adressen[w],"%s/%s",dir,gelesen->lines[i]);
   else
     sprintf(liste->adressen[w],"%s%s",dir,gelesen->lines[i]);
   w++;
 }
 liste->anz= w;

 /* the popen result is fully exploited */
 Lies_popen_cleanup(gelesen,0); gelesen= NULL;

 chdir(wdmerk);
 ret= Fileliste_read_properties(liste,((flag&1)<<2)|(flag&2));
ex:;
 chdir(wdmerk);
 Lies_popen_cleanup(gelesen,0); gelesen= NULL;
 if(ret<=0) return(ret);

 return(liste->anz);
}


int Fileliste_lines(liste,filec,files,flag)
struct FilelistE *liste;
int filec;
char **files;
int flag;
/*
 bit0= no Fileliste_read_properties()
 bit3= do not separate targets from sources (files[i] is all source)
*/
{
 int ret= 0,i;

 if(filec>liste->maxanz) return(0);
 /* eventually attach a new line vector */
 if(Fileliste_resize(liste,filec+1,0)<=0) return(-1);
 liste->anz= 0;
 for(i=0;i<filec;i++) {
   ret= Fileliste_append(liste,files[i],flag&(1|8));
   if(ret<=0) return(ret);
 }
 return(1);
}


int Fileliste_single(liste,name,typ,groesse,flag)
struct FilelistE **liste;
char *name;
int typ;
double groesse;
int flag;
/*
 bit0= map file to root (set liste->flags:bit0 )
 bit1= set target_inp_trenner to '='  
 bit3= do not separate target from source (name is all source)
*/
{
 if(Fileliste_new(liste,1,"",0)<=0) return(0);
 if(flag&2)
   Fileliste_set_target_inp_trenner(liste,"=",0);
 Fileliste_lines(*liste,1,&name,1|(flag&8));
 (*liste)->typen[0]= typ;
 (*liste)->flags[0]|= (flag&1);
 (*liste)->groessen[0]= groesse;
 return(1);
}


int Fileliste_set_crc(liste,idx,crc,md5,flag)
struct FilelistE *liste;
int idx;
unsigned long crc;
unsigned char md5[16];
int flag;
/*
 bit0= make crc invalid
 bit12= set md5 rather than crc
*/
{
 int ret;

 if(liste==NULL)
   return(0);
 if(liste->anz<=idx)
   return(0);
 if(flag&1)
   liste->flags[idx]&= ~8;
 else {
   if(liste->crcs[idx]==NULL) {
     ret= Filelist_crc_rec_new(&liste->crcs[idx],!!(flag&(1<<12)));
     if(ret<=0)
       return(ret);
   }
   if(flag&(1<<12)) {
     memcpy(liste->crcs[idx]->md5,md5,16);
     liste->flags[idx]|= 16;
   } else
     liste->crcs[idx]->crc= crc;
   liste->flags[idx]|= 8;
 }
 return(1);
}


int Fileliste_set_xprop(liste,idx,mod,own,grp,date,flag)
struct FilelistE *liste;
int idx;
int mod;
char *own;
char *grp;
time_t date;
int flag;
/*
*/
{
 int ret;

 if(liste==NULL)
   return(0);
 if(liste->anz<=idx)
   return(0);
 if(liste->xprops[idx]==NULL) {
   ret= Filelist_xprop_rec_new(&liste->xprops[idx],0);
   if(ret<=0)
     return(ret);
 }
 liste->xprops[idx]->mod= mod;
 if(Sregex_string(&(liste->xprops[idx]->own),own,0)<=0)
   return(-1);
 if(Sregex_string(&(liste->xprops[idx]->grp),grp,0)<=0)
   return(-1);
 liste->xprops[idx]->date= date;
 return(1);
}


int Fileliste_read_properties(liste,flag)
struct FilelistE *liste;
int flag;
/*
 bit0= do not complain about missing file
 bit1= no determination of size necessary
 bit2= determine size of excluded directories
 bit3= determine size even if the list item already has a size recorded
 bit4= accept all file types
 bit5= return 1 if there were 0s and 1s
*/
{
 int i,ret,ret_one= 0,ret_zero= 0;

 /* determine file types and sizes */
 for(i=0;i<liste->anz;i++) {
   liste->groessen[i]= -1;
   ret= Fileliste_read_properties_idx(liste,i,flag);
   if(ret<0) 
     return(ret);
   if(ret==0)
     ret_zero++;
   else
     ret_one++;
 }
 if(flag&32)
   if(ret_one)
     ret_zero=0;
 return(!ret_zero);
}


double Fileliste_size(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;
 double groesse;

 groesse= 0.0;
 for(i=0;i<liste->anz;i++)
   if(liste->typen[i]==1||liste->typen[i]==2)
     if(liste->groessen[i]>0)
       groesse+= liste->groessen[i];
 return(groesse);
}


double Fileliste_dir_size(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
/*
 bit0= determine size even if directory is excluded (else 0 byte )
 bit1= attach subdir 
*/
{
 struct Lies_popeN *gelesen,gelesen_data;
 struct FilelistE *sub;
 double kbytes,groesse= -1.0;
 char line[4096],txs_buf[4096];
 int i,ret,gleich=0,echt_drueber=0;

 gelesen= &gelesen_data;

 if(idx>liste->anz) return(-1.0);

 /* attempt to reuse already recorded sizes */
 if(liste->groessen[idx]>0)
   return(liste->groessen[idx]);
 if(liste->subdirs!=NULL)
   if(liste->subdirs[idx]!=NULL){
     groesse= Fileliste_size(liste->subdirs[idx],0);
     return(groesse);
   }

 if(Fileliste_search_file(liste->sperrliste,liste->adressen[idx],0)>=0)
   gleich= 1;
 if(Fileliste_search_file(liste->sperrliste,liste->adressen[idx],1|4)>=0)
   echt_drueber= 1;
  
 if( gleich  && !(flag&1) ){ /* is the directory itself excluded ? */
   return(-1.0);
 }else if(echt_drueber){ /* is a sub directory excluded ? */
   /* new file list from directory */;
   if(Fileliste_new(&sub,0,liste->main_dir,0)<=0) 
     return(-1.0);
   Fileliste_set_sperrliste(sub,liste->sperrliste,0);
   if(Fileliste_dir(sub,liste->adressen[idx],0)<=0) 
     goto sub_ex;
   groesse= Fileliste_size(sub,0);
   /* we do not throw this information away */
   if(flag&4) Fileliste_set_subdir(liste,idx,sub,0);
sub_ex:;
   /* this only deletes sub if it was not attached as a sub list */
   Fileliste_destroy(&sub,0);
   return(groesse);
 }else{
   if(liste->dir_size_internal) {
     ret= Fileliste_determine_treesize(liste,idx,0,&groesse,0);
     if(ret<=0)
       groesse= -1.0;
   } else {
     /* determine size by  dir_size_cmd  */
     Text_shellsafe(liste->adressen[idx],txs_buf,0);
     sprintf(line,"%s %s",liste->dir_size_cmd,txs_buf);
     if(Lies_popen_init(line,gelesen,0)<=0) return(-1.0);
     if(gelesen->anz>0) {
       if(sscanf(gelesen->lines[0],"%lf",&kbytes)==1)
         groesse= kbytes*1024.0;
     }
     Lies_popen_cleanup(gelesen,0);
   }
 }
 return(groesse);
}


int Fileliste_read_properties_idx(liste,i,flag)
struct FilelistE *liste;
int i;
int flag;
/*
 bit0= do not complain about missing file
 bit1= no determination of size necessary
 bit2= determine size of excluded directories
 bit3= determine size even if the list item already has a size recorded
 bit4= accept all file types
 >>> bit5= determine xprop (mod,own,grp,date)
 >>> bit11= determine checksum
 >>> bit12= checksum MD5 rather than CRC32
*/
{
 double kbytes,groesse;
 char target[4096];
 int adrlen;
 struct Lies_popeN *gelesen,gelesen_data;
 struct Cd_backup_planer_struct_staT stbuf;
 static int verbous= 0;

 gelesen= &gelesen_data;

 liste->typen[i]= 0;
 if((flag&8)){
   Fileliste_set_subdir(liste,i,NULL,0);
   liste->groessen[i]= -1.0;
 }
 if(Cd_backup_planer_call_lstaT(liste->adressen[i],&stbuf)==-1) {
   if(!(flag&1))
     fprintf(stderr,
             "lstat(%s) failed. %s\n",liste->adressen[i],strerror(errno));
   return(0);
 }
 if(S_ISREG(stbuf.st_mode)) {
   /* Data file */
   liste->typen[i]= 1;

   /* >>> if the file is not readable : complain and block out */

   liste->groessen[i]= stbuf.st_size;
 } else if (S_ISDIR(stbuf.st_mode)) {
   /* Directory */
   if(flag&2){
     liste->typen[i]= 2;
   }else{

     /* >>> if the directory is not readable : complain and block out */

     groesse= Fileliste_dir_size(liste,i,4|((flag>>2)&1));
     if(groesse>=0.0) {
       liste->groessen[i]= groesse;
       liste->typen[i]= 2;
     }
   }
 } else if (S_ISLNK(stbuf.st_mode)) {
   /* Softlink */
   liste->typen[i]= 3;
   liste->groessen[i]= stbuf.st_size;
 } else {
   liste->typen[i]= 0x40000000;
   if(flag&16) 
     liste->groessen[i]= -1;
   else
     liste->groessen[i]= stbuf.st_size;
 }

 /* eventually add name length and node_overhead */
 if(liste->groessen[i]>=0 && liste->typen[i]!=2 ) {
   if(Fileliste_node_addrl) {
     if(Fileliste_target(liste,i,target,0)<=0)
       adrlen= strlen(liste->adressen[i]);
     else
       adrlen= strlen(target);
   } else
     adrlen= 0;
   liste->groessen[i]+= adrlen+Fileliste_node_overhead;
 }

 if(verbous)
   fprintf(stderr,"%.lf %s\n",liste->groessen[i],liste->adressen[i]);

 return(1);
}


int Fileliste_filetyp(adr,flag)
char *adr;
int flag;
{
 struct FilelistE *liste;
 int ret;

 if(Fileliste_single(&liste,adr,0,0.0,0)<=0)
   return(-1);
 ret= Fileliste_read_properties(liste,1|2);
 if(ret>0) ret= liste->typen[0];
 Fileliste_destroy(&liste,0);
 return(ret);
}


int Fileliste_defunct_item(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
{
 if(idx>=liste->anz)
   return(0);
 if(liste->typen[idx]>0)
   liste->typen[idx]= -liste->typen[idx];
 return(1);
}


int Fileliste_copy_item(von,von_idx,nach,nach_idx,flag)
struct FilelistE *von;
int von_idx;
struct FilelistE *nach;
int nach_idx;
int flag;
/*
 flag: for Fileliste_copy_item_hd
  blocked is :  bit0 
*/
{
 return(Fileliste_copy_item_hd(von,von_idx,nach,nach_idx,flag&~1));
}


/* was once more different from Fileliste_copy_item(). Now : only bit1 */
int Fileliste_copy_item_hd(von,von_idx,nach,nach_idx,flag)
struct FilelistE *von;
int von_idx;
struct FilelistE *nach;
int nach_idx;
int flag;
/*
 bit0= determine von.target write it explicitely into nach
       ("von" = from , "nach" = to )
 bit1= move allocated data rather than copying them
*/
{
 int ret;
 char target[4096],*targetpt;
 targetpt= target;

 if(von_idx>=von->anz) return(0);
 if(nach_idx>=nach->maxanz)
   if(Fileliste_enlarge(nach,nach_idx+1,0)<=0) return(-1);

 /* if nach already hashlinked then unlink */
 if(nach_idx<nach->anz)
   Fileliste_hashlink(nach,nach_idx,1);

 nach->typen[nach_idx]= von->typen[von_idx];
 nach->flags[nach_idx]= von->flags[von_idx];
 nach->groessen[nach_idx]= von->groessen[von_idx];
 if(nach->adressen[nach_idx]!=NULL) 
   free(nach->adressen[nach_idx]);
 if(flag&2) {
   nach->adressen[nach_idx]= von->adressen[von_idx];
   von->adressen[von_idx]= NULL;
   if(nach->targets[nach_idx]!=NULL) 
     free(nach->targets[nach_idx]);
   nach->targets[nach_idx]= von->targets[von_idx];
   von->targets[von_idx]= NULL;
   if(nach->crcs[nach_idx]!=NULL) 
     free((char *) nach->crcs[nach_idx]);
   nach->crcs[nach_idx]= von->crcs[von_idx];
   von->crcs[von_idx]= NULL;
   if(nach->xprops[nach_idx]!=NULL) 
     free((char *) nach->xprops[nach_idx]);
   nach->xprops[nach_idx]= von->xprops[von_idx];
   von->xprops[von_idx]= NULL;
   von->flags[von_idx]&= ~16;
   Fileliste_destroy(&(nach->subdirs[nach_idx]),2);
   nach->subdirs[nach_idx]= von->subdirs[von_idx];
   von->subdirs[von_idx]= NULL;
   if(nach->subdirs[nach_idx]!=NULL) {
     nach->subdirs[nach_idx]->boss= nach;
     nach->subdirs[nach_idx]->boss_idx= nach_idx;
   }
 } else {
   if(von->adressen[von_idx]==NULL){
     nach->adressen[nach_idx]= NULL;
   }else{
     nach->adressen[nach_idx]= TSOB_FELD(char,strlen(von->adressen[von_idx])+1);
     if(nach->adressen[nach_idx]==NULL) return(0);
     strcpy(nach->adressen[nach_idx],von->adressen[von_idx]);
   }
   if(flag&1){
     Fileliste_target(von,von_idx,target,0);
   }else{
     targetpt= von->targets[von_idx];
   }
   Fileliste_set_target(nach,nach_idx,targetpt,1);
   if(von->crcs[von_idx]!=NULL) {
     ret= Fileliste_set_crc(nach,nach_idx,
                            von->crcs[von_idx]->crc,von->crcs[von_idx]->md5,
                            (!!(von->flags[von_idx]&16))<<12);
     if(ret<=0)
       return(0);
   }
   Fileliste_set_subdir(nach,nach_idx,von->subdirs[von_idx],0);
 }
 if(nach->anz<=nach_idx) nach->anz= nach_idx+1;
 Fileliste_hashlink(nach,nach_idx,0);
 return(1);
}


int Fileliste_truncate(liste,anz,flag)
struct FilelistE *liste;
int anz;
int flag;
{
 int i;

 for(i=anz;i<liste->anz;i++){
   Fileliste_destroy_item(liste,i,0);
 }
 liste->anz= anz;
 return(1);
} 


int Fileliste_garbage(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i,w;
 w= 0;
 for(i=0;i<liste->anz;i++){
   if(liste->typen[i]<0 || liste->adressen[i]==NULL)
 continue;
   if(w!=i)
     Fileliste_copy_item(liste,i,liste,w,0);
   w++;
 }
 Fileliste_truncate(liste,w,0);
 return(1);
}


int Fileliste_append(liste,filename,flag)
struct FilelistE *liste;
char *filename;
int flag;
/*
 bit0= no Fileliste_read_properties_idx()
 bit1= later map to root
 bit2= record only if the file exists
 bit3= do not separate target
*/
{
 int i,ret;
 char quelle[4096],ziel[4096],*cpt;

 if(liste->target_inp_trenner && !(flag&8))
   cpt= strchr(filename,liste->target_inp_trenner);
 else
   cpt= NULL;
 if(cpt==NULL){
   strcpy(quelle,filename);
   ziel[0]= 0;
 } else {
   strcpy(quelle,cpt+1);
   strncpy(ziel,filename,cpt-filename);
   ziel[cpt-filename]= 0;
 }

 if(flag&4){
   struct stat stbuf;
   if(Cd_backup_planer_call_lstaT(quelle,&stbuf)==-1) 
     return(2);
 }
 i= liste->anz;
 if(i>=liste->maxanz)
   if(Fileliste_enlarge(liste,i+1,0)<=0) return(-1);
 if(liste->adressen[i]!=NULL)
   free(liste->adressen[i]);
 liste->adressen[i]= TSOB_FELD(char,strlen(quelle)+1);
 if(liste->adressen[i]==NULL) return(-1);
 strcpy(liste->adressen[i],quelle);
 Fileliste_hashlink(liste,i,0);
 if(liste->targets[i]!=NULL)
   free(liste->targets[i]);
 if(ziel[0]){
   liste->targets[i]= TSOB_FELD(char,strlen(ziel)+1);
   if(liste->targets[i]==NULL) return(-1);
   strcpy(liste->targets[i],ziel);
 }else{
   liste->targets[i]= NULL;
 }
 liste->flags[i]= 0;
 if(flag&2) liste->flags[i]|= 1;
 liste->anz++;
 if(flag&1){
   liste->groessen[i]= -1;
   liste->typen[i]= 0;
 }else{
   ret= Fileliste_read_properties_idx(liste,i,0);
   if(ret<0) return(ret);
 }
 return(1);
}


int Fileliste_relativ_adr(oberdir,adr,rel_adr,flag)
char *oberdir;
char *adr;
char *rel_adr;
int flag;
{
 int l,slash_ende=0;

 rel_adr[0]= 0;
 l= strlen(oberdir);
 if(l>0) slash_ende= (oberdir[l-1]=='/');
 if(slash_ende) l--;
 
 if(strncmp(oberdir,adr,l)!=0) return(0);
 if(strlen(adr)==l) 
   return(2);
 if(adr[l]=='/'){
   strcpy(rel_adr,adr+l+1);
   return(1);
 }
 return(0);
}


int Fileliste_search_file(liste,filename,flag)
struct FilelistE *liste;
char *filename;
int flag;
/*
 bit0= test wether filename is a parent directory of a listed file 
 bit1= test wether filename is a sub directory of a listed file
 bit2= do not accept exact equality
 bit3= only accept items with  liste->typen[i]>=0 
*/
{
 int i,l,slash_ende=0;
 char rel_adr[4096];

 if(liste==NULL) return(-1);
 if(flag&2)
   goto unter_suche;
 if(liste->use_hashing && (flag&7)==0)
   goto hashed_search;

 for(i=0;i<liste->anz;i++){
     if(liste->adressen[i]==NULL)
 continue;
   if(flag&8)
     if(liste->typen[i]<0)
 continue;
   if(Fileliste_relativ_adr(filename,liste->adressen[i],rel_adr,0)>0) {
     if(rel_adr[0]==0 && !(flag&4))
 break;
     if(rel_adr[0]!=0 && (flag&1))
 break;
   }
 }
 if(i<liste->anz) return(i);
 return(-1);

unter_suche:;
 for(i=0;i<liste->anz;i++){
     if(liste->adressen[i]==NULL)
 continue;
   if(flag&8)
     if(liste->typen[i]<0)
 continue;
   if(Fileliste_relativ_adr(liste->adressen[i],filename,rel_adr,0)>0) {
     if(rel_adr[0] || !(flag&4))
 break;
   }
 }
 if(i<liste->anz) return(i);
 return(-1);

hashed_search:;
 i= Fileliste__compute_hashvalue(filename,0);
 for(i= liste->hashstart[i]; i>=0; i= liste->hashlinks[i]) {
   if(flag&8)
     if(liste->typen[i]<0)
 continue;
  if(liste->adressen[i]==NULL)
 continue;
   if(strcmp(filename,liste->adressen[i])==0)
     return(i);
 }
 return(-1);
}


int Fileliste_set_dir_size_cmd(liste,cmd,flag)
struct FilelistE *liste;
char *cmd;
int flag;
{
 int ret;

 if(liste==NULL)
   return(0);
 ret= Sregex_string(&(liste->dir_size_cmd),cmd,0);
 if(strncmp(cmd,"-internal",9)==0)
   liste->dir_size_internal= 1;
 else
   liste->dir_size_internal= 0;
 return(ret);
}


int Fileliste_default_dir_size_cmd(cmd,flag)
char *cmd;
int flag;
{
 char *cpt;
 double wert= 0.0;
 int separator_minus= 0;

 if(strlen(cmd)>sizeof(Fileliste_dir_size_cmd)-1) return(0);
 strcpy(Fileliste_dir_size_cmd,cmd);
 if(strncmp(cmd,"-internal",9)==0)
   Fileliste_dir_size_internal= 1;
 else
   Fileliste_dir_size_internal= 0;
 if(strncmp(cmd,"-internal",9)==0 && strlen(cmd)>9) {
   cpt= strchr(cmd,'+');
   if(cpt==NULL) {
     cpt= strchr(cmd+1,'-'); /* +1 = skip first '-' */
     separator_minus= 1;
   }
   if(cpt!=NULL) {
     if(*(cpt+1)!=0) {
       sscanf(cpt+1,"%lf",&wert);
       if(wert>=0 && wert<=1000000000)
         Fileliste_node_overhead= wert;
     }
     if(separator_minus)
       Fileliste_node_addrl= 0;
     else
       Fileliste_node_addrl= 1;
   }
 }
 return(1);
}


int Fileliste_set_impl_igno(liste,text,flag)
struct FilelistE *liste;
char *text;
int flag;
{
 if(liste==NULL) return(0);
 if(liste->impl_igno!=NULL) free((char *) liste->impl_igno);
 liste->impl_igno= TSOB_FELD(char,strlen(text)+1);
 strcpy(liste->impl_igno,text);
 return(1);
}


int Fileliste_default_impl_igno(text,flag)
char *text;
int flag;
{
 if(strlen(text)>sizeof(Fileliste_impl_igno)-1) return(0);
 strcpy(Fileliste_impl_igno,text);
 return(1);
}


int Fileliste_set_target_inp_trenner(liste,trenner,flag)
struct FilelistE *liste;
char *trenner;
int flag;
{
 if(liste==NULL) return(0);
 liste->target_inp_trenner= trenner[0];
 return(1);
}


int Fileliste_default_target_inp_trenner(trenner,flag)
char *trenner;
int flag;
{
 Fileliste_target_inp_trenner= trenner[0];
 return(1);
}


int Fileliste_set_sperrliste(liste,sperrliste,flag)
struct FilelistE *liste,*sperrliste;
int flag;
{
 if(liste==NULL) return(0);
 liste->sperrliste= sperrliste;
 return(1);
}


int Fileliste_set_subdir(liste,idx,sub,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *sub;
int flag;
/*
 bit0= do not set boss pointer of sub
*/
{
 if(liste->subdirs==NULL)
   return(0);
 if(liste->subdirs[idx]!=NULL)
   Fileliste_destroy(&(liste->subdirs[idx]),2);
 if(sub!=NULL) {
   sub->ref_count++;
   if(!(flag&1)) {
     sub->boss= liste;
     sub->boss_idx= idx;
   }
 }
 liste->subdirs[idx]= sub;
 return(1);
}


int Fileliste_hauptdir(hauptdir,liste,flag)
char *hauptdir;
struct FilelistE *liste;
int flag;
{
 int i,j,slmerk;

 if(liste==NULL) return(0);
 hauptdir[0]= 0;
 if(liste->anz<=0) return(1);
 strcpy(hauptdir,liste->adressen[0]);
 for(i=1;i<liste->anz;i++){
   slmerk= 0;
   for(j=0;hauptdir[j]!=0;j++){
     if(hauptdir[j]!=liste->adressen[i][j]) {
       if(hauptdir[j]=='/' && liste->adressen[i][j]==0) slmerk= j;
       hauptdir[slmerk]= 0;
   break;
     }
     if(hauptdir[j]=='/') slmerk= j;
   }
 if(hauptdir[0]==0) break;
 }
 return(1);
}


int Fileliste_mapped_to_root(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
{
 if(liste==NULL) return(0);
 if(idx>=liste->anz) return(0);
 return(!!(liste->flags[idx]&1));
}


int Fileliste_is_impl_igno(liste,adr,flag)
struct FilelistE *liste;
char *adr;
int flag;
{
 int k;

 for(k=0;k<strlen(liste->impl_igno);k++)
   if(strchr(adr,liste->impl_igno[k])!=NULL) break;
 if(k<strlen(liste->impl_igno)) 
   return(1);
 return(0);
}


int Fileliste_set_target_dir(liste,dir,flag)
struct FilelistE *liste;
char *dir;
int flag;
{
 if(Sregex_string(&(liste->target_dir),dir,0)<=0)
   return(-1);
 return(1);
}


int Fileliste_set_main_dir(liste,dir,flag)
struct FilelistE *liste;
char *dir;
int flag;
{
 if(Sregex_string(&(liste->main_dir),dir,0)<=0)
   return(-1);
 return(1);
}


int Fileliste_target(liste,idx,target,flag)
struct FilelistE *liste;
int idx;
char *target;
int flag;
/*
 bit0= do not mark directories by a trailing  (eventually remove it)
*/
{
 int ret,l;
 char *cpt,*tpt,*hauptdir;

 hauptdir= liste->main_dir;
 target[0]= 0;
 if(liste->targets[idx]!=NULL)
  if(liste->targets[idx][0]){
   target[0]= 0;
   if(liste->targets[idx][0]!='/')
     strcat(target,"/");
   strcat(target,liste->targets[idx]);
   ret= 2;
   goto fertig;
 }
 if(liste->target_dir[0]!=0) {
   target[0]= 0;
   if(liste->target_dir[0]!='/')
     strcat(target,"/");
   strcat(target,liste->target_dir);
   if(target[strlen(target)-1]!='/') 
     strcat(target,"/");
   Fileliste_leafname(liste->adressen[idx],target+strlen(target),0);
   ret= 3;
   goto fertig;
 }

 cpt= liste->adressen[idx];
 while(1){
   if(cpt[0]=='.' && cpt[1]=='/') cpt+= 2;
   else if(cpt[0]=='.' && cpt[1]=='.' && cpt[2]=='/') cpt+=3;
   else 
 break;
 }

 tpt= cpt;
 if(Fileliste_mapped_to_root(liste,idx,0)) 
   tpt+= strlen(tpt);
 if(strlen(hauptdir))
   if(strncmp(cpt,hauptdir,strlen(hauptdir))==0)
     if(tpt[strlen(hauptdir)]=='/' || tpt[strlen(hauptdir)]==0)
       tpt+= strlen(hauptdir);
 if(liste->typen[idx]==2){
   if(tpt[0]=='.'&&tpt[1]==0) tpt++;
   else if(tpt[0]!='/' && tpt[0]) strcat(target,"/");
 }else{
   if(tpt[0]!='/' && tpt[0]) strcat(target,"/");
   if(tpt[0]==0 && tpt!=cpt){ /* single file and -main_dir -auto */
     /* look for the start of the file name */
     do{
       if(*(tpt-1)=='/') 
     break;
     }while((--tpt)>cpt);
   }
 }
 strcat(target,tpt);
 if(!(flag&1) && liste->typen[idx]==2)
   strcat(target,"/");

 ret= 1;

fertig:;
 l= strlen(target);
 if(flag&1){
   if(l>0)
     if(target[l-1]=='/') 
       target[l-1]= 0;
 } else {
   if(liste->typen[idx]==2 && l>0)
     if(target[l-1]!='/')
       strcat(target,"/");
 }

 /* replace any '=' in the target address */
 for(cpt=target;*cpt!=0;cpt++) 
   if(*cpt==Fileliste_target_inp_trenner) *cpt= '_';
 return(ret);
}


int  Fileliste_leafname(adr,leafname,flag)
char *adr,*leafname;
int flag;
{
 char *cpt;
 cpt=strrchr(adr,'/');
 if(cpt==NULL) cpt= adr;
 else          cpt++;
 strcpy(leafname,cpt);
 return(1);
}


int Fileliste_set_target(liste,idx,target,flag)
struct FilelistE *liste;
int idx;
char *target;
int flag;
/*
 bit0= do not check index
*/
{
 if(!(flag&1))
   if(idx<0||idx>=liste->anz)
     return(0);
 if(liste->targets[idx]!=NULL)
   free(liste->targets[idx]);
 if(target==NULL){
   liste->targets[idx]= NULL;
 }else{
   liste->targets[idx]= TSOB_FELD(char,strlen(target)+1);
   if(liste->targets[idx]==NULL) return(-1);
   strcpy(liste->targets[idx],target);
 }
 return(1);
}


int Fileliste_inherit_target(liste,idx,sub,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *sub;
int flag;
/*
 bit0= only set target at directory items
*/
{
 int ret,i;
 char target[4096],bosstarget[4096];

 if(liste->targets[idx]!=NULL) if(liste->targets[idx][0]) {
   Fileliste_target(liste,idx,bosstarget,0);
   for(i=0;i<sub->anz;i++){
     if((flag&1) && sub->typen[i]!=2)
   continue;
     strcpy(target,bosstarget);
     Fileliste_leafname(sub->adressen[i],target+strlen(target),0);
     if(sub->typen[i]==2)
       strcat(target,"/");
     ret= Fileliste_set_target(sub,i,target,0);
     if(ret<=0)
       return(ret);
   }
 } else {

   /* >>> eventually release existing targets of sub */;

 }
 return(1);
}


int Fileliste_is_collision_sub(liste,ober_idx,sperrliste,
                               coll_idx,coll_rel_adr,flag)
struct FilelistE *liste;
int ober_idx;
struct FilelistE *sperrliste;
int coll_idx;
char *coll_rel_adr;
int flag;
/*
 bit8= directory opened virtually, call Fileliste_is_impl_igno() in any case 
*/
{
 char rel_adr[4096],target[4096],target_i[4096],line[4096];
 struct FilelistE *sub1=NULL,*sub2=NULL,*sub=NULL;
 int i,k,ret=0,typ;

 i= coll_idx;
 Fileliste_target(liste,ober_idx,target,1);
 Fileliste_target(liste,i,target_i,1);
 if(Fileliste_relativ_adr(target_i,target,coll_rel_adr,1)<=0)
   return(0);
 if(coll_rel_adr[0]==0) {
   if(flag&256)
     if(Fileliste_is_impl_igno(liste,coll_rel_adr,0)>0)
       {ret= 0; goto ex;}
   /* both target addresses are equal */;
   if(liste->typen[i]!=2 || liste->typen[ober_idx]!=2) {
     /* one of the items is no directory : collision detected */
     ret= 2;
     goto ex;
   }
   strcpy(line,liste->adressen[i]);
 } else {

   /* check wether coll_rel_adr would be rejected when found within a 
      directory */
   if(Fileliste_is_impl_igno(liste,coll_rel_adr,0)>0)
     {ret= 0; goto ex;}

   /* check wether liste->adressen[i]/coll_rel_adr  exists */;
   strcpy(line,liste->adressen[i]);
   if(line[strlen(line)-1]!='/') strcat(line,"/");
   strcat(line,coll_rel_adr);
   if(Fileliste_search_file(sperrliste,line,2)>=0)
     {ret= 0; goto ex;}
   typ= Fileliste_filetyp(line,0);
   if(typ<0)
     goto kaputt;
   if(typ<=0)
     {ret= 0; goto ex;}
   if(typ!=2 || liste->typen[ober_idx]!=2) {
     /* is not a directory : collision detected */
     ret= 3; 
     goto ex;
   }
 }
 if(Fileliste_new(&sub1,0,liste->main_dir,0)<=0) 
   goto kaputt;
 if(Fileliste_dir(sub1,line,2)<=0) 
   goto kaputt;
 if(Fileliste_new(&sub2,0,liste->main_dir,0)<=0) 
   goto kaputt;
 if(Fileliste_dir(sub2,liste->adressen[ober_idx],2)<=0) 
   goto kaputt;
 if(Fileliste_new(&sub,sub1->anz+sub2->anz+1,liste->main_dir,0)<=0)
   goto kaputt;
 for(i=0;i<sub1->anz;i++)
   Fileliste_copy_item(sub1,i,sub,i,0);
 for(i=0;i<sub2->anz;i++)
   Fileliste_copy_item(sub2,i,sub,i+sub1->anz,0);
 if(sub1!=NULL) Fileliste_destroy(&sub1,0); sub1= NULL;
 if(sub2!=NULL) Fileliste_destroy(&sub2,0); sub2= NULL;
 ret= Fileliste_collision_test(sub,sperrliste,1|256);
 if(ret==1) 
   goto ex;
 if(ret<0)
   goto kaputt;

 ret= 0; 
ex:;
 if(sub1!=NULL) Fileliste_destroy(&sub1,0);
 if(sub2!=NULL) Fileliste_destroy(&sub2,0);
 if(sub!=NULL) Fileliste_destroy(&sub,0);
 return(ret);
kaputt:
 ret= -1;
 goto ex;
}


int Fileliste_is_collision(liste,target_idx,sperrliste,
                           coll_idx,coll_rel_adr,flag)
struct FilelistE *liste;
int target_idx;
struct FilelistE *sperrliste;
int *coll_idx;
char *coll_rel_adr;
int flag;
/*
 bit0= check for inclusion in both directions
       (else : wether target is included in a list element )
 bit2= check even if excluded by sperrliste 
 bit8= directory opened virtually
*/
{
 char rel_adr[4096],target[4096],target_i[4096],line[4096];
 struct FilelistE *sub1=NULL,*sub2=NULL,*sub=NULL;
 int i,ret=0,typ;

 *coll_idx= -1;
 coll_rel_adr[0]= 0;
 if(Fileliste_search_file(sperrliste,liste->adressen[target_idx],
                          2|(flag&4))>=0)
   return(0);

 Fileliste_target(liste,target_idx,target,1);
 if(target[0]==0 || (target[0]=='/' && target[1]==0)){
   if(liste->typen[target_idx]!=2) 
     /* Pseudo collision: only directories may be mapped to / */
     return(4);
 }

 for(i=0;i<liste->anz;i++){
   *coll_idx= i;
   if(liste->typen[i]<0)
 continue;
   if(i==target_idx) 
 continue;
   if(Fileliste_search_file(sperrliste,liste->adressen[i],2)>=0)
 continue;
   ret= Fileliste_is_collision_sub(liste,target_idx,sperrliste,
                                   *coll_idx,coll_rel_adr,flag&256);
   if(ret!=0) return(ret);
   if(flag&1){
     ret= Fileliste_is_collision_sub(liste,*coll_idx,sperrliste,
                                     target_idx,coll_rel_adr,flag&256);
     if(ret!=0) return(ret);
   }
 }
 return(0);
}


int Fileliste_collision_avoid(liste,idx,sperrliste,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *sperrliste;
int flag;
{
 int l,zaehler,coll_idx;
 char coll_rel_adr[4096],target[4096];
 
 Fileliste_target(liste,idx,target,1);
 l= strlen(target);
 zaehler= 0;
 do{
   sprintf(target+l,"_%-3.3d",zaehler++);
   Fileliste_set_target(liste,idx,target,0);
 }while(Fileliste_is_collision(liste,idx,sperrliste,
                               &coll_idx,coll_rel_adr,1)>0);
 
 return(1);
}


int Fileliste_redirect_links(liste,flag)
struct FilelistE *liste;
int flag;
/*
 bit0= do not try to redirect but flatly remove the link from the list
*/
{
 int i,ret,typ,sret,stage,stage_limit= 10;
 char filename[4096],buff[4096],*cpt,target[4096];
 static char typcodes[]={"?fdl"};

 for(i=0;i<liste->anz;i++){
   typ= Fileliste_filetyp(liste->adressen[i],0);
   if(typ!=3)
 continue;
   if(flag&1) {
     fprintf(stderr,"--- Symbolic link detected in file list and excluded\n");
     fprintf(stderr,"       :   %s\n",liste->adressen[i]);
     Fileliste_defunct_item(liste,i,0);
 continue;
   } else if(typ==3) {
     fprintf(stderr,"--- Symbolic link detected\n");
     fprintf(stderr,"       :   %s\n",liste->adressen[i]);

     /* try to redirect source redirect to link target */
     strcpy(filename,liste->adressen[i]);
     for(stage= 0; stage<stage_limit && typ==3;stage++) {
       sret= Sfile_readlink(filename,buff,sizeof(buff),0);
       if(sret<=0) {
         fprintf(stderr,"       :   cannot read link target from link\n");
     break;
       }
       if(buff[0]=='/')
         filename[0]= 0;
       else {
         cpt= strrchr(filename,'/');
         if(cpt!=NULL)
           *(cpt+1)= 0;
         else
           filename[0]= 0;
       }
       strcat(filename,buff);
       typ= Sfile_type(filename,1);
       if(typ<=0) {
         fprintf(stderr,"   -> ?:   %s\n",filename);
         if(typ<0)
           fprintf(stderr,"       :   link target not found\n");
         else
           fprintf(stderr,
                      "       :   link target not a recognizable file type\n");
         typ= 3;
     break;
       }
       if(typ<0||typ>3)
         typ= 0;
       fprintf(stderr,"   -> %c:   %s\n",typcodes[typ],filename);
     }
     if(stage>=stage_limit) {
       fprintf(stderr,"       :   too many link levels (max. %ld)\n",
                      stage_limit);
     }
     if(typ==3 || typ<=0) {
       /*  no valid link target found */
       Fileliste_defunct_item(liste,i,0);
       fprintf(stderr,
                 " SORRY :   could not redirect source to link target\n");
 continue;
     } else {
       /* only change source address but not target.  */
       Fileliste_target(liste,i,target,0);
       if(Fileliste_set_address(liste,i,filename,0)<=0)
         return(-1);
       if(Fileliste_set_target(liste,i,target,0)<=0)
         return(-1);
       liste->typen[i]= typ;
       fprintf(stderr,
                 " OK    :   source successfully redirected to link target\n");
     }
   }
 }
 return(1);
}


int Fileliste_collision_test(liste,sperrliste,flag)
struct FilelistE *liste;
struct FilelistE *sperrliste;
int flag;
/*
 bit0= exit after first collision, do not report collision to stderr
 bit1= change targets to avoid collisions
 bit2= check even if excluded by sperrliste 
 bit3= like bit2 but only with the first item
 bit8= directory opened virtually
*/
{
 int i,ret,coll_idx,swarwas=0,typ,t_flag;
 char coll_rel_adr[4096],target[4096];
 static char typcodes[]={"?fdl"};

 for(i=0;i<liste->anz;i++){
   if((flag&4)||((flag&8)&&i==0)) t_flag= 4;
   else                           t_flag= 0;
   t_flag|= (flag&256);
   ret= Fileliste_is_collision(liste,i,sperrliste,
                               &coll_idx,coll_rel_adr,t_flag);
   if(ret>0) {
     swarwas++;
     if(flag&1) return(1);
     if(coll_rel_adr[0] || i>coll_idx){
       fprintf(stderr,"--- Target collision detected:\n"); 
       typ= Fileliste_filetyp(liste->adressen[i],0);
       if(typ<0||typ>3) typ= 0;
       fprintf(stderr,"    1 %c:   %s\n",typcodes[typ],liste->adressen[i]); 
       if(coll_idx>=0){
         typ= Fileliste_filetyp(liste->adressen[coll_idx],0);
         if(typ<0||typ>2) typ= 0;
         fprintf(stderr,"    2 %c:   %s",typcodes[typ],
                                         liste->adressen[coll_idx]);
         if(coll_rel_adr[0]) 
           fprintf(stderr," / %s\n",coll_rel_adr);
         else
           fprintf(stderr,"\n");
       } else {
         fprintf(stderr,
                 "       :   not a usable target for data files\n"); 
       }
       /* report target address */
       Fileliste_target(liste,i,target,1);
       fprintf(stderr,"    CD :   %s\n",target);

       if(flag&2){
         if(Fileliste_collision_avoid(liste,i,sperrliste,0)>0){
           Fileliste_target(liste,i,target,0);
           fprintf(stderr,"    1! :   %s=%s\n",target,liste->adressen[i]);
         } else {
           fprintf(stderr,"    failed to find new target address\n");
         }
       }
     }
   }
   if(ret<0) return(-1); 
 } 
 return(swarwas>0);
}


int Fileliste_determine_treesize(liste,idx,dirlen,size,flag)
/* Note : does not care for exclusion lists. Fileliste_dir_size() does that */
struct FilelistE *liste;
int idx;
int dirlen;    /* length of (target) directory name */
double *size;
int flag;
/*
 bit0= recursion
*/
{
 int i,typ,ret,adrlen= 0;
 double sub_size= 0.0;
 struct FilelistE *sub= NULL;
 char *name,*cpt,target[4096];

 *size= 0.0;
 typ= Type_and_size(liste->adressen[idx],size,1);
 if(typ<=0)
   {ret=0; goto ex;}
 if(Fileliste_node_addrl) {
   if(!(flag&1)) { /* top level : set adrlen */
     if(Fileliste_target(liste,idx,target,0)<=0) {
       adrlen= strlen(liste->adressen[idx]);
     } else {
       adrlen= strlen(target);
     }
   } else {
     name= liste->adressen[idx];
     for(cpt= liste->adressen[idx]; *cpt!=0; cpt++)
       if(*cpt=='/')
         if(*(cpt+1)!=0)
           name= cpt+1;
     adrlen= dirlen+strlen(name);
   }
 } else 
   adrlen= 0;
 *size+= adrlen+Fileliste_node_overhead;

file_object_done:;
 if(typ!=2)
   {ret= 1; goto ex;}
 /* create file list and read directory content */
 if(Fileliste_new(&sub,0,liste->main_dir,0)<=0)
   {ret= -1; goto ex;}
 Fileliste_dir(sub,liste->adressen[idx],2);
 for(i=0;i<sub->anz;i++) {
   ret= Fileliste_determine_treesize(sub,i,adrlen,&sub_size,1);
   if(ret<0)
     goto ex;
   if(ret>0)
     *size+= sub_size;
 }
 ret= 1;
ex:;
 if(sub!=NULL)
   Fileliste_destroy(&sub,0);
 return(ret);
}


int Fileliste_leanfind(liste,idx,parent_dev,fp,linecount,flag)
/* replacement for shell command  "find"  */
struct FilelistE *liste;
int idx;
dev_t parent_dev;
FILE *fp;
int *linecount;
int flag;
/*
 bit0= recursion
 bit1= use relative addresses
        /x/x/x/   absolute directory address
        /x/x/x/   absolute file address
        ./x/x/x/  relative directory address
        ./x/x/x   relative file name
        ../../..  one or more directories up
        ../       to start directory
        #=# xxx   additional program information
       top_level addresses get prepended the target
 bit2= do not follow mount points
 bit3= do not follow net mount points
 bit4= print linecount
 bit5= generate dirlist (with property information)
 bit6= with bit 5 : generate dir_script (chown, chmod)

 <<< bit7= ignore names listed in liste->sperrliste

 bit9= prepend type
 bit10= prepend length
 bit11= preprend crc
 bit12= use MD5 rather than CRC32
*/
{
 char *name,*npt,slash[2],dotslash[3],typetext[4],lengthtext[80],crctext[80];
 char adr[4096];
 unsigned char md5[16];
 int i,swarwas= 0,ret,is_dir= 0,typ,read_ret= 0,report_step= 1000;
 double size;
 unsigned long crc;
 struct FilelistE *sub;
 dev_t st_dev;

 is_dir= (liste->typen[idx]==2);
 if((flag&32) && !is_dir)
   return(2);
 if(flag&2048)
   report_step= 100;
 /* test name */
 name= liste->adressen[idx];
 if(name==NULL)
   return(0);
 if(name[0]==0)
   return(0);
 if(is_dir&& (flag&1)) 
   if(Is_forbidden_mountpoint(name,parent_dev,&st_dev,
                              16|((flag>>2)&3)))
     return(2);
/*
 if(flag&128) {
   if(Fileliste_search_file(liste->sperrliste,name,0)>=0)
     return(2);
 }
*/

 /* obtain requested prepend information */
 typetext[0]= lengthtext[0]= crctext[0]= 0;
 read_ret= 2|4|8;
 typ= liste->typen[idx];
 if((flag&1024)||typ==0) {
   typ= Type_and_size(name,&size,1);
   if(typ<=0)
     read_ret&= ~(2|4);
 }
 if((flag&2048)) {
   if(typ==0)
     ret= 0;
   else if(flag&(1<<12)) {
     read_ret|= 16;
     ret= Determine_file_checksum(name,NULL,NULL,&md5,&md5,&size,1<<12);
   } else {
     ret= Determine_file_checksum(name,NULL,NULL,&crc,&crc,&size,0);
   }
   if(ret<=0)
     read_ret&= ~(8|16);
 }
 Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                     read_ret&(2|4|8|16),flag&(512|1024|2048));

 if(flag&2) {
   slash[0]= slash[1]= 0;
   if(is_dir)
     slash[0]= '/';
   if(flag&1) {
     /* ./basename[/] , ignore any trailing slash */;
     for(npt= name+strlen(name)-1;npt>name;npt--)
       if(*npt=='/')
     break;
     if(*npt=='/')
       npt++;
     sprintf(adr,"./%s%s",npt,slash);
   } else {
     dotslash[0]= dotslash[1]= dotslash[2]= 0;
     if(name[0]!='/')
       strcpy(dotslash,"./");
     sprintf(adr,"%s%s%s",dotslash,name,slash);
   }
 } else
   strcpy(adr,name);

 if(flag&64) {
   /* script format has to record target addresses */
   Fileliste_target(liste,idx,adr,1);
   if(adr[0]==0)
     strcpy(adr,".");
 } else {
   if((flag&2)&&!(flag&1)) { /* at top recursion level */
     char target[4096];
     /* prepend  target  to adr */;
     Fileliste_target(liste,idx,target,0);
     if(liste->target_inp_trenner!=0) {
       sprintf(target+strlen(target),"%c%s",Fileliste_target_inp_trenner,adr);
       strcpy(adr,target);
     }
   }
 }
 if(flag&32) {
   ret= Record_dir_properties(fp,name,adr,lengthtext,crctext,!!(flag&64));
   if(ret<=0)
     return(ret);
 } else
   fprintf(fp,"%s%s%s%s\n",typetext,lengthtext,crctext,adr);

file_object_done:;
 (*linecount)++;
 if(flag&16) 
   if(((*linecount)%(report_step/100))==0 || (*linecount)==2)
     Print_pacifier(*linecount,"",2);

 /* if not a directory : done */
 if(!is_dir)
   return(1);

 /* create file list and read directory content */
 if(Fileliste_new(&sub,0,liste->main_dir,0)<=0) 
   return(-1);
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 Fileliste_dir(sub,name,2);
 if(flag&32)
   Fileliste_inherit_target(liste,idx,sub,1);
 
 /* examine found file objects */
 ret= 0;
 for(i=0;i<sub->anz;i++) {
   /* do not examine excluded objects */
   if(Fileliste_search_file(liste->sperrliste,sub->adressen[i],0)>=0)
 continue;
   ret|= Fileliste_leanfind(sub,i,st_dev,fp,linecount,1|flag);
 }

 /* mark end of (sub-)directory */
 if(flag&2) {
   if(flag&1)
     fprintf(fp,"..\n",name);
   else
     fprintf(fp,"../\n",name);
 }

 Fileliste_destroy(&sub,0);
 return(1);
}


int Fileliste_merge_trees(liste,idx,parent_dev,linecount,
                          source_prefix,target_prefix,removed_prefix,old_tree,
                          flag)
/* move one file tree into another one */
struct FilelistE *liste;
int idx;
dev_t parent_dev;
int *linecount;
char *source_prefix;
char *target_prefix;
char *removed_prefix;
struct FilelistE *old_tree;
int flag;
/*
 bit0= recursion
 bit1= precheck mode : do not create or move objects
 bit2= do not follow mount points
 bit3= do not follow net mount points
 bit4= print linecount
 bit5= do not abort at first error
 bit6= copy file objects rather than moving them
 bit8-11= merge mode
       0= never overwrite existing target objects
       1= always overwrite existing target objects
       2= overwrite existing target objects if older than source object
 bit12= initialize rootlist with source_prefix (if not recursion)
 bit13= make target objects writable by owner
*/
{
 char *name,*srcpt;
 char target[4096],link_target[4096],new_link_target[4096],removed[4096];
 char line[2*4096];
 int i,ret,is_dir= 0,is_link= 0,typ,report_step= 1000, merge_mode;
 int target_type,link_limit= 100,link_count= 0,directory_link= 0;
 int new_dir_created= 0,errn,rec_ret,is_dir_link,prop_flag= 0,old_mode=0700;
 double size;
 struct FilelistE *sub= NULL,*rootlist= NULL;
 struct stat stbuf;
 dev_t st_dev;

 uid_t uid;
 gid_t gid;
 time_t date;
 struct FilelistE *found_list= NULL;
 int found_idx;


 merge_mode= (flag>>8)&15;

 if((flag&(1<<12)) && ! (flag&1)) {
   typ= Type_and_size(source_prefix,&size,1);
   if(typ<0) {
no_source_prefix:;
     fprintf(stderr,
             "Fileliste_merge_trees : source_prefix does not exist\n");
     fprintf(stderr,"source_prefix=%s\n",source_prefix);
     ret= 0; goto ex;
   }
   ret= Fileliste_single(&rootlist,source_prefix,typ,size,0);
   if(ret<=0)
     goto ex;
   liste= rootlist;
   idx= 0;
   *linecount= 0;
   if(Cd_backup_planer_call_lstaT(source_prefix,&stbuf)==-1)
     goto no_source_prefix;
   parent_dev= stbuf.st_dev;
 }

 is_dir= (liste->typen[idx]==2);
 is_link= (liste->typen[idx]==3);
 if(is_link)
   is_dir_link= (Type_and_size(liste->adressen[idx],&size,4)==2);
 /* test name */
 name= liste->adressen[idx];
 if(name==NULL)
   {ret= 0; goto ex;}
 if(name[0]==0)
   {ret= 0; goto ex;}
 if(is_dir&& (flag&1)) 
   if(Is_forbidden_mountpoint(name,parent_dev,&st_dev,
                              16|((flag>>2)&3)))
     {ret= 2; goto ex;}

 /* compose target address */;
 srcpt= strstr(name,source_prefix);
 if(srcpt!=name) {
   fprintf(stderr,
           "Fileliste_merge_trees : list item does not match source_prefix\n");
   fprintf(stderr,"item[%7.7ld]=%s\n",idx,name);
   fprintf(stderr,"source_prefix=%s\n",source_prefix);
   {ret= 0; goto ex;}
 }
 strcpy(target,target_prefix);
 strcat(target,name+strlen(source_prefix));
 target_type= Type_and_size(target,&size,1);
 removed[0]= 0;
 if(!(strcmp(removed_prefix,"-")==0 || removed_prefix[0]==0)) {
   strcpy(removed,removed_prefix);
   strcat(removed,name+strlen(source_prefix));
 }

 if(is_dir) {

   /* does target address exist ? */
   if(target_type!=-1) { 
     /* is it a directory or a link to a directory ? */;
     if(target_type==3) {
       strcpy(new_link_target,target);
       while(target_type==3 && link_count++<link_limit) {
         strcpy(link_target,new_link_target);
         ret= Sfile_readlink(link_target,
                             new_link_target,sizeof(new_link_target),1);
         if(ret<=0)
       break;
         typ= Type_and_size(new_link_target,&size,1);
         if(typ==2) {
           directory_link= 1;
       break;
         } else if(typ!=3)
       break;
       }
     }
     if(target_type==2 || directory_link) {

       /* ok */;

     } else {
       if(merge_mode==1) {
         /*  replace by directory */;
         if(!(flag&2)) {
           ret= Remove_file(target,removed,0);
           if(ret<=0)
             goto ex;
         }
         goto create_new_directory;
       } else {
         fprintf(stderr,"%s : %s",
                        "Fileliste_merge_trees",
                        "directory target exists but is not a directory\n");
         fprintf(stderr,"source=%s\n",name);
         fprintf(stderr,"target=%s\n",target);
         {ret= 0; goto ex;}
       }
     }
   } else {
create_new_directory:;
     if(!(flag&2)) {
       ret= mkdir(target,0700);
       if(ret==-1) {
         fprintf(stderr,
                "Fileliste_merge_trees : creation of directory failed :\n");
         fprintf(stderr,"target=%s\n",target);
         fprintf(stderr,"%s\n",strerror(errno));
       }
       new_dir_created= 1;
     }
   }

 } else if(is_link) {

   /* >>> how to handle softlinks ? */;
     /* >>> readjust link content */;
 
 } else { /* any other filetype */

   /* does target address exist ? */
   if(target_type!=-1) { 
     /* decide wether to overwrite or not */
     if(merge_mode==0)
       {ret= 2; goto ex;}
     else if(merge_mode==2) {
       if(Is_older_file(target,name,0)<=0)
         {ret= 2; goto ex;}
     }
   } 

   /* move file object */;
   if(!(flag&2)) {
     if(target_type!=-1) {
       ret= Remove_file(target,removed,0);
       if(ret<=0)
         goto ex;
     }
     /* copy or move source to target */
     if(flag&64) {
       ret= Sfile_copy_file(name,target,&errn,line,0);
       if(ret<=0) {
         fprintf(stderr,
              "Fileliste_merge_trees : copy from source to target failed :\n");
         fprintf(stderr,"source=%s\n",name);
         fprintf(stderr,"target=%s\n",target);
         fprintf(stderr,"%s\n",line);
         goto ex;
       }
       Sfile_copy_properties(name,target,1|4|(flag&(1<<13)));
     } else {
       ret= Sfile_move_file(name,target,&errn,line,0);
       if(ret<=0) {
         fprintf(stderr,
              "Fileliste_merge_trees : move from source to target failed :\n");
         fprintf(stderr,"source=%s\n",name);
         fprintf(stderr,"target=%s\n",target);
         fprintf(stderr,"%s\n",line);
         goto ex;
       }
     }
   }
 }

file_object_done:;
 (*linecount)++;
 if(flag&16) 
   if(((*linecount)%(report_step/100))==0 || (*linecount)==2) {
     sprintf(line,"  %-10ld : %s",*linecount,name);
     Print_pacifier(0,line,4|16|32);
   }

 /* if not a directory : done */
 if(!is_dir)  /* ???  allow directory links at recursion level 0 ? */
   {ret= 1; goto ex;}

next_recursion_level:;
 /* create file list and read directory content */
 if(Fileliste_new(&sub,0,liste->main_dir,0)<=0) 
   {ret= -1; goto ex;}
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 Fileliste_dir(sub,name,2);

 /* cautious settings while installing files */
 if(!(flag&2))
 if(merge_mode==1 || merge_mode==2 || new_dir_created) {
   if(new_dir_created) {
     Cd_backup_planer_call_staT(name,&stbuf);
     prop_flag= 1|2|4; /* set permissions for user+group+others */
   } else {
     Cd_backup_planer_call_staT(target,&stbuf);
     prop_flag= 1;     /* only set permissions for user */
   }
   old_mode= stbuf.st_mode&0xfff;
   Sfile_set_properties(target,0700,(uid_t) 0,(gid_t) 0,(time_t) 0,(time_t) 0,
                        (time_t) 0,prop_flag); 
 }

 /* examine found file objects */
 ret= 0;
 for(i=0;i<sub->anz;i++) {
   /* do not examine excluded objects */
   if(Fileliste_search_file(liste->sperrliste,sub->adressen[i],0)>=0)
 continue;
   rec_ret= Fileliste_merge_trees(sub,i,st_dev,linecount,
                                  source_prefix,target_prefix,removed_prefix,
                                  old_tree,1|flag);
   ret|= rec_ret;
   if(rec_ret<0)
 break;
   if(rec_ret==0 && !(flag&32))
 break;
 }

 /* eventually try to copy directory properties */
 if(!(flag&2)) {
   if(merge_mode==1 || merge_mode==2 || new_dir_created) {
     ret= Sfile_copy_properties(name,target,1|4|(flag&(1<<13)));
     if(ret<=0) {
       fprintf(stderr,"%s : %s",
                      "Fileliste_merge_trees",
                      "cannot copy directory properties\n");
       fprintf(stderr,"source=%s\n",name);
       fprintf(stderr,"target=%s\n",target);
       {ret= 0; goto ex;}
     }
     prop_flag= 0; /* old_mode is invalid now */

#ifndef Not_yeT
     ret= 0;
     /* try to get source permissions as they should be */;
     if(old_tree!=NULL)
       ret= Fileliste_search_branch(old_tree,&found_list,&found_idx,name,0);
     if(ret<0)
       goto ex;
     if(ret>0 && found_list->xprops[found_idx]!= NULL) {
       ret= Get_user_group_id(found_list->xprops[found_idx]->own,
                              found_list->xprops[found_idx]->grp,&uid,&gid,0);
       if(ret>0) {
         date= found_list->xprops[found_idx]->date;
         Sfile_set_properties(target,found_list->xprops[found_idx]->mod,
                              uid,gid,date,date,date,1|2|4|8|16|32|128);
       }
     }
#endif /* !  Not_yeT */

   }

   /* Try to remove the source directory */
   if(!(flag&64))
     Remove_dir(name,4);
 }

 Fileliste_destroy(&sub,0);
 ret= 1; goto ex;

ex:;
 if(prop_flag!=0) /* directory properties have been altered temporarily */
   Sfile_set_properties(target,old_mode,(uid_t) 0,(gid_t) 0,
                               (time_t) 0,(time_t) 0,(time_t) 0,prop_flag); 
 if(rootlist!=NULL)
    Print_pacifier(*linecount,"file objects processed",1|2|4|8|16);
 Fileliste_destroy(&rootlist,0);
 return(ret);
}


int Fileliste_find_path_list(path_list,flag)
char *path_list;
int flag;
{
 char line[4096],*fret;
 struct FilelistE *find_liste= NULL;
 FILE *fp= NULL; 
 int ret,linecount= 0;

 fp= Sfile_fopen(path_list,"r",0);
 if(fp==NULL)
   {ret= 0; goto ex;}

 while(1) {
   fret= Sfile_fgets(line,sizeof(line)-1,fp);
   if(fret==NULL)
 break;
   if(line[0]==0)
 continue;
   if(Fileliste_single(&find_liste,line,0,0.0,8)>0){
     if(Fileliste_read_properties(find_liste,2)>0) 
       Fileliste_leanfind(find_liste,0,dev_dummy,stdout,&linecount,0);
     Fileliste_destroy(&find_liste,0);
   }
 }
 ret= 1;
ex:;
 if(fp!=NULL)
   Sfile_fclose(fp);
 return(ret);
}


int Fileliste_ist_zu_alt(liste,idx,altersgrenze,size,flag)
struct FilelistE *liste;
int idx;
time_t altersgrenze;
double *size;
int flag;
/*
 bit0= also take into respect st_ctime (directory operations)
 bit1= do not record in  liste->groessen
*/ 
{
 struct Cd_backup_planer_struct_staT stbuf;
 time_t ges_datum;

 if(Cd_backup_planer_call_lstaT(liste->adressen[idx],&stbuf)==-1)
   return(-1);
 *size= stbuf.st_size;
 if(!(flag&2))
   liste->groessen[idx]= *size;
 ges_datum= stbuf.st_mtime;
 if(flag&1)
   if(stbuf.st_ctime>ges_datum)
     ges_datum= stbuf.st_ctime;
 if(ges_datum<altersgrenze)
   return(1);
 return(0);
}


int Fileliste_lohnender_split(zu_alt,junge_ausnahmen,number_of_files,
                              min_sbpf,flag)
struct FilelistE *zu_alt,*junge_ausnahmen;
int number_of_files;
double min_sbpf;
int flag;
{
 double saved_bytes;
 int i;

 if(zu_alt->anz+junge_ausnahmen->anz<=0)
   return(2);
 if(number_of_files<=0)
   return(3);
 saved_bytes= 0.0;
 for(i=0;i<zu_alt->anz;i++)
   if(zu_alt->groessen[i]>0)
     saved_bytes+= zu_alt->groessen[i];
 for(i=0;i<junge_ausnahmen->anz;i++)
   if(junge_ausnahmen->groessen[i]>0)
     saved_bytes-= junge_ausnahmen->groessen[i];
 if(saved_bytes/((double) number_of_files)<min_sbpf) {
   /* exclusion would not be rewarding enough */
   Fileliste_truncate(zu_alt,0,0);
   Fileliste_truncate(junge_ausnahmen,0,0);
   return(0);
 }
 return(1);
}


int Fileliste_detect_change(liste,idx,old_tree,old_tree_idx,size,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *old_tree;
int old_tree_idx;
double size;
int flag;
/*
 bit0= do not search but use  old_tree  and  old_tree_idx  as found
 bit8= check for existence in old_tree (else assume no change)
 bit9= with bit8 : check type in old_tree
 bit10= with bit8 : check size in old_tree
 bit11= with bit8 : check crc in old_tree
return:
 <0 = error
  0 = no change detected
  1 = change detected
  2 = not found in old_tree
*/
{
 
 int found_idx,ret;
 unsigned long lcrc= 0,fcrc= 0;
 unsigned char *lmd5= NULL,*fmd5= NULL;
 struct FilelistE *found_list;

 if(!(flag&256))
   return(0);
 if(flag&1) {
   found_list= old_tree;
   found_idx= old_tree_idx;
 } else {
   if(Fileliste_search_branch(old_tree,&found_list,&found_idx,
                              liste->adressen[idx],0)<=0)
     return(2);
 }

 if(liste->crcs[idx]!=NULL) {
   lcrc= liste->crcs[idx]->crc;
   lmd5= liste->crcs[idx]->md5;
 }
 if(found_list->crcs[found_idx]!=NULL) {
   fcrc= found_list->crcs[found_idx]->crc;
   fmd5= found_list->crcs[found_idx]->md5;
 }
 ret= File_is_changed(liste->adressen[idx],liste->flags[idx],liste->typen[idx],
                      liste->groessen[idx],
                      lcrc,lmd5,
                      found_list->flags[found_idx],
                      found_list->typen[found_idx],
                      found_list->groessen[found_idx],
                      fcrc,fmd5,
                      flag&(512|1024|2048));
 return(ret);
}


int Fileliste_precheck_sub(liste,idx,old_tree,
                           namen_krank,kein_zugriff,
                           mount_punkte,parent_dev,
                           altersgrenze,zu_alt,junge_ausnahmen,
                           max_ausnahmen,min_sbpf,
                           exclude_size,too_big,
                           linecount,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *old_tree;
struct FilelistE *namen_krank,*kein_zugriff,*mount_punkte;
dev_t parent_dev;
time_t altersgrenze;
struct FilelistE *zu_alt,*junge_ausnahmen;
int max_ausnahmen;
double min_sbpf;
double exclude_size;
struct FilelistE *too_big;
int *linecount;
int flag;
/*
 bit0= recursion
 bit1= consider files with st_dev != parent_dev to be unreadable
 bit2= don't check for real readability
 bit3= check for age
 bit4= consider remote (e.g. NFS) mountpoints to be unreadable 
 bit5= with bit3 and bit8 : mistrust directory timestamps and crcs
 bit6= ignore timestamps
 bit7= with bit3 : parent directory has old timestamp
 bit8= with bit3 : check for existence in old_tree
 bit9= with bit8 : check type in old_tree
 bit10= with bit8 : check size in old_tree
 bit11= with bit8 : check crc in old_tree
 bit12= put files larger than exclude_size into too_big 
 bit13= with bit12 : eventually use size of link target 

return:
 bit0= ill name
 bit1= not readable
 bit2= mountpoint of unwanted type
 bit3= too old
 bit4= too big
*/
{
 char *name,*npt,link_target[4096],line[4096+100];
 int i,swarwas= 0,ret,fret,jung_impl,l_max_ausnahmen,jung_merk,lret,skip= 0;
 int big_file= 0,crc_check= 0,force_show= 0,hflag,read_link_target= 0;
 double size;
 FILE *fp;
 struct stat stbuf;
 struct FilelistE *sub= NULL,*l_zu_alt= NULL,*l_junge_ausnahmen= NULL;

 /* test name */
 name= liste->adressen[idx];
 if(name==NULL) return(0);

 /* recorded in exclusion list ? */
 if(Fileliste_search_file(liste->sperrliste,name,0)>=0)
   return(0);

 big_file= (liste->typen[idx]==1 && liste->groessen[idx]>5*1024*1024);
 if((flag&(8|256|2048))==(8|256|2048))
   if((flag&32)|!(flag&128))
     crc_check= 1;
 if((*linecount%10)==0 || (*linecount)==2 || crc_check || big_file) {
   if(big_file) 
     if(crc_check) 
       force_show= 1;

   if(crc_check)
     ret= 0;

   sprintf(line,"  %-10ld %s: %s",*linecount,
                (crc_check?"crc check":"checking "),liste->adressen[idx]);
     
   Print_pacifier(0,line,force_show|4|16|32);
 }
 (*linecount)++;

 npt= strrchr(name,'/');
 if(npt==NULL) npt= name;
 else          npt++;
 for(i=0;i<strlen(liste->impl_igno);i++)
   if(strchr(npt,liste->impl_igno[i])!=NULL)
     break;
 if(i<strlen(liste->impl_igno)){
   Fileliste_copy_item(liste,idx,namen_krank,namen_krank->anz,0);
   swarwas|= 1;
 }

 if(Is_forbidden_mountpoint(name,parent_dev,NULL,
                            ((flag&2)>>1)|((flag&16)>>3))) {
   Fileliste_copy_item(liste,idx,mount_punkte,mount_punkte->anz,0);
   swarwas|= 4;
 }

 if(flag&(1<<12) && liste->typen[idx]!=2) {
   read_link_target= 0;
   if(flag&(1<<13))
     read_link_target= 4;
   Type_and_size(name,&size,1|read_link_target);
   if(size>=exclude_size) {
     Fileliste_copy_item(liste,idx,too_big,too_big->anz,0);
     swarwas|= 16; goto ex;
   }
 }

 /* Test readbility */
 /* A00728_0054_dave_erickson */
 if((liste->typen[idx]>=1 && liste->typen[idx]<=3) && !(flag&4)){
   if(liste->typen[idx]==3){
     if(readlink(name,link_target,sizeof(link_target)-1)==-1) 
       goto not_readable;
   } else {
     if((fp= fopen(name,"r"))==NULL){
not_readable:;
       Fileliste_copy_item(liste,idx,kein_zugriff,kein_zugriff->anz,0);
       swarwas|= 2;
     }else
       fclose(fp);
   }
 }

 if(flag&8){ 
   /* if age test enabled and not a directory : perform age test */
   if(liste->typen[idx]!=2 && liste->typen[idx]>=0) {
     if((flag&64) || Fileliste_ist_zu_alt(liste,idx,altersgrenze,&size,1)>0) {
       if((flag&128) && !(flag&32))
         /* parent directory is trustfully old. No moves possible */
         goto really_old;
       hflag= flag&(256|512|1024|2048);
       ret= Fileliste_detect_change(liste,idx,old_tree,-1,size,hflag);
       if(ret>0)
         goto young_enough;

really_old:;
       liste->flags[idx]|= 4;
       /* if large enough or top level : exclude */
       if(liste->groessen[idx]>2.0*min_sbpf||!(flag&1))
         Fileliste_copy_item(liste,idx,zu_alt,zu_alt->anz,0);
       swarwas|= 8;
     } else {
young_enough:;
       /* young enough or not verified in old tree */
       liste->flags[idx]|= 2;
     }
   }
 } else /* (flag&8) */
   liste->flags[idx]|= 2; /* never tested for age. thus young enough */

 /* if designated to be skipped or if not a directory : check is done */
 if(liste->typen[idx]!=2 || (swarwas&6))
   {ret= swarwas; goto ex;}

 if(flag&8) {
   /* check wether in old_tree */;
   hflag= flag&(256|512);
   /* currently directory content changes are not reflected by the content
      lists of higher levels. Therefore the directory checksums in old_tree
      may be in the state of level 0. If not absolutely necessary, one should
      not take a mismatch of size or checksum as a sign for recent changes.
      A directory update report will be necessary in the future.
   */
   if(flag&64)
     hflag|= (1024|2048);
   ret= Fileliste_detect_change(liste,idx,old_tree,-1,size,hflag);
   if(ret==2) {
     /* not in old_tree : entire subtree is young enough */;
     liste->flags[idx]|= 2;
     /* disable agetest for subsequent directories */
     flag&= ~8;
     /* need to continue for other kinds of tests */
   }
   flag&= ~128;
   if(ret==0 && !(flag&32)) {
     /* no changes detected:
        check wether directory is too old to contain moves
     */
     if((flag&64) || Fileliste_ist_zu_alt(liste,idx,altersgrenze,&size,1|2)>0)
       flag|= 128;
   }
 }

 /* create file list and read content of directory */
 if(Fileliste_new(&sub,0,liste->main_dir,0)<=0) 
   {ret= -1; goto ex;}
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 Fileliste_set_impl_igno(sub,liste->impl_igno,0);
 Fileliste_dir(sub,name,2);
 Fileliste_inherit_target(liste,idx,sub,0);
 if(Fileliste_new(&l_zu_alt,0,liste->main_dir,0)<=0) 
   {ret= -1; goto ex;}
 if(Fileliste_new(&l_junge_ausnahmen,0,liste->main_dir,0)<=0) 
   {ret= -1; goto ex;}
 
 /* test all files */
 ret= 0;
 jung_impl= 0;
 if(sub->anz<=0)
   goto ex;
 for(i=0;i<sub->anz;i++) {
   l_max_ausnahmen= max_ausnahmen-junge_ausnahmen->anz;
   jung_merk= l_junge_ausnahmen->anz;
   fret= Fileliste_precheck_sub(sub,i,old_tree,
                                namen_krank,kein_zugriff,
                                mount_punkte,parent_dev,
                                altersgrenze,l_zu_alt,l_junge_ausnahmen,
                                l_max_ausnahmen,min_sbpf,
                                exclude_size,too_big,
                                linecount,1|flag);
   if(fret<0)
     {ret= -1; goto ex;}
   if(!(fret&8)) {
     if(l_junge_ausnahmen->anz==jung_merk)
       jung_impl++; /* objects wich are in none of both lists */
   }
   ret|= fret;
 }

 if(flag&8){ /* age test */
   liste->groessen[idx]= -1;
   for(i=0;i<sub->anz;i++) 
     if(sub->typen[i]>=0)
       if(sub->groessen[i]>0)
         liste->groessen[idx]+= sub->groessen[i];
   if(!(ret&8)) {                          /* if everything was young enough */
     /* count this directory as "young" */;
     liste->flags[idx]|= 2;
     if((ret&16)) {        /* if old files have been accepted for efficiency */
       liste->flags[idx]|= 4;
     }
   } else if (jung_impl+l_junge_ausnahmen->anz<=0) {
                                               /* if everything was too old  */
     liste->flags[idx]|= 4;
     Fileliste_truncate(l_zu_alt,0,0);   /* they will be replaced as a whole */
     Fileliste_copy_item(liste,idx,l_zu_alt,l_zu_alt->anz,0);     /* by this */
   } else if(l_junge_ausnahmen->anz+jung_impl<l_max_ausnahmen) {
                              /* if there is still room for local youngsters */

     /* count this directory as "old" */;
     liste->flags[idx]|= 6;
     /* append youngsters from  sub  to  l_junge_ausnahmen */
     if(jung_impl>0) {
       for(i=0;i<sub->anz;i++) 
        if(sub->flags[i]&2)
         if(Fileliste_search_file(l_zu_alt,sub->adressen[i],0)<0)
          if(Fileliste_search_file(l_junge_ausnahmen,sub->adressen[i],0)<0)
           Fileliste_copy_item(sub,i,
                               l_junge_ausnahmen,l_junge_ausnahmen->anz,0);
     }
     Fileliste_truncate(l_zu_alt,0,0);   /* they will be replaced as a whole */
     Fileliste_copy_item(liste,idx,l_zu_alt,l_zu_alt->anz,0);     /* by this */
     jung_impl= 0;
   } else { 
     liste->flags[idx]|= 6;
     /* count this directory as "young" , mark unclean */;
     ret&= ~8;
     ret|= 16;
   }

   /* evaluate effectiveness and eventually prevent split */
   lret= Fileliste_lohnender_split(l_zu_alt,l_junge_ausnahmen,
                                   l_junge_ausnahmen->anz+jung_impl,
                                   min_sbpf,0);
   if(lret==0){
     ret&= ~8;
     ret|= 16;
   }
   /* append to the "old" list l_zu_alt 
      and to the "young" list l_junge_ausnahmen
   */
   for(i=0;i<l_zu_alt->anz;i++)
     Fileliste_copy_item(l_zu_alt,i,zu_alt,zu_alt->anz,0);
   for(i=0;i<l_junge_ausnahmen->anz;i++) 
     Fileliste_copy_item(l_junge_ausnahmen,i,
                             junge_ausnahmen,junge_ausnahmen->anz,0);
 }

ex:;
 if(sub!=NULL) Fileliste_destroy(&sub,0);
 if(l_zu_alt!=NULL) Fileliste_destroy(&l_zu_alt,0);
 if(l_junge_ausnahmen!=NULL) Fileliste_destroy(&l_junge_ausnahmen,0);
 return(ret);
}


int Fileliste_precheck(liste,old_tree,
                       namen_krank,kein_zugriff,mount_punkte,
                       altersgrenze,zu_alt,junge_ausnahmen,
                       max_ausnahmen,min_sbpf,
                       exclude_size,too_big,
                       flag)
struct FilelistE *liste;
struct FilelistE *old_tree;
struct FilelistE *namen_krank,*kein_zugriff,*mount_punkte;
time_t altersgrenze;
struct FilelistE *zu_alt,*junge_ausnahmen;
int max_ausnahmen;
double min_sbpf;
double exclude_size;
struct FilelistE *too_big;
int flag;
/*
 bit1= consider files with st_dev != parent_dev to be unreadable
 bit2= don't check for real readability
 bit3= check for age
 bit4= consider remote (e.g. NFS) mountpoints to be unreadable 
 bit5= with bit8 : mistrust directory timestamps 
 bit6= ignore timestamps

 bit8= check for existence in old_tree
 bit9= check type in old_tree
 bit10= check size in old_tree
 bit11= check crc in old_tree
 bit12= consider files larger than exclude_size unreadable 
 bit13= with bit12 : eventually use size of link target 
*/
{
 int i,subflag;
 dev_t parent_dev;
 struct stat stbuf;
 int linecount= 0;

 for(i=0;i<liste->anz;i++) {
   subflag= flag&(4|8|32|64|256|512|1024|2048|4096|8192);
   if(flag&(2|16))
     if(Cd_backup_planer_call_lstaT(liste->adressen[i],&stbuf)!=-1) {
       parent_dev= stbuf.st_dev;
       subflag|= (flag&(2|16));
     }
   Fileliste_precheck_sub(liste,i,old_tree,
                          namen_krank,kein_zugriff,
                          mount_punkte,parent_dev,
                          altersgrenze,zu_alt,junge_ausnahmen,
                          max_ausnahmen,min_sbpf,
                          exclude_size,too_big,
                          &linecount,subflag);
 }
 Print_pacifier(linecount,"file objects checked",1|2|4|8|16);
 fprintf(stderr,"\n");
 return(1);
}


int Fileliste_count_dir_files(liste,dir,flag)
struct FilelistE *liste;                             /* does not get altered */
char *dir;
int flag;
{
 struct FilelistE *sub;
 int count= -1;

 if(Fileliste_new(&sub,0,liste->main_dir,0)<=0)
   return(-1);
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 if(Fileliste_dir(sub,dir,0)<=0) {
   Fileliste_destroy(&sub,0);
   return(-1);
 }
 count= sub->anz;
 Fileliste_destroy(&sub,0);
 return(count);
}


int Fileliste_build_branch(liste,new_current,line_in,target,
                           typ,size,crc,md5,xprop,flag)
struct FilelistE *liste;
struct FilelistE **new_current;
char *line_in;
char *target;                    /* use only if neither NULL nor empty text */
int typ;
double size;
unsigned long crc;
unsigned char md5[16];
struct Filelist_xprop_reC *xprop;
int flag;
/*
 bit0= handle as directory address not as file object
       set *new_current to last resulting list
 bit4= delete depicted object rather than create it
       (do not delete if bit0 is set)
 bit8= line valid
 bit9= typ valid
bit10= size valid
bit11= crc valid
bit12= crc is MD5 rather than CRC32
*/
{
 int ret,idx,nodecount= 0,i;
 char *cpt,*npt= NULL,line[4096];
 struct FilelistE *current,*sub;

 strcpy(line,line_in);
 current= liste;

 npt= cpt= line;

 while(npt!=NULL) {
   if(nodecount)
     cpt= npt+1;
   nodecount++;
   idx= -1;
   npt= strchr(cpt,'/');
   if(npt!=NULL)
     *npt= 0;
   if(cpt[0]==0)
 continue;
   if(cpt[0]=='.'&&cpt[1]==0)
 continue;
   idx= Fileliste_search_file(current,cpt,0);
   if(idx>=0) {
     /* already existing */
     if(current->subdirs[idx]==NULL) { 
       if(npt!=NULL) {
         fprintf(stderr,"--- branch node #%ld not a directory '%s'\n",
                        nodecount,cpt);
         ret= 0; goto ex;
       }
/* break; */
     }
     if(npt!=NULL) {
       current= current->subdirs[idx];
 continue;
     } else {
       if(flag&512)
         current->typen[idx]= typ; 
       if(flag&1024)
         current->groessen[idx]= size; 
       if(flag&2048) {
         ret= Fileliste_set_crc(current,idx,crc,md5,flag&(1<<12));
         if(ret<=0)
           goto ex;
       }
       if(xprop!=NULL) {
         ret= Fileliste_set_xprop(current,idx,xprop->mod,xprop->own,xprop->grp,
                                  xprop->date,0);
         if(ret<=0)
           goto ex;
       }
       if(flag&1)
         current= current->subdirs[idx];
     }
   } else {
     /* not existing yet */;
     if((flag&16) && !(flag&1))
       {ret= 2; goto ex;}
     /* append new list item */
     ret= Fileliste_append(current,cpt,1|8);
     if(ret<=0) {
cannot_append:;
       fprintf(stderr,"--- cannot append to file list\n");
       goto ex;
     }
     if(npt==NULL) {
       /* this is the tip of the branch */
       if(target!=NULL) if(*target!=0) {
         ret= Fileliste_set_target(current,current->anz-1,target,1);
         if(ret<=0)
           goto cannot_append;
       }
       if(flag&512)
         current->typen[current->anz-1]= typ; 
       if(flag&1024)
         current->groessen[current->anz-1]= size; 
       if(flag&2048) 
          Fileliste_set_crc(current,current->anz-1,crc,md5,flag&(1<<12));
       if(xprop!=NULL) {
         ret= Fileliste_set_xprop(current,current->anz-1,
                                  xprop->mod,xprop->own,xprop->grp,
                                  xprop->date,0);
         if(ret<=0)
           goto ex;
       }
     } else
       current->typen[current->anz-1]= 2;

     /* if dir : create new sub list */
     if(npt!=NULL || (flag&1)) {
       char sub_target[4096];

       ret= Fileliste_new(&sub,0,current->main_dir,Fileliste_hash_sizE>0);
       if(ret<=0) {
cannot_create_new_list:;
         fprintf(stderr,"--- cannot create new file list\n");
         goto ex;
       }
       Fileliste_set_subdir(current,current->anz-1,sub,0);
       Fileliste_set_sperrliste(sub,liste->sperrliste,0);

       Fileliste_target(current,current->anz-1,sub_target,0);
       if(Fileliste_set_target_dir(sub,sub_target,0)<=0)
         {ret= -1; goto cannot_create_new_list;}

       current= sub;
     }
   }
 }

 if((flag&16) && idx>=0 && !(flag&1)) {
   if(current->subdirs[idx]!=NULL) {
     /* directory node : do only remove if empty */;
     sub= current->subdirs[idx];
     for(i=0;i<sub->anz;i++)
       if(sub->adressen[idx]!=NULL)
     break;
     if(i<sub->anz) {
       fprintf(stderr,"--- branch node subdirectory not empty on delete\n");
       ret= 0; goto ex;
     }
   } 
   Fileliste_destroy_item(current,idx,0);
 }

 ret= 1;
ex:;
 if(ret>0 && (flag&1))
   *new_current= current;
 return(ret);
}


int Fileliste_search_branch(liste,found_list,found_idx,adr_in,flag)
struct FilelistE *liste;
struct FilelistE **found_list;
int *found_idx;
char *adr_in;
int flag;
/*
 bit0= in case of liste==NULL return 1 rather than 0
*/
{
 int ret,idx,nodecount= 0,i;
 char *cpt,*npt= NULL,line[4096];
 struct FilelistE *current,*sub;

 if(liste==NULL)
   return(flag&1);
 strcpy(line,adr_in);
 current= liste;

 npt= cpt= line;

 while(npt!=NULL) {
   if(nodecount)
     cpt= npt+1;
   nodecount++;
   idx= -1;
   npt= strchr(cpt,'/');
   if(npt!=NULL)
     *npt= 0;
   if(cpt[0]==0)
 continue;
   if(cpt[0]=='.'&&cpt[1]==0)
 continue;
   idx= Fileliste_search_file(current,cpt,0);
   if(idx>=0) {
     /* already existing */
     if(current->subdirs[idx]==NULL) { 
       if(npt!=NULL) 
         { ret= 0; goto ex; }
 break;
     }
     if(npt!=NULL) {
       current= current->subdirs[idx];
 continue;
     }
   } else {
     ret= 0; goto ex;
   }
 }

 ret= (idx>=0);
ex:;
 if(ret>0) {
   *found_list= current;
   *found_idx= idx;
 }
 return(ret);
}


int Fileliste_next_in_tree(liste_pt,idx_pt,current_dir,line,typ,size,crc,md5,
                           flag)
struct FilelistE **liste_pt;
int *idx_pt;
char current_dir[4096];
char line[4096];
int *typ;
double *size; 
unsigned long *crc;
unsigned char md5[16];
int flag;
/*
 bit13=compose line from target and source
*/
/*
 return:
  <=0 = error
    bit0= line valid
    bit1= typ valid
    bit2= size valid
    bit3= crc valid
    bit4= crc is MD5 rather than CRC32
    bit13= target has been prepended
*/
{
 int ret= 0,idx,l;
 char *cpt;
 struct FilelistE *liste;

 liste= *liste_pt;
 idx= *idx_pt;
 line[0]= 0;
 if(liste==NULL)
   return(0);
 if(liste->boss==NULL) {
   strcpy(line,"/");
   current_dir[0]= 0;
 } else
   strcpy(line,"./");

 for(;idx<liste->anz;idx++) {
   if(liste->typen[idx]<0 || liste->adressen[idx]==NULL)
 continue;
   strcat(line,liste->adressen[idx]);
   /* is it necessary to prepend a target ? */
   if((Fileliste_target_inp_trenner!=0 && flag&(1<<13))
      && (liste->boss==NULL || liste->targets[idx]!=NULL)) {
     /* prepend target */
     char target[4096],*cpt;
     Fileliste_target(liste,idx,target,1); 
     l=strlen(target);
     for(cpt= line+strlen(line);cpt>=line;cpt--)
       *(cpt+l+1)= *cpt;
     strcpy(line,target);
     *(cpt+l+1)= Fileliste_target_inp_trenner; /* glues together both parts */
     /* give signal that target has been prepended */;
     ret|= (1<<13);
   }
   ret|= 1;
   *typ= liste->typen[idx];
   if(*typ>0)
     ret|= 2;
   *size= liste->groessen[idx];
   if(*typ>0 && *size>=0)                      /* ??? why was :  && *typ!=2 */
     ret|= 4;
   if(liste->crcs[idx]==NULL) {
     strcpy(md5,"no md5         ");
     *crc= 0;
   } else  if(liste->flags[idx]&16) {
     memcpy(md5,liste->crcs[idx]->md5,16);
     ret|= 16;
   } else 
     *crc= liste->crcs[idx]->crc;
   if(liste->flags[idx]&8)
     ret|= 8;
   if(liste->subdirs[idx]!=NULL) {
     strcat(line,"/");
     sprintf(current_dir+strlen(current_dir),"/%s",liste->adressen[idx]);
     /* enter sub directory */
     *liste_pt= liste->subdirs[idx];
     *idx_pt= 0;
   } else
     *idx_pt= idx+1;
   goto ex;
 }
 /* list done , go upwards */
 cpt= strrchr(current_dir,'/');
 if(cpt!=NULL)
   *cpt= 0;
 else
   current_dir[0]= 0;
 if(liste->boss==NULL) 
   {ret= 0; goto ex;}
 *idx_pt= liste->boss_idx+1;
 *liste_pt= liste->boss;
 *typ= 2;
 strcpy(line,"..");
 ret= 3;
ex:;
 return(ret);
}


int Fileliste_read_serialized_line(liste,idx,current_dir,fpin,
                                   line,typ,size,crc,md5,xprop,flag)
struct FilelistE **liste;
int *idx;
char current_dir[4096];
FILE *fpin;
char *line;
int *typ;
double *size;
unsigned long *crc;
unsigned char md5[16];
struct Filelist_xprop_reC **xprop; /* *xprop has to be NULL or freeable */
int flag;
/*
 bit9= prepended is type
bit10= prepended is length
bit11= prepended is crc
bit12= reserved (MD5 bit is automated)
(a fixed field format is recognized by type "D")
bit13= return raw address even if with '='
bit14= if with '=' choose target addresses rather than source addresses
       change to relative addresses with ./ prepended
return:
  <=0 = error
    bit0= line valid
    bit1= typ valid
    bit2= size valid
    bit3= crc valid
    bit4= crc is MD5 (in md5) rather than CRC32 (in *crc)
    bit13= target has been prepended
*/
{
 int ret=0,i,x,is_big_format= 0,valid= 0,hret,codelen;
 char *cpt,*npt,buf[3],*date_str= NULL;

 line[0]= 0;
 *typ= 0;
 *size= -1;
 *crc= 0;
 if(*liste!=NULL) {
   ret= Fileliste_next_in_tree(liste,idx,current_dir,line,typ,size,crc,md5,
                               flag&(1<<13));
 } else {
   if(Sfile_fgets(line,4096,fpin)==NULL) 
     {ret= 0; goto ex;}
   ret= 1;
   if(line[0]=='#' && line[1]=='=' && line[2]=='#' && line[3]==' ')
     {ret= 1; goto ex;}
   if(line[0]=='.' && line[1]=='.')
     {ret= 1; goto ex;}
   /* eventually split into : type size crc name */
   /* resp : Type permissionbits owner group time lengthtext crctext adr */
   cpt= line;
   if(flag&512) {
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     ret|= 2;
     if(isalpha(*cpt))
       if(toupper(*cpt)==*cpt)
         is_big_format= 1;
     if(*cpt=='0')
       *typ= 0;
     if(*cpt=='1' || *cpt=='f' || *cpt=='F' || *cpt=='-') 
       *typ= 1;
     else if(*cpt=='2' || *cpt=='d' || *cpt=='D')
       *typ= 2;
     else if(*cpt=='3' || *cpt=='l' || *cpt=='L')
       *typ= 3;
     else if(*cpt=='4' || *cpt=='p' || *cpt=='P')
       *typ= 4;
     else if(*cpt=='5' || *cpt=='m' || *cpt=='M')
       *typ= 5;
     else
       ret&= ~2;
     cpt= npt+1;
   }
   if(is_big_format) {

#ifdef NIX
     /* skip :  permissionbits owner group time */
     for(i=0;i<4;i++) {
       npt= strchr(cpt,' ');
       if(npt==NULL)
         {ret= 0; goto ex;}
       cpt= npt+1;
     }
#else /* NIX */
     if((*xprop)==NULL) {
       hret= Filelist_xprop_rec_new(xprop,0);
       if(hret<=0)
         goto ex;
     }
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     sscanf(cpt,"%lo",&((*xprop)->mod));
     cpt= npt+1;
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     if(Sregex_string_cut(&((*xprop)->own),cpt,npt-cpt,0)<=0)
       {ret= -1; goto ex;}
     cpt= npt+1;
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     if(Sregex_string_cut(&((*xprop)->grp),cpt,npt-cpt,0)<=0)
       {ret= -1; goto ex;}
     cpt= npt+1;
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     if(Sregex_string_cut(&date_str,cpt,npt-cpt,0)<=0)
       {ret= -1; goto ex;}
     hret= Zeitdefinition(date_str,&valid,&((*xprop)->date),0);
     if(hret<=0)
       goto ex;
     cpt= npt+1;
#endif /* ! NIX */

   }
   if(flag&1024) {
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     while(*(npt+1)==' ')
       npt++;
     sscanf(cpt,"%lf",size);
     if(*size>=0)
       ret|= 4;
     cpt= npt+1;
   }
   if(flag&2048) {
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     codelen= npt-cpt;
     while(*(npt+1)==' ')
       npt++;
     if(*cpt!='-' && *cpt!=0) {
       if(codelen==32) {
         buf[2]= 0;
         for(i=0;i<16;i++) {
           buf[0]= cpt[2*i];
           buf[1]= cpt[2*i+1];
           sscanf(buf,"%x",&x);
           md5[i]= x;
         }
         ret|= 16;
       } else
         sscanf(cpt,"%lx",crc);
       ret|= 8;
     }
     cpt= npt+1;
   }

   /* handle address part */
   if(flag&(1<<13))
     npt= NULL;                           /* return raw address */
   else
     npt= strchr(cpt,Fileliste_target_inp_trenner);
   if(npt!=NULL) {
     if(flag&(1<<14)) {
       /* cut off source part */
       *npt= 0;
       /* make relative addresses */
       if(cpt[0]=='.' && cpt[1]=='/') 
         ;
       else {
         if(cpt-line<2) {
           for(npt= cpt+strlen(cpt); npt>=cpt; npt--)
             *(npt+2)= *npt;
           cpt+= 2;
         }
         if(cpt[0]=='/') {
           /* prepend . */;
           *(--cpt)= '.';
         } else {
           /* prepend ./ */;
           *(--cpt)= '/'; /* last first */
           *(--cpt)= '.';
         }
       }
     } else
       /* skip target part */
       cpt= npt+1;
   }
   if(cpt!=line) {
     for(npt= line;*cpt!=0;cpt++)
       *(npt++)= *cpt;
     *npt= 0;
   }
 }
ex:;
 Sregex_string(&date_str,NULL,0);
 return(ret);
}


int Fileliste_decode_serialized_head(liste,line,waiting_for_wd,
                                     initial_wd,initial_wd_list,current_list,
                                     list_type,
                                     layout_flag,got_wd,at_top_level,flag)
struct FilelistE *liste;
char *line;
int *waiting_for_wd;         /* in/out: status of header interpretation */
char *initial_wd;            /* out|const (see flag) */
struct FilelistE **initial_wd_list; /* out */
struct FilelistE **current_list;    /* out */
int *list_type;              /* out : <0= invalid / unusable
                                       0= raw
                                       1= complete backup area file list
                                       2= backup content list
                                       3= complete backup area directory list
                                       4= vanished file objects list
*/
int *layout_flag;            /* out */
int *got_wd;                 /* out */
int *at_top_level;           /* out */
int flag;
/*
 bit0-3= mode :
          2=graft files into a tree under liste
          other modes only  interesting to Fileliste_read_serialized_tree()
 bit4= in mode 2 : delete files from liste tree rather than to insert them
 bit5= do not process wd line when finally found
 bit6= do not accept content lists but only complete and raw formats
       (eventually return(2) )
 bit14= do not overwrite initial_wd
*/
/*
 return:
  <=0 = error
    1 = ok
    2 = not an acceptable list format
    3 = line is a valid address, do not skip
*/
{
 int ret,ignore_wd,mode,del;

 if(*waiting_for_wd==0)
   return(1);
 del= flag&16;
 mode= flag&15;
 ignore_wd= !!(flag&32);
 if(*waiting_for_wd==-1) {
   *waiting_for_wd= 1;
   *list_type= -1;
   if(strcmp(line,"#=# complete backup area file list")==0)
     *list_type= 1;
   else if( strcmp(line,"#=# backup content list")==0)
     *list_type= 2;
   else if(strcmp(line,"#=# complete backup area directory list")==0)
     *list_type= 3;
   else if(strcmp(line,"#=# vanished file objects list")==0)
     *list_type= 4;
   else if(strncmp(line,"#=# ",4)==0) {
     fprintf(stderr,
               "--- this file seems to be in a list format from the future\n");
     ret= 0; goto ex;
   } else if(line[0]=='#' && line[1]=='!') {
     fprintf(stderr,"--- this file seems to be an interpreter script\n");
     ret= 0; goto ex;
   } else {
     *waiting_for_wd= 0;
     *list_type= 0;
     ret= 3; goto ex;
   }
   if(flag&64) 
     if(*list_type>1)
       {ret= 2; goto ex;}
 } else if(*waiting_for_wd==1) { 
   /* gathering header information until working directory is encountered */
   if(strncmp(line,"#=# line layout :",17)==0) {
     *layout_flag= 0;
     if(strstr(line+17," type ")!=NULL)
       *layout_flag|= (1<<9);
     if(strstr(line+17," size ")!=NULL)
       *layout_flag|= (1<<10);
     if(strstr(line+17," crc ")!=NULL || 
        strstr(line+17," md5 ")!=NULL ||
        strstr(line+17," chk ")!=NULL)
         *layout_flag|= (1<<11);
   } else if(strcmp(line,"#=# next line tells current working directory")==0)
     *waiting_for_wd= 2;
 } else if(*waiting_for_wd==2) {
   line[strlen(line)-1]= 0;
   if(!ignore_wd) {
     if(!(flag&(1<<14)))
       strcpy(initial_wd,line);
     if(mode==2) {
       ret= Fileliste_build_branch(liste,initial_wd_list,initial_wd,"",
                                      0,-1.0,0,NULL,NULL,1|del);
       if(ret<=0) {
         fprintf(stderr,"--- cannot build file list branch '%s'\n",initial_wd);
         goto ex;
       }
       *current_list= *initial_wd_list;
     }
   }
   *waiting_for_wd= 0;
   *got_wd= 1;
   *at_top_level= 1;
 }

 ret= 1;
ex:;
 return(ret);
}


int Fileliste_read_serialized_tree(liste,fpin,ftree,fpout,
                                   initial_wd,initial_wd_list,prepend_dir,flag)
struct FilelistE *liste;
FILE *fpin;                             /* to be read from if ftree == NULL */
struct FilelistE *ftree;                     /* to be read from if not NULL */
FILE *fpout;
char initial_wd[4096];
struct FilelistE **initial_wd_list;
char *prepend_dir;
int flag;
/*
 bit0-3= mode :
          0=check existence and print absolute addresses of vanished files
            to fpout (print target and source )
          1=check existence and print vanished files as shell commands rm and
            rmdir to fpout (use target address only)
          2=graft files into a tree under liste
          3=check existence, compare with checksum and delete if it matches
          4=check existence, compare with checksum and report if it matches
 bit4= in mode 2 : delete files from liste tree rather than to insert them
 bit5= initial_wd and initial_wd_list are already set
 bit6= do not accept content lists but only complete and raw formats
       (eventually return(2) )
 bit7= expect target=source pairs and process only the source part
       do not expect incremental tree notation
       with mode 4 : print the whole input line
 bit9= prepend type
bit10= prepend length
bit11= preprend crc
bit12= use MD5 rather than CRC32
bit13= allow mode 3 with insecure parameters
bit14= choose target addresses rather than source addresses
       Use text given with  initial_wd  as prefix. Ignore prepend_dir

return:
 <=0 = error
   1 = complete list
   2 = content list
   3 = raw list
   4 = old_tree
*/
{
 char dir[4096],line[3*4096],adr[4096],*spt,shell_adr[3*4096],read_dir[4096];
 char *target= "",target_dir[4096],target_adr[4096];
 char typetext[4],lengthtext[80],crctext[80],*cpt,*wpt,*fapt;
 int waiting_for_wd= -1,ret= 0,slashno,absolute= 0,is_dir,got_wd= 0,popened= 0;
 int mode, complete_list= 0,ignore_wd= 0,del,linecount= 0,dircount= 0;
 int noprint= 0,read_idx,typ,read_ret,layout_flag= 0,hflag,was_mismatch= 0;
 int is_vanished_file_list= 0,hashtype,rec_fits_dir,i,rec_nam_l;
 int malformed= 0,line_flag= 0,at_top_level= 1,tree_target_prepended= 0;
 int list_type;
 double size;
 unsigned long crc;
 unsigned char md5[16];
 struct FilelistE *current,*read_list;
 struct CheckrecorD *dir_rec_stack= NULL,*dir_rec;
 struct Filelist_xprop_reC *xprop= NULL;
 time_t dummy_time= 0;

 static int dummy_delete= 0;

 del= flag&16;
 mode= flag&15;
 dir[0]= 0;
 if((flag&32) && ftree==NULL) {
   ignore_wd= 1;
   strcpy(dir,initial_wd);
 } else {
   if(!(flag&(1<<14)))
     strcpy(initial_wd,"");
   if(mode==2)
     *initial_wd_list= liste;
   if(ftree!=NULL)
     waiting_for_wd= 0;
 }

 if(mode==2) {
   if(liste==NULL)
     {ret= -1; goto silent_ex;}
   current= liste;
   if(flag&32)
     if(*initial_wd_list!=NULL)
       current= *initial_wd_list;
 }

 if(ftree!=NULL)
   Print_pacifier(0,"  0          tree nodes read",1|4);
 else
   Print_pacifier(0,"  0          file list lines read",1|4);

 read_list= ftree;
 read_idx= 0;
 read_dir[0]= 0;
 if(flag&(1<<14))
   line_flag|= (1<<14);
 else
   line_flag|= (1<<13);
 while(1) {

   line_flag= (line_flag&~layout_flag)|(layout_flag*!!got_wd);
   read_ret= Fileliste_read_serialized_line(&read_list,&read_idx,read_dir,fpin,
                                            line,&typ,&size,&crc,md5,&xprop,
                                            line_flag);
   if(read_ret<=0) {
     if(got_wd || waiting_for_wd>0)
       goto unexpected_end;
     goto expected_end;
   }
   tree_target_prepended= !!(read_ret&(1<<13));

   linecount++;
   if((linecount%10)==0 || linecount==2)
     Print_pacifier(linecount,"",2);

   if(line[0]==0)
 continue;

   if(waiting_for_wd) {
     /* try to interpret line as part of file header */
     ret= Fileliste_decode_serialized_head(liste,line,&waiting_for_wd,
                                     initial_wd,initial_wd_list,&current,
                                     &list_type,
                                     &layout_flag,&got_wd,&at_top_level,
                                     flag&(15|16|32|64|(1<<14)));
     if(ret<=0 || ret==2)
       goto ex;
     complete_list= (list_type==1);
     is_vanished_file_list= (list_type==4);
     if(waiting_for_wd==0) {
       strcpy(dir,initial_wd);
       at_top_level= 1;
     }
     if(ret!=3)
 continue;
   }

   if(got_wd) {
     if(strcmp(line,"#=# end")==0)
 break;
     if(line[0]=='#' && line[1]=='=' && line[2]=='#' && line[3]==' ')
 continue;
   }

   target= "";
   fapt= line;
   /* if a raw address was fetched : split it */
   if((flag&128 || at_top_level || tree_target_prepended) && !(flag&(1<<14))) {
     cpt= strchr(line,Fileliste_target_inp_trenner);
     if(cpt!=NULL) {
       fapt= cpt+1;
       if(!(mode==4 && (flag&128)))
         *cpt= 0;
       target= line;
     }
   }
   if(strcmp(fapt,"./")==0)
 continue;

   /* >>> this does not work correctly with (flag&128) */
   /* >>> ( but maybe it is better this way ) */
   if(fapt[strlen(fapt)-1]=='/') {
     is_dir= 1;
     fapt[strlen(fapt)-1]= 0;
     if(fapt[0]!='.' || fapt[1]!='.')
       dircount++;
     at_top_level= 0;
   } else
     is_dir= 0;

   if(flag&(1<<14))
     strcpy(adr,"");
   else
     strcpy(adr,prepend_dir);
   strcpy(target_adr,""); /* ??? should this be prepend_dir rather than "" ? */

   if(flag&128) {
     strcat(adr,fapt);
     strcat(target_adr,fapt);
   } else if(fapt[0]=='/') {
     absolute= 1;
     if(mode==2) {
       ret= Fileliste_build_branch(liste,&current,fapt,target,typ,size,crc,md5,
                                   xprop,is_dir|del|(read_ret<<8));
       if(ret<=0) {
could_not_build_branch:;
         fprintf(stderr,"--- cannot build file list branch '%s'\n",fapt);
         goto ex;
       }
     }
     if(is_dir) {
       strcpy(dir,fapt);
       if(target[0]!=0)
         strcpy(target_dir,target);
       else
         strcpy(target_dir,fapt);
 continue;
     }
     strcat(adr,fapt);
     strcat(target_adr,fapt);
   } else if(fapt[0]=='.' && fapt[1]=='/') {
     if(mode==2) {
       ret= Fileliste_build_branch(current,&current,fapt,target,
                                   typ,size,crc,md5,xprop,
                                   is_dir|del|(read_ret<<8));
       if(ret<=0) 
         goto could_not_build_branch;
     }
     if(is_dir) {
       strcat(dir,fapt+1);
       if(target[0]=='/')
         target_dir[0]= 0;
       if(target[0]!=0)
         strcat(target_dir,target);
       else
         strcat(target_dir,fapt+1);

       /* push directory stack */
       if(!(read_ret&4))
         size= -1;
       hashtype= -1;
       if(read_ret&8)
         hashtype= 0;
       if(read_ret&16)
         hashtype= 1; 
       ret= Checkrecord_new(&dir_rec,fapt+2,dummy_time,size,hashtype,
                            crc,md5,0);
       if(ret<=0) {
         fprintf(stderr,"--- not enough memory for directory stack\n");
         goto ex;
       }
       Checkrecord_link(dir_rec,dir_rec_stack,0);
       dir_rec_stack= dir_rec;
 continue;
     } else {
       strcat(adr,dir);
       strcat(adr,fapt+1);
       strcat(target_adr,target_dir);
       strcat(target_adr,fapt+1);
     }
   } else if(fapt[0]=='.' && fapt[1]=='.') {
     if(is_dir) 
       slashno= 0;                      /* do not check the parent directory */
     else 
       slashno= strlen(fapt)/3;         /* computes "/" in "../../../.." */
     for(;slashno>=0;slashno--) {
       /* obtain size and crc */
       dir_rec= dir_rec_stack;
       if(dir_rec==NULL)
     break;
       dir_rec_stack= dir_rec_stack->prev;      /* pop directory stack */
       Checkrecord_link(dir_rec,NULL,0);

       /* does it fit the end of dir ? */
       rec_fits_dir= 0;
       Checkrecord_get_name(dir_rec,&spt,0);

       /* ??? >>> target_adr ??? (adr looks like a temporary text) */
       strcpy(adr,spt);
       rec_nam_l= strlen(adr);

       if(strlen(dir)>rec_nam_l && rec_nam_l!=0)
         if(strcmp(dir+(strlen(dir)-strlen(adr)),adr)==0)
           rec_fits_dir= 1;
       if(!rec_fits_dir)
     break;
       for(i=strlen(adr)-1;i>=0;i--)
         if(adr[i]=='/')
           slashno--;
       read_ret= 2;
       typ= 2;
       Checkrecord_get_size(dir_rec,&size,0);
       if(size>=0)
         read_ret|= 4;
       Checkrecord_get_checksum(dir_rec,&hashtype,&crc,md5,0);
       if(hashtype==0) 
         read_ret|= 8;
       else if(hashtype==1)
         read_ret|= (16|8);
       Checkrecord_destroy(&dir_rec,0);
       if(mode==0 || mode==1) {
         /* files done, finally check existence of dir */
         ret= Type_and_size(dir,NULL,1);
         if(ret==-1) {
           /* Vanished directory ! */
           if(flag&1) {
             /* write relative to / */
             fprintf(fpout,"rmdir %s\n",
                           Text_shellsafe(target_dir+(target_dir[0]=='/'),
                                          shell_adr,0));
           } else {
             hflag= flag&(512|1024|2048);
             Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                                 read_ret&(2|4|8|16),hflag);
             fprintf(fpout,"%s%s%s%s=%s\n",typetext,lengthtext,crctext,
                                           target_dir,dir);
           }
         }
       } else if(mode==2) {
         /* move current upward */
         if(current->boss!=NULL) {
           current= current->boss;
         } else {
           current= *initial_wd_list;
         }
       } else if(mode==3 || mode==4) {
         if(flag&(1<<14))
           strcpy(adr,"");
         else
           strcpy(adr,prepend_dir);
         strcat(adr,dir);
         ret=  Remove_dir(adr,(mode==4)|((!!dummy_delete)<<1));
         if(ret<=0)
           was_mismatch= 1;
       }
       dir[strlen(dir)-1-rec_nam_l]= 0;
       if(strlen(target_dir)-1-rec_nam_l>0)
         target_dir[strlen(target_dir)-1-rec_nam_l]= 0;
       else
         target_dir[0]= 0;
       if(strcmp(dir,initial_wd)==0 && !absolute)
     break;
     }
     if(is_dir) {
       at_top_level= 1;
       strcpy(dir,initial_wd);
       strcpy(target_dir,initial_wd);
       if(mode==2)
         current= *initial_wd_list;
       absolute= 0;
       Checkrecord_destroy_all(&dir_rec_stack,0);
     }
 continue;
   } else {
     if(malformed==0) {
       fprintf(stderr,"--- malformed line of serialized tree : %s\n",fapt);
       fprintf(stderr,"--- ( please contact scdbackup@gmx.net )\n");
       malformed= 1;
 continue;
     }
     strcat(adr,fapt);
   }

   if(mode==0 || mode==1) {
     /* test existence of adr */
     ret= Type_and_size(adr,NULL,1);
     if(ret==-1) {
       /* Vanished file ! */;
       if(flag&1) {
         /* write with address relative to / */
         if((read_ret&2) && typ==2) {
           /* directory given without trailing slash */
           fprintf(fpout,"rmdir %s\n",
                   Text_shellsafe(target_adr+(target_adr[0]=='/'),shell_adr,0));
         } else {
           fprintf(fpout,"rm %s\n",
                   Text_shellsafe(target_adr+(target_adr[0]=='/'),shell_adr,0));
         }
       } else {
         hflag= flag&(512|1024|2048);
         Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                             read_ret&(2|4|8|16),hflag);
         fprintf(fpout,"%s%s%s%s=%s\n",typetext,lengthtext,crctext,
                                       target_adr,adr);
       }
     }
   } else if((mode==3 || mode==4) && (read_ret&2) && typ==2) {
     /* directory given without trailing slash */
     ret= Remove_dir(adr,(mode==4)|((!!dummy_delete)<<1));
     if(ret<=0)
       was_mismatch= 1;
   } else if(mode==3 || mode==4) {
     if(mode==3 &&
        (layout_flag&((1<<9)|(1<<10)|(1<<11)))!=((1<<9)|(1<<10)|(1<<11))  && 
        !(flag&(1<<13)) ) {
       fprintf(stderr,
        "--- will not delete files without type,size,checksum verification !\n"
              );
       ret= 0; goto ex;
     }
     if(mode==3 && (!is_vanished_file_list) && !(flag&(1<<13)) ) {
       fprintf(stderr,
         "--- will delete files only if recorded by  -vanished_list_adr !\n"
              );
       ret= 0; goto ex;
     }
     /* test existence and eventually type, size, checksum 
        In non-dangerous mode 3 , count missing information as mismatch */
     ret= File_is_changed(adr,0,0,0.0,0,NULL,
                          read_ret&(8|16),typ,size,crc,md5,
                          1 | ((mode==3 && !(flag&(1<<13)))<<1)
                            | ((read_ret&14)<<8));
     if(ret<0)
       goto ex;
     if(ret==0 && mode==3 && (read_ret&(4|8))!=(4|8) && !(flag&(1<<13)) ) {
       fprintf(stderr,"--- unclear  : %s\n",adr);
     } else if(ret==0) {
       /* unchanged */
       if(mode==3) {
         /* delete file */
         if(dummy_delete)
           fprintf(stdout,"rm %s\n",Text_shellsafe(adr,shell_adr,0));
         else {
           if(unlink(adr)==-1) {
             fprintf(stderr,"--- Cannot remove : %s \n    %s",
                            adr,strerror(errno));
             ret= 0; goto ex;
           }
         }
       } else {
         if(flag&128)
           fprintf(stdout,"%s\n",line);
         else
           fprintf(stdout,"%s\n",adr);
       }
     } else if(ret==1) {
       fprintf(stderr,"--- mismatch : %s\n",adr);
       was_mismatch= 1;
     } else if(ret==2) {
       fprintf(stderr,"--- missing  : %s\n",adr);
       was_mismatch= 1;
     }
   }
 }
expected_end:;
 ret= 1;
 if(complete_list)
   ret= 2;
 if(!got_wd)
   ret= 3;
 if(ftree!=NULL)
   ret= 4;
 if(mode==3 || mode==4)
   if(was_mismatch)
     ret= 0;
ex:;
 if(noprint)
   fprintf(stderr,"\r                                 \n");
 else {
   Print_pacifier(linecount,"",1|2|8);
   if(ftree!=NULL)
     sprintf(line,"  %-10ld directory nodes",dircount);
   else
     sprintf(line,"  %-10ld directory lines",dircount);
   Print_pacifier(0,line,1|4|8);
 }
silent_ex:;
 Filelist_xprop_rec_destroy(&xprop,0);
 return(ret);

unexpected_end:;
 fprintf(stderr,"--- unexpected end of input\n");
 ret= 0;
 goto ex;
}


/* ----------------- Fileliste - end ----------- */


/* ----------------- Backup_stueck --------------- */


struct Backup_stuecK {               /* english : Backup_piecE */
 double bytes;
 int max_items;
 double max_size;
 struct FilelistE *liste;            /* english : flist */
 struct Backup_stuecK *next,*prev;
};


int Backup_stueck_new(stueck,prev,maxanz,max_size,main_dir,flag)
struct Backup_stuecK **stueck,*prev;
int maxanz;
double max_size;
char *main_dir;                         /* for liste */
int flag;
{
 int i,l_maxanz;
 *stueck= TSOB_FELD(struct Backup_stuecK,1);
 if(*stueck==NULL) return(0);
 (*stueck)->bytes= 0;
 (*stueck)->max_items= maxanz;
 (*stueck)->max_size= max_size;
 if(maxanz<1000) l_maxanz= maxanz;
 else            l_maxanz= 1000;
 if(Fileliste_new(&((*stueck)->liste),l_maxanz,main_dir,0)<=0)
   return(0);
 (*stueck)->prev= prev;
 (*stueck)->next= NULL;
 if(prev!=NULL) prev->next= *stueck;
 return(1); 
}


int Backup_stueck_destroy(stueck,flag)
struct Backup_stuecK **stueck;
int flag;
{
 int i;
 Fileliste_destroy(&((*stueck)->liste),0);
 free((char *) (*stueck));
 *stueck= NULL;
 return(1);
}


int Backup_stueck_report_zeile(ziel,line,flag)
char *ziel;
char *line;
int flag;
/*
 bit0= do only count , do not append
*/
{
 int l;
 l= strlen(line);
 if(flag&1) return(l);
 strcat(ziel,line);
 return(l);
}


int Backup_stueck_append_file(stueck,liste,idx,flag)
struct Backup_stuecK *stueck;
struct FilelistE *liste;
int idx;
int flag;
/*
 bit0= do not append , make invalid
*/
{
 struct FilelistE *ziel;
 if(!(flag&1)){
   ziel= stueck->liste;
   Fileliste_copy_item(liste,idx,ziel,ziel->anz,0);
   if(liste->groessen[idx]>0)
     stueck->bytes+= liste->groessen[idx];
 }
 liste->typen[idx]= -liste->typen[idx];
 return(1);
}


int Backup_stueck_report(ziel,cmd,liste,volume,total_volumes,flag)
char **ziel;
char *cmd;
struct FilelistE *liste;
int volume,total_volumes;
int flag;
/*
 bit0-7= output form: 0= Backup_plan_skript target=source \
                      1= Backup_plan_stdout
                      2= Backup_plan_skript source \
                      3= Backup_plan_skript <<+ source +
*/
{
 int i,l= 0,pass;
 char *cpt,*tpt,line[4096],target[4096],txs_buf[4096];


 for(pass=0;pass<2;pass++){
   sprintf(line,"%s",cmd);
   if (volume >= 0) 
     sprintf(line+strlen(line)," -v %d",volume);
   if (total_volumes >= 0) 
     sprintf(line+strlen(line)," -t %d",total_volumes);
   if((flag&255)==3) {
     strcat(line," <<'+'\n");
   } else {
     strcat(line," \\\n");
   }
   l+= Backup_stueck_report_zeile(*ziel,line,pass==0); 
 
   for(i=0;i<liste->anz;i++){
     line[0]= 0;
     cpt= liste->adressen[i];
     if(cpt[0]=='.'){
       if(cpt[1]=='/') cpt+= 2;
     }
     if((flag&255)==2) {
       sprintf(line,"%s \\\n",Text_shellsafe(cpt,txs_buf,0));
     } else if((flag&255)==3) {
       sprintf(line,"%s\n",cpt);
     } else {
       if(Fileliste_target(liste,i,target,0)<=0)
   continue;
       sprintf(line+strlen(line),"%s=",
                                 Text_shellsafe(target,txs_buf,0));
       sprintf(line+strlen(line),"%s \\\n",
                                 Text_shellsafe(cpt,txs_buf,0));
     }
     l+= Backup_stueck_report_zeile(*ziel,line,pass==0); 
   }
   if((flag&255)==3) {
     sprintf(line,"+\n");
     l+= Backup_stueck_report_zeile(*ziel,line,pass==0);
   } else {
     sprintf(line,"\n");
     l+= Backup_stueck_report_zeile(*ziel,line,pass==0);
   }
   sprintf(line,"\ncmd_return=$?\n");
   l+= Backup_stueck_report_zeile(*ziel,line,pass==0);
   if(pass==0){
     *ziel= malloc(l+1);
     if(*ziel==NULL) return(0);
     (*ziel)[0]= 0;
   }
 }
 return(1);
}


int Backup_stueck_report_pl(fp,verify_source_filter,
                            cmd,liste,volume,total_volumes,
                            max_items_per_tile,path_list,flag)
FILE *fp;
char *verify_source_filter,*cmd;
struct FilelistE *liste;
int volume,total_volumes;
int max_items_per_tile;
char *path_list;
int flag;
/*
 bit0-7= output form:
         xxx 0= Backup_plan_skript target=source \
         xxx 1= Backup_plan_stdout
         xxx 2= Backup_plan_skript source \
         xxx 3= Backup_plan_skript <<+ source +
         xxx 4= Backup_plan_stdout(super_raw)
         5= Backup_plan_skript cat target=source >path_list
                               cmd -path-list path_list target1=source1
         6= Backup_plan_skript cat source >path_list
                               cmd -path-list path_list
*/
{
 int i,first_item= -1,verify_source= 1;
 char *cpt,line[4096],target[4096],txs_buf[4096],*first_cpt= NULL;

 Text_shellsafe(path_list,txs_buf,0);
 fprintf(fp,"test -e %s && rm %s\n",txs_buf,txs_buf);
 fprintf(fp,"( cat <<'+'\n");
 
 for(i=0;i<liste->anz;i++){
   cpt= liste->adressen[i];
   if(cpt[0]=='.'){
     if(cpt[1]=='/') cpt+= 2;
   }
   if((flag&255)==5) {
     if(Fileliste_target(liste,i,target,0)<=0)
 continue;
     if(first_item==-1) {
       first_item= i;
       first_cpt= cpt;
       /* this item will be used as command line seed */
 continue;
     }
     fprintf(fp,"%s=%s\n",target,cpt);
   } else {                              /* if((flag&255)==6) */
     fprintf(fp,"%s\n",cpt);
   }
   if((i%max_items_per_tile)==0)
     fprintf(fp,"+\ncat <<'+'\n");
 }

 if(verify_source_filter==NULL) 
   verify_source= 0;
 else if(strlen(verify_source_filter)==0) 
   verify_source= 0;
 if(verify_source) {
   fprintf(fp,"+\n) | \\\n");
   fprintf(fp,"%s \\\n",verify_source_filter);
   fprintf(fp,"> %s\n",path_list);
 } else { 
   fprintf(fp,"+\n) >%s\n",path_list);
 }

 fprintf(fp,"%s \\\n",cmd);
 fprintf(fp,"-path-list %s \\\n",txs_buf);
 if(first_item>=0 && (flag&255)==5) {
   Fileliste_target(liste,first_item,target,0);
   fprintf(fp,"%s=%s \\\n",target,first_cpt);
 }
 fprintf(fp,"\ncmd_return=$?\n");
 fprintf(fp,"\n");
 fprintf(fp,"rm %s\n",txs_buf);
 return(1);
}


/* ----------------- Backup_stueck - end -------- */


/* ----------------- Backup_plan ----------------- */


#define Backup_plan_max_size_stackL 10


struct Backup_plaN {

 /* Task */

 int max_items;
 int max_items_per_tile;                         /* for output_form==5 */
 char path_list[4096];
 double max_size;
 double default_max_size;
 int save_stack_media;            /* 1= after trying to fill max_size from
                                        stack : do not use max_size_stack media 
                                        if default_max_size would suffice */
 double max_size_stack[Backup_plan_max_size_stackL];
 int max_size_counts[Backup_plan_max_size_stackL];
 int max_size_stackpt;
 double min_size_fak;
 int scattering;
 char cmd[4096];
 char verify_source_filter[4096];
 struct FilelistE *startliste;            /* english : startlist */
 struct FilelistE *auslassen;             /* english : excludelist */
 int auslassen_nicht_startliste;          /* english : exclude_not_startlist */
 char dir_size_cmd[4096];
 char target_inp_trenner;                 /* english : target_inp_separator */
 char shell_to_use[4096];
 int accept_all_filetypes;

 char hauptdir[4096];                     /* english : maindir */
 int hauptdir_auto;                       /* english : maindir_auto */

 int auskunft_an;                         /* english : info_enabled */
 struct FilelistE *auskunft_name;         /* english : info_name */
 double auskunft_groesse;                 /* english : info_size */
 int auskunft_include;                    /* english : info_include */
 int auskunft_individuell;                /* english : info_individual */
 int auskunft_skript_root;                /* english : info_skript_root */

 char skript_status_file[4096];           /* english : script_status_file */
 char prompt_skript[4096];                /* english : prompt_script */

 char conf_dir[4096];
 int level;         /* -2:current -1:next >=0:level number */
 int configuration; /* 0=ignore , 1=create , 2=use , 3=update */

 int list_output_format;           /* bit0=adr bit1=type bit2=size bit3=crc 
                                      bit4=MD5 rather than CRC32 */
 char tmp_dir[4096];
 char complete_list_adr[4096];
 char append_complete_target[4096];
 char dir_list_adr[4096];
 char dir_list_target[4096];
 char dir_script_adr[4096];
 char dir_script_target[4096];
 char content_list_adr[4096];
 char append_content_target[4096];
 struct FilelistE *old_tree;
 char old_complete_list_adr[4096];
 char old_vanish_list_adr[4096];
 char vanish_list_adr[4096];
 char vanish_list_target[4096];
 char vanish_script_adr[4096];
 char vanish_script_target[4096];
 time_t start_time;

 double pre_checksum_size;

 /* Result */

 int anz_stuecke;                         /* english : piece_count */
 struct Backup_stuecK *stuecke;           /* english : pieces */
 struct Backup_stuecK *aktstueck;         /* english : current_piece */

 int output_form; /* 0= Backup_plan_skript target=source \
                     1= Backup_plan_stdout
                     2= Backup_plan_skript source \
                     3= Backup_plan_skript <<+ source +
                     4= Backup_plan_stdout(super_raw)
                     5= Backup_plan_skript 
                        ( cat <<+ target=source + ) | command -path-list -
                   */
 int collision_test; /* 0=do not test
                        1=report collisions 
                      */
 int collision_avoid; /* 0=leave collisions as they are
                         1=change target in case of collision
                         2=exit in case of collision
                       */
 int precheck;        /* 0=no check of file addresses
                           (may be overridden by precheck_xdev)
                         1=check file addresses
                         2=end program after check
                       */
 int precheck_expl;   /* 0=warn only
                         1=append impl_igno addresses to start list 
                         2=abort
                       */
 int precheck_expl_replace;  /*
                         0=keep problem characters in target address
                         1=replace by _XX_ (XX=hex ASCII Code)
                       */
 int precheck_unread; /* 0=warn only
                         1=append unreadable addresses to exclusion list
                         2=abort    
                       */
 int precheck_xdev;   /* 0= no special handling of mount points
                         1= add mount points to exclusion list
                         2= add remote mount points (e.g. NFS) to exclusion list
                       */
 int precheck_agetest; /* bit0= enable age test
                          bit5= with bit8 : mistrust directory timestamps
                          bit6= ignore timestamps
                          bit8= check for existence in old_tree
                          bit9= check type in old_tree
                         bit10= check size in old_tree
                         bit11= check crc in old_tree
                         bit12= reserved
                       */
 int precheck_sizetest; /* bit0= apply exclude_size test to single files 
                           bit1= eventually test size of link target 
                       */
 double exclude_size;

 int volume_info;     /* 1=Pass current and total volume numbers to backup
                           script:  -v current_volume -t total_volumes 
                       */

 time_t precheck_age;
 int precheck_max_exempt;
 

 struct FilelistE *uebergrosse;           /* english : oversized */

 /* pacifieres and dialog */
 int compact_num;             /* wether to display form "512.8k" or "525107" */

};


int Backup_plan_init(plan,max_items,max_size,cmd,flag)
struct Backup_plaN *plan;
int max_items;
double max_size;
char *cmd;
int flag;
{
 plan->max_items= max_items;
 plan->max_items_per_tile= 100;
 plan->path_list[0]= 0;
 plan->default_max_size= max_size;
 plan->save_stack_media= 1;
 plan->max_size= max_size;
 plan->max_size_stackpt= -1;
 plan->min_size_fak= 0.8;
 plan->scattering= 0;
 strcpy(plan->cmd,cmd);
 plan->verify_source_filter[0]= 0;
 if(Fileliste_new(&(plan->startliste),0,"",0)<=0)
   return(0);
 if(Fileliste_new(&(plan->auslassen),0,"",0)<=0)
   return(0);
 Fileliste_set_sperrliste(plan->startliste,plan->auslassen,0);
 plan->auslassen_nicht_startliste= 0;
 strcpy(plan->dir_size_cmd,"du -sk");
 plan->target_inp_trenner= 0;
 strcpy(plan->shell_to_use,"/bin/sh");
 plan->accept_all_filetypes= 0;
 plan->hauptdir[0]= 0;
 plan->hauptdir_auto= 0;
 plan->auskunft_an= 1;
 if(Fileliste_single(&(plan->auskunft_name),"",0,0.0,0)<=0)
   return(0);
 plan->auskunft_groesse= 102400.0;
 plan->auskunft_include= 0;
 plan->auskunft_individuell= 1;
 plan->auskunft_skript_root= 0;
 plan->skript_status_file[0]= 0;
 plan->prompt_skript[0]= 0;
 plan->conf_dir[0]= 0;
 plan->level= -1;
 plan->configuration= 0;

 plan->list_output_format= 1|2|4|8;
 strcpy(plan->tmp_dir,"/tmp");
 plan->complete_list_adr[0]= 0;
 plan->append_complete_target[0]= 0;
 plan->dir_list_adr[0]= 0;
 plan->dir_list_target[0]= 0;
 plan->dir_script_adr[0]= 0;
 plan->dir_script_target[0]= 0;
 plan->content_list_adr[0]= 0;
 plan->append_content_target[0]= 0;
 plan->old_tree= NULL;
 plan->old_complete_list_adr[0]= 0;
 plan->old_vanish_list_adr[0]= 0;
 plan->vanish_list_adr[0]= 0;
 plan->vanish_list_target[0]= 0;
 plan->vanish_script_adr[0]= 0;
 plan->vanish_script_target[0]= 0;
 plan->start_time= time(0);
 plan->pre_checksum_size= 0;                    /* 1024.0*1024.0; */

 plan->anz_stuecke= 0;
 plan->stuecke= NULL;
 plan->aktstueck= plan->stuecke;
 plan->output_form= 0;
 plan->collision_test= 1;
 plan->collision_avoid= 1;
 plan->precheck= 0;
 plan->precheck_expl= 0;
 plan->precheck_expl_replace= 0;
 plan->precheck_unread= 0;
 plan->precheck_xdev= 0;
 plan->precheck_agetest= 0;
 plan->precheck_age= 0;
 plan->precheck_sizetest= 0;
 plan->exclude_size= -1.0;
 plan->volume_info= 0;
 plan->precheck_max_exempt= -1;
 if(Fileliste_new(&(plan->uebergrosse),0,"",0)<=0)
   return(0);
 plan->compact_num= 1;
 return(1);
}


int Backup_plan_cleanup(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 struct Backup_stuecK *akt,*next;

 akt= plan->stuecke;
 while(akt!=NULL) {
   next= akt->next;
   free((char *) akt);
   akt= next;
 }
 plan->anz_stuecke= 0;
 plan->stuecke= NULL;
 plan->aktstueck= NULL;
 Fileliste_destroy(&(plan->uebergrosse),0);
 Fileliste_destroy(&(plan->auskunft_name),0);
 Fileliste_destroy(&(plan->auslassen),0);
 Fileliste_destroy(&(plan->startliste),0);
 Fileliste_destroy(&(plan->old_tree),0);
 return(1);
}


int Backup_plan_is_planned(plan,liste,idx,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int idx;
int flag;
{
 struct Backup_stuecK *stueck;
 struct FilelistE *sl;
 char target[4096],s_target[4096];
 int i,j;

 if(Fileliste_target(liste,idx,target,0)<=0)
   return(-1);
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   sl= stueck->liste;
   for(j=0;j<sl->anz;j++){
     if(strcmp(liste->adressen[idx],sl->adressen[j])!=0)
   continue;
     if(Fileliste_target(sl,j,s_target,0)<=0)
   continue;
     if(strcmp(target,s_target)==0)
       return(1);
   }
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 return(0);
}


int Backup_plan_split(plan,liste,idx,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int idx;
int flag;
{
 struct FilelistE *sub=NULL;
 int i,ret,slashende,subloeschen= 0,j,il;
 char target[4096];

 fprintf(stderr,"... splitting %s\n",liste->adressen[idx]);
 /* read file list from directory and process it */
 if(liste->subdirs!=NULL) {
   sub= liste->subdirs[idx];
 }
 if(sub==NULL){
   if(Fileliste_new(&sub,0,liste->main_dir,0)<=0)
     return(-1);
   if(Fileliste_set_main_dir(sub,liste->main_dir,0)<=0)
     return(-1);
   Fileliste_target(liste,idx,target,0);
   if(Fileliste_set_target_dir(sub,target,0)<=0)
     return(-1);
   Fileliste_set_sperrliste(sub,liste->sperrliste,0);
   if(Fileliste_dir(sub,liste->adressen[idx],0)<=0) {
     Fileliste_destroy(&sub,0);
     return(-1);
   }
   subloeschen= 1;
 } 
 /* Do not process names which are ignored by the writer program */
 il= strlen(liste->impl_igno);
 for(i=0;i<sub->anz;i++){
   if(Fileliste_is_impl_igno(liste,sub->adressen[i],0)>0)
     sub->typen[i]= -sub->typen[i];
 }
 
 Fileliste_inherit_target(liste,idx,sub,0);
 Fileliste_set_impl_igno(sub,liste->impl_igno,0);

 ret= Backup_plan_compose(plan,sub,1);
 if(subloeschen)
   Fileliste_destroy(&sub,0);
 if(ret<=0)
   return(ret);
 liste->typen[idx]= -liste->typen[idx];
 return(1);
}


int Backup_plan_stueck_header(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= only test wether a info_skript would be prepended 
*/
{
 struct FilelistE *liste= NULL;
 char auskunft_name[4096];
 int ret,idx,i;

 if(plan->auskunft_include && strlen(plan->auskunft_name->adressen[0]) && 
    plan->auskunft_an) {
   if(flag&1)
      {ret= 1; goto ex;}  /* info_skript would be prepended */
   idx= 0;
   if(plan->anz_stuecke >= 1) {
     /* use a number suffix for info script name */
     ret= Fileliste_copy_item(plan->auskunft_name,0,plan->auskunft_name,1,0);
     if(ret<=0)
       goto ex;
     sprintf(auskunft_name,"%s.%ld",plan->auskunft_name->adressen[0],
                                    plan->anz_stuecke);
     ret= Fileliste_set_address(plan->auskunft_name,1,auskunft_name,0);
     if(ret<=0)
       goto ex;
     idx= 1;
   }
   ret= Backup_stueck_append_file(plan->aktstueck,plan->auskunft_name,idx,0);
   if(ret<=0)
     goto ex;

   /*  
    '/ASKME_level_0'='$conf_dir/ASKME_level_0' \
    ...
    '/ASKME_level_${level-1}'='$conf_dir/ASKME_level_${level-1}' \
   */
   if(plan->configuration==2 && plan->level>0) {
     for(i=0;i<plan->level;i++) {
       sprintf(auskunft_name,"%s/%s_level_%ld",plan->conf_dir,"ASKME",i);
       ret= Fileliste_single(&liste,auskunft_name,0,0.0,0);
       if(ret<=0)
         goto ex;
       ret= Fileliste_read_properties(liste,0);
       if(ret<=0)
         goto ex;
       sprintf(auskunft_name,"/%s_level_%ld","ASKME",i);
       ret= Fileliste_set_target(liste,0,auskunft_name,0);
       if(ret<=0)
         goto ex;
       ret= Backup_stueck_append_file(plan->aktstueck,liste,0,0);
       if(ret<=0)
         goto ex;
       Fileliste_destroy(&liste,0);
     }
   }
 } else {
   if(flag&1)
     {ret= 0; goto ex;} /* no info_skript was prepended */
 }
 ret= 1;
ex:;
 if(liste!=NULL)
   Fileliste_destroy(&liste,0);
 return(ret);
}


int Backup_plan_is_acceptable(plan,type,flag)
struct Backup_plaN *plan;
int type;
int flag;
{
 if(type<=0)
   return(0);
 if(plan->accept_all_filetypes)
   return(1);
 if(type==1 || type==2)
   return(1);
 return(0);
}


int Backup_plan_list_main_dirs(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 if(Fileliste_set_main_dir(plan->startliste,plan->hauptdir,0)<=0)
   return(-1);
 if(Fileliste_set_main_dir(plan->auslassen,plan->hauptdir,0)<=0)
   return(-1);
 if(Fileliste_set_main_dir(plan->auskunft_name,plan->hauptdir,0)<=0)
   return(-1);
 if(Fileliste_set_main_dir(plan->uebergrosse,plan->hauptdir,0)<=0)
   return(-1);
 return(1);
}


int Backup_plan_stacked_max_size(plan,max_size,count,flag)
struct Backup_plaN *plan;
double *max_size;
int *count;
int flag;
/*
 bit0= push max_size,count (else advance reading by one)
 bit1= reset stack to empty
 bit2= re-increment last used stack item
*/
{
 if(flag&4) { /* re-increment last used stack item */
   if(plan->max_size_stackpt>=0)
     plan->max_size_counts[plan->max_size_stackpt]++;
   return(1);
 }
 if(flag&2) { /* reset stack to empty */
   plan->max_size_stackpt= -1;
   *max_size= plan->default_max_size;
   *count= -1;
   return(1);
 }
 if(flag&1) { /* push max_size,count to stack */
   if(plan->max_size_stackpt>=0)
     if(plan->max_size_counts[plan->max_size_stackpt]<=0)
       plan->max_size_stackpt--;
   if(plan->max_size_stackpt+1>=Backup_plan_max_size_stackL) {
     fprintf(stderr,"--- Overflow of max_size stack\n");
     return(0);
   }
   plan->max_size_stackpt++;
   plan->max_size_stack[plan->max_size_stackpt]= *max_size;
   plan->max_size_counts[plan->max_size_stackpt]= *count;
   return(1);
 }
 if(plan->max_size_stackpt>=0)
   if(plan->max_size_counts[plan->max_size_stackpt]<=0)
     plan->max_size_stackpt--;
 if(plan->max_size_stackpt>=0) {
   if(plan->max_size_counts[plan->max_size_stackpt]>0) {
     plan->max_size_counts[plan->max_size_stackpt]--;
     *max_size= plan->max_size_stack[plan->max_size_stackpt];
     *count= plan->max_size_counts[plan->max_size_stackpt];
     return(1);
   }
 }
 *max_size= plan->default_max_size;
 *count= -1;
 return(1);
}


int Backup_plan_compose(plan,liste,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int flag;
/*
 bit0= recursion
 bit1= read properties of liste
 bit2= do not print final message
 bit3= do nothing but to print final message
 bit4= print initial message
*/
{
 struct Backup_stuecK *stueck;
 struct FilelistE *ziel,*sub;
 double frei,best_size,basisgroesse;
 int best,i,fileda,ret,sonderspaltung=0,pass,passes,nichts_teilbares;
 int auslassen,count,stack_count;
 char num[81];

 if(flag&16)
   fprintf(stderr,"\nbeginning to compose backup volumes\n");
 if((flag&2)){
   if(Fileliste_read_properties(liste,
                                8|(!!plan->auslassen_nicht_startliste)<<2)<0)
     return(0);
 }
 if(plan->stuecke==NULL) {
   Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,0);
   if(Backup_stueck_new(&(plan->stuecke),NULL,plan->max_items,plan->max_size,
                        plan->hauptdir,0)<=0)
     return(0);
   stueck= plan->aktstueck= plan->stuecke;
   plan->anz_stuecke= 1;
   if(Backup_plan_stueck_header(plan,0)<=0)
     return(0); 
 }
 stueck= plan->aktstueck;
 if(stueck==NULL)  
   return(0);
 if(flag&8)
   goto final_message;
 ziel= stueck->liste;
 basisgroesse= plan->auskunft_groesse*plan->auskunft_include;

 if(Fileliste_set_main_dir(liste,plan->hauptdir,0)<=0)
  return(-1);

 /* leave out excluded objects */
 if((flag&1)||!plan->auslassen_nicht_startliste)
   for(i=0;i<liste->anz;i++) {
     if(Fileliste_search_file(plan->auslassen,liste->adressen[i],0)>=0)
       liste->typen[i]= -liste->typen[i];
   }

 do {
   ziel= stueck->liste;

   nichts_teilbares= 0;
   if(stueck->bytes<=basisgroesse ||
      sonderspaltung){ /* the backup piece is still too small */
     passes= 1+!!sonderspaltung;
     sonderspaltung= 0;
     for(pass=0;pass<passes;pass++){
       /* at start : find oversized directory and split */;
       best= -1;
       best_size= 0;
       for(i=0;i<liste->anz;i++){
         if(!Backup_plan_is_acceptable(plan,liste->typen[i],0))
       continue;
         /* With pass 0 : only take those which need to get split anyway */
         if(liste->groessen[i]<=
                          plan->max_size-basisgroesse-stueck->bytes*(!!pass))
       continue;
         if(liste->typen[i]!=2) {
           if(pass>0)
       continue;
           fprintf(stderr,"oversized file (%.lf>%.lf) : %s\n",
                         liste->groessen[i],
                         plan->max_size-basisgroesse-stueck->bytes*(!!pass),
                         liste->adressen[i]);
           Fileliste_copy_item(liste,i,plan->uebergrosse,
                               plan->uebergrosse->anz,0);
           liste->typen[i]= -liste->typen[i];
       continue;
         }

         /* check wether the directory contains too many files and would
            make the next piece unnecessarly small */
         if(pass>0){
           count= Fileliste_count_dir_files(liste,liste->adressen[i],0);
           if(count<=0)
       continue;
           if(count>plan->max_items/2)
       continue;
         }

         if(best_size<liste->groessen[i]){ 
           best= i;
           best_size= liste->groessen[i]; 
         }
       }
       if(best>=0){
         ret= Backup_plan_split(plan,liste,best,0);
         if(ret<=0) 
           return(ret);
     break; /* pass */
       }
       if(pass==passes-1 && passes>1)
         nichts_teilbares= 1;
     }
   }

   stueck= plan->aktstueck;
   if(stueck==NULL) return(0);
   ziel= stueck->liste;

   frei= plan->max_size-stueck->bytes;
   best= -1;
   best_size= -1.0;
   fileda= 0;
   for(i=0;i<liste->anz;i++){
     if(!Backup_plan_is_acceptable(plan,liste->typen[i],0))
   continue;
     if(liste->groessen[i]<0)
   continue;
     fileda= 1;
     if(liste->groessen[i]<=frei){
       if(best_size<liste->groessen[i]){ 
         best= i;
         best_size= liste->groessen[i]; 
       }
     }
   }
 if(!fileda) break;

   if(best<0) {
     /* only objects left which do not fit */

     /* check wether the current piece is still too small */
     if(stueck->bytes<plan->min_size_fak*plan->max_size) {
       if(nichts_teilbares) {
         if(plan->scattering && (flag&1) && liste!=plan->startliste) {
           /* delay the remaining objects and process others first */
           char target[4096]; int aufgeschoben= 0;
           /* dirty hack : append all remaining objects to start list */
           for(i=0;i<liste->anz;i++){
             if(!Backup_plan_is_acceptable(plan,liste->typen[i],0))
           continue;
             if(Fileliste_copy_item_hd(liste,i,plan->startliste,
                                       plan->startliste->anz,1)<=0)
               return(0);
             /* mark as "done"  */;
             liste->typen[i]= -liste->typen[i];
             aufgeschoben++;
             fprintf(stderr,"deferred by -scattering -on : %s\n",
                            liste->adressen[i]);
           }
           /* leve this level of recursion */
           if(aufgeschoben>0) 
             return(1);
         }
       }else{
         /* eventually resume the old piece by a part of a large object */
         sonderspaltung= 1;
 continue;
       }
     }

     if(plan->save_stack_media)
       /* if suitable : revoke count as stacked media */
       if(plan->max_size>plan->default_max_size)
         if(plan->default_max_size>stueck->bytes) {
           Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,4);
           plan->max_size= plan->default_max_size;
         }
     if(plan->compact_num)
       Sfile_scale(stueck->bytes,num,7,1e5,1);
     else
       sprintf(num,"%.lf",stueck->bytes);
     fprintf(stderr,"########### %ld  (%ld items  %s bytes)\n",
                    plan->anz_stuecke,stueck->liste->anz,num);
     /* start new piece */
     Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,0);
     if(Backup_stueck_new(&stueck,stueck,plan->max_items,plan->max_size,
                          plan->hauptdir,0)<=0)
       return(0);
     plan->aktstueck= stueck;
     plan->anz_stuecke++;
     if(Backup_plan_stueck_header(plan,0)<=0)
       return(0); 
 continue;
   }

   if(ziel->anz>=plan->max_items) { /*  >=ziel->maxanz */
     /* the current piece is full */
     if(plan->save_stack_media)
       /* if suitable : revoke count as stacked media */
       if(plan->max_size>plan->default_max_size)
         if(plan->default_max_size>stueck->bytes) {
           Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,4);
           plan->max_size= plan->default_max_size;
         }
     if(plan->compact_num)
       Sfile_scale(stueck->bytes,num,7,1e5,1);
     else
       sprintf(num,"%.lf",stueck->bytes);
     fprintf(stderr,"########### %ld  (%ld items  %s bytes)\n",
                    plan->anz_stuecke,stueck->liste->anz,num);
     /* start new piece */
     Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,0);
     if(Backup_stueck_new(&stueck,stueck,plan->max_items,plan->max_size,
                          plan->hauptdir,0)<=0)
       return(0);
     plan->aktstueck= stueck;
     plan->anz_stuecke++;
     if(Backup_plan_stueck_header(plan,0)<=0)
       return(0); 
   }

   if(Fileliste_search_file(plan->auslassen,liste->adressen[best],1)>=0) {
     /* parent address of an excluded path */
     ret= Backup_plan_split(plan,liste,best,0);
     if(ret<=0) return(ret);
   } else {

     /* check wether already included with the same target Liste */
     ret= Backup_plan_is_planned(plan,liste,best,0);
     auslassen= 0;
     if(ret<0) {
       if(verbous)
         fprintf(stderr,"---------- Backup_plan_is_planned() failed: %s\n",
                        liste->adressen[best]);
       auslassen= 1;
     } else if(ret>0) {
       if(verbous)
         fprintf(stderr,"---------- already in backup plan: %s\n",
                        liste->adressen[best]);
       auslassen= 1;
     }
     Backup_stueck_append_file(stueck,liste,best,auslassen);
   }
 } while(1);
final_message:;
 if(!(flag&(1|4))) {
   if(plan->compact_num)
     Sfile_scale(stueck->bytes,num,7,1e5,1);
   else
     sprintf(num,"%.lf",stueck->bytes);
   fprintf(stderr,"########### %ld  (%ld items  %s bytes)\n\n",
                  plan->anz_stuecke,stueck->liste->anz,num);
 }
 return(1);
}


int Backup_plan_collision_test(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= check  plan->startliste rather  than  plan->stuecke
*/
{
 struct Backup_stuecK *stueck;
 struct FilelistE *liste= NULL;
 int i,zaehl=0,w=0,j,ret,ct_flag;

 if(plan->collision_avoid==1)      ct_flag= 2;
 else                              ct_flag= 0;
 if(plan->auslassen_nicht_startliste) ct_flag|= 4;
 else                                 ct_flag|= 8;
 if(flag&1){
   if(Fileliste_new(&liste,plan->startliste->anz+1,plan->hauptdir,0)<=0)
     return(-1);
   w= 0;
   if(plan->auskunft_include && strlen(plan->auskunft_name->adressen[0]) &&
      plan->auskunft_an) {
     Fileliste_copy_item(plan->auskunft_name,0,liste,w++,0);
   }
   for(j=0;j<plan->startliste->anz;j++){
     Fileliste_copy_item(plan->startliste,j,liste,w++,0);
   }  
 }else{
   /* determine total number */
   stueck= plan->stuecke; 
   for(i=0;i<plan->anz_stuecke;i++){
     liste= stueck->liste;
     zaehl+= stueck->liste->anz;
     stueck= stueck->next;
   }
   /* consolidate in one list */
   if(Fileliste_new(&liste,zaehl+1,plan->hauptdir,0)<=0)
     return(-1);
   stueck= plan->stuecke; 
   for(i=0;i<plan->anz_stuecke;i++){
     for(j=0;j<stueck->liste->anz;j++){
       Fileliste_copy_item(stueck->liste,j,liste,w++,0);
     }
     stueck= stueck->next;
   }
 }
 if(Fileliste_set_main_dir(liste,plan->hauptdir,0)<=0)
   return(-1);

 /* collisiontest */
 ret= Fileliste_collision_test(liste,plan->auslassen,ct_flag);
 if(plan->collision_avoid==1){
   /* copy back */
   w= 0;
   if(flag&1){
     if(plan->auskunft_include && strlen(plan->auskunft_name->adressen[0]) &&
        plan->auskunft_an) {
       Fileliste_copy_item(liste,w++,plan->auskunft_name,0,0);
     }
     for(j=0;j<plan->startliste->anz;j++){
       Fileliste_copy_item(liste,w++,plan->startliste,j,0);
     }  
   }else{
     stueck= plan->stuecke;
     for(i=0;i<plan->anz_stuecke;i++){
       for(j=0;j<stueck->liste->anz;j++){
         Fileliste_copy_item(liste,w++,stueck->liste,j,0);
       }
       stueck= stueck->next;
     }
   }
 }

 if(liste!=NULL) Fileliste_destroy(&liste,0);
 if(ret>0 && plan->collision_avoid==2) {
   ret= -2; /* abort because of collision  */
 }
 return(ret);
ex:;
 if(liste!=NULL) Fileliste_destroy(&liste,0);
 return(-1);
}


int Backup_plan_precheck(plan,flag)
struct Backup_plaN *plan;
int flag;
/*

return:
 bit0= ill name found 
 bit1= unreadable found

 bit15= abort !!!
*/
{
 char *rpt,*wpt,target[4096],new_target[4096];
 struct FilelistE *kranke_namen,*kein_zugriff,*mount_punkte;
 struct FilelistE *zu_alt,*junge_ausnahmen,*too_big;
 int i,ret,idx,exclude_anyway= 0;
 double precheck_min_sbpf; /* Minimum required : saved bytes per file */
 double size;

 fprintf(stderr,"performing file checks\n");

 Fileliste_new(&kranke_namen,0,plan->hauptdir,0);
 Fileliste_new(&kein_zugriff,0,plan->hauptdir,0);
 Fileliste_new(&mount_punkte,0,plan->hauptdir,0);
 Fileliste_new(&zu_alt,0,plan->hauptdir,0);
 Fileliste_new(&junge_ausnahmen,0,plan->hauptdir,0);
 Fileliste_new(&too_big,0,plan->hauptdir,0);
 precheck_min_sbpf= plan->max_size/((double) plan->max_items);

 Fileliste_redirect_links(plan->startliste,0);
 Fileliste_precheck(plan->startliste,plan->old_tree,
                    kranke_namen,kein_zugriff,mount_punkte,
                    plan->precheck_age,zu_alt,junge_ausnahmen,
                    plan->precheck_max_exempt,precheck_min_sbpf,
                    plan->exclude_size,too_big,
                    ((plan->precheck_xdev==1)<<1)|
                    ((plan->precheck_xdev==2)<<4)|
                    ((!!plan->precheck)<<2)|
                    ((plan->precheck_agetest&1)<<3)|
                    (plan->precheck_agetest&(32|64|256|512|1024|2048))|
                    ((plan->precheck_sizetest&(1|2))<<12)
                    );
 for(i=0;i<kranke_namen->anz;i++){
   /* check wether the name is already in start list */
   if(Fileliste_search_file(plan->startliste,kranke_namen->adressen[i],0)>=0)
 continue;
   if(plan->precheck_expl==1 && plan->startliste->impl_igno!=NULL){
     fprintf(stderr,"explicitely appended : %s\n",kranke_namen->adressen[i]);

     Fileliste_copy_item(kranke_namen,i,
                         plan->startliste,plan->startliste->anz,0);
     /* if enabled : replace evil characters in target address */
     if(plan->precheck_expl_replace){
       Fileliste_target(plan->startliste,plan->startliste->anz-1,target,1);
       wpt= new_target;
       for(rpt=target;*rpt!=0;rpt++)
         if(strchr(plan->startliste->impl_igno,*rpt)!=NULL){
           sprintf(wpt,"+%-2.2X+",*rpt);
           wpt+= strlen(wpt);
         } else
           *(wpt++)= *rpt;
       *wpt= 0;
       Fileliste_set_target(plan->startliste,plan->startliste->anz-1,
                            new_target,0);
     }
   }else{
     fprintf(stderr,"+++ name will be ignored by backup command : %s\n",
             kranke_namen->adressen[i]);
   }
 }
 for(i=0;i<kein_zugriff->anz;i++){
   if(plan->precheck)
     fprintf(stderr,"+++ no permission to read %s : %s\n",
             (kein_zugriff->typen[i]==2?"directory":"file"),
             kein_zugriff->adressen[i]);
   if(plan->precheck_unread==1){
     fprintf(stderr,"explicitely excluded : %s\n",kein_zugriff->adressen[i]);
     Fileliste_copy_item(kein_zugriff,i,
                         plan->auslassen,plan->auslassen->anz,0);
   }
 }
 for(i=0;i<too_big->anz;i++){
   Type_and_size(too_big->adressen[i],&size,1|((plan->precheck_sizetest&2)<<1));
   fprintf(stderr,
           "+++ single %s too large (%.lf >= -exclude_size %s%.lf) :\n   %s\n",
           (too_big->typen[i]==3?"link target":"file"),size,
           (too_big->typen[i]==3?"-":""),plan->exclude_size,
           too_big->adressen[i]);
     exclude_anyway= 1;
   fprintf(stderr,"explicitely excluded : %s\n",too_big->adressen[i]);
   Fileliste_copy_item(too_big,i,
                       plan->auslassen,plan->auslassen->anz,0);
 }
 for(i=0;i<mount_punkte->anz;i++){
   if(plan->precheck_xdev>0){
     fprintf(stderr,"+++ mountpoint to filesystem of unwanted type : %s\n",
             mount_punkte->adressen[i]);
     fprintf(stderr,"explicitely excluded : %s\n",mount_punkte->adressen[i]);
     Fileliste_copy_item(mount_punkte,i,
                         plan->auslassen,plan->auslassen->anz,0);
   }
 }

 /* react on outcome of age test */
 if(plan->precheck_agetest&1)
   fprintf(stderr,"Options added by -younger_than :\n");
 for(i=0;i<zu_alt->anz;i++){
   fprintf(stderr,"-not %s\n",zu_alt->adressen[i]);
   while(1){ /* remove all occurences of this object from start list */
     idx= Fileliste_search_file(plan->startliste,zu_alt->adressen[i],8);
     if(idx<0)
   break;
     Fileliste_defunct_item(plan->startliste,idx,0);
   }
   if(Fileliste_search_file(plan->startliste,zu_alt->adressen[i],2)>=0){
     Fileliste_copy_item(zu_alt,i,
                         plan->auslassen,plan->auslassen->anz,0);
   }
 }
 Fileliste_garbage(plan->startliste,0);
 /* append junge_ausnahmen to start list if not already listed */;
 for(i=0;i<junge_ausnahmen->anz;i++){
   /* fprintf(stderr,"-yes %s\n",junge_ausnahmen->adressen[i]); */
   if(Fileliste_search_file(plan->startliste,junge_ausnahmen->adressen[i],0)<0)
     Fileliste_copy_item(junge_ausnahmen,i,
                         plan->startliste,plan->startliste->anz,0);
 }
 if(junge_ausnahmen->anz>0)
   fprintf(stderr,"%ld items added with -yes\n",junge_ausnahmen->anz);
 if(plan->precheck_agetest&1)
   plan->auslassen_nicht_startliste= 1;

 ret= (kranke_namen->anz>0)|((kein_zugriff->anz)<<1);
 Fileliste_destroy(&kranke_namen,0);
 Fileliste_destroy(&kein_zugriff,0);
 Fileliste_destroy(&mount_punkte,0);
 Fileliste_destroy(&zu_alt,0);
 Fileliste_destroy(&junge_ausnahmen,0);
 if((ret&1) && plan->precheck_expl==2)
   ret|= 1<<15;
 if((ret&2) && plan->precheck_unread==2)
   ret|= 1<<15;
 return(ret);
}


int Backup_plan_stdout(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= stderr rather than stdout
 bit1= super raw : simple file list, empty lines separate the pieces 
*/
{
 int i,j;
 struct Backup_stuecK *stueck;
 struct FilelistE *liste;
 FILE *out;

 out= stdout;
 if(flag&1) out= stderr;


 if(flag&2) goto super_raw;

 liste= plan->uebergrosse;
 for(j=0;j<liste->anz;j++){
   fprintf(out,"+++++++++++ oversized file (%.lf bytes) : %s\n",
               liste->groessen[j],liste->adressen[j]);
 }
 if(liste->anz>0) fprintf(out,"\n");

 stueck= plan->stuecke; 
 for(i=0;i<plan->anz_stuecke;i++){
   liste= stueck->liste;
   fprintf(out,"########### %ld  (%ld items  %.lf bytes)\n",
               i+1,liste->anz,stueck->bytes);
   for(j=0;j<liste->anz;j++) {
     fprintf(out,"%c %9.lf  %s\n",(liste->typen[j]==2?'D':'F'),
                                  liste->groessen[j],liste->adressen[j]);
   }
   stueck= stueck->next;
 }
 return(1);

super_raw:;
 stueck= plan->stuecke; 
 for(i=0;i<plan->anz_stuecke;i++){
   liste= stueck->liste;
   for(j=0;j<liste->anz;j++) {
     fprintf(out,"%s\n",liste->adressen[j]);
   }
   fprintf(out,"\n");
   stueck= stueck->next;
 }
 return(1);
}


int Backup_plan_info_skript(plan,nr,flag)
struct Backup_plaN *plan;
int nr;
int flag;
/*
 bit0= create .pnr file with a placeholder text for piece number 
*/
{
 int ret,i,extra= 0;
 struct FilelistE *liste;
 struct Lies_popeN *gelesen,gelesen_data;
 struct Backup_stuecK *stueck;
 char *buffer;
 FILE *fp;
 char line[4096],auskunft_name[4096],num[81];

 gelesen= &gelesen_data;

 fp= NULL;
 strcpy(auskunft_name,plan->auskunft_name->adressen[0]);
 if(auskunft_name[0]) {
   if(flag&1) strcat(auskunft_name,".pnr");
   unlink(auskunft_name);
   fp= fopen(auskunft_name,"w");
   if(fp==NULL){
     fprintf(stderr,"can't open info_script '%s' for writing.\n",
                    auskunft_name);
     fprintf(stderr,"reason given by OS : %s\n",strerror(errno));
   }
 } else {
   fp= stdout;
 }
 extra= 0;
 if(fp!=NULL && fp!=stdout) extra= 1;
 else {
   /* included ASKME is no longer supported */
   if(fp==stdout) return(0);
 }

 if(fp==NULL) return(1);

 if(!extra) {
   fprintf(fp,"if test \"$#\" -gt 0 ; then\n\n");
 } else
   fprintf(fp,"#!%s\n\n",plan->shell_to_use);

 /* identification */
 fprintf(fp,"if test \"$1\" = -id -o  \"$1\" = -help -o $# = 0 -o  \"$1\" = -hilfe ; then\n");
 fprintf(fp,"\n");
 fprintf(fp,"  echo 'cd_backup_planer %s : information script'\n",PROG_VERSION);
 fprintf(fp,"  echo\n");
 /* date */
 if(Lies_popen_init("date",gelesen,0)>0) {
   if(gelesen->anz>0)
     fprintf(fp,"  echo 'backup date %s'\n",gelesen->lines[0]);
   Lies_popen_cleanup(gelesen,0);
 }
 if(plan->precheck_agetest&1) {
   fprintf(fp,"  echo '");
   if(plan->level>0)
     fprintf(fp,"-level %ld   ",plan->level);
   fprintf(fp,"-younger_than %s",Sfile_datestr(plan->precheck_age,1|2|256));
   fprintf(fp,"'\n");
 }
 if(flag&1){
   fprintf(fp,"  echo 'BACKUP_PLANER_PART_MESSAGE_PLACEHOLDER'\n");
 } else if(nr>0) {
   fprintf(fp,"  echo 'this is backup part no %ld of %ld'\n",
              nr,plan->anz_stuecke);
 }else{
   fprintf(fp,"  echo 'this backup consists of %ld parts'\n",
              plan->anz_stuecke);
 }
 fprintf(fp,"  echo\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");

 fprintf(fp,"if test \"$1\" = -help -o $# = 0 ; then\n");
 fprintf(fp,"  echo \"usage: $0  [option] file [file ...]\"\n");
 fprintf(fp,
 "  echo 'options:'\n");
 fprintf(fp,
 "  echo '  -help   prints this text'\n");
 fprintf(fp,
 "  echo '  -hilfe  gibt diesen Text in deutscher Sprache aus'\n");
 fprintf(fp,
 "  echo '          (prints this text in german language)'\n");
 fprintf(fp,
 "  echo '  -id     prints a message about this backup'\n");
 fprintf(fp,
 "  echo '  -count  prints the number of parts'\n");
 fprintf(fp,
 "  echo '  -list   prints the parts and their items'\n");
 fprintf(fp,
 "  echo '  -levelall lookup all levels (else: only youngest match)'\n");
 fprintf(fp,
 "  echo '  -suball lookup all sub items and levels (else: one per part)'\n");
 fprintf(fp,
 "  echo '  -fast   do not look for sub items'\n");
 fprintf(fp,
 "  echo '  files   (respectively directories) are looked up in the'\n");
 fprintf(fp,
 "  echo '          parts'\"'\"' item lists. If the filename or a'\n");
 fprintf(fp,
 "  echo '          directory above is found then the part number, the'\n");
 fprintf(fp,
 "  echo '          item and the filename re printed. If a filename below'\n");
 fprintf(fp,
 "  echo '          (a sub item) is found, then the action depends on'\n");
 fprintf(fp,
 "  echo '          -suball and -fast'\n");
 fprintf(fp,
 "  echo\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");

 fprintf(fp,"if test \"$1\" = -hilfe ; then\n");
 fprintf(fp,"  echo \"Anwendung: $0  [option] file [file ...]\"\n");
 fprintf(fp,
 "  echo 'Optionen :'\n");
 fprintf(fp,
 "  echo '  -hilfe  gibt diesen Text aus'\n");
 fprintf(fp,
 "  echo '  -help   prints this text in english language'\n");
 fprintf(fp,
 "  echo '          (gibt diesen Text in englischer Sprache aus)'\n");
 fprintf(fp,
 "  echo '  -id     gibt eine Meldung ueber diesen Backup aus'\n");
 fprintf(fp,
 "  echo '  -count  gibt die Zahl der Teile aus'\n");
 fprintf(fp,
 "  echo '  -list   gibt die Teile und ihre Filelisten aus'\n");
 fprintf(fp,
 "  echo '  -levelall alle -level zeigen (sonst: nur den juengsten Fund)'\n");
 fprintf(fp,
 "  echo '  -suball alle Unteradressen zeigen (sonst: eine pro Stueck)'\n");
 fprintf(fp,
 "  echo '  -fast   nicht nach Unteradressen suchen'\n");
 fprintf(fp,
 "  echo '  files   (bzw. Directories) werden in den Filelisten der'\n");
 fprintf(fp,
 "  echo '          Stuecke gesucht. Wenn ein Filename oder ein darueber-'\n");
 fprintf(fp,
 "  echo '          liegendes Directory gefunden  wird, dann werden die'\n");
 fprintf(fp,
 "  echo '          Stuecknummer, die gefundene Adresse und der Filename'\n");
 fprintf(fp,
 "  echo '          ausgegeben. Wenn eine Fileadresse gefunden wird, die'\n");
 fprintf(fp,
 "  echo '          unterhalb des Filenamens liegt (Unteradresse), dann'\n");
 fprintf(fp,
 "  echo '          haengt die Reaktion von -suball und -fast ab.'\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");


 fprintf(fp,"if test \"$1\" = -id -o \"$1\" = -help -o $# = 0 -o \"$1\" = -hilfe ;");
 fprintf(fp," then exit 0 ; fi\n\n");

 fprintf(fp,"if test \"$1\" = -count ; then\n");
 fprintf(fp,"  echo %ld\n",plan->anz_stuecke);
 fprintf(fp,"  exit 0\n");
 fprintf(fp,"fi\n\n");

 /* list funktions */
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   fprintf(fp,"cd_liste_%ld() {\n",i+1);
   ret= Backup_stueck_report(&buffer,"cat ",stueck->liste,-1,-1,3);
   if(ret<=0) return(ret);
   fprintf(fp,"%s",buffer);
   free(buffer);
   fprintf(fp,"}\n\n");
   stueck= stueck->next;
 if(stueck==NULL) break;
 }

 /* search funktions */
 fprintf(fp,"such_in_cd() {\n");
 fprintf(fp,"such_cd=$1\n");
 fprintf(fp,"for i in $(cd_liste_$such_cd)\n");
 fprintf(fp,"do\n");
 fprintf(fp,"  if grep -q '^'\"$i\"/ <<+\n");
 fprintf(fp,"$name\n");
 fprintf(fp,"+\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test \"$i\" = \"$name\"\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"done\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");
 fprintf(fp,"such_in_cd_sub() {\n");
 fprintf(fp,"such_cd=$1\n");
 fprintf(fp,"if test $sub = 1 ; then subsperr=0 ; else subsperr=1 ; fi\n");
 fprintf(fp,"for i in $(cd_liste_$such_cd)\n");
 fprintf(fp,"do\n");
 fprintf(fp,"  if grep -q '^'\"$i\"/ <<+\n");
 fprintf(fp,"$name\n");
 fprintf(fp,"+\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test \"$i\" = \"$name\"\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test $suball = 1 -o $subsperr = 0\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    if grep -q '^'\"$name\" <<+\n");
 fprintf(fp,"$i\n");
 fprintf(fp,"+\n");
 fprintf(fp,"    then\n");
 fprintf(fp,"      ok=0\n");
 fprintf(fp,"      if grep -q '^'\"$name\"/ <<+\n");
 fprintf(fp,"$i\n");
 fprintf(fp,"+\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        ok=1\n");
 fprintf(fp,"      elif grep -q '^'\"$name\"'$' <<+\n");
 fprintf(fp,"$i\n");
 fprintf(fp,"+\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        ok=1\n");
 fprintf(fp,"      fi\n");
 fprintf(fp,"      if test $ok = 1\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"        fund=1\n");
 fprintf(fp,"        subsperr=1\n");
 fprintf(fp,"        if test $suball = 0 ; then return ; fi\n");
 fprintf(fp,"      fi\n");
 fprintf(fp,"    fi\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"done\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");

 /* process arguments */
 fprintf(fp,"if test \"$1\" = -list ; then\n");
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   if(plan->compact_num)
     Sfile_scale(stueck->bytes,num,7,1e5,1);
   else
     sprintf(num,"%.lf",stueck->bytes);
   fprintf(fp,"  echo '########## part %ld  (%ld items  estim. %s bytes)'\n",
              i+1,stueck->liste->anz,num);
   fprintf(fp,"  cd_liste_%ld\n",i+1);
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 fprintf(fp,"  exit 0\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");

 fprintf(fp,"level_opt=\n");
 fprintf(fp,"sub=1\n");
 fprintf(fp,"if test \"$1\" = -fast\n");
 fprintf(fp,"then\n");
 fprintf(fp," shift 1\n");
 fprintf(fp,"  sub=0\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");
 fprintf(fp,"suball=0\n");
 fprintf(fp,"if test \"$1\" = -suball\n");
 fprintf(fp,"then\n");
 fprintf(fp,"  shift 1\n");
 fprintf(fp,"  suball=1\n");
 fprintf(fp,"  level_opt=\"${level_opt} -suball\"\n");
 fprintf(fp,"  sub=1\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"levelall=0\n");
 fprintf(fp,"if test \"$1\" = -levelall\n");
 fprintf(fp,"then\n");
 fprintf(fp,"  shift 1\n");
 fprintf(fp,"  levelall=1\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"nolevel=0\n");
 fprintf(fp,"if test \"$1\" = -nolevel\n");
 fprintf(fp,"then\n");
 fprintf(fp,"  shift 1\n");
 fprintf(fp,"  nolevel=1\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");
 fprintf(fp,"fund=0\n");
 fprintf(fp,"for name in \"$@\"\n");
 fprintf(fp,"do\n");
 fprintf(fp,"  namefund=0\n");
 fprintf(fp,"  for cd in");
 for(i=0;i<plan->anz_stuecke;i++)
   fprintf(fp," %ld",i+1);
 fprintf(fp,"\n");
 fprintf(fp,"  do\n");
 fprintf(fp,"    if test $namefund = 0\n");
 fprintf(fp,"    then\n");
 fprintf(fp,"      if test $sub = 1\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        such_in_cd_sub $cd\n");
 fprintf(fp,"      else\n");
 fprintf(fp,"        such_in_cd $cd\n");
 fprintf(fp,"      fi\n");
 fprintf(fp,"    fi\n");
 fprintf(fp,"  done\n");
 fprintf(fp,"done\n");
 fprintf(fp,"if test $fund = 1 -a $levelall = 0 ; then exit 0 ; fi\n");
 fprintf(fp,"if test $nolevel = 1 ; then exit 1 ; fi\n");

 fprintf(fp,"if test \"$level_opt\" = \"\" \n");
 fprintf(fp,"then\n");
 fprintf(fp,"  level_opt=\"-nolevel\" \n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");
 if(plan->configuration==2 && plan->level>0) {
   fprintf(fp,"for lev in");
   for(i=plan->level-1;i>=0;i--) 
     fprintf(fp," %ld",i);
   fprintf(fp,"\n");
   fprintf(fp,"do\n");
   fprintf(fp,"  if $(dirname $0)/%s_level_$lev $level_opt \"$@\" 2>&1 >/dev/null\n",
              "ASKME");
   fprintf(fp,"  then\n");
   fprintf(fp,"    fund=1\n");
   fprintf(fp,"    echo '+++ on backup level '$lev\n");
   fprintf(fp,"    $(dirname $0)/%s_level_$lev $level_opt \"$@\"\n",
              "ASKME");
   fprintf(fp,"    if test $levelall = 0\n");
   fprintf(fp,"    then\n");
   fprintf(fp,"      exit 0\n");
   fprintf(fp,"    fi\n");
   fprintf(fp,"  fi\n");
   fprintf(fp,"done\n");
 }

 fprintf(fp,"if test $fund = 1 ; then exit 0 ; fi\n");
 fprintf(fp,"exit 1\n");
 fprintf(fp,"\n");
 
 if(!extra) fprintf(fp,"fi\n");

 if(fp!=stdout && fp!=stdout && strlen(auskunft_name)>0) {
   /* give x-permissions */
   sprintf(line,"chmod a+x %s",auskunft_name);
   system(line);
   fclose(fp);
 }

 return(0);
}


int Backup_plan_write_line_layout(plan,fp,flag)
struct Backup_plaN *plan;
FILE *fp;
int flag;
/*
 bit0= write fixed line layout of directory list
*/
{
 fprintf(fp,"#=# line layout : ");
 if(flag&1) {
   fprintf(fp,"type chmod chown chgrp date size crc adr \n");
 } else {
   if(plan->list_output_format&2)
     fprintf(fp,"type ");
   if(plan->list_output_format&4)
     fprintf(fp,"size ");
   if(plan->list_output_format&8)
     fprintf(fp,"crc ");
   fprintf(fp,"adr \n");
 }
 return(1);
}


int Backup_plan_list_header(plan,fp,title,flag)
struct Backup_plaN *plan;
FILE *fp;
char *title;
int flag;
/*
 bit0= write fixed line layout of directory list
*/
{
 char hname[4096],pwd[4096];

 fprintf(fp,"#=# %s\n",title);
 if(gethostname(hname,sizeof(hname))==-1)
   strcpy(hname,"localhost"); 
 fprintf(fp,"#=# hostname %s\n",hname);
 fprintf(fp,"#=# date     %s\n",Sfile_datestr(plan->start_time,1|2|256));
 Backup_plan_write_line_layout(plan,fp,flag&1);
 if(getcwd(pwd,sizeof(pwd))!=NULL) {
   if(strcmp(pwd,"/")==0)
     pwd[0]= 0;
   fprintf(fp,"#=# next line tells current working directory\n%s/\n",pwd);
 }
 return(1);
}


int Backup_plan_complete_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= generate dirlist (with status information) 
 bit1= with bit 0 : generate dir_script (chown, chmod)
 bit2= unlink target only, do not create new file
*/
{
 int ret,i,j,popened= 0,linecount= 0,sparse_adr= 2,list_header_flag= 0;
 int contains_banned_files= 0;
 char *listdescription="",*list_adr="",*option_name="";
 FILE *fp= NULL;

 if(flag&1) {
   if(flag&2) {
     listdescription= "script to restore directory properties";
     list_adr= plan->dir_script_adr;
     option_name= "-dir_script_adr";
   } else {
     listdescription= "complete backup area directory list";
     list_adr= plan->dir_list_adr;
     option_name= "-dir_list_adr";
     list_header_flag|= 1;
   }
 } else {
   listdescription= "complete backup area file list";
   list_adr= plan->complete_list_adr;
   option_name= "-complete_list_adr";
 }
 if(list_adr[0]==0)
   return(2);
 if(!(flag&4))
   fprintf(stderr,"generating  %s %s\n",option_name,list_adr);

 ret= Open_listfile(&fp,list_adr,"w",1|(2*!!(flag&4)));
 if(flag&4)
   goto ex;
 popened= (ret==2);
 if(fp==NULL) {
   fprintf(stderr,"--- %s %s\n",listdescription,option_name,
           "Failed to write %s (%s).",
           "Program aborted.");
   ret= 0; goto ex;
 }
 if((flag&1) && (flag&2)) {
   Backup_plan_script_header(plan,fp,listdescription,
                             "SCDBACKUP_RESTORE_ROOT",1);
 } else {
   Backup_plan_list_header(plan,fp,listdescription,list_header_flag);
 }
 Print_pacifier(0,"  0          file items written",1|4);

 if((flag&3)==3)
   sparse_adr= 0;
 for(i=0;i<plan->startliste->anz;i++) {
   contains_banned_files= 0;
   if(!plan->auslassen_nicht_startliste) {
     if(Fileliste_search_file(plan->startliste->sperrliste,
                              plan->startliste->adressen[i],0)>=0)
 continue;
     if(Fileliste_search_file(plan->startliste->sperrliste,
                              plan->startliste->adressen[i],1|4|8)>=0)
       contains_banned_files= 128;
   }
   ret= Fileliste_leanfind(plan->startliste,i,
                           dev_dummy,fp,&linecount,
                           ((flag&3)<<5)|sparse_adr|16|
                            (plan->precheck_xdev<<2)|
                            contains_banned_files|
                            (plan->list_output_format<<8));
   if(ret<0)
     {ret= 0; goto ex;}
 }
 fprintf(fp,"#=# end\n");
 Sfile_fclose(fp); fp= NULL;
 Print_pacifier(linecount,"",1|2|8);
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(fp!=NULL){
   if(popened)
     pclose(fp);
   else
     Sfile_fclose(fp);
 }
 return(ret);
}


int Backup_plan_content_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit1= unlink target only, do not create new file
*/
{
 int ret,i,j,popened,linecount= 0;
 FILE *fp= NULL;
 struct Backup_stuecK *stueck;

 if(plan->content_list_adr[0]==0)
   return(2);
 if(!(flag&2))
   fprintf(stderr,"\ngenerating -content_list_adr %s\n",
                  plan->content_list_adr);

 ret= Open_listfile(&fp,plan->content_list_adr,"w",1|(flag&2));
 if(flag&2)
   return(ret);
 popened= (ret==2);
 if(fp==NULL) {
   fprintf(stderr,"--- %s %s\n",
           "Failed to write content file list (-content_list_adr).",
           "Program aborted.");
   ret= 0; goto ex;
 }
 Backup_plan_list_header(plan,fp,"backup content list",0);
 Print_pacifier(0,"  0          file items written",1|4);
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++) {
   fprintf(fp,"#=# volume %ld\n",i);
   for(j=0;j<stueck->liste->anz;j++) {
     ret= Fileliste_leanfind(stueck->liste,j,
                             dev_dummy,fp,&linecount,
                             2|16|(plan->list_output_format<<8));
     if(ret<0)
       {ret= 0; goto ex;}
   }
   stueck= stueck->next;
   if(stueck==NULL) 
 break;
 }
 fprintf(fp,"#=# end\n");
 Sfile_fclose(fp); fp= NULL;
 Print_pacifier(linecount,"",1|2|8);
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(fp!=NULL){
   if(popened)
     pclose(fp);
   else
     Sfile_fclose(fp);
 }
 return(ret);
}


int Backup_plan_script_header(plan,fpout,scriptdescription,rootvar,flag)
struct Backup_plaN *plan;
FILE *fpout;
char *scriptdescription;
char *rootvar;
int flag;
/*
 bit0= include a superuser stopper
*/
{
 char hname[4096];

 fprintf(fpout,"#!%s\n#\n",plan->shell_to_use);
 fprintf(fpout,"# %s\n",scriptdescription);
 if(gethostname(hname,sizeof(hname))==-1)
   strcpy(hname,"localhost"); 
 fprintf(fpout,"# hostname %s\n",hname);
 fprintf(fpout,"# date     %s\n",Sfile_datestr(plan->start_time,1|2|256));
 if(flag&1) {
   fprintf(fpout,"# Use With Extreme Care !\n#\n");
   fprintf(fpout,"if test \"$SCDBACKUP_RESTORE_ALLOW_SUPERUSER\" = Y\n");
   fprintf(fpout,"then\n");
   fprintf(fpout,"  dummy=dummy\n");
   fprintf(fpout,"elif test \"$UID\" = 0\n");
   fprintf(fpout,"then\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo 'You are superuser. This is quite dangerous.'\n");
   fprintf(fpout,"  echo 'If you have to insist being superuser, execute'\n");
   fprintf(fpout,"  echo 'export SCDBACKUP_RESTORE_ALLOW_SUPERUSER=Y'\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo 'Sie sind Superuser. Das ist arg gefaehrlich.'\n");
   fprintf(fpout,
               "  echo 'Wenn Sie darauf bestehen muessen, fuehren Sie aus'\n");
   fprintf(fpout,"  echo 'export SCDBACKUP_RESTORE_ALLOW_SUPERUSER=Y'\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  exit 1\n");
   fprintf(fpout,"fi\n\n");
 } else {
   fprintf(fpout,"# Use With Care\n#\n");
 }
 fprintf(fpout,"if test \"$%s\" = \"\"\n",rootvar);
 fprintf(fpout,"then\n");
 fprintf(fpout,"  echo\n");
 fprintf(fpout,"  echo 'Variable %s not set or empty'\n",rootvar);
 fprintf(fpout,"  echo 'If you are restoring your backup area under'\n");
 fprintf(fpout,"  echo 'directory /restore/home , execute :'\n");
 fprintf(fpout,"  echo '  export %s=/restore/home'\n",rootvar);
 fprintf(fpout,"  echo\n");
 fprintf(fpout,"  echo 'Variable %s ist leer. Wenn'\n",rootvar);
 fprintf(fpout,"  echo 'Sie ihren Backupbereich unterhalb von Directory'\n");
 fprintf(fpout,"  echo '/restore/home wiederherstellen, fuehren Sie aus :'\n");
 fprintf(fpout,"  echo '  export %s=/restore/home'\n",rootvar);
 fprintf(fpout,"  echo\n");
 fprintf(fpout,"  exit 2\n");
 fprintf(fpout,"else\n");
 fprintf(fpout,"  if cd \"$%s\"\n",rootvar);
 fprintf(fpout,"  then\n");
 fprintf(fpout,"    dummy=dummy\n");
 fprintf(fpout,"  else\n");
 fprintf(fpout,"    echo '--- cannot enter directory'\n");
 fprintf(fpout,"    echo '--- kann nicht ins Directory gehen'\n");
 fprintf(fpout,"    echo '%s='$%s\n",rootvar,rootvar);
 fprintf(fpout,"    exit 3\n");
 fprintf(fpout,"  fi\n");
 fprintf(fpout,"fi\n\n");
 return(1);
}


int Backup_plan_vanished_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= write shell script
*/
{
 FILE *fpin= NULL,*fpout= NULL;
 char adr[4096],*spt,initial_wd[4096],*adr_arg,*option_name;
 int ret= 0,in_popened= 0,out_popened= 0,hflag= 0;

 if(flag&1) {
   adr_arg= plan->vanish_script_adr;
   option_name= "-vanish_script_adr";
   hflag= 1;
 } else {
   adr_arg= plan->vanish_list_adr;
   option_name= "-vanish_list_adr";
 }
 if(plan->old_complete_list_adr[0]==0 || adr_arg[0]==0) 
   {ret= 2; goto ex;}
 fprintf(stderr,"generating %s %s %s\n",
                option_name,plan->old_complete_list_adr,adr_arg);

 if(strchr(adr_arg,'/')==NULL && plan->conf_dir[0]
    && plan->level>=0) {
   sprintf(adr,"%s/level_%ld/%s",plan->conf_dir,plan->level,adr_arg);
   strcpy(adr_arg,adr);
 }
 out_popened= Open_listfile(&fpout,adr_arg,"w",1);
 if(out_popened!=2)
   out_popened= 0;
 if(fpout==NULL) {
   fprintf(stderr,"--- %s %s%s%s\n","Failed to write list of vanished files",
                  "(",option_name,"). Program aborted.");
   ret= 0; goto ex;
 }
 if(!(flag&1))
   Backup_plan_list_header(plan,fpout,"vanished file objects list",0);

 if(strchr(plan->old_complete_list_adr,'/')==NULL && plan->conf_dir[0]) {
   if(plan->level>=1) {
     sprintf(adr,"%s/level_%ld/%s",plan->conf_dir,plan->level-1,
                                   plan->old_complete_list_adr);
     strcpy(plan->old_complete_list_adr,adr);
   } else {
/*      plan->old_complete_list_adr[0]= 0; */
     if(flag&1) {
       fprintf(fpout,"#!%s\n#\n",plan->shell_to_use);
       fprintf(fpout,
            "# -level 0 dummy of script that removes vanished files\n");
       fprintf(fpout,
            "# See incremental backups of higher levels for the real thing\n");
       fprintf(fpout,"exit 0\n");
     }
     ret= 2; goto ex;
   }
 }
 in_popened= Open_listfile(&fpin,plan->old_complete_list_adr,"r",1);
 if(in_popened!=2)
   in_popened= 0;
 if(fpin==NULL) {
   fprintf(stderr,"--- %s %s\n",
           "Failed to read complete file list (from ",option_name,").",
           "Program aborted.");
   ret= 0; goto ex;
 }

process_old_list:;
 if(flag&1) {
   Backup_plan_script_header(plan,fpout,
                    "script to remove files that vanished between two backups",
                    "SCDBACKUP_RESTORE_ROOT",1);
 }
 ret= Fileliste_read_serialized_tree(NULL,fpin,plan->old_tree,fpout,
                                     initial_wd,NULL,"",
                                     hflag|(plan->list_output_format<<8));
 fprintf(stderr,"\n");
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 if(fpout!=NULL && ret>0 && !flag&1)
   fprintf(fpout,"#=# end\n");
 if(in_popened) {
   if(fpin!=NULL)
     pclose(fpin);
 } else
   Sfile_fclose(fpin);
 if(out_popened) {
   if(fpout!=NULL)
     pclose(fpout);
 } else
   Sfile_fclose(fpout);
 return(ret);
}


int Backup_plan_read_old_list(plan,adr,initial_wd,initial_wd_list,flag)
struct Backup_plaN *plan;
char *adr;
char initial_wd[4096];
struct FilelistE **initial_wd_list;
int flag;
/*
 bit4= delete files from liste tree rather than to insert them
 bit5= this initial_wd and initial_wd_list are already set
 bit6= demand complete list and no content list
 bit14= choose target addresses rather than source addresses
       use text given with  initial_wd  as prefix_dir
*/
{
 FILE *fpin= NULL;
 int ret= 0,popened= 0;
 char prefix_dir[4096];

 if(adr[0]==0)
   return(0);

 popened= Open_listfile(&fpin,adr,"r",1);
 if(popened!=2)
   popened= 0;
 if(fpin==NULL) {
   fprintf(stderr,"--- %s%s%s\n",
           "Failed to read file list (from -load_old_tree : ",adr,
           "). Program aborted.");
   ret= 0; goto ex;
 }
 fprintf(stderr,"List file opened :\n%s\n",adr);
 ret= Fileliste_read_serialized_tree(plan->old_tree,fpin,NULL,NULL,
                                     initial_wd,initial_wd_list,"",
                                     2|(flag&(16|32|64|(1<<14))));
 if(ret<=0)
   goto ex;
 if((flag&64) && ret==2) 
   goto ex;

 ret= 1;
ex:;
 if(popened) {
   if(fpin!=NULL)
     pclose(fpin);
 } else
   Sfile_fclose(fpin);
 return(ret);
}


int Backup_plan_load_old_tree(plan,old_list_adr,old_vanish_adr,tree_prefix,
                              flag)
struct Backup_plaN *plan;
char *old_list_adr;
char *old_vanish_adr;
char *tree_prefix;
int flag;
/*
 >>> bit14= choose target addresses rather than source addresses
       use tree_prefix
*/
{
 int ret= 0,demand_complete= 0,not_from_scratch= 0,level_mode= 0,i;
 int target_mode=0;
 char adr[4096],initial_wd[4096];
 struct FilelisteE *initial_wd_list= NULL;

 initial_wd[0]= 0;
 if(old_list_adr[0]==0 || old_vanish_adr[0]==0)
   return(2);
 strcpy(adr,old_list_adr);
 fprintf(stderr,"performing -load_old_tree %s %s\n",
                plan->old_complete_list_adr,plan->old_vanish_list_adr);

 /* first try wether this is a complete list */
 if(strchr(old_list_adr,'/')==NULL && plan->conf_dir[0] && plan->level>=0) {
   if(plan->level==0)
     return(2);
   sprintf(adr,"%s/level_%ld/%s",plan->conf_dir,plan->level-1,old_list_adr);
   level_mode= 1;
 }
 if(plan->old_tree==NULL) {
   ret= Fileliste_new(&(plan->old_tree),0,plan->hauptdir,Fileliste_hash_sizE>0);
   if(ret<=0)
     goto ex;
   not_from_scratch= 0;
 } else
   not_from_scratch= 32;
 demand_complete= (plan->level>1 && level_mode)<<6;
 target_mode= (flag&(1<<14));
 if(target_mode)
   strcpy(initial_wd,tree_prefix);
 ret= Backup_plan_read_old_list(plan,adr,initial_wd,&initial_wd_list,
                             target_mode | not_from_scratch | demand_complete);
 if(ret<=0)
   goto ex;
 if(demand_complete && ret==2) { 
   fprintf(stderr,"\nGenerating old_tree from layered content/vanish lists\n");
   /* reconstruct complete list tree from content lists and vanish lists */
   for(i=0;i<plan->level;i++) {
     /* read content list from level */;
     sprintf(adr,"%s/level_%ld/%s",plan->conf_dir,i,old_list_adr);
     ret= Backup_plan_read_old_list(plan,adr,initial_wd,&initial_wd_list,
                                    target_mode | not_from_scratch);
     if(ret<=0)
       goto ex;
     not_from_scratch= 32;
     if(i>0) {
       /* read vanished list from level */;
       sprintf(adr,"%s/level_%ld/%s",plan->conf_dir,i,old_vanish_adr);
       ret= Backup_plan_read_old_list(plan,adr,initial_wd,&initial_wd_list,
                                      16 | target_mode | not_from_scratch);
       if(ret<=0)
         goto ex;
     }
   }
 }
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(ret<=0) {
   fprintf(stderr,"--- failed : -load_old_tree %s %s\n",
                  old_list_adr,old_vanish_adr);
 }
 return(ret);
}


int Backup_plan_skript(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int ret,i,j;
 struct FilelistE *liste;
 struct Backup_stuecK *stueck;
 char *buffer,*auskunft_name,pwd[4096],shell_safe[4096],num[81];
 FILE *fp;

 auskunft_name= plan->auskunft_name->adressen[0];

 /* ---------------------- Information --------------------- */

 if(plan->auskunft_an) {
   ret= Backup_plan_info_skript(plan,0,0);
   if(ret) return(ret);
   if(plan->auskunft_individuell){
     ret= Backup_plan_info_skript(plan,0,1);
     if(ret) return(ret);
   }
 }
 

 /* ------------------- Production -------------------- */

 if(plan->output_form==1 || plan->output_form==4) {
   ret= Backup_plan_stdout(plan,(plan->output_form==4)<<1);
   if(ret<=0) ret= 1;
   else       ret= 0;
   return(ret);
 }

 fp= stdout;
 fprintf(fp,"#!%s\n\n",plan->shell_to_use);

 if(getcwd(pwd,sizeof(pwd))!=NULL)
   fprintf(fp,"\ncd %s\n\n",Text_shellsafe(pwd,shell_safe,0));

 fprintf(fp,"export SCDBACKUP_ACTIVE_SCRIPT\n");
 fprintf(fp,"export SCDBACKUP_PIECENO SCDBACKUP_PIECES\n");
 fprintf(fp,"SCDBACKUP_ACTIVE_SCRIPT=\"$0\"\n");
 fprintf(fp,"SCDBACKUP_PIECES=%ld\n",plan->anz_stuecke);

 fprintf(fp,"is_enabled() {\n");
 fprintf(fp,"  part=\"$1\"\n");
 fprintf(fp,"  if test \"$part\" = \"\" ; then return 0 ; fi\n");
 if(strlen(plan->skript_status_file)){
   fprintf(fp,"  if test \"$part\" = AUTO ; then\n");
   fprintf(fp,"    part=\n");
   fprintf(fp,"    if test -r %s ; then\n",plan->skript_status_file);
   fprintf(fp,"      part=$(cat %s)\n",plan->skript_status_file);
   fprintf(fp,"    fi\n");
   fprintf(fp,"    if test \"$part\" = \"\" ; then\n");
   fprintf(fp,"      echo 'no usable status file : '%s >&2\n",
              plan->skript_status_file);  
   fprintf(fp,"      echo 'number of start volume needed' >&2\n");
   fprintf(fp,"      exit 1\n");
   fprintf(fp,"    fi\n");
   fprintf(fp,"  fi\n");
 }  
 fprintf(fp,"  if test \"$part\" -lt 0 ; then\n");
 fprintf(fp,"    if test \"$part\" -eq -\"$2\" ; then return 0 ; fi\n");
 fprintf(fp,"    if test -%ld -gt \"$part\" ; then exit 11 ; fi\n",
            plan->anz_stuecke);
 fprintf(fp,"    return 1\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test \"$part\" -le \"$2\" ; then return 0 ; fi\n");
 fprintf(fp,"  if test %ld -lt \"$part\" ; then exit 11 ; fi\n",
            plan->anz_stuecke);
 fprintf(fp,"  return 1\n");
 fprintf(fp,"}\n\n");

 liste= plan->uebergrosse;
 for(i=0;i<liste->anz;i++){
   fprintf(fp,"echo '+++++++++++ oversized file (%.lf bytes) : %s'\n",
              liste->groessen[i],liste->adressen[i]);
 }
 if(liste->anz>0) fprintf(fp,"\n");

 fprintf(fp,"if test \"$1\" = -count ; then\n");
 fprintf(fp,"  echo %ld\n",plan->anz_stuecke);
 fprintf(fp,"  exit 0\n");
 fprintf(fp,"fi\n\n");

 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   fprintf(fp,"if is_enabled \"$1\" %ld ; then\n",i+1);
   if(strlen(plan->skript_status_file))
     fprintf(fp,"echo %ld >%s\n",i+1,plan->skript_status_file);
   fprintf(fp,"SCDBACKUP_PIECENO=%ld\n",i+1);
   if(plan->compact_num)
     Sfile_scale(stueck->bytes,num,7,1e5,1);
   else
     sprintf(num,"%.lf",stueck->bytes);
   fprintf(fp,"echo '########## part %ld of %ld (%ld items  %s bytes)'\n",
          i+1,plan->anz_stuecke,stueck->liste->anz,num);
   if(plan->prompt_skript[0]!=0) {
     fprintf(fp,"if test -x %s ; then\n",plan->prompt_skript);
     fprintf(fp,"  %s\n",plan->prompt_skript);
     fprintf(fp,"fi\n");
   }
   if(plan->precheck_agetest&1) {
     fprintf(fp,"echo '           ");
     if(plan->level>0)
       fprintf(fp,"-level %ld   ",plan->level);
     fprintf(fp,"-younger_than %s",Sfile_datestr(plan->precheck_age,1|2|256));
     fprintf(fp,"'\n");
   }
   if(plan->auskunft_an && plan->auskunft_individuell &&
     strlen(auskunft_name)>0) {
     fprintf(fp,"sed -e 's/BACKUP_PLANER_PART_MESSAGE_PLACEHOLDER/");
     fprintf(fp,"this is backup part no %ld of %ld/' < %s.pnr > %s.%ld\n",
             i+1,plan->anz_stuecke,auskunft_name,auskunft_name,i+1);
     fprintf(fp,"chmod a+x %s.%ld\n",auskunft_name,i+1);
     fprintf(fp,"cp %s.%ld %s\n",auskunft_name,i+1,auskunft_name);
   }
   if(plan->output_form==5 || plan->output_form==6) {
     if(plan->volume_info) {
       ret= Backup_stueck_report_pl(fp,plan->verify_source_filter,plan->cmd,
                                    stueck->liste,i+1,plan->anz_stuecke,
                                    plan->max_items_per_tile,plan->path_list,
                                    plan->output_form&255);
     } else {
       ret= Backup_stueck_report_pl(fp,plan->verify_source_filter,plan->cmd,
                                    stueck->liste,-1,-1,
                                    plan->max_items_per_tile,plan->path_list,
                                    plan->output_form&255);
     }
     if(ret<=0)
       return(18);
   } else {
     if(plan->volume_info) {
       ret= Backup_stueck_report(&buffer,plan->cmd,
                                 stueck->liste,i+1,plan->anz_stuecke,
                                 plan->output_form&255);
     } else {
       ret= Backup_stueck_report(&buffer,plan->cmd,
                                 stueck->liste,-1,-1,plan->output_form&255);
     }
     if(ret<=0)
       return(19);
     fprintf(fp,"%s",buffer);
     free(buffer);
   }

   if(auskunft_name[0]!=0) {
     fprintf(fp,"rm %s.%ld\n",auskunft_name,i+1);
     fprintf(fp,"\n");
   }
   fprintf(fp,"  if test \"$cmd_return\" = \"10\" ; then\n");
   fprintf(fp,"    echo aborted\n");
   fprintf(fp,"    exit 10\n");
   fprintf(fp,"  fi\n");
   fprintf(fp,"fi\n\n");
   fprintf(fp,"\n");
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 if(strlen(plan->skript_status_file))
   fprintf(fp,"rm %s\n",plan->skript_status_file);
 fprintf(fp,"SCDBACKUP_PIECENO=\n");
 if(plan->auskunft_an && plan->auskunft_individuell &&
    strlen(auskunft_name)>0) {
   fprintf(fp,"sed -e 's/BACKUP_PLANER_PART_MESSAGE_PLACEHOLDER/");
   fprintf(fp,"this backup consists of %ld parts/' < %s.pnr > %s\n",
               plan->anz_stuecke,auskunft_name,auskunft_name);
   fprintf(fp,"chmod u+x %s\n",auskunft_name);
 }
 fprintf(fp,"\n");
 fprintf(fp,"echo '########## parts in total : %ld'\n",plan->anz_stuecke);
 if(plan->precheck_agetest&1) {
   fprintf(fp,"echo '           ");
   if(plan->level>0)
     fprintf(fp,"-level %ld   ",plan->level);
   fprintf(fp,"-younger_than %s",Sfile_datestr(plan->precheck_age,1|2|256));
   fprintf(fp,"'\n");
 }
 fprintf(fp,"echo 'backup script ended normally'\n");
 if(fp!=NULL && fp!=stdout && fp!=stdout) fclose(fp);

 return(0);
}


int Backup_plan_set_skript_shell(argc,argv,newshell,flag)
int argc;
char **argv;
char *newshell;
int flag;
{
 FILE *fpr= NULL,*fpw= NULL;
 char newname[4096],line[4096];
 int i;

 for(i=0;i<argc;i++) {
   fpr= fopen(argv[i],"r");
   if(fpr==NULL) {
     fprintf(stderr,"Cannot open file for reading : %s\n   %s\n",
                    argv[i],strerror(errno));
     goto loopend;
   }
   if(Sfile_fgets(line,sizeof(line),fpr)==NULL) 
     goto loopend;
   if(strncmp(line,"#!",2)!=0)
     goto loopend;
   sprintf(newname,"%s.new",argv[i]);
   fpw= fopen(newname,"w");
   if(fpr==NULL) {
     fprintf(stderr,"Cannot open file for writing : %s\n   %s\n",
                    newname,strerror(errno));
     goto loopend;
   }
   fprintf(fpw,"#!%s\n",newshell);
   while(Sfile_fgets(line,sizeof(line),fpr)!=NULL) {
     fprintf(fpw,"%s\n",line);
   }
   fclose(fpw);
   fpw= NULL;
   sprintf(line,"%s.old",argv[i]);
   if(rename(argv[i],line)==-1){
     fprintf(stderr,"Cannot rename %s to %s\n   %s",
                    argv[i],line,strerror(errno));
     goto loopend;
   }
   if(rename(newname,argv[i])==-1){
     fprintf(stderr,"Cannot rename %s to %s\n   %s",
                    newname,argv[i],strerror(errno));
     goto loopend;
   }
   if(unlink(line)==-1) {
     fprintf(stderr,"Cannot remove %s \n   %s",
                    line,strerror(errno));
     goto loopend;
   }
   sprintf(line,"chmod a+x '%s'",argv[i]);
   system(line);
loopend:;
   if(fpr!=NULL) fclose(fpr);
   fpr= NULL;
   if(fpw!=NULL) fclose(fpw);
   fpw= NULL;
 }
 return(1); 
}


int Backup_plan_level_create(plan,level,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= it is no fault if the directory already exists
*/
{
 int ret;
 char filename[4096],level_dir[4096];
 struct stat stbuf;

 if(level>0) {
   /* >>> test wether previous level exists */;
 }

 sprintf(level_dir,"%s/%s_%ld",plan->conf_dir,"level",level);
 if(Cd_backup_planer_call_staT(level_dir,&stbuf)!=-1) {
   if(((stbuf.st_mode)&S_IFMT) != S_IFDIR) {
     fprintf(stderr,"-level %ld : file object should be a directory\n");
     fprintf(stderr,"%s\n",level_dir);
     ret= 0; goto ex;
   }
   if(flag&1)
     goto after_creation;
 }

 ret= mkdir(level_dir,0x1FF);
 if(ret==-1) {
   fprintf(stderr,"-level %ld : creation of directory failed :\n",level);
   perror(level_dir);
   ret= 0; goto ex;
 }
after_creation:;
 ret= Backup_plan_level_write(plan,level,0,0,"",0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_level_write(plan,level,run_date,volumes,askme_adr,flag)
struct Backup_plaN *plan;
int flag;
time_t run_date;
int volumes;
char *askme_adr;
{
 int ret,errn;
 char filename[4096],level_dir[4096],datestring[80],volumetext[80];
 char errmsg[SfileadrL];

 sprintf(level_dir,"%s/%s_%ld",plan->conf_dir,"level",level);

 /* format and write run date */
 sprintf(filename,"%s/%s",level_dir,"last_run_date");
 strcpy(datestring,Sfile_datestr(run_date,1|2|256));
 ret= Sfile_string_to_file(filename,datestring,1);
 if(ret<=0) {
   fprintf(stderr,"-level %ld : cannot write date string\n",level);
perr:;
   perror(filename);
   ret= 0; goto ex;
 }
 /* number of volumes */
 sprintf(filename,"%s/%s",level_dir,"number_of_volumes");
 sprintf(volumetext,"%ld",volumes);
 ret= Sfile_string_to_file(filename,volumetext,1);
 if(ret<=0) {
   fprintf(stderr,"-level %ld : cannot write number of volumes\n",level);
   goto perr;
 }
 sprintf(filename,"%s/%s_level_%ld",plan->conf_dir,"ASKME",plan->level);
 if(askme_adr[0]!=0) {
   /* copy ASKME to conf_dir */
   ret= Sfile_copy_file(askme_adr,filename,&errn,errmsg,0);
   if(ret<=0) {
     fprintf(stderr,"-level %ld : cannot copy info script\n",level);
     goto perr;
   } 
 } else {
   /* create ASKME dummy which exits 1 */;
   ret= Sfile_string_to_file(filename,"exit 1\n",1);
   if(ret<=0) {
     fprintf(stderr,"-level %ld : cannot write default info script\n",level);
     goto perr;
   } 
 }
 chmod(filename,S_IRUSR | S_IWUSR | S_IXUSR | 
                S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_conf_dir_default(plan,progname,flag)
struct Backup_plaN *plan;
char *progname;
int flag;
{
 char *cpt;
 int l;

 cpt= strrchr(progname,'/');
 if(cpt==NULL) {
   strcpy(plan->conf_dir,".");
 } else {
   l= cpt-progname;
   if(l>0)
     strncpy(plan->conf_dir,progname,l);
   plan->conf_dir[l]= 0;
 }
 strcat(plan->conf_dir,"/default_conf_dir");
 return(1);
}


int Backup_plan_conf_dir_check(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int ret;
 struct stat stbuf;

 /*  test wether -conf_dir and at least level_${level-1} exists */
 if(Cd_backup_planer_call_staT(plan->conf_dir,&stbuf)==-1) {
   fprintf(stderr,"-level ... : cannot access configuration directory\n");
perr:;
   perror(plan->conf_dir);
   ret= 0; goto ex;
 }
 if(((stbuf.st_mode)&S_IFMT) != S_IFDIR) {
   fprintf(stderr,"-level ... : -conf_dir is not a directory\n");
   fprintf(stderr,"%s\n",plan->conf_dir);
   ret= 0; goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_conf_create(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
/*
 bit0= update : do not create -conf_dir directory
*/
{
 int ret,i,first_line= 1;
 char filename[4096],pwd[4096],text[80],*buf= NULL,*cpt,*wpt;
 static char defaultargs[][81]= {
   "@@@start safe_incremental",
   "-load_old_tree","content_file_list","vanished_file_list",
   "-changetest_options","type:size:crc:adr",
   "-list_output_format","type:size:crc:adr",
   "-content_list_adr",
               "/added_by_scdbackup/content_file_list.gz=content_file_list.gz",
   "-vanish_script_adr","content_file_list",
                           "/added_by_scdbackup/VANISHED.sh.gz=VANISHED.sh.gz",
   "-vanish_list_adr","content_file_list",
             "/added_by_scdbackup/vanished_file_list.gz=vanished_file_list.gz",
   "-dir_list_adr",
               "/added_by_scdbackup/complete_dir_list.gz=complete_dir_list.gz",
   "@@@end safe_incremental",
   "@end_default_args@"
 };

 /* create -conf_dir */
 if(plan->conf_dir[0]==0) {
   Backup_plan_conf_dir_default(plan,argv[0],0);
 }
 if(flag&1) {
   ret= Backup_plan_conf_dir_check(plan,0);
   if(ret<=0)
     goto ex;
   fprintf(stderr,"Updating configuration directory :\n");
 } else {
   ret= mkdir(plan->conf_dir,0x1FF);
   if(ret==-1) {
     fprintf(stderr,
          "%s -level -create_configuration : creation of directory failed :\n",
             argv[0]);
     strcpy(filename,plan->conf_dir);
perr:;
     perror(filename);
     ret= 0; goto ex;
   }
   fprintf(stderr,"Configuration directory created :\n");
 } 
 fprintf(stderr,"%s\n",plan->conf_dir);

 /* write specific_arguments */
 sprintf(filename,"%s/%s",plan->conf_dir,"specific_arguments");

 /* write defaults for handling of vanished files */
 ret= Sfile_string_to_file(filename,"-working_directory",1|4);
 if(ret<=0)
   goto write_failed;
 first_line= 0;
 getcwd(pwd,sizeof(pwd));
 ret= Sfile_string_to_file(filename,pwd,1|2|4);
 if(ret<=0)
   goto write_failed;

 for(i=0;strcmp(defaultargs[i],"@end_default_args@")!=0; i++) {

   /* >>> safe and sparse default setup */;

   if(strncmp(defaultargs[i],"@@@",3)==0)
 continue;
   ret= Sfile_string_to_file(filename,defaultargs[i],1|((!first_line)<<1)|4);
   if(ret<=0) {
write_failed:;
     fprintf(stderr,
         "%s -level -create_configuration : recording of arguments failed :\n",
             argv[0]);
     goto perr;
   }
   first_line= 0;
 }

 /* write arguments */
 for(i=1;i<argc;i++) {
   /* skip arguments that interfere with the configuration system */
   if(strcmp(argv[i],"-conf_dir")==0 ||
      strcmp(argv[i],"-level")==0 ||
      strcmp(argv[i],"-younger_than")==0) {
     i++;
 continue;
   }

   /* escape NewLine */
   if(buf!=NULL)
     Smem_freE(buf);
   buf= Smem_malloC(strlen(argv[i])*2+1);
   if(buf==NULL) {
     fprintf(stderr,"%s out of memory\n", argv[0]);
     ret= -1; goto ex;
   }
   wpt= buf;
   for(cpt= argv[i];*cpt!=0;cpt++) {
     if(*cpt=='\\') {
       *(wpt++)= '\\';
       *(wpt++)= '\\';
     } else if(*cpt=='\n') {
       *(wpt++)= '\\';
       *(wpt++)= 'n';
     } else
       *(wpt++)= *cpt;
   }
   *(wpt)= 0;

   ret= Sfile_string_to_file(filename,buf,1|((!first_line)<<1)|4);
   if(ret<=0)
     goto write_failed;
   first_line= 0;
 }

 /* create and init level directores up to plan's level : level_#  */
 if(plan->level<0)
   plan->level= 0;
 for(i=0;i<=plan->level;i++) {
   ret= Backup_plan_level_create(plan,i,(flag&1));
   if(ret<=0)
     goto ex;
 }

 sprintf(filename,"%s/%s",plan->conf_dir,"highest_level");
 sprintf(text,"%ld",plan->level);
 ret= Sfile_string_to_file(filename,text,1);
 if(ret<=0) {
   fprintf(stderr,
       "%s -level -create_configuration : cannot write highest level number\n",
           argv[0]);
   goto perr;
 }

 ret= 1;
ex:;
 if(buf!=NULL)
   Smem_freE(buf);
 return(ret);
}


int Backup_plan_prepend_level_address(plan,adr,level,flag)
struct Backup_plaN *plan;
char adr[4096];
int level;
int flag;
{
 char name[4096];

 if(adr[0]) {
   if(strchr(adr,'/')==NULL) {
     if(plan->conf_dir[0] && level>=0)
       sprintf(name,"%s/level_%ld/%s",plan->conf_dir,level,adr);
     else 
       sprintf(name,"%s/%s",plan->tmp_dir,adr);
     strcpy(adr,name);
   }
 }
 return(1);
}


int Backup_plan_conf_use(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
/*
 bit0= allow some command line arguments to override or extend configured 
       arguments
*/
{
 int ret,i,first_line= 1,hargc= 0,level,configuration,add_to_level;
 int highest_level;
 char filename[4096],text[80],**hargv= NULL,level_dir[4096],date_mark[80];
 char cmd[4096],*yargv[3],conf_dir[4096],*cpt,*wpt,name[4096];
 struct stat stbuf;

 if(plan->conf_dir[0]==0) 
   Backup_plan_conf_dir_default(plan,argv[0],0);

 level= plan->level;
 configuration= plan->configuration;
 strcpy(conf_dir,plan->conf_dir);

 /*  test wether -conf_dir and at least level_${level-1} exists */
 ret= Backup_plan_conf_dir_check(plan,0);
 if(ret<=0)
   goto ex;
 sprintf(filename,"%s/highest_level",conf_dir);
 ret= Sfile_file_to_int(filename,&highest_level,1);
 if(ret<=0) {
   fprintf(stderr,"%s -level : cannot read highest level mark",argv[0]); 
perr:;
   perror(filename);
   ret= 0; goto ex;
 }
 if(level<0) { /* last used level or automatically incremented level */
   if(level==-1)
     add_to_level= 1;
   else
     add_to_level= 0;
   if(highest_level<0) {
     fprintf(stderr,"%s -level -next : illegal content %ld in file\n",
                    argv[0],highest_level); 
     fprintf(stderr,"%s\n",filename);
     ret= 0; goto ex;
   }
   level= highest_level+add_to_level;
   plan->level= level;
 }

 if(highest_level<level-1) {
   fprintf(stderr,"%s -level %ld is too high, at most possible : -level %ld\n",
                    argv[0],level,highest_level+1); 
   ret= 0;
   goto ex;
 }

 if(level>0) {
   sprintf(level_dir,"%s/level_%ld",conf_dir,level-1);
   sprintf(filename,"%s/%s",level_dir,"last_run_date");
   ret= Sfile_file_to_text(filename,date_mark,sizeof(date_mark),0);
   if(ret<=0) {
     fprintf(stderr,
           "%s -level %ld : cannot read previous level's date mark\n",
           argv[0],level);
     goto perr;
   }
 }

 /* eventually create level directory level_$level */
 ret= Backup_plan_level_create(plan,level,1);
 if(ret<=0)
   goto ex;

 /* reprogram with arguments from specific_arguments 
    plus -younger_than from level_$level/last_run_date
    if bit0 : plus argv again 
  */
 /* read specific_arguments */
 sprintf(filename,"%s/%s",plan->conf_dir,"specific_arguments");

 ret= Sfile_read_argv(argv[0],filename,&hargc,&hargv,0); 
 if(ret<=0) {
   fprintf(stderr,
           "%s -level %ld : cannot read recorded arguments\n",
           argv[0],level);
   goto perr;
 }

 for(i=1;i<hargc;i++) {
   /* unescape NewLine */;
   wpt= hargv[i];
   for(cpt= hargv[i]; (*cpt)!=0; cpt++) {
     if(*cpt=='\\') {
       if(*(cpt+1)=='\\') {
         cpt++;
       } else if(*(cpt+1)=='n') {
         *(wpt++)= '\n';
         cpt++;
   continue;
       }
     }
     *(wpt++)= *cpt;
   }
   *wpt= 0;
 }

 /* copy cmd because it is cleared during _init() */
 strcpy(cmd,plan->cmd);

 Backup_plan_cleanup(plan,0);
 ret= Backup_plan_init(plan,plan->max_items,plan->max_size,cmd,0);
 if(ret<=0)
   goto ex;
 /* preliminary setting of conf_dir and level. will be repeated. */
 strcpy(plan->conf_dir,conf_dir);
 plan->level= level;
 ret= Backup_plan_setup(plan,hargc,hargv,1);
 if(ret!=0)
   {ret= 0; goto ex;}
 if(level>0) {
   yargv[0]= argv[0];
   yargv[1]= "-younger_than";
   yargv[2]= date_mark;
   ret= Backup_plan_setup(plan,3,yargv,1);
   if(ret!=0)
     {ret= 0; goto ex;}
 }
 if(flag&1) {
   ret= Backup_plan_setup(plan,argc,argv,1);
   if(ret!=0)
     {ret= 0; goto ex;}
 }
 /* close the plan */
 yargv[0]= argv[0];
 ret= Backup_plan_setup(plan,1,yargv,0);
 if(ret!=0)
   {ret= 0; goto ex;}

 strcpy(plan->conf_dir,conf_dir);
 plan->level= level;
 plan->configuration= configuration;

 fprintf(stderr,"\nIncremental backup :\n");
 fprintf(stderr,"   -conf_dir %s\n",plan->conf_dir);
 fprintf(stderr,"   -level    %ld",plan->level);
 if(plan->level>0)
   fprintf(stderr,  "    ( -younger_than %s )",date_mark);
 fprintf(stderr,"\n\n");

 ret= 1;
ex:;
 Sfile_destroy_argv(&hargc,&hargv,0);
 return(ret);
}


int Backup_plan_conf_update(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int ret;
 char filename[4096],text[80];

 ret= Backup_plan_level_write(plan,plan->level,
                              plan->start_time,
                              plan->anz_stuecke,
                              plan->auskunft_name->adressen[0],0);
 if(ret<=0)
   goto ex;
 sprintf(filename,"%s/%s",plan->conf_dir,"highest_level");
 sprintf(text,"%ld",plan->level);
 ret= Sfile_string_to_file(filename,text,1);
 if(ret<=0) {
   fprintf(stderr,
           "-level %ld : cannot write highest level number\n",plan->level);
   perror(filename);
   ret= 0; goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_append_and_exclude(plan,liste,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int flag;
/*
 bit0= map to root directory
*/
{
char absolut[4096],relativ[4096],punktslash[4096],haupt[4096],ziel[4096];
char auskunft_name[4096],*cpt;
FILE *fp;
int ret;

 Fileliste_target(liste,0,auskunft_name,1);
 /* touch the file */
 if((fp=fopen(auskunft_name,"w"))!=NULL) 
   fclose(fp);
 ret= Fileliste_append(plan->auslassen,auskunft_name,1);
 if(ret<=0) 
   return(ret);
 cpt= strrchr(auskunft_name,'/');
 if(cpt==NULL) cpt= auskunft_name;
 else          cpt++;
 if(flag&1)
   strcpy(ziel,cpt);
 else
   strcpy(ziel,auskunft_name);
 if(ziel[0]=='/'){
   strcpy(absolut,ziel);
   strcpy(relativ,ziel+1);
   sprintf(punktslash,".%s",ziel);
 }else if(ziel[0]=='.' && ziel[1]=='/') {
   strcpy(absolut,ziel+1);
   strcpy(relativ,ziel+2);
   strcpy(punktslash,ziel);
 }else{
   sprintf(absolut,"/%s",ziel);
   strcpy(relativ,ziel);
   sprintf(punktslash,"./%s",ziel);
 }
 ret= Fileliste_append(plan->auslassen,absolut,1|4);
 if(ret<=0)
   return(ret);
 ret= Fileliste_append(plan->auslassen,relativ,1|4);
 if(ret<=0)
   return(ret);
 ret= Fileliste_append(plan->auslassen,punktslash,1|4);
 if(ret<=0)
   return(ret);
 if(plan->hauptdir[0] && strcmp(plan->hauptdir,"/")!=0){
   sprintf(haupt,"%s%s",plan->hauptdir,absolut);
   ret= Fileliste_append(plan->auslassen,haupt,1|4);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Backup_plan_append_simple(plan,source,target,flag)
struct Backup_plaN *plan;
char *source;
char *target;
int flag;
{
 int ret;
 struct FilelistE *liste= NULL;

 ret= Fileliste_single(&liste,source,0,0.0,0);
 if(ret<=0)
   goto ex;
 ret= Fileliste_set_target(liste,0,target,0);
 if(ret<=0)
   goto ex;
 ret= Backup_plan_compose(plan,liste,2|4);
 if(ret<=0)
   goto ex;
 
 ret= 1;
ex:;
 Fileliste_destroy(&liste,0);
 return(ret);
}


int Backup_plan_generate_lists(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0-3= stage
    0= after setup of arguments
    1= after file check 
*/
{
 int ret,stage;
 struct FilelistE *liste= NULL;

 stage= (flag&15);
 if(stage==0) {

   ret= Backup_plan_load_old_tree(plan,plan->old_complete_list_adr,
                                       plan->old_vanish_list_adr,"",0);
   if(ret<=0)
     return(-22);
   ret= Backup_plan_vanished_list(plan,0);
   if(ret<=0)
     {ret= -19; goto ex;}
   ret= Backup_plan_vanished_list(plan,1);
   if(ret<=0)
     {ret= -24; goto ex;}

   /* complete_list_adr */
   Backup_plan_complete_list(plan,0);
   /* dir_list_adr */
   Backup_plan_complete_list(plan,1);
   /* dir_script_adr */
   Backup_plan_complete_list(plan,3);

   /* remove any old result files */
   /* content_list_adr */
   Backup_plan_content_list(plan,2);

 } else if(stage==1) {

#ifdef NIX
   /* complete_list_adr */
   ret= Backup_plan_complete_list(plan,0);
   if(ret<=0)
     return(-21);
   /* dir_list_adr */
   ret= Backup_plan_complete_list(plan,1);
   if(ret<=0)
     return(-25);
   /* dir_script_adr */
   ret= Backup_plan_complete_list(plan,3);
   if(ret<=0)
     return(-26);
#endif /* NIX */

   ret= Backup_plan_content_list(plan,0);
   if(ret<=0)
     {ret= -15; goto ex;}
   if(plan->complete_list_adr[0]!=0 && plan->append_complete_target[0]!=0) {
     ret= Backup_plan_append_simple(plan,plan->complete_list_adr,
                                         plan->append_complete_target,0);
     if(ret<=0)
       goto ex;
   }
   if(plan->dir_list_adr[0]!=0 && plan->dir_list_target[0]!=0) {
     ret= Backup_plan_append_simple(plan,plan->dir_list_adr,
                                         plan->dir_list_target,0);
     if(ret<=0)
       goto ex;
   }
   if(plan->content_list_adr[0]!=0 && plan->append_content_target[0]!=0) {
     ret= Backup_plan_append_simple(plan,plan->content_list_adr,
                                         plan->append_content_target,0);
     if(ret<=0)
       goto ex;
   }
   if(plan->vanish_list_adr[0]!=0 && plan->vanish_list_target[0]!=0) {
     ret= Backup_plan_append_simple(plan,plan->vanish_list_adr,
                                         plan->vanish_list_target,0);
     if(ret<=0)
       goto ex;
   }
   if(plan->vanish_script_adr[0]!=0 && plan->vanish_script_target[0]!=0) {
     ret= Backup_plan_append_simple(plan,plan->vanish_script_adr,
                                         plan->vanish_script_target,0);
     if(ret<=0)
       goto ex;
   }
   if(plan->dir_script_adr[0]!=0 && plan->dir_script_target[0]!=0) {
     ret= Backup_plan_append_simple(plan,plan->dir_script_adr,
                                         plan->dir_script_target,0);
     if(ret<=0)
       goto ex;
   }
 }
 ret= 1;
ex:;
 Fileliste_destroy(&liste,0);
 return(ret);
}



/* bug patch A40701 . -level -create_configuration produced double backup.
   Thanks to Samuele E. Locatelli. */
int Backup_plan_conf_main(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
{
 int ret,extendable= 1;

 if(plan->configuration==1 || plan->configuration==3) {
   ret= Backup_plan_conf_create(plan,argc,argv,(plan->configuration==3));
   if(ret<=0)
     return(-1);
   /* switch to -configuration use */
   extendable= 0; /* do not read arguments again */
   plan->configuration= 2;
 }

 if(plan->configuration==2) {
   Fileliste_destroy(&(plan->old_tree),0);
   ret= Backup_plan_conf_use(plan,argc,argv,extendable);
   if(ret<=0 || plan->startliste->anz<=0)
     return(-1);
 }
 /* prepend level_directory or tmp_dir */
 Backup_plan_prepend_level_address(plan,plan->complete_list_adr,
                                   plan->level,flag);
 Backup_plan_prepend_level_address(plan,plan->dir_list_adr,
                                   plan->level,flag);
 Backup_plan_prepend_level_address(plan,plan->dir_script_adr,
                                   plan->level,flag);
 Backup_plan_prepend_level_address(plan,plan->content_list_adr,
                                   plan->level,flag);
 return(1);
}


int Backup_plan_match_checksums(plan,cmd,listadr,
                                s_name,s_old,s_young,flag)
struct Backup_plaN *plan;
char *cmd,*listadr;
char *s_name,*s_old,*s_young;
int flag;
/*
  bit8= compare name
  bit9= compare date limits

 bit12= work with MD5 rather than CRC32
*/
{
 int ret,found_line= 0,valid,i,rec_count,use_handle= 512,hashtype;
 FILE *fp= NULL;
 char checksum_text[256],*rec_name;
 unsigned char md5[16],rec_md5[16];
 unsigned long crc32,rec_crc32;
 double read_size,rec_size,size,match_size;
 time_t old_limit= 0,young_limit= 0,rec_date= 0;
 struct CheckrecorD *checklist= NULL,*c_rec;
 char *handle= NULL;   /* should be void *  */

 ret= Checkrecord_read_list(&checklist,listadr,&rec_count,1);
 if(ret<=0)
   goto ex;
 
 fprintf(stderr,"Loaded %ld checksum records from %s\n",rec_count,listadr);

 /* decode time limits */;
 if(strcmp(s_old,"-")==0) {
   flag&= ~256;
 } else {
   ret= Zeitdefinition(s_old,&valid,&old_limit,0);
   if(ret<=0) {
bad_limit:;
     fprintf(stderr,
          "\n%s : NOT A USABLE TIME RANGE.  UNBRAUCHBARE ZEITANGABE :\n",cmd);
     fprintf(stderr, "%s %s\n",s_old,s_young);
     {ret= 0; goto ex;}
   }
 }
 if(strcmp(s_young,"-")==0) {
   flag&= ~512;
 } else {
   ret= Zeitdefinition(s_young,&valid,&young_limit,0);
   if(ret<=0) 
     goto bad_limit;
 }
 if(old_limit>young_limit) {
   rec_date= old_limit;
   old_limit= young_limit;
   young_limit= rec_date;
 }

 read_size= 0;
 for(c_rec= checklist;c_rec!=NULL;c_rec= c_rec->next) {
   /* crawl along the list */
   Checkrecord_get_name(c_rec,&rec_name,0);
   Checkrecord_get_size(c_rec,&rec_size,0);
   Checkrecord_get_date(c_rec,&rec_date,0);
   Checkrecord_get_checksum(c_rec,&hashtype,&rec_crc32,rec_md5,0);
   if((hashtype==1) != !!(flag&(1<<12)))
 continue;
   if(flag&(1<<8))
     if(strcmp(rec_name,s_name)!=0)
 continue;
   if(flag&(1<<9)) {
     if(rec_date<old_limit)
 continue;
     if(rec_date>young_limit)
 continue;
   }
   /* crawl along the file */
   size= rec_size-read_size;
   if(size>0) {
     if(flag&(1<<12))
       ret= Determine_file_checksum("-",stdin,&handle,md5,md5,&size,
                                    1|2|4|8|use_handle|(1<<12));
     else
       ret= Determine_file_checksum("-",stdin,&handle,&crc32,&crc32,&size,
                                    1|2|4|8|use_handle);
     if(ret<=0)
        goto ex;
     use_handle|= 256;  /* mark handle as already open for next cycle */
     if(size<rec_size-read_size) {
       read_size+= size;
 break;
     }
     read_size= rec_size;
   }
   /* compare checksums */
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       if(md5[i]!=rec_md5[i])
     break;
     if(i<16)
 continue;
   } else {
     if(crc32!=rec_crc32)
 continue;
   }

   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       sprintf(checksum_text+2*i,"%2.2x",rec_md5[i]);
   } else
     sprintf(checksum_text,"%X",rec_crc32);
   fprintf(stderr,"\n\nFound matching record :\n%s %s %.lf %s\n\n",
                  rec_name,Sfile_datestr(rec_date,1|2|256),
                  rec_size,checksum_text);
   match_size= read_size;
   found_line= 1;
 }
 if(!found_line) {
   fprintf(stderr,
       "\n\n%s : NO MATCHING RECORD FOUND. KEIN PASSENDER RECORD GEFUNDEN :\n",
        cmd);
   fprintf(stderr,"  %s %s %s\n\n",s_name,s_old,s_young);
   {ret= 0; goto ex;}
 } else {
   if(read_size-match_size)
     fprintf(stderr,"Note: read %.lf surplus bytes\n\n",read_size-match_size);
   fprintf(stderr,
             "OK.  Matching Record found . Passender Record gefunden .\n\n");
 }

 ret= (!!found_line);
ex:;
 if(handle!=NULL)
   free((char *) handle);
 Sfile_fclose(fp,0);
 return(ret);
}


int Backup_plan_find_checksum(plan,cmd,listadr,
             s_name,s_old,s_young,s_size,s_checksum,
             f_name,f_date,f_size,f_checksum,f_md5,f_crc32,
             flag)
struct Backup_plaN *plan;
char *cmd,*listadr;
char *s_name,*s_old,*s_young;
double s_size;
char *s_checksum;
char f_name[256],f_date[256];
double *f_size;
char f_checksum[256];
unsigned char f_md5[16];
unsigned long *f_crc32;
int flag;
/*
 bit1= only print matching record lines but do not check stdin

  bit8= compare name
  bit9= compare date limits
 bit10= compare size 
 bit11= compare checksum
*/
{
 int ret,found_line= 0,valid,i,is_md5;
 FILE *fp= NULL;
 char buf[256],age_limit[14],young_limit[14];
 char read_name[256],read_date[256],read_length[256],read_sum[256];
 double read_size;
 time_t moment;

 fp= Sfile_fopen(listadr,"r",0);
 if(fp==NULL)
   {ret= 0; goto ex;}

 /* decode and reformat time limits */;
 if(strcmp(s_old,"-")==0) {
   strcpy(age_limit,"-");
 } else {
   ret= Zeitdefinition(s_old,&valid,&moment,0);
   if(ret<=0) {
bad_limit:;
     fprintf(stderr,
          "\n%s : NOT A USABLE TIME RANGE.  UNBRAUCHBARE ZEITANGABE :\n",cmd);
     fprintf(stderr, "%s %s\n",s_old,s_young);
     {ret= 0; goto ex;}
   }
   strcpy(age_limit,Sfile_datestr(moment,1|2|256));
 }
 if(strcmp(s_young,"-")==0) {
   strcpy(young_limit,"-");
 } else {
   ret= Zeitdefinition(s_young,&valid,&moment,0);
   if(ret<=0) 
     goto bad_limit;
   strcpy(young_limit,Sfile_datestr(moment,1|2|256));
 }
 if(strcmp(age_limit,"-")!=0 && strcmp(young_limit,"-")!=0 && 
    strcmp(age_limit,young_limit)>0) {
   strcpy(buf,age_limit);
   strcpy(age_limit,young_limit);
   strcpy(young_limit,buf);
 }

 /* try to find matching line (keep last match) */;
 while(Sfile_fgets(buf,sizeof(buf)-1,fp)!=NULL) {
   read_name[0]= read_date[0]= 0;
   sscanf(buf,"%s %s %s %s",read_name,read_date,read_length,read_sum);
   read_size= -1;
   sscanf(read_length,"%lf",&read_size);
   if(read_name[0]==0 || strlen(read_date)!=13)
 continue;
   if(flag&(1<<8)) {
     if(strcmp(s_name,"-")!=0 && strcmp(read_name,s_name)!=0)
 continue;
   }
   if(flag&(1<<9)) {
     if(strcmp(age_limit,"-")!=0 && strcmp(age_limit,read_date)>0)
 continue;
     if(strcmp(young_limit,"-")!=0 && strcmp(young_limit,read_date)<0)
 continue;
   }
   if(flag&(1<<10)) {
     if(read_size!=s_size)
 continue;
   }
   if(flag&(1<<11)) {
     if(strcmp(s_checksum,"-")!=0 && strcmp(s_checksum,read_sum)!=0)
 continue;
   }
   if(flag&2)
     fprintf(stdout,"%s\n",buf);
   strcpy(f_name,read_name);
   strcpy(f_date,read_date);
   *f_size= read_size;
   strcpy(f_checksum,read_sum);

   /* determine type of checksum and set flag bit12 */;
   if(strlen(f_checksum)==32) {
     for(i=0;i<16;i++) {
       buf[0]= f_checksum[2*i];
       buf[1]= f_checksum[2*i+1];
       buf[2]= 0;
       sscanf(buf,"%x",&ret);
       f_md5[i]= ret;
     }
     is_md5= 1;
   } else {
     *f_crc32= 0;
     sscanf(f_checksum,"%X",f_crc32);
     is_md5= 0;
   }
   found_line= 1;
 }
 if(!found_line) {
   fprintf(stderr,
        "\n%s : NO MATCHING RECORD FOUND. KEIN PASSENDER RECORD GEFUNDEN :\n",
        cmd);
   fprintf(stderr,"  %s %s %s\n",s_name,s_old,s_young);
   {ret= 0; goto ex;}
 }

 ret= (!!found_line)+is_md5;
ex:;
 Sfile_fclose(fp,0);
 return(ret);
}


int Backup_plan_filter_checksum(plan,cmd,target,
                                idx_text,age_limit_text,young_limit_text,
                                flag)
struct Backup_plaN *plan;
char *cmd,*target,*idx_text,*age_limit_text,*young_limit_text;
int flag;
/*
 bit0= read from target and compare with stdin (else write stdin sum to target)
 bit1= only print matching record lines but do not check stdin
 bit2= with bit 0 : search matching record
                    either by pre_checksum or by sorted match 
 bit12= work with MD5 rather than CRC32
*/
{
 int ret,i,read_idx,found_line= 0,cut,valid,pre_crc= 0,use_handle= 0;
 FILE *fp= NULL;
 unsigned char md5[16],old_md5[16],pre_md5[16];
 char hexmd5[33],buf[256],old_date[256],idx[256];
 char old_sum[256],old_idx[256],rec_name[256];
 unsigned long crc32,old_crc32,pre_crc32;
 double size= 0.0,old_size= 0.0;
 time_t moment;
 char *handle= NULL;   /* should be void *  */

 if(strlen(idx_text)>80) {
   fprintf(stderr,"\n%s : RECORD NAME TOO LONG (max. 80 char)\n",cmd);
   {ret= 0; goto ex;}
 }
 sscanf(idx_text,"%s",idx);
 if(idx[0]==0) {
   fprintf(stderr,"\n%s : NO USABLE RECORD NAME\n",cmd);
   {ret= 0; goto ex;}
 }

 if(flag&4) {
   if(plan->pre_checksum_size==0) {
     ret= Backup_plan_match_checksums(plan,cmd,target,
                 idx,age_limit_text,young_limit_text,(flag&(1<<12)));
     goto ex;
   }
   use_handle= 512; /* let checksum handle open on first call of
                        Determine_file_checksum()  */
 } else if(flag&(1|2)) {
   ret= Backup_plan_find_checksum(plan,cmd,target,
             idx,age_limit_text,young_limit_text,0.0,"-",
             old_idx,old_date,&old_size,old_sum,old_md5,&old_crc32,
             (flag&2)|256|512);
   if(ret<=0)
     goto ex;
   if(ret==2)
     flag|= (1<<12);
   else
     flag&= ~(1<<12);

   if(flag&2) 
     {ret= 1; goto ex;}
   fprintf(stderr,"\nComparing with record :\n%s %s %.lf %s\n\n",
                  old_idx,old_date,old_size,old_sum); 
 } else {
   fp= Sfile_fopen(target,"a",0);
   if(fp==NULL)
     {ret= 0; goto ex;}
 }

 if(flag&(1|4)) {
   cut= 4;
   if(flag&4) {
     size= plan->pre_checksum_size;
     pre_crc= 8;
   } else
     size= old_size;
 } else {
   cut= 0;
   if(plan->pre_checksum_size>0.0) {
     size= plan->pre_checksum_size;
     pre_crc= 8;
   }
 }

determine_schecksum:;
 if(flag&(1<<12))
   ret= Determine_file_checksum("-",stdin,&handle,md5,pre_md5,&size,
                                1|2|cut|pre_crc|use_handle|(1<<12));
 else 
   ret= Determine_file_checksum("-",stdin,&handle,&crc32,&pre_crc32,&size,
                                1|2|cut|pre_crc|use_handle);
 if(ret<=0)
   goto ex;

 if((flag&4) && pre_crc) { /* only on first pass */
   /* to erase possible pacifier messages of program raedchen */
   fprintf(stderr,"\r                                                   \r");
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       sprintf(hexmd5+2*i,"%2.2x",pre_md5[i]);
   } else {
     sprintf(hexmd5,"%X",pre_crc32);
   }
   if(strcmp(idx,"-")==0)
     strcpy(rec_name,idx);
   else
     sprintf(rec_name,"%s_pre",idx);
   ret= Backup_plan_find_checksum(plan,cmd,target,
             rec_name,age_limit_text,young_limit_text,
             plan->pre_checksum_size,hexmd5 ,
             old_idx,old_date,&old_size,old_sum,old_md5,&old_crc32,
             (flag&2)|256|512|1024|2048);
   if(ret<0)
     goto ex;
   if(flag&2) 
     {ret= 1; goto ex;}
   if(ret==0) {

     {ret= 0; goto ex;}
   }
   fprintf(stderr,"\nFound matching preview record :\n%s %s %.lf %s\n",
                  old_idx,old_date,old_size,old_sum);

   /* continue verification with record found */;
   /* obtain real record */;
   ret= Backup_plan_find_checksum(plan,cmd,target,
             idx,old_date,old_date,0.0,"-",
             old_idx,old_date,&old_size,old_sum,old_md5,&old_crc32,
             (flag&2)|256|512);
   if(ret<=0)
     goto ex;
   if(ret==2)
     flag|= (1<<12);
   else
     flag&= ~(1<<12);
   if(flag&2) 
     {ret= 1; goto ex;}
   fprintf(stderr,"\nFinally comparing with record :\n%s %s %.lf %s\n\n",
                  old_idx,old_date,old_size,old_sum); 

   size= old_size - plan->pre_checksum_size;
   use_handle= 256;
   pre_crc= 0;
   goto determine_schecksum;

 } else if(flag&1) {
   /* to erase possible pacifier messages of program raedchen */
   fprintf(stderr,"\r                                                   \r");
   /* compare length and checksum */
   if(flag&4) /* add length consumed by preview */
     size+= plan->pre_checksum_size;
   if(size!=old_size) {
     fprintf(stderr,"SIZE MISMATCH. FALSCHE GROESSE :\n",old_size,size);
     fprintf(stderr,"record=%.lf  stdin=%.lf\n\n",old_size,size);
     ret= 0; goto ex;
   }
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       if(md5[i]!=old_md5[i])
     break;
     if(i<16) {
       fprintf(stderr,"MD5 CHECKSUM MISMATCH. FALSCHE MD5 PRUEFSUMME :\n");
       fprintf(stderr,"record= ");
       for(i=0;i<16;i++)
         fprintf(stderr,"%2.2x",old_md5[i]);
       fprintf(stderr,"\nstdin = ");
       for(i=0;i<16;i++)
         fprintf(stderr,"%2.2x",md5[i]);
       fprintf(stderr,"\n\n");
       ret= 0; goto ex;
     }
   } else {
     if(crc32!=old_crc32) {
       fprintf(stderr,"CRC32 CHECKSUM MISMATCH. FALSCHE CRC32 PRUEFSUMME :\n");
       fprintf(stderr,"record= %X\nstdin = %X\n",old_crc32,crc32);
       ret= 0; goto ex;
     }
   }
   fprintf(stderr,
             "OK.  Record and data input match. Daten passen zum Record.\n\n");
 } else {
   if(size<=0) {
     fprintf(stderr,"\n%s : EOF at stdin without any data\n",cmd);
     ret= 0; goto ex;
   }

   /* eventually print idx_pre date plan->pre_checksum_size pre_checksum */
   moment= time(0);
   if(plan->pre_checksum_size>0.0 && size>=plan->pre_checksum_size) {
     fprintf(fp,"%s_pre %s %.lf ",idx,Sfile_datestr(moment,1|2|256),
                plan->pre_checksum_size);
     if(flag&(1<<12)) {
       for(i=0;i<16;i++)
         sprintf(hexmd5+2*i,"%2.2x",pre_md5[i]);
       fprintf(fp,"%s\n",hexmd5);
     } else {
       fprintf(fp,"%X\n",pre_crc32);
     }
   }
   /* print idx date length checksum */
   fprintf(fp,"%s %s %.lf ",idx,Sfile_datestr(moment,1|2|256),size);
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       sprintf(hexmd5+2*i,"%2.2x",md5[i]);
     fprintf(fp,"%s\n",hexmd5);
   } else {
     fprintf(fp,"%X\n",crc32);
   }

   /* ??? >>> also report this to stderr */
 
 }
 ret= 1;
ex:;
 if(handle!=NULL)
   free((char *) handle);
 fflush(stdout);
 fflush(stderr);
 Sfile_fclose(fp,0);
 return(ret);
}


int Backup_plan_verify_filelist(plan,listadr,prepend_dir,flag)
struct Backup_plaN *plan;
char *listadr;
char *prepend_dir;
int flag;
/*
 bit0= delete all verified files (DANGEROUS !!!)
 bit7= expect  target=source  and process only the source part
 bit13= with bit0 : allow insecure parameters
 bit14= process target addresses rather than source addresses
*/
{
 FILE *fp;
 int ret= 0,popened= 0,pacifier;
 char initial_wd[4096];

 pacifier= Cd_backup_planer_pacifier_intervaL;
 Cd_backup_planer_pacifier_intervaL= -1;
 popened= Open_listfile(&fp,listadr,"r",1);
 if(popened!=2)
   popened= 0;
 if(flag&(1<<14))
   strcpy(initial_wd,prepend_dir); /* with bit14, prepend_dir is ignored */
 ret= Fileliste_read_serialized_tree(NULL,fp,NULL,NULL,initial_wd,NULL,
                                     prepend_dir,
                                     (4-(flag&1))|
                                     (flag&((1<<7)|(1<<13)|(1<<14))));

ex:;
 Cd_backup_planer_pacifier_intervaL= pacifier;
 if(popened) {
   if(fp!=NULL)
     pclose(fp);
 } else
   Sfile_fclose(fp);
 return(ret);
}


int Backup_plan_setup(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
/*
 bit0= do not finalize , other settings will follow
 bit1= with plan->configuration==2 allow configuration changing commands
*/
{
 int i,ret,not_modus= 0,j,linecount= 0;
 double wert;
 char c,line[4096],*cpt;

 for(i=1;i<argc;i++){
   if(strcmp(argv[i],"--")==0 && i+1<argc){
     i++;
     goto eintragen;
   } else if(strcmp(argv[i],"-accept_all_filetypes")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->accept_all_filetypes= 1;
   } else if(strcmp(argv[i],"-alldev")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_xdev= 0;
   }else if(strcmp(argv[i],"-backup_cmd")==0 && i+1<argc){
     i++;
     strcpy(plan->cmd,argv[i]);
   }else if(strcmp(argv[i],"-verify_source_filter")==0 && i+1<argc){
     i++;
     strcpy(plan->verify_source_filter,argv[i]);
   }else if(strcmp(argv[i],"-collision_test")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"-off")==0){
       plan->collision_test= 0;
       plan->collision_avoid= 0;
     }else if(strcmp(argv[i],"-on")==0){
       plan->collision_test= 1;
       plan->collision_avoid= 0;
     }else if(strcmp(argv[i],"-redirect")==0){
       plan->collision_test= 1;
       plan->collision_avoid= 1;
     }else if(strcmp(argv[i],"-abort")==0){
       plan->collision_test= 1;
       plan->collision_avoid= 2;
     } else {
       fprintf(stderr,"%s : %s unknown sub option : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
   }else if(strcmp(argv[i],"-complete_list_adr")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->append_complete_target,
                       plan->complete_list_adr,0);
   }else if(strcmp(argv[i],"-dir_list_adr")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->dir_list_target,
                       plan->dir_list_adr,0);
   }else if(strcmp(argv[i],"-dir_script_adr")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->dir_script_target,
                       plan->dir_script_adr,0);
   }else if(strcmp(argv[i],"-content_list_adr")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->append_content_target,
                       plan->content_list_adr,0);
   }else if(strcmp(argv[i],"-vanish_list_adr")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->old_complete_list_adr,argv[i-1]);
     Split_fileaddress(argv[i],plan->vanish_list_target,
                       plan->vanish_list_adr,0);
   }else if(strcmp(argv[i],"-vanish_script_adr")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->old_complete_list_adr,argv[i-1]);
     Split_fileaddress(argv[i],plan->vanish_script_target,
                       plan->vanish_script_adr,0);
   }else if(strcmp(argv[i],"-conf_dir")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->conf_dir,argv[i]);
   }else if(strcmp(argv[i],"-dir_size_cmd")==0 && i+1<argc){
     i++;
     for(j=0;j<strlen(argv[i]);j++) if(!isspace(argv[i][j])) break;
     if(j<strlen(argv[i])){
       Fileliste_default_dir_size_cmd(argv[i],0);
       Fileliste_set_dir_size_cmd(plan->startliste,argv[i],0);
       Fileliste_set_dir_size_cmd(plan->auslassen,argv[i],0);
       Fileliste_set_dir_size_cmd(plan->uebergrosse,argv[i],0);
     } else {
       fprintf(stderr,"%s : -dir_size_cmd '%s' = no usable command\n",
                      argv[0],argv[i]); 
       return(1);
     }
   }else if(strcmp(argv[i],"-shell_to_use")==0 && i+1<argc){
     i++;
     strcpy(plan->shell_to_use,argv[i]);
   }else if(strcmp(argv[i],"-set_shell_in_scdbackup_scripts")==0 
            && i+2<argc){
     i++;
     Backup_plan_set_skript_shell(argc-i-1,argv+i+1,argv[i],0);
     i= argc;
   }else if(strcmp(argv[i],"-implicite_ignored")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strlen(argv[i])>255){
       fprintf(stderr,"%s : -implicite_ignored '%s' = list too long \n",
                      argv[0],argv[i]); 
       return(1);
     }
     Fileliste_default_impl_igno(argv[i],0);
     Fileliste_set_impl_igno(plan->startliste,argv[i],0);
     Fileliste_set_impl_igno(plan->auslassen,argv[i],0);
     Fileliste_set_impl_igno(plan->uebergrosse,argv[i],0);
   }else if(strcmp(argv[i],"-info_exclude")==0){
     plan->auskunft_include= 0;
   }else if(strcmp(argv[i],"-volume_info")==0){
     plan->volume_info= 1;
   }else if(strcmp(argv[i],"-info_script")==0 && i+1<argc){
     i++;
     sscanf(argv[i],"%s",line);
     plan->auskunft_include= 0;
     Fileliste_destroy(&plan->auskunft_name,0);
     if(strcmp(line,"-off")==0) {
       plan->auskunft_an= 0;
       if(Fileliste_single(&plan->auskunft_name,"",0,0.0,2)<=0)
         return(1);
     } else if(strcmp(line,"-inline")==0) {
       if(Fileliste_single(&plan->auskunft_name,"",0,0.0,2)<=0)
         return(1);
     } else {
       plan->auskunft_an= 1;
       plan->auskunft_include= 1;
       if(Fileliste_single(&plan->auskunft_name,line,1,plan->auskunft_groesse,
                           2)<=0)
         return(1);
     }
   }else if(strcmp(argv[i],"-filecheck")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"-off")==0) {
       plan->precheck= 0;
       plan->precheck_expl= 0;
       plan->precheck_unread= 0;
     } else if(strcmp(argv[i],"-on")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 0;
       plan->precheck_unread= 0;
     } else if(strcmp(argv[i],"-only")==0) {
       plan->precheck= 2;
       plan->precheck_expl= 0;
       plan->precheck_unread= 0;
     } else if(strcmp(argv[i],"-append_hexcode")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 1;
       plan->precheck_expl_replace= 1;
     } else if(strcmp(argv[i],"-append")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 1;
       plan->precheck_expl_replace= 0;
     } else if(strcmp(argv[i],"-exclude")==0) {
       plan->precheck= 1;
       plan->precheck_unread= 1;
     } else if(strcmp(argv[i],"-abort")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 2;
       plan->precheck_unread= 2;
     } else {
       fprintf(stderr,"%s : %s unknown sub option : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
   }else if(strcmp(argv[i],"-find")==0 && i+1<argc){
     struct FilelistE *find_liste;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(Fileliste_single(&find_liste,argv[i],0,0.0,8)>0){
       if(Fileliste_set_main_dir(find_liste,plan->hauptdir,0)<=0) {
         Fileliste_destroy(&find_liste,0);
         return(-1);
       }
       if(Fileliste_read_properties(find_liste,2)>0)
         Fileliste_leanfind(find_liste,0,
                            dev_dummy,stdout,&linecount,0);
       Fileliste_destroy(&find_liste,0);
     }
   }else if(strcmp(argv[i],"-find_path_list")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Fileliste_find_path_list(argv[i],0);
   }else if(strcmp(argv[i],"-form_script")==0){
     plan->output_form= 0;
   }else if(strcmp(argv[i],"-form_script_raw")==0){
     plan->output_form= 2;
   }else if(strcmp(argv[i],"-form_script_here")==0){
     plan->output_form= 3;
   }else if(strcmp(argv[i],"-form_raw")==0){
     plan->output_form= 4;
     if(strlen(plan->auskunft_name->adressen[0])==0) plan->auskunft_an= 0;
   }else if(strcmp(argv[i],"-form_listing")==0){
     plan->output_form= 1;
     if(strlen(plan->auskunft_name->adressen[0])==0) plan->auskunft_an= 0;
   }else if((strcmp(argv[i],"-form_path_list")==0 ||
             strcmp(argv[i],"-form_path_list_mkisofs")==0 ||
             strcmp(argv[i],"-form_path_list_afio")==0)
             && i+2<argc){
     i+= 2;
     if(strcmp(argv[i-2],"-form_path_list_afio")==0)
       plan->output_form= 6;
     else
       plan->output_form= 5;
     strcpy(plan->path_list,argv[i-1]);
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=2e9) wert= 2e9;
     if(wert>1) 
       plan->max_items_per_tile= wert;
   }else if(strcmp(argv[i],"-help")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(argc!=2) {
       fprintf(stderr,"%s : option -help ignored because not alone.\n",argv[0]);
     } else {
       Backup_plan_help(argv[0],0);
       return(-1);
     }
   }else if(strcmp(argv[i],"-hilfe")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(argc!=2) {
       fprintf(stderr,
               "%s : option -hilfe wurde ignoriert wegen weiterer Optionen.\n",
               argv[0]);
     } else {
       Backup_plan_hilfe(argv[0],0);
       return(-1);
     }
   }else if(strcmp(argv[i],"-version")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     printf("scdbackup %s\n",PROG_VERSION);
     return(-1);
   }else if(strcmp(argv[i],"-level")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"-create_configuration")==0 || 
        strcmp(argv[i],"create_configuration")==0) {
       plan->configuration= 1;
       plan->level= 0;
     } else if(strcmp(argv[i],"-update_configuration")==0 ||
               strcmp(argv[i],"update_configuration")==0) {
       plan->configuration= 3;
       plan->level= 0;
     } else if(strcmp(argv[i],"-none")==0 || strcmp(argv[i],"-off")==0 ||
               strcmp(argv[i],"none")==0 || strcmp(argv[i],"off")==0 ) {
       plan->configuration= 0;
       plan->level= 0;
     } else if(strcmp(argv[i],"-next")==0 || strcmp(argv[i],"next")==0) {
       plan->configuration= 2;
       plan->level= -1;
     } else if(strcmp(argv[i],"-last")==0 || strcmp(argv[i],"-current")==0 ||
               strcmp(argv[i],"last")==0 || strcmp(argv[i],"current")==0) {
       plan->configuration= 2;
       plan->level= -2;
     } else {
       plan->configuration= 2;
       wert= -1;
       sscanf(argv[i],"%lf",&wert);
       if(wert>=2e9) wert= 2e9;
       if(wert>=0)
         plan->level= wert; 
       else {
         fprintf(stderr,"%s : %s illegal level number : %s\n",
                        argv[0],argv[i-1],argv[i]);
         return(1);
       }
     }
   }else if(strcmp(argv[i],"-list_output_format")==0 && i+1<argc){
     i+= 1;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->list_output_format= 1;
     sprintf(line,":%s:",argv[i]);
     if(strstr(line,":type:")!=NULL)
       plan->list_output_format|= 2;
     if(strstr(line,":size:")!=NULL)
       plan->list_output_format|= 4;
     if(strstr(line,":crc:")!=NULL)
       plan->list_output_format= (plan->list_output_format&~16)|8;
     if(strstr(line,":md5:")!=NULL)
       plan->list_output_format|= 8|16;
   }else if(strcmp(argv[i],"-load_old_tree")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->old_complete_list_adr,argv[i-1]);
     strcpy(plan->old_vanish_list_adr,argv[i]);
     if(plan->old_complete_list_adr[0]!=0 && plan->old_vanish_list_adr[0]!=0)
       plan->precheck_agetest|= 256|512|1024|2048;
     else
       plan->precheck_agetest&= ~256;
   }else if(strcmp(argv[i],"-main_dir")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     sscanf(argv[i],"%s",plan->hauptdir);
     plan->hauptdir_auto= 0;
     if(strcmp(plan->hauptdir,"-off")==0) {
       plan->hauptdir[0]= 0;
     } else if(strcmp(plan->hauptdir,"-auto")==0) {
       plan->hauptdir_auto= 1;
       plan->hauptdir[0]= 0;
     }
     if(Backup_plan_list_main_dirs(plan,0)<=0)
       return(-1);
   }else if(strcmp(argv[i],"-max_items")==0 && i+1<argc){
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=2e9) wert= 2e9;
     if(wert>1) 
       plan->max_items= wert;
   }else if(strcmp(argv[i],"-exclude_size")==0 && i+1<argc){
     int skip_char= 0;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(argv[i][0]=='-')
       skip_char= 1;
     wert= Scanf_io_size(argv[i]+skip_char,1);
     if(wert>=1.0) {
       plan->exclude_size= wert;
       plan->precheck_sizetest|= 1|((!!skip_char)<<1);
     } else
       plan->precheck_sizetest&= ~1;
   }else if(strcmp(argv[i],"-max_size")==0 && i+1<argc){
     i++;
     wert= Scanf_io_size(argv[i],1);
     if(wert>=1.0) {
       plan->default_max_size= wert;
       plan->max_size= wert;
     }
   }else if(strcmp(argv[i],"-max_size_stack")==0 && i+2<argc){
     int count;

     i+= 2;
     sscanf(argv[i],"%ld",&count);
     wert= Scanf_io_size(argv[i-1],1);
     if(wert>=1.0) {
       Backup_plan_stacked_max_size(plan,&wert,&count,1);
     } else if(count<=0) {
       Backup_plan_stacked_max_size(plan,&wert,&count,2);
     }
   }else if(strcmp(argv[i],"-max_size_stack_save")==0 && i+1<argc){
     i++;
     if(strcmp(argv[i],"-off")==0) {
       plan->save_stack_media= 0;
     } else if(strcmp(argv[i],"-on")==0) {
       plan->save_stack_media= 1;
     }
   }else if(strcmp(argv[i],"-max_young_append")==0 && i+1<argc){
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=2e9) wert= 2e9;
     if(wert>=0)
       plan->precheck_max_exempt= wert; 
   }else if(strcmp(argv[i],"-min_size_percent")==0 && i+1<argc){
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=0. && wert<=100.0) {
        plan->min_size_fak= wert/100.0;
     }
   }else if(strcmp(argv[i],"-no")==0 || strcmp(argv[i],"-not")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     not_modus= 1;
   }else if(strcmp(argv[i],"-scattering")==0 && i+1<argc){
     i++;
     if(strcmp(argv[i],"-off")==0) {
       plan->scattering= 0;
     } else if(strcmp(argv[i],"-on")==0) {
       plan->scattering= 1;
     }
   }else if(strcmp(argv[i],"-prompt_script")==0 && i+1<argc){
     i++;
     strcpy(plan->prompt_skript,argv[i]);
     if(strcmp(plan->prompt_skript,"-")==0)
       plan->prompt_skript[0]= 0;
   }else if(strcmp(argv[i],"-status_file")==0 && i+1<argc){
     i++;
     strcpy(plan->skript_status_file,argv[i]);
     if(strcmp(plan->skript_status_file,"-")==0)
       plan->skript_status_file[0]= 0;
   }else if(strcmp(argv[i],"-target_separator")==0 && i+1<argc){
     i++;
     plan->target_inp_trenner= argv[i][0];
     /* Fileliste_default_target_inp_trenner(argv[i],0); */
     Fileliste_set_target_inp_trenner(plan->startliste,argv[i],0);
     Fileliste_set_target_inp_trenner(plan->auslassen,argv[i],0);
   }else if(strcmp(argv[i],"-tmp_dir")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->tmp_dir,argv[i]);
     if(strcmp(plan->tmp_dir,"-")==0)
       plan->tmp_dir[0]= 0;
   }else if(strcmp(argv[i],"-working_directory")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= chdir(argv[i]);
     if(ret==-1) {
       fprintf(stderr,"Cannot set -working_directory : %s\n",argv[i]);
       fprintf(stderr,"errno= %ld , '%s'\n",errno,strerror(errno));
       return(21);
     }
   } else if(strcmp(argv[i],"-xdev")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_xdev= 1;
   } else if(strcmp(argv[i],"-localdev")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_xdev= 2;
   }else if(strcmp(argv[i],"-yes")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     not_modus= 0;
     plan->auslassen_nicht_startliste= 1;
   }else if(strcmp(argv[i],"-younger_than")==0 && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_agetest|= 1;
     if(Zeitdefinition(argv[i],
                       &(plan->precheck_agetest),&(plan->precheck_age),0)<=0){
       fprintf(stderr,"%s : %s unknown time definition : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
     if(plan->precheck_agetest&1)
       if(plan->precheck_max_exempt<0)
         plan->precheck_max_exempt= 2000000000; /* unlimited */
   }else if(strcmp(argv[i],"-disable_safe_leveling")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->complete_list_adr[0]= 0;
     plan->dir_list_adr[0]= 0;
     plan->dir_script_adr[0]= 0;
     plan->content_list_adr[0]= 0;
     plan->old_complete_list_adr[0];
     plan->vanish_list_adr[0]= 0;
     plan->vanish_script_adr[0]= 0;
     plan->old_vanish_list_adr[0]= 0;
   }else if(strcmp(argv[i],"-changetest_options")==0 && i+1<argc){
     i+= 1;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_agetest&= ~(32|64|512|1024|2048);
     plan->precheck_agetest|= 256;
     sprintf(line,":%s:",argv[i]);
     if(strstr(line,":type:")!=NULL)
       plan->precheck_agetest|= 512;
     if(strstr(line,":size:")!=NULL)
       plan->precheck_agetest|= 1024;
     if(strstr(line,":crc:")!=NULL ||
        strstr(line,":md5:")!=NULL ||
        strstr(line,":chk:")!=NULL)
       plan->precheck_agetest|= 2048;
     if(strstr(line,":mistrust:")!=NULL)
       plan->precheck_agetest|= 32;
     if(strstr(line,":assume_old:")!=NULL)
       plan->precheck_agetest|= 32;
     if(strstr(line,":off:")!=NULL)
       plan->precheck_agetest&= ~256;
     if(strstr(line,":adr:")!=NULL)
       plan->precheck_agetest|= 256;
   }else if(strcmp(argv[i],"-pacifier")==0 && i+1<argc){
     i++;
     if(strcmp(argv[i],"-off")==0) {
       Cd_backup_planer_pacifier_intervaL= -1;
     } else if(strcmp(argv[i],"-on")==0) {
       Cd_backup_planer_pacifier_intervaL= 1;
     } else if(argv[i][0]>='0' && argv[i][0]<='9') {
       sscanf(argv[i],"%ld",&Cd_backup_planer_pacifier_intervaL);
     }
   }else if((strcmp(argv[i],"-filter_pipe")==0) && i+1<argc){
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Stdout_pipe_broker(argv[i],strlen(argv[i]),1);
   }else if((strcmp(argv[i],"-filter_through")==0)){
     char buf[4096];
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     while(1) {
       ret=  fread(buf,1,sizeof(buf),stdin);
       if(ret<=0)
     break;
       Stdout_pipe_broker(buf,ret,0);
     }
     Stdout_pipe_broker("",0,1<<15);
   }else if((strcmp(argv[i],"-filter_crc32")==0 || 
             strcmp(argv[i],"-filter_md5")==0)  && i+2<argc){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_filter_checksum(plan,argv[i],argv[i+1],argv[i+2],"","",
                                     (strcmp(argv[i],"-filter_md5")==0)<<12,0);
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-compare_checksum")==0 && i+4<argc){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_filter_checksum(plan,argv[i],argv[i+1],argv[i+2],
                                           argv[i+3],argv[i+4],1);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if((strcmp(argv[i],"-search_crc32")==0 ||
             strcmp(argv[i],"-search_md5")==0)  && i+4<argc){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_filter_checksum(
                         plan,argv[i],argv[i+1],argv[i+2], argv[i+3],argv[i+4],
                         1|4|((strcmp(argv[i],"-search_md5")==0)<<12));
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-list_checksum")==0 && i+4<argc){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_filter_checksum(plan,argv[i],argv[i+1],argv[i+2],
                                           argv[i+3],argv[i+4],1|2);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-pre_checksum_size")==0 && i+1<argc){
     i++;
     wert= Scanf_io_size(argv[i],1);
     if(wert>=0.0) {
       plan->pre_checksum_size= wert;
     }
   }else if(strcmp(argv[i],"-verify_targetlist")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],1<<14);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-verify_filelist")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],0);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-verify_sources")==0 && i+1<argc){
     i+= 1;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i],"",128);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-delete_targetlist")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],1|(1<<14));
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-delete_filelist")==0 && i+2<argc){
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],1);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-merge_trees")==0 && i+4<argc){
     int hflag= 0,tree_flag= (1<<14);
     char tree_prefix[4096],*cpt,*dpt;
     i+= 4;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;

     hflag= 16|32|(1<<8)|(1<<12);
     strcpy(tree_prefix,argv[i-3]);
     strcpy(line,":");
     strcat(line,argv[i]);
     strcat(line,":");
     if(strstr(line,":no_overwrite:")!=NULL)
       hflag= (hflag&~(15<<8))|(0<<8);
     if(strstr(line,":overwrite:")!=NULL)
       hflag= (hflag&~(15<<8))|(1<<8);
     if(strstr(line,":overwrite_older:")!=NULL)
       hflag= (hflag&~(15<<8))|(2<<8);
     if(strstr(line,":move:")!=NULL)
       hflag&= ~64;
     if(strstr(line,":copy:")!=NULL)
       hflag|= 64;
     if(strstr(line,":owner_plus_w:")!=NULL)
       hflag|= (1<<13);
     if(strstr(line,":chmod_original:")!=NULL)
       hflag&= ~(1<<13);
     cpt= strstr(line,":tree_prefix=");
     if(cpt!=NULL) {
       cpt+= strlen(":tree_prefix=");
       dpt= strchr(cpt,':');
       strncpy(tree_prefix,cpt,dpt-cpt);
       tree_prefix[dpt-cpt]= 0;
     }
     if(strstr(line,":tree_source:")!=NULL)
       tree_flag&= ~(1<<14);
     if(strstr(line,":tree_target:")!=NULL)
       tree_flag|= (1<<14);
     ret= Backup_plan_load_old_tree(plan,plan->old_complete_list_adr,
                                    plan->old_vanish_list_adr,tree_prefix,
                                    tree_flag);
     if(ret<=0)
       exit(1);

     ret= Fileliste_merge_trees(NULL,0,(dev_t) 0,&linecount,
                          argv[i-3],argv[i-2],argv[i-1],plan->old_tree,hflag);

     if(ret<=0)
       exit(1);
     exit(0);
   }else if(argv[i][0]=='-'){
     fprintf(stderr,"%s : unknown or incomplete option %s\n",argv[0],argv[i]);
     return(1);
   } else {
eintragen:;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;

     /* ??? remove trailing / ??? */

     if(not_modus)
       ret= Fileliste_append(plan->auslassen,argv[i],1);
     else
       ret= Fileliste_append(plan->startliste,argv[i],1);
     if(ret<=0) return(2);
   }
ignore_silently:;
 }

 if(!(flag&1)){
   if(plan->hauptdir_auto){
     /* determine common parent directory */ ;
     Fileliste_hauptdir(plan->hauptdir,plan->startliste,0);
   }
   if(Backup_plan_list_main_dirs(plan,0)<=0)
     return(-1);

   /* finalize setup */
   if(Fileliste_read_properties(plan->startliste,
                                2|(!!plan->auslassen_nicht_startliste)<<2)<0)
     return(2);
   if(!plan->accept_all_filetypes) {
     Fileliste_redirect_links(plan->startliste,0);

/*
     for(i=0;i<plan->startliste->anz;i++){
       if(plan->startliste->typen[i]==3) {
         fprintf(stderr,"--- symbolic link ignored : %s\n",
                      plan->startliste->adressen[i]);
       }
     }
*/

   }
   if(plan->output_form==0 && Backup_plan_stueck_header(plan,1)){
     /* The info script will be included in any piece.
        Its target competitors need to be excluded */
     ret= Backup_plan_append_and_exclude(plan,plan->auskunft_name,
                                         !!plan->auskunft_skript_root);
     if(ret<=0)
       return(2);
   }
 } 
 return(0);
}


#define MELDEN(x) fprintf(stdout,"%s\n",x)


int Backup_plan_hilfe(progname,flag)
char *progname;
int flag;
{
 static char hilfetext[][80]={
"usage :  %s  [option | file] ...",
"",
"Gewoehnlich teilt dieses Programm eine Liste von Directories in Teile auf,",
"von denen jedes einzelne auf eine CD passt.",
"Es erzeugt dazu ein Skript mit Backupkommandos. Die Listen der Files oder",
"Directories in den einzelnen Stuecken werden anhand von Obergrenzen fuer",
"die Groesse in Bytes und die Anzahl der einzeln erwaehnten Fileadressen",
"festgelegt.",
"",
"Eine solche Ergebnisliste kann nur Directories und Datenfiles enthalten.",
"Andere Filetypen koennen in Directories enthalten sein, aber die meisten",
"ISO9660-CD-Schreibprogramme werden sie nicht verarbeiten.",
"",
"Die Ergebnislisten bestehen aus Ziel- und Quelladresse in einer Form, die",
"sich fuer das Programm  mkisofs  in seinen modernen Versionen eignet.",
"  Wenn die Quelle ein Directory ist:  /Zieladresse/=Fileadresse",
"  Wenn die Quelle ein File ist:       /Zieladresse=Fileadresse",
"Diese Listen werden an ein Backupkommando angehaengt, damit dieses",
"dann eine CD fuellt. Wenn nicht ausdruecklich angegeben und auch nicht",
"zur Vermeidung einer Kollision umgeleitet, lautet die Zieladresse",
"gleich wie die Fileadresse. Das Zeichen \"=\" wird zu \"_\" gewandelt.",
"Es koennen auch andere Ausgabeformate als das hier beschriebene bestellt",
"werden.",
"",
"Weil das Arbeitsdirectory, das Rootdirectory und evtl. das -main_dir auf",
"das Rootdirectory der CD abgebildet werden, koennen Kollisionen von Ziel-",
"adressen auftreten. Ein eingebauter Kollisionstest kann diese erkennen und",
"durch Umnennen der Zieladressen vermeiden.",
"",
"Eine Pruefung kann durchgefuehrt werden, um problematische Files zu finden",
" - solche, die von mkisofs-1.12b3 ignoriert werden, wenn sie nicht",
"   ausdruecklich als Argumente gennat werden.",
" - solche, die nicht gelesen werden koennen (Files oder Directories)",
"   Das kann die Schaetzung der Ergebnisgroessen verfaelschen.",
"Die Kombination von -filecheck -append -filecheck -exclude kann helfen,",
"diese Probleme zu vermeiden.",
"",
"Inkrementelle Backups koennen mit beliebig vielen Ebenen gemacht werden :",
" -level 0 macht einen vollen Backup ohne Zeiteinschraenkung.",
" -level N macht einen Backup der Files, die seit dem letzten -level N-1",
"  geaendert wurden. -level N-1 muss bereits existieren.",
" - Ein Backup mit -level N macht alle darueber liegenden -level ungueltig.",
"Die noetigen Daten der Ebenen werden in einem gesonderten Directory",
"verwaltet, das mit -conf_dir gewaehlt wird. Dort werden auch alle Parameter",
"des Backups wie Fileliste, Maximalgroessen usw. gespeichert.",
"",
"Inkrementelle Backups geben sich automatisch die Option -younger_than vor.",
"Durch sie werden unter Umstaenden auch aeltere Files mit aufgenommen.",
"Leider verursachen reine Namensaenderungen im allgemeinen Linux Filesystem",
"keine Aenderung an den Zeitmarken des Files. (Aber zB. chmod u+r tut es)",
"",
"Weil eine zu grosse Zahl einzelner Filenamen Probleme mit alten Versionen",
"von mkisofs bereiten wuerde, versucht das Programm hauptsaechlich, ganze",
"Directories auszuschliessen, die nur sehr wenige junge Files enthalten.",
"Diese Ausnahmen werden dann ausdruecklich in den Backup aufgenommen.",
"Es muss ausserdem eine gewisse Quote von eingesparten Bytes pro",
"hinzukommendem Filenamen erfuellt sein, damit es zum Ausschluss von",
"Directories kommt. Diese ist umso hoeher, je kleiner -max_items gewaehlt ist.",
"",
"Es kann ein gesondertes Skript erzeugt werden, das Auskunft ueber den",
"Backup gibt. Optionen dieses Skriptes sind dann:",
"-hilfe, -id, -count, -list, -levelall , -suball, -fast, -help",
"Andere Argumente werden als Filenamen in den Listen der Stuecke gesucht.",
"Wenn ein passendes Directory oder der Filename gefunden wird, dann wird",
"die Nummer des Stuecks, der gefundene und der gesuchte Name ausgegeben.",
"Wenn nichts gefunden wird, ist der exit-Wert des Skriptes 1.",
"",
"Dieses sog. Infoskript wird nur dann auf der Platte gespeichert und in",
"jedes einzelne Backupstueck eingefuegt, wenn ein Filename mit der",
"Option -info_script angegeben wurde (siehe unten).",
"Dabei wird seine Groesse ueppig mit 100kB geschaetzt. Falls Ihr Infoskript",
"wirklich groesser werden sollte, muss dass durch eine entsprechend",
"kleinere -max_size (siehe unten) ausgeglichen werden.",
"Wenn das Infoskript in die Stuecke eingefuegt wird und die Ausgabgeform",
"fuer mkisofs verwendet wird (Grundeinstellung bzw. -form_script) dann",
"werden alle Fileadressen vom Backup ausgeschlossen, die zum selben",
"Namen auf der CD fuehren wuerden.",
"",
"Wenn eine Zahl als Argument uebergeben wird, ueberspringt das Skript alle",
"Stuecke mit niedrigerer Nummer. Um zB. einen Backup fortzusetzen, der bei",
"der Abfrage fuer Stueck 3 unterbrochen wurde:",
"  skriptname 3",
"So ueberspringt das Skript die Stuecke 1 und 2.",
"Wenn das Skript mit der Option -status_file erzeugt wurde, dann kann es das",
"betroffene Stueck automatisch anhand des Statusfiles finden. Diese Automatik",
"wird durch das Argument \"AUTO\" ausgeloest.",
"",
"Argumente, die nicht mit '-' beginnen, sind Fileadressen, die entweder",
"in die Backupliste oder in die Ausschlussliste aufgenommen werden.",
"(siehe auch Optionen -yes und -no)",
"Vor den Fileadressen koennen Zieladressen stehen. Das ist nuetzlich, um",
"eine CD aus weit verstreuten Daten zusammenzustellen. Dazu muss mit der",
"Option -target_separator vorher ein Trennzeichen eingestellt werden.",
"Beispiel: -target_separator = /bilder=/home/privat/bilder ...",
"",
"Alle Fileaddressen zusammen legen das \"Backupgebiet\" fest, in dem Teil-",
"bereiche durch die Option -not ausgeblendet sein koennen. Das Backupgebiet",
"kann durch Pruefungen (zB. -younger_than) ausgeduennt werden und bildet",
"danach den \"Backupinhalt\".",
"",
"Optionen:",
"",
" -hilfe            diesen Text ausgeben und Programm beenden",
" -help             print this text and exit program",
"                   (dieser Text in englischer Sprache und Programmende)",
" -version          Versionskennung ausgeben und Programm beenden.",
" --                das naechste Argument ist ein File und keine Option",
"",
" -accept_all_filetypes",
"                   normalerweise werden nur Directories und regulaere",
"                   Datenfiles in das Ergebnis aufgenommen. Diese Option",
"                   gibt auch alle anderen Filetypen dafuer frei.",
" -alldev           gemountete Filesysteme nicht ausschliessen.",
"                   (Grundeinstellung, siehe -xdev und -localdev)",
" -backup_cmd text  der angegebene Text wird in dem vom Programm erzeugten",
"                   Skript als Backup-Kommando verwendet.",
"                   (Grundeinstellung: cd_backup_kommando)",
" -changetest_options  code:code:...",
"                   waehlt die Eigenschaften, die waehrend des Tests von",
"                   -younger_than beruecksichtigt werden sollen. Es werden",
"                   dabei die Objekte des Backupgebiets mit dem Speicherbaum",
"                   verglichen, der mit -old_tree eingelesen wurden.",
"                   Die einzelnen Codes werden mit \":\" verbunden :",
"                     type  Filetyp (zB. Directory, Datenfile, Softlink)",
"                     size  Groesse des Inhaltes in Bytes",
"                     crc   CRC-32 Pruefsumme des Inhalts",
"                     off   Vergleich mit altem Speicherbaum wird ausgelassen",
"                     adr   Vergleich wird gemacht (Grundeinstellung)",
"                     mistrust    vermeidet eine schnelle Abkuerzung, die auf",
"                                 Directory Zeitmarken und CRCs beruht.",
"                     assume_old  alle Files als zu alt werten.",
"                   HINWEIS: -load_old_tree mit nichtleeren Adressen setzt",
"                   -changetest_options auf \"type:size:crc:adr\"",
"                   Setzen Sie Ihre -changetest_options nach -load_old_tree",
" -collision_test modus  steuert den Kollisionstest fuer Zielnamen",
"                   modus kann sein (Grundeinstellung: -redirect): ",
"                     -off       keinen Kollisionstest machen",
"                     -on        bei Kollisionen warnen",
"                     -redirect  bei Kollisionen Zielnamen aendern",
"                     -abort     bei Kollisionen Programm abbrechen",
" -compare_checksum listenfile suchname altersgrenze jugendgrenze",
"                   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Aus listenfile",
"                   wird ein Record gelesen, der zu den anderen drei",
"                   Argumenten passt. Er sollte von -filter_md5 oder",
"                   von -filter_crc32 stammen. Damit er passt, muss der",
"                   suchname \"-\" sein oder genau der Name des Records. Wenn",
"                   altergrenze nicht \"-\" ist, dann darf er nicht aelter",
"                   sein als diese Grenze. jugendgrenze verlangt dagegen ein",
"                   gewisses Mindestalter (siehe -younger_than wegen des",
"                   Formats von Zeitangaben). Der letzte passende Record aus",
"                   listenfile wird verwendet. Die in ihm verzeichnete",
"                   Datenmenge wird von stdin gelesen und zu einer Pruefsumme",
"                   verarbeitet, die mit der des Records verglichen wird.",
"                   Liefert stdin genuegend Daten und stimmt die Pruefsumme",
"                   ueberein, dann endet das Programm mit Rueckgabewert 0,",
"                   in allen anderen Faellen mit 1.",
"                   Ueberschuessige Daten an stdin werden ignoriert !",
" -complete_list_adr fileadresse",
"                   schreibt eine komplette Fileliste des Backupgebiets in",
"                   den angegebenen File. Wenn die fileadresse kein \"/\"",
"                   enthaelt und eine -level Option angegeben ist, dann wird",
"                   der File im level_# Directory erzeugt (siehe -conf_dir).",
"                   Ohne -level ensteht ein solcher File im -tmp_dir.",
"                   Wenn fileadresse mit .gz endet, dann wird die Ausgabe",
"                   durch das programm gzip komprimiert. Wenn eine Zieladresse",
"                   angegeben ist, dann wird der Ergebnisfile zum letzten",
"                   Stueck des Backups hinzugefuegt. (zB. /compl.gz=compl.gz)",
"                   Die Liste wird in einem Format geschrieben, das Platz",
"                   spart aber immer noch von Menschen lesbar ist. Jede",
"                   Listenzeile besteht aus den mit -list_output_format",
"                   festgelegten Feldern, wobei die Fileadresse immer am",
"                   Zeilenende steht. (Felderliste siehe \"#=# line layout\")",
"                   Die ersten und letzten Zeichen einer Adresse bestimmen",
"                   ihren Typ :",
"                     /x/x/x/   absolute Directoryadresse",
"                     /x/x/x    absolute Fileadresse",
"                     ./x/      tiefergehen in Directory x",
"                     ./x       File x im aktuellen Directory",
"                     ..        ein Directory nach oben gehen",
"                     ../       zurueck zum Startdirectory",
"                     #=# xxxx  Zusatzinformationen",
"                   Die erste angegebene Directoryadresse ist absolut und",
"                   bestimmt das Startdirectory fuer relative Adressen. Mit",
"                   ./x/ und /x/x/x kann der Directorycursor aus dem",
"                   Startdirectory herausbewegt werden. Solange er in im",
"                   Startdirectory ist, wird den Adressen eine Zieladresse",
"                   vorangestellt, die dann auch fuer die darunterliegenden",
"                   relativen Adressen gilt.",
"                   Die Ergebnisliste eignet sich fuer -vanish_list_adr .",
" -content_list_adr fileaddress",
"                   schreibt eine Fileliste des Backupinhalts. Diese Option",
"                   listet nur diejenigen Files, die Teil eines Backupstuecks",
"                   werden. Das Ausgabeformat ist wie bei -complete_list_adr .",
"                   Die Inhaltslisten und die Schwundlisten aller Levels",
"                   zusammen enthalten die selbe Information wie eine einzelne",
"                   Komplettliste. Eine Inhaltsliste markiert auch die Grenzen",
"                   der Backupstuecke.",
" -conf_dir directoryadresse",
"                   legt die Adresse einer Backupkonfiguration fest. Sie",
"                   enthaelt Files mit Informationen, die von inkrementellen",
"                   Backups erzeugt und genutzt werden. (siehe -level)",
"                     specific_arguments : Optionenliste des vollen Backups",
"                                          jede Zeile ist ein Argument. ",
"                     highest_level      : Hoechste gueltige Levelnummer",
"                     ASKME_level_#      : Infoskript zu -level #",
"                     level_#/           : Leveldirectory mit folgenden Files",
"                       last_run_date    : Datumskennung des letzten Backups",
"                       number_of_volumes : Anzahl der Backupstuecke",
"                       content_file_list.gz : siehe -content_list_adr",
"                                              und   -load_old_tree",
"                       vanished_file_list.gz: siehe -vanish_list_adr",
"                       VANISHED.sh.gz       : siehe -vanish_script_adr",
" -delete_filelist  listenfile  oberdirectory",
"                   HINWEIS: fuer das Entfernen von Files waehrend der",
"                            Wiederherstellung eines backups eignet sich",
"                            -delete_targetlist besser.",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Aus listenfile (\"-\" = stdin) werden Fileadressen mit",
"                   Typ, Groesse und Pruefsumme gelesen.",
"                   Diese Angaben werden mit dem Ist-Zustand der Fileadresse",
"                   verglichen. Dabei wird oberdirectory jeder Adresse",
"                   vorangestellt. Existiert die Adresse und passen die",
"                   anderen Angaben, dann wird der File an dieser Adresse",
"                   mit oberdirectory geloescht. Directories werden geloescht",
"                   wenn sie in der Liste vorkommen und im Filesystem leer",
"                   sind. Ansonsten geht eine Fehlermeldung an stderr.",
"                   listenfile muss mit der Option -vanish_list_adr erzeugt",
"                   worden sein. Er muss Typ, Groesse, Checksumme enthalten.",
"                   War der gesamte Programmlauf fehlerfrei, dann ist der",
"                   Exitwert 0, sonst 1.",
" -delete_targetlist  listenfile  oberdirectory",
"                   wie -delete_filelist. Jedoch werden die Zieladressen",
"                   verwendet, was einen Unterschied macht, wenn beim Backup",
"                   mit ziel=quelle Umleitungen bestellt wurden.",
" -dir_list_adr fileadresse",
"                   schreibt eine komplette Directoryliste des Backupgebiets",
"                   in den angegebenen File. Das Ausgabeformat ist aehnlich",
"                   wie bei -complete_list_adr. Allerdings ist es festgelegt",
"                   auf \"type:chmod:chown:chgrp:date:size:crc:adr\" wobei",
"                   chmod:chown:chgrp:date Zugriffsrechte, Besitzer und Datum",
"                   angeben. Wenn bei -list_output_format MD5-Pruefsummen",
"                   bestellt sind, werden sie auch hier verwendet.",
" -dir_script_adr fileadresse",
"                   schreibt ein Shellscript zur Wiederherstellung der",
"                   Zugriffsrechte der Directories des Backupgebiets in",
"                   in den angegebenen File. Es werden die Zieladressen",
"                   relativ zum /-Directory der CDs eingetragen.",
" -disable_safe_leveling  ist aequivalent zu",
"                   -load_old_tree \"\" \"\"",
"                   -content_list_adr \"\" -complete_list_adr \"\"",
"                   -vanish_script_adr \"\" \"\" -vanish_list_adr \"\" \"\"",
" -exclude_size groessengrenze  alle files ausschliessen, die mindestens die",
"                   angegebene Zahl von Bytes enthalten. groessengrenze 0",
"                   schaltet diesen Test ab. Wenn groessengrenze mit \"-\"", 
"                   beginnt, gilt bei Softlinks die Groesse des Linkziels.",
"                   'k' 'm' und 'g' werden erkannt wie bei -max_size .",
" -dir_size_cmd text der angegebene Text wird als Kommando zur Bestimmung",
"                   der Groesse eines Directory verwendet. Das Kommando",
"                   muss die Adresse eines Directory erwarten und als",
"                   erstes Wort auf der Standardausgabe dessen Groesse in",
"                   Kilobyte angeben (wie  du -sk ). Grundeinstellung:",
"                   \"-internal\" ruft kein externes Kommando auf, sondern",
"                   summiert rekursiv die einzelnen Filegroessen und die",
"                   Laenge ihrer Adressen. Dazu kommt ein Zuschlag von 256",
"                   Bytes pro File. Dieser kann mit \"-internal+$zahl\"",
"                   anders gewaehlt werden. So rechnet \"-internal+10\" pro",
"                   File 10 Bytes hinzu.  \"-internal-$zahl\" schaltet das",
"                   Hinzuaddieren der Adressenlaenge ab.",
" -implicite_ignored text  Abhilfe gegen eine Spezialitaet von mkisofs-1.12b3",
"                   Der angegebene Text meldet diejenigen Zeichen an, die",
"                   das Backup-Kommando veranlassen, einen File auszulassen,",
"                   sofern sein Name eines der Zeichen enthaelt und er",
"                   nicht ausdruecklich dem Backup-Kommando genannt wird.",
"                   Bei mkisofs-1.12b3 sollte der Text lauten: \"~#\"",
" -info_script name  legt den Filenamen fuer das getrennt zu speichernde",
"                   Infoskript fest. Es werden zwei Files erzeugt: einer mit",
"                   dem angegebenen Namen und einer mit dem Zusatz .pnr am",
"                   Namensende (zB. ASKME und ASKME.pnr)",
"                   Namen mit besonderer Bedeutung sind;",
"                    -off     kein Infoskript erzeugen",
"                   Alle anderen Namen werden als Speichernamen fuer das",
"                   Infoskript verwendet und in jedes einzelne Backupstueck",
"                   eingefuegt.",
"                   Anders als bei den Fileadressen ist hier als Trenn-",
"                   zeichen fuer Zieladressen immer fest \"=\" eingestellt.",
"                   (Beispiel: /ASKME=/home/cd_backup_planer_dir/ASKME)",
" -info_exclude     Das Infoskript nicht in die Backupstuecke einfuegen,",
"                   auch wenn vorher ein Filename mit -info_script fest-",
"                   gelegt wurde.",
" -filecheck modus  steuert die Voruntersuchung aller betroffenen Files.",
"                   modus kann sein (Grundeinstellung: -off): ",
"                    -off     keine Voruntersuchung (spart viel Zeit)",
"                    -on      warnen vor -implicite_ignored Zeichen und",
"                             bei nicht lesbaren Files",
"                    -append  wie -on. Die -implicite_ignored Namen werden",
"                             der Backupliste ausdruecklich hinzugefuegt.",
"                    -append_hexcode  wie -append. Die Problemzeichen werden",
"                             mit +XY+ kodiert, wobei XY der ASCII-Code in",
"                             hex ist.",
"                    -abort   wie -on. Wenn eine Warnung ausgegeben wurde",
"                             bricht das Programm ab.",
"                    -only    wie -on. Programm nach Untersuchung beenden.",
"                             (exit 16 = Warnung gemeldet , exit 0 = OK)",
" -filter_crc32 listenfile recordname",
"                   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Dabei zaehlt es",
"                   die Bytes und berechnet eine CRC32 Pruefsumme. Zusammen",
"                   mit dem recordname und einer Zeitmarke wird daraus ein",
"                   Record, der an listenfile angehaengt wird. recordname darf",
"                   bis zu 80 Zeichen (ohne Leerzeichen) enthalten.",
"                   Solche Records werden von -compare_checksum verwendet.",
"                   Wenn -pre_checksum_size nicht 0 ist, dann wird zunaechst",
"                   ein Vorschau-Record fuer -search_crc32 geschrieben.",
"                   CRC32 hat 32 Bits und versagt bei einem von 4e9 Tests.",
"                   Das Programm gibt 0 zurueck, wenn Daten gelesen wurden.",
"                   Wenn keine Bytes aus stdin kommen, wird kein Record",
"                   angehaengt und der Exitwert ist 1.",
" -filter_md5 listenfile recordname",
"                   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Es macht das",
"                   gleiche wie -filter_crc32, jedoch mit \"RSA Data Security",
"                   MD5 Message-Digest Algorithm\". MD5 Pruefsummen haben",
"                   128 Bit und versagen bei einem von 3e38 Tests.",
" -filter_pipe shellkommando",
"                   Hiermit wird vor den Befehlen -filter_md5 , -filter_crc32",
"                   oder -filter_through ein Shellkommando festgelegt, das",
"                   die Ausgabe dieser Befehle weiterverarbeitet. Es startet",
"                   dann beim ersten Byte das auf stdin empfangen wird.",
"                   Werden keine Bytes empfangen, dann wird Shellkommando",
"                   nicht ausgefuehrt.",
" -filter_through   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Sinnvoll wenn",
"                   es mit -filter_pipe kombiniert wird.",
" -find directory   gibt den Namen des Directory sowie aller Files und",
"                   Directories darunter aus. Wie das Programm 'find' aber",
"                   ohne jede Interpretation des Directorynamens.",
" -find_path_list filename",
"                   liest die einzelne Zeile aus File filename als Fileadresse",
"                   und wendet Option -find auf sie an.",
" -form_listing     die Ergebniszeilen werden in folgender Form gezeigt",
"                     Typ Groesse Adresse",
"                   und das Backupkommando wird nicht benutzt.",
" -form_path_list_afio listenfilename harmlose_zahl_von_argumenten",
"                   Spezialversion von -form_script_raw. Die Fileadressen",
"                   werden keine Argumente sondern in listenfilename",
"                   geschrieben (siehe auch -form_path_list_mkisofs) und",
"                   dieser listenfilename wird mit Option -path-list an das",
"                   Schreibkommando uebergeben. Es sollte den File dann mit",
"                   -find_path_list zur vollstaendigen fileliste expandieren.",
" -form_path_list_mkisofs listenfilename harmlose_zahl_von_argumenten",
"                   Spezialversion von -form_script . Geeignet fuer Versionen",
"                   von mkisofs, die die Option -path-list kennen.",
"                   Mit dieser Form sind sehr grosse Werte bei -max_items",
"                   zulaessig (z.B. 100000), weil die Uebergabe als Shell-",
"                   argumente vermieden wird.",
"                   listfilename wird erzeugt, um die Liste der Fileadressen",
"                   zu speichern. Deswegen ist Schreibrecht an seinem Wirts-",
"                   directory noetig.",
"                   harmlose_zahl_von_argumenten sollte klein genug sein,",
"                   um die Shell nicht zu ueberfordern (zB. 100).",
" -form_raw         die Ergebniszeilen enthalten nur die Adresse. Jedes",
"                   einzelne Stueck wird durch eine Leerzeile beended.",
" -form_script      die Ergebniszeilen werden als Argumente der folgenden",
"                   Form an das Backupkommando angehaengt",
"                     /addresse[/]=addresse",
"                   dies ist die Grundeinstellung fuer das Ausgabeformat.",
" -form_script_here die Ergebniszeilen werden als Standardeingabe der",
"                   folgenden Form an das Backupkommando angehaengt",
"                     addresse",
"                   Das geschieht durch ein here-document mit dem",
"                   Schlusszeichen \"+\".",
" -form_script_raw  die Ergebniszeilen werden als Argumente der folgenden",
"                   Form an das Backupkommando angehaengt",
"                     addresse",
" -level kennwort   inkrementeller Backup mit der Konfiguration, die durch",
"                   -conf_dir angewaehlt wird. Das Kennwort bestimmt ganz",
"                   wesentlich den weiteren Verlauf des Backups :",
"                     -create_configuration erzeugt eine neues -conf_dir ,",
"                      traegt dort die anderen Parameter des Laufs ein",
"                      und startet einen Level 0 Lauf.",
"                     -update_configuration traegt neue Parameter in ein",
"                      vorhandenes -conf_dir ein und startet -level 0.",
"                     -off schaltet den inkrementellen Modus ab.",
"                     -last waehlt das zuletzt benutzte Level.",
"                     -next waehlt das naechste ungenutzte Level (-last+1).",
"                     0,1,2,... direkte Anwahl des Levels durch seine Nummer.",
"                   -last, -next und Levelnummern fuehren dazu, dass die",
"                   meisten Optionen ausser -conf_dir und -level ignoriert und",
"                   durch die aus -conf_dir/specific_arguments ersetzt werden.",
"                   Einige Kommandozeilenoptionen setzen sich jedoch durch :",
"                     -backup_cmd -dir_size_cmd -form_* -info_* -max_* -min_*",
"                     -prompt_script -scattering -status_file",
"                     -target_separator -shell_to_use -verify_source_filter",
"                     (d.h. das Backupgebiet kann so nicht geaendert werden.)",
"                   Wenn der File specific_arguments von diesem Programm",
"                   erzeugt oder geaendert wird, beginnt er mit Zeilen die",
"                   folgende Optionen voreinstellen :",
"                     -working_directory ...Adresse...",
"                     -load_old_tree content_file_list",
"                                    vanished_file_list",
"                     -changetest_options type:size:crc:adr",
"                     -list_output_format type:size:crc:adr",
"                     -content_list_adr content_file_list.gz",
"                     -vanish_script_adr content_file_list",
"                                  /tmp/VANISHED.sh.gz=VANISHED.sh.gz",
"                     -vanish_list_adr content_file_list",
"                                      vanished_file_list.gz",
"                   Wenn Sie keine Inhaltslisten, Schwundlisten und",
"                   Loeschskripte bei inkrementellen Backups wollen, dann kann",
"                   das mit Option -disable_safe_leveling bestellt werden",
" -list_checksum listenfile suchname altersgrenze jugendgrenze",
"                   Das Programm endet sofort nachdem es diejenigen Records",
"                   aus listenfile ausgegeben hat, auf die die anderen drei",
"                   Argumente passen. Der letzte dieser Records wuerde von",
"                   -compare_checksum (siehe dort) verwendet werden.",
"                   Wird mindestens ein Record gefunden, dann endet das",
"                   Programm mit Exitwert 0, sonst gibt es 1 zurueck.",
" -list_output_format  code:code:...",
"                   legt fest, welche Informationen in die Listenzeilen",
"                   geschrieben werden sollen, die durch -*_list_adr Optionen",
"                   entstehen.",
"                   Die einzelnen Codes werden mit \":\" verbunden :",
"                     type  d = Directory , l = Softlink",
"                           - = Datenfile , 0 = unbekannt",
"                     size  Groesse des Inhaltes in Bytes (-1 = unbekannt)",
"                     crc   CRC32 Pruefsumme des Inhalts (-1 = unknown)",
"                     md5   MD5 (128 Bit) statt CRC32 (32 Bit)",
"                     adr   Fileadresse (immer an, Code wird ignoriert)",
"                   Die Reihenfolge ist immer  type size crc adr . zB:",
"                     d 91286 A929B47C /home/thomas/",
" -load_old_tree  altlisten_adresse  schwund_listenadresse",
"                   laedt eine Komplettliste oder die Inhaltslisten und",
"                   Schwundlisten eines Backups mit mehreren Leveln um einen",
"                   Speicherbaum aufzubauen, der alten den Zustand des",
"                   Backupgebietes wiedergibt, wie er beim Backup des",
"                   naechsttieferen Levels gewesen ist. Der Baum wird von den",
"                   Optionen -younger_than und -vanish_list_adr benutzt, um",
"                   Files zu finden, die seitdem geloescht oder umbenannt",
"                   wurden. Wenn eine -level Option vorhanden ist und die",
"                   Adressen kein \"/\" enthalten, dann werden die Files aus",
"                   den tieferen level_# Directories gelesen. Daher geschieht",
"                   bei Level 0 durch diese Option garnichts.",
"                   Wenn altlisten_adresse eine Komplettliste ist, dann muss",
"                   sonst nichts gelesen werden und schwund_listenadresse darf",
"                   irgendein nichtleerer Text sein. Wenn altlisten_adresse",
"                   nicht existiert, aber mit Zusatz .gz gefunden wird, dann",
"                   wird dieser .gz File verwendet. Wenn ein File mit .gz",
"                   gelesen wird, geschieht das durch das Programm gunzip.",
"                   WARNUNG: So ein Baum kann viel Speicher verbrauchen.",
"                   HINWEIS: altlisten_adresse von dieser Option ist identisch",
"                   mit altlisten_adresse von -vanish_list_adr. Wenn beide",
"                   Optionen angegeben sind, setzt sich die altlisten_adresse",
"                   der letzten durch durch.",
"                   HINWEIS: -load_old_tree mit nichtleeren Adressen setzt",
"                   -changetest_options auf \"type:size:crc:adr\"",
" -localdev         (betrifft zZt. nur NFS auf SGI-IRIX Systemen)",
"                   alle Filesysteme auslassen, die per Netzwerk angeschlossen",
"                   sind. Gegenteil: -alldev , siehe auch -xdev .",
" -main_dir text    Name eines Directory, das auf das root-Directory",
"                   der CD abgebildet werden soll. Besonders nuetzlich",
"                   ist der Text  -auto . Dann wird das gemeinsame Ober-",
"                   directory der Backupliste verwendet. Der Text  -off",
"                   schaltet diese Abbildung aus (Grundeinstellung).",
" -max_items zahl   die Maximalzahl von einzeln erwaehnten Files und",
"                   Directories pro Backupstueck wird festgelegt.",
"                   (Grundeinstellung: 400)",
" -max_size zahl    die Maximale Groesse eines Backupstuecks in Byte",
"                   wird festgelegt. Die Zeichen 'k', 'm' bzw. 'g' am",
"                   Ende der Zahl multiplizieren diese mit 1024,",
"                   1024*1024 bzw. 1024*1024*1024. ",
"                   (Grundeinstellung: 600m)",
" -max_size_stack groesse anzahl",
"                   meldet eine Anzahl von Backupstuecken (d.h. Datentraeger)",
"                   mit abweichender Kapazitaet an. Das kann mehrere Male",
"                   geschehen, wobei die zuletzt angemeldete Gruppe mit",
"                   hoechster Prioritaet verplant wird. Beachten Sie dann die",
"                   Groessenanmeldung der einzelnen Stuecke, um die richtigen",
"                   Medien auszuwaehlen. Zum Beispiel: 25 CDs zu 700 MB, der",
"                   Rest hat 650 MB :  -max_size_stack 685m 25 -max_size 640m",
" -max_size_stack_save -on|off",
"                   Es kann vorkommen, dass groessere Stuecke nur soweit",
"                   ausgelastet sind, dass auch -max_size ausreichen wuerde.",
"                   Normalerweise (\"-on\") gilt das -max_size_stack Stueck",
"                   dann als nicht verbraucht. Das kann dazu fuehren, dass",
"                   ein grosser Datentraeger erst unerwartet spaet angefordert",
"                   wird. Im Modus \"-off\" werden die -max_size_stack Stuecke",
"                   stur der Reihe nach verplant.",
" -merge_trees quelle ziel papierkorb optionen",
"                   Das Programm endet sofort wenn diese Option ausgefuehrt",
"                   worden ist. Sie bewegt quelle und alle Fileobjekte",
"                   darunter auf die neue Adresse ziel und analoge Adressen",
"                   darunter. Die Bewegung darf die Grenzen von Filesystemen",
"                   ueberschreiten.",
"                   ziel und eventuell schon darunter vorhandene Directories",
"                   uebernehmen die Zugangsrechte von quelle und seinen",
"                   Unterdirectories. Files unter ziel, die ueberschrieben",
"                   werden, koennen in einer analogen Adressstruktur unter",
"                   Directory papierkorb aufgefangen werden. Ist papierkorb",
"                   ein einzelner Bindestrich \"-\" so werden vorhandene Files",
"                   unter ziel beim Ueberschreiben geloescht.",
"                   Ist vor dieser Option mit -load_old_tree die Adresse einer",
"                   Directoryliste (erzeugt mit -dir_list_adr) angegeben, dann",
"                   werden die dort verzeichneten Eigenschaften ggf. auf die",
"                   Directories unter ziel uebertragen. Ansonsten werden die",
"                   Eigenschaften von quelle abkopiert.",
"                   Folgende Optionen werden erkannt :",
"                     overwrite        vorhandene Files immer ueberschreiben",
"                     no_overwrite     vorhandene Files nie ueberschreiben",
"                                      Rechte existierender Directories nicht",
"                                      aendern.",
"                     overwrite_older  vorhandene Files ueberschreiben, wenn",
"                                      in ziel aelter als in quell",
"                     move             Files bewegen (in quelle loeschen)",
"                     copy             Files kopieren (in quelle erhalten)",
"                     owner_plus_w     immer w-Recht fuer Besitzer erteilen",
"                     chmod_original   Rechte ungeaendert uebernehmen",
"                     tree_prefix=<path>  voranzustellendes Directory beim",
"                                      Aufbau von -load_old_tree. (Nur noetig",
"                                      wenn <path> ungleich quelle ist.)",
"                     tree_source      bei -load_old_tree die Quelladressen",
"                                      einlesen.",
"                     tree_target      bei -load_old_tree die Zieladressen",
"                                      einlesen.",
"                   Optionen koennen mit Doppelpunkt verbunden werden. Die",
"                   leere Optionsliste \"-\" ist gleich :",
"                     \"overwrite:move:chmod_original:tree_target\" .",
" -min_size_percent zahl  Prozentanteil von -max_size, der mindestens ge-",
"                   fuellt sein muss, bevor ein neues Stueck eroeffnet wird.",
"                   Wenn dieser Anteil nicht gefuellt ist, kann ein",
"                   groesseres Directory aufgespalten werden.",
"                   (Grundeinstellung:  80.0)",
" -no               alle folgenden Fileadressen sollen nicht mit in",
"                   den Backup aufgenommen werden. Diese Option kann",
"                   zu einer langen Ergebnisliste fuehren, weil das",
"                   Auslassen eines Directories die Erwaehnung all",
"                   seiner Nachbarn erfordert.",
" -not              wirkt genau wie -no",
"",
" -pacifier modus|rythmus  einige Optionen um inkrementelle Backups geben",
"                   Meldungen ueber ihren Fortschritt auf stderr aus. Wenn",
"                   Sie stderr in einen File leiten, kann das laestig sein.",
"                   -off schaltet die Meldungen aus, -on schaltet einen",
"                   Rythmus von etwa 1 Meldung pro Sekunde ein. Eine Zahl",
"                   waehlt einen Rythmus in Sekunden.",
" -pre_checksum_size Zahl  legt die Groesse des Anfangsstuecks im Datenstrom",
"                   fest, das fuer die Vorschaupruefsumme verwendet wird.",
"                   Wenn diese Zahl groesser als 0 ist, schreiben -filter_md5",
"                   und -filter_crc32 zwei Records. Der erste mit dem Anhang",
"                   \"_pre\" am Namen kann von den Optionen -search_md5 und",
"                   -search_crc32 genutzt werden, um den zweiten, vollen",
"                   Record zu finden. Grundeinstellung ist \"0\".",
" -scattering -on|-off  Modus -on erlaubt das Verlagern grosser Files auf",
"                   andere CDs als die, die ihre Directorynachbarn aufnehmen.",
"                   Das hilft -min_size_percent zu erfuellen, kann aber zu",
"                   weit verstreuten Directoryinhalten fuehren.",
" -search_crc32 listenfile suchname altersgrenze jugendgrenze",
"                   wie -compare_checksum aber mit automatischer Erkennung des",
"                   Datenstroms und nur fuer Records mit CRC32 Pruefsummen.",
"                   Die Suchmethode haengt von -pre_checksum_size ab. Wenn",
"                   sie 0 ist, dann werden alle Records in den Arbeitsspeicher",
"                   geladen, nach Groesse sortiert und mit Zwischenpruefsummen",
"                   des Datenstroms verglichen. Der Datenstrom wird bis zu",
"                   seinem Ende gelesen. Mehr als eine Uebereinstimmung kann",
"                   gefunden werden (besonders mit Vorschaurecords).",
"                   Wenn -pre_checksum_size >0 ist, dann wird zunaechst ein",
"                   passender Vorschaurecord gesucht, um den zugehoerigen",
"                   Record mit der vollen Pruefsumme zu finden. Die Records",
"                   fuer die Vorschau werden von -filter_md5 und -filter_crc32",
"                   erzeugt, wenn -pre_checksum_size nicht 0 ist. Diese",
"                   Vorschaugroesse muss bei filter_ und search_ gleich sein.",
" -search_md5 listenfile suchname altersgrenze jugendgrenze",
"                   wie -search_crc32 aber mit denjenigen Records, die MD5",
"                   Pruefsummen haben.",
" -set_shell_in_scdbackup_scripts shellname filename ... filename",
"                   eine besondere Option, die nur fuer die Konfiguration",
"                   verwendet wird. Die gegebenen Files werden geprueft,",
"                   ob sie mit \"#!\" beginnen. Wenn ja, wird der Rest der",
"                   Zeile durch shellname ersetzt. Es duerfen keine Optionen",
"                   mehr folgen, weil der Rest der Argumente als Liste von",
"                   Filenamen verstanden wird.",
" -shell_to_use progname  der progname wird nach \"#!\" in die erste Zeile",
"                   von Skripten gesetzt, die von diesem Programm erzeugt",
"                   werden. Grundeinstellung: /bin/sh",
" -status_file filename  Wenn der Filename nicht leer und nicht \"-\" ist,",
"                   dann wird das Ergebnisskript die aktuelle Stuecknummer in",
"                   diesen File schreiben, bevor es das -backup_cmd aufruft.",
"                   Es wird dann in der Lage sein, einen Backup am Anfang des",
"                   unterbrochenen Stueckes wieder aufzunehmen. Das geschieht,",
"                   wenn das Argument keine Stuecknummer sondern \"AUTO\" ist.",
" -target_separator text  das erste Zeichen des Textes wird als Trenn-",
"                   zeichen zwischen Zieladressen und den Fileadressen",
"                   verwendet. (Wirkt nicht auf -info_script)",
"                   Ein leerer Text schaltet die Zieladressen-Erkennung",
"                   ab. Das ist auch die Grundeinstellung.",
"                   Beispiel: Bilde /home/buero auf das /-Directory der CD ab",
"                    -target_separator ':' /:/home/buero -target_separator ''",
" -tmp_dir directory_adresse",
"                   bestimmt ein Directory, in dem temporaere Files abgelegt",
"                   werden koennen. Zur Zeit wird es fuer Listenfileadressen",
"                   ohne '/' verwendet, die eigentlich in ein Leveldirectory",
"                   sollen, wenn kein -level gesetzt ist (-complete_list_adr).",
" -vanish_list_adr  altlisten_adresse  resultat_adresse",
"                   erzeugt eine Schwundliste durch Vergleich mit den",
"                   Aufzeichnungen im File altlisten_adresse. Objekte, die",
"                   nicht mehr im Filesystem vorkommen, werden im File",
"                   resultat_adresse aufgelistet. Wenn eine -level Option",
"                   vorhanden ist und resultat_adresse kein \"/\" enthaelt,",
"                   dann wird der File im level_# Directory erzeugt (siehe",
"                   conf_dir). Wenn resultat_adresse mit .gz endet, dann wird",
"                   die Ausgabe durch das Programm gzip geleitet.",
"                   Wenn eine Zieladresse angegeben ist, dann wird der",
"                   Ergebnisfile zum letzten Stueck des Backups hinzugefuegt.",
"                   (zB: /schwund.gz=schwund.gz)",
"                   altlisten_adresse sollte ein File sein, der von",
"                   -complete_list_adr erzeugt wurde. Wenn mit -load_old_tree",
"                   ein Baum geladen wird, dann darf es auch eine Inhaltsliste",
"                   sein. Der Baum ersetzt dann die Komplettliste.",
"                   HINWEIS: altlisten_adresse von dieser Option ist identisch",
"                   mit altlisten_adresse von -load_old_tree. Wenn beide",
"                   Optionen angegeben sind, setzt sich die altlisten_adresse",
"                   der letzten durch.",
"                   WARNUNG: Ohne geladenen Baum oder bei -changetest_options",
"                   \"off\" kann die Schwundliste alte Namen von umbenannten",
"                   Files enthalten, obwohl deren neue Namen *nicht* im",
"                   inkrementellen Backup enthalten sind. Am besten ist es,",
"                   bei -changetest_options auch CRCs zu bestellen.",
" -vanish_script_adr  altlisten_adresse  result_address",
"                   WARNUNG: Das Ergebnisskript ist nicht ungefaehrlich !",
"                            Vor allem, wenn es ohne CRC-Test erzeugt wird.",
"                   Wie -vanish_list_adr , aber mit einem Shellskript als",
"                   Ergebnis, das Fileobjekte aus der Schwundliste loescht.",
"                   Solch eine Loeschung ist noetig, wenn man ein hoeheres",
"                   Level bei der Wiederherstellung des Backupgebiets ueber",
"                   das vorherige Level kopiert. Leider ist dieses Skript zu",
"                   hirnlos, um seine Opfer vorher genau zu pruefen.",
"                   HINWEIS: das Skript enthaelt nur die Zieladressen.",
" -verify_filelist  listenfile  oberdirectory",
"                   HINWEIS: Zum Pruefen von Files auf CD oder waehrend des",
"                            Wiederherstellens eignet sich -verify_targetlist",
"                            besser.",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Aus listenfile (\"-\" = stdin) werden Fileadressen und",
"                   ggf. auch Typ, Groesse und Pruefsumme gelesen.",
"                   Diese Angaben werden mit dem Ist-Zustand der Fileadresse",
"                   verglichen. Dabei wird oberdirectory jeder Adresse",
"                   vorangestellt. Existiert die Adresse und passen die",
"                   anderen Angaben, dann wird die Fileadresse einschliesslich",
"                   oberdirectory auf stdout ausgegeben.",
"                   Ansonsten geht eine Fehlermeldung an stderr.",
"                   War der gesamte Programmlauf fehlerfrei, dann ist der",
"                   Exitwert 0, sonst 1.",
" -verify_sources  listenfile",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Diese Option arbeitet fast wie -verify_filelist. Es werden",
"                   aber Adressen der Form  ziel=quelle  erwartet und nur die",
"                   Quelle wird ueberprueft. Ein Oberdirectory wird nicht",
"                   vorangestellt.",
" -verify_source_filter  shell_filter",
"                   Im Ausgabeformat -form_path_list_mkisofs koennen aus der",
"                   Fileliste jeder CD unmittelbar vor der der Aufforderung,",
"                   die CD einzulegen, solche Files entfernt werden, die",
"                   der shell_filter nicht durchlaesst. Will man zB. alle",
"                   Files weglassen, die seit der Planung von der Platte",
"                   verschwunden sind, ist ein geeigneter Filter :",
"                     ..../cd_backup_planer -verify_sources - ",
"                   Ein Leertext als shell_filter schaltet diese Pruefung ab.",
" -verify_targetlist  listenfile  oberdirectory",
"                   wie -verify_filelist. Jedoch werden die Zieladressen",
"                   verwendet, was einen Unterschied macht, wenn beim Backup",
"                   mit ziel=quelle Umleitungen bestellt wurden.",
" -volume_info      Uebergibt an das Backup-Kommando vor den Fileadressen",
"                   zusaetzliche Argumente:  -v Stuecknummer -t Stueckzahl",
" -working_directory  directory_adresse",
"                   stellt das aktuelle Directory auf die angegebene Adresse.",
"                   Diese Option sollte (wenn ueberhaupt) nur einmal vor der",
"                   ersten Fileadresse angegeben werden.",
" -xdev             alle Filesysteme auslassen, die unterhalb der Directories",
"                   in der Backupliste gemountet sind. Gegenteil: -alldev .",
" -yes              Die folgenden Fileadressen werden in die Backupliste",
"                   aufgenommen. Sie werden also noetigenfalls zerlegt und",
"                   dem Backupkommando uebergeben. Diese Option beendet",
"                   die Wirkung von \"-no\". Sie ist die Grundeinstellung.",
"                   Wenn die Option -yes mindestens einmal vorkommt, dann",
"                   werden die Fileadressen der Backupliste auf jeden Fall",
"                   ins Ergebnis uebernommen, auch wenn eine gleichlautende",
"                   in der Ausschlussliste vorkommt.",
"                    ( -yes file -not file  bringt file in den Backup",
"                      file -not file       laesst file aus) ",
" -younger_than zeitangabe  Es wird versucht, moeglichst nur Files in den",
"                   Backup aufzunehmem, deren Inhalt oder inode nach dem",
"                   angegebenen Zeitpunkt geaendert wurde. Bewegungen im",
"                   selben Filesystem bleiben leider unberuecksichtigt !",
"                   Die Zeitangaben koennen folgende Form haben :",
"                   Absolute Zeitpunkte koennen durch das Eingabeformat des",
"                   des UNIX-Befehls 'date' MMDDhhmm[[CC]YY][.ss]] festgelegt",
"                   werden. Zum Beispiel :",
"                     020616122000 heisst '6. Februar 2000 16:12:00 Ortszeit'",
"                   Ausserdem wird das Format YYMMDD[.hhmm[ss]] verstanden:",
"                     2.Jan 2000 = A00102 , 11.Mai 2001 23:45 = A10511.2345",
"                     1.Mrz 2010 = B00301 , 14.Jul 1998 = 980714",
"                   Relativangaben beginnen mit '-' oder '+'. Sie werden zum",
"                   aktuellen Zeitpunkt addiert. '-' zeigt also in die",
"                   Vergangenheit, '+' in die Zukunft. Die Grundeinheit ist",
"                   die Sekunde. Sie ist durch einen Buchstaben am Ende der",
"                   Zahl aenderbar:",
"                     Sekunden 1s=1 , Stunden 1h=3600s , Tage 1d=24h ,",
"                     Wochen 1w=7d , Monate 1m=31d, Jahre y : Xy=X*365.25d+1d",
"                   Bitte beachten: 'm' und 'y' sind oft etwas groesser als",
"                   der genaue Kalenderzeitraum.",
"                   Beispiel:",
"                     -3w  heisst 'vor 3 Wochen' = -1814400s .",
"",
"Beispiele:",
"",
"Den Inhalt des Arbeitsdirectory in Stuecke zu 625 MB und maximal 200",
"einzelne Adressen zerteilen. Das Infoskript ASKME im Arbeitsdirectory",
"erzeugen und in jedes einzelne Stueck des Backup aufnehmen.",
"",
"  %s \\",
"    -max_items 200 -max_size 625m -info_script ASKME . >bskript",
"  chmod u+x bskript",
"",
"Um den Backup dann wirklich durchzufuehren :",
"  bskript",
"",
"",
"Mit der folgenden Option wird das Infoskript an einer separaten Stelle",
"erzeugt und in das Rootdirectory der CD abgebildet.",
"  -info_script /ASKME=/home/backup_planer_dir/ASKME",
"",
"",
"Zerteile die Directories  \"buero\" und \"bilder\" in Stuecke von",
"Standardgroesse (600 MB / 400 Adressen). Nicht die Unterdirectories",
"\"buero/privat\" und \"bilder/freundinnen\" aufnehmen.",
"Anstelle von \"bilder/freundinnen\" eine zweite Kopie von \"bilder/katzen\"",
"einfuegen.",
"",
"  %s \\",
"    buero bilder -target_separator '=' /bilder/freundinnen=bilder/katzen \\",
"    -not buero/privat bilder/freundinnen \\",
"  >bskript",
"",
"",
"DC-6250 Baender mit komprimierten tar-Archiven fuellen.",
"",
"  %s \\",
"    -backup_cmd 'read dummy ; tar cvzf /dev/rmt0' \\",
"    -form_script_raw -max_size 245m -max_items 100 \\",
"    -info_script ASKME \\",
"    . \\",
"  >bskript",
"",
"",
"Eine Konfiguration fuer inkrementelle Backups erzeugen und den ersten Backup",
"-level 0 planen lassen. Durch -level -create_configuration wird ein neues",
"Directory angelegt, das (wegen -conf_dir) $HOME/my_backup_conf heisst.",
"",
"  %s \\",
"    -conf_dir $HOME/my_backup_conf -level -create_configuration \\",
"    ... alle Argumente, die den vollen Backup definieren ... \\",
"  >bskript",
"",
"Diese Konfiguration kann man nutzen, indem man einfach das Directory und die",
"gewuenschte Ebene angibt. Die anderen Parameter des Backups werden aus dem",
"-conf_dir gelesen. (File specific_arguments : eine Zeile = ein Argument)",
"",
"Man sollte sich ein festes Schema fuer Backups auferlegen und durchhalten.",
"Angenommen, wir haetten am Sonntag der 0. Woche den Backup konfiguriert",
"und also auch den ersten -level 0 Backup gemacht :",
"",
"Ein Schema, das von Montag bis Samstag nur die Aenderungen des Tages",
"aufzeichnet und am Sonntag einen vollen Backup vorsieht.",
"  Montag:   -conf_dir $HOME/my_backup_conf -level -next",
"  ...",
"  Samstag:  -conf_dir $HOME/my_backup_conf -level -next",
"  Sonntag:  -conf_dir $HOME/my_backup_conf -level 0",
"Nun koennen die Datentraeger von Montag bis Samstag recycelt werden.",
"",
"Ein Schema, das etwa vierteljaehrlich volle Backups macht, sowie nach vier",
"Wochen und nach jeder einzelnen Woche auffrischt. (-conf_dir wie oben)",
"Dieses Schema macht Aenderungen an Files ueber laengere Zeit nachvollziehbar.",
"  1. Woche :  -level 1",
"  2. Woche :  -level 2",
"  3. Woche :  -level 2",
"  4. Woche :  -level 2",
"  5. Woche :  -level 1",
"  6. Woche :  -level 2",
"  7. Woche :  -level 2",
"  8. Woche :  -level 2",
"  9. Woche :  -level 1",
" 10. Woche :  -level 2",
" 11. Woche :  -level 2",
" 12. Woche :  -level 2",
" 13. Woche :  -level 0",
"Danach beginnt man wieder bei Woche 1.",
"Der Backup von Woche 12 sichert auch die Wochen 11 und 10 ab.",
"Der Backup von Woche 9 sichert Wochen 8 bis 1 ab.",
"Die Levels 0 und 1 sollte man laenger aufheben.",
"",
"Um beide gezeigte Schemata zu kombinieren, sollte man im ersten Schema den",
"Sonntagslauf durch den woechentlichen Lauf des zweiten Schemas ersetzen.",
"",
"Wenn Sie je einen inkrementellen Backup mit mehreren Levels wiederherstellen",
"muessen, beginnen Sie mit Level 0. Nachdem das auf der Platte ist, muessen",
"die Files geloescht werden, die in der Woche bis zum Level 1 verschwunden",
"sind.",
"Zu diesem Zweck finden Sie im letzten Backupstueck jedes Levels einen File",
"namens /tmp/VANISHED.sh.gz . Er hat keine x-Rechte, um ueble Unfaelle zu",
"vermeiden. Man kann ihm nur trauen, wenn er durch Vergleich mit einem",
"aufgezeichneten Speicherbaum unter Verwendung von CRCs erstellt wurde. Das",
"ist zwar Grundeinstellung aber die kann geaendert werden mit Optionen wie :",
"-content_list_adr , -list_output_format , -load_old_tree ,",
"-changetest_options , -vanish_script_adr ",
"Besonders misstrauische Datenbesitzer verwenden bei -list_output_format die",
"Option \"md5\" statt \"crc\".",
"",
"Man kann das Skript auf die Platte kopieren, es mit gunzip auspacken und",
"ihm mit  chmod u+x  Ausfuehrungserlaubnis erteilen und es dann starten.",
"Man kann es aber auch mit dieser Shellpipe direkt ausfuehren :",
"  gunzip </cdrom/tmp/VANISHED.sh.gz | /bin/sh",
"Man sollte aber vorher auf jeden Fall einen Blick auf seinen Inhalt werfen",
"(mit \"less\" statt \"/bin/sh\"). Letzte Chance diese Files zu begnadigen.",
"",
"Das VANISHED.sh Skript von Level 1 kann irgendwann nach der Installation von",
"Level 0 und vor der Installation von Level 2 ausgefuehrt werden. Wenn",
"Plattenplatz knapp ist, macht man es besser vor der Installation von Level 1.",
"Allerdings ist es etwas sicherer, das Skript auszufuehren, nachdem Level 1",
"auf die Platte kopiert wurde. Man hat dann eher zuviele als zuwenig Files,",
"wenn man unterbrochen wird.",
"Wenn man dann die Levels 2,3,4,... installiert, darf man nicht vergessen,",
"auch jeweils deren VANISHED.sh Skripte rechtzeitig auszufuehren. Eventuell",
"geraten Sie auch beim Installieren auf die Platte und sollten dort besser",
"wieder geloescht werden.",
"",
"",
"",
"@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@"
 };
 char line[4096];
 int i;
 
 MELDEN("");
 sprintf(line,"cd_backup_planer Version %s",PROG_VERSION);
 MELDEN(line);
 MELDEN("");
 for(i=0;strcmp(hilfetext[i],
                "@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@"
                )!=0;i++){
   sprintf(line,hilfetext[i],progname);
   MELDEN(line);
 }
 return(1);
}


int Backup_plan_help(progname,flag)
char *progname;
int flag;
{
 static char hilfetext[][80]={
"usage :  %s  [option | file] ...",
"",
"Usually this command splits a list of directories into volumes",
"which fit on CD-R media.",
"It generates a script with backup commands. The lists of files",
"or directories are determined according to maximum capacities",
"in terms of total size and number of items.",
"",
"A result list can contain only directories and datafiles.",
"Other file types may be included in directories but most CD",
"writer programs will not process them.",
"",
"The result lists consist of target and source addresses suitable",
"for the program  mkisofs  in its more modern versions:",
"  If fileaddress is a directory:  /targetaddress/=fileaddress",
"  If it is a data file:           /targetaddress=fileaddress",
"These lists are appended to the backup command in order to fill",
"one CD medium. If not given explicitely and not redirected by a",
"collision, the targetaddress is the same as the fileaddress.",
"The character \"=\" is changed to \"_\".",
"Other output forms of the result lists are available too.",
"",
"Since the working directory, the root directory and possibly",
"the -main_dir all are mapped to the root directory collisions",
"of targetadresses may occur. A builtin collision test can detect",
"and avoid them by renaming one of the affected targetadresses.",
"",
"A filecheck can be made to find problematic files:",
" - those which are ignored by mkisofs-1.12b3 if not given",
"   explicitely as arguments.",
" - those which cannot be read (files or directories)",
"   this can mislead the estimation of resulting directory sizes.",
"The combined options -filecheck -append -filecheck -exclude can",
"help to work around these problems.",
"",
"Incremental backups are possible with an arbitrary number of levels",
" -level 0 does a full backup with no time constraints.",
" -level N backups those files that have changed since the last run",
"  of -level N-1 . -level N-1 has to exist already.",
" - A backup with -level N makes invalid all levels above N.",
"The data necessary for keeping track of levels are kept in a",
"dedicated directory which is addressed by -conf_dir. It also hosts",
"all parameters of the backup like filelist or maximum sizes.",
"",
"Incremental backups use the option -younger_than which possibly",
"also includes some older files in the backup. Regrettably, mere",
"name changes do not change the file's time stamps in the general",
"Linux filesystem model. (other than e.g.:  chmod u+r )",
"",
"Since a large number of single filenames could make problems with",
"older versions of mkisofs, the incremental backup mainly tries to",
"exclude directories which contain only very few young files and to",
"mention only these few exemption files. Also a certain effectivity",
"is required in terms of byte gain versus number of filenames. This",
"required quote is high if -max_items is set to a small value.",
"",
"An additional script can be generated which is able to print",
"several informations about the backup.",
"Special options are:",
"-hilfe, -id, -count, -list, -levelall , -suball, -fast, -help",
"Other arguments are looked up as file names in the lists.",
"If a name or a directory above this name is found, then the part",
"number, the item and the file name are printed. If none of the",
"files can be found this way, then the script's exit value is 1.",
"",
"This script is stored on the disk and included in each part of",
"backup if a name is given with option -info_script (see below).",
"It is estimated to have 100k size (in case that your scripts",
"actually exceed this size reduce -max_size accordingly).",
"If the info script gets included and the output form for mkisofs",
"(default or -form_script) is used then all possible fileadresses",
"which may lead to the same name on the CD are excluded.",
"",
"A number as optional argument tells the backup script to skip all",
"volumes with lower numbers. To continue a backup which has been",
"interrupted while volume 3 was ready to be written:",
"  scriptname 3",
"Then the script skips volumes 1 and 2.",
"If the script has been generated with -status_file then it can find",
"the appropriate volume automatically by reading the status file",
"This automatic is triggered with the argument \"AUTO\".",
"",
"Arguments which do not begin with '-' are file addresses.",
"Fileadresses may be prefixed by target adresses. This is useful",
"for composing a CD from various scattered sources. To enable",
"explicit target addresses a separating character has to be chosen",
"with the option -target_separator.",
"Example: -target_separator '=' /images=/home/private/images ...",
"",
"All fileaddresses together define the \"backup area\" from which some",
"sub trees may be excluded by options -no, -not, -xdev, -localdev .",
"This set of file objects may get examined and reduced further",
"(e.g. by -younger_than) and so becomes the \"backup content\".",
"",
"Options:",
" ",
" -help             print this text and exit program",
" -hilfe            diesen Text ausgeben und Programm beenden",
"                   (like -help , use german language)",
" -version          print version identification and exit program",
" --                the next argument is a file, not an option.",
"",
" -accept_all_filetypes",
"                   usually only directories and regular data files",
"                   get mentioned in the result script. This option",
"                   enables all filetypes",
" -alldev           do not exclude mounted filesystems. (default)",
" -backup_cmd text  the given text is used as command in the",
"                   script generated by this program.",
"                   (default: cd_backup_kommando)",
" -changetest_options  code:code:...",
"                   chooses properties to be compared during check",
"                   -younger_than. Compared are the file objects on",
"                   the disk with those obtained by -load_old_tree.",
"                   Multiple codes are separated by \":\". ",
"                     type  e.g. directory, data file, softlink",
"                     size  bytecount",
"                     crc   CRC-32 content checksum",
"                     off   disable comparison with old tree",
"                     adr   enable comparison with old tree (default)",
"                     mistrust    avoid a fast shortcut depending on",
"                                 directory timestamps and CRCs.",
"                     assume_old  assume all timestamps as old enough",
"                   NOTE: -load_old_tree with nonempty addresses",
"                   sets -changetest_options to \"type:size:crc:adr\"",
"                   Set own -changetest_options after -load_old_tree",
" -collision_test mode  controls the collision test for target names",
"                   Modes can be (default is -redirect ) :",
"                     -off       do not perform collision test",
"                     -on        warn about collisions",
"                     -redirect  change target names if necessary",
"                     -abort     abort if a collision is detected",
" -compare_checksum recordlist searchname old_limit young_limit",
"                   the program ends immediately when stdin closes.",
"                   It passes stdin to stdout without changes.",
"                   Reads from file recordlist a record matching the",
"                   other three arguments. The record may have been",
"                   written by -filter_crc32 or -filter_md5. In order",
"                   to match, searchname has to be \"-\" or exactly",
"                   the same as the name of the record. If old_limit",
"                   is not \"-\" all older records do not match. If",
"                   young_limit is not \"-\" then all younger ones do",
"                   not match. (see -younger_than for time formats)",
"                   The last matching record in the list is read and",
"                   its recorded data amount is read from stdin to",
"                   compute a checksum which is compared with the",
"                   recorded checksum. If stdin delivers enough",
"                   bytes and the checksums are equal, then the",
"                   return value of the program is 0. In case of",
"                   not enough input or wrong checksum exit is 1.",
"                   Any surplus bytes at stdin are ignored.",
" -complete_list_adr fileaddress",
"                   writes a complete file list of the backup area",
"                   to the given fileaddress. If fileaddress contains no \"/\"",
"                   and a -level is given then the file is created in the",
"                   level_# directory. (See below : -conf_dir) If no -level",
"                   is given then such a file goes to -tmp_dir.",
"                   If the given fileaddress ends with .gz then the",
"                   output gets piped through program gzip.",
"                   If a target is given in fileaddress then the",
"                   resulting file will be appended to the last",
"                   backup volume. (e.g. /complete.gz=complete.gz)",
"                   The list is written in a special format to save",
"                   disk space while still being human readable. The",
"                   first and the last characters of a line determine",
"                   its type",
"                     /x/x/x/   absolute directory address",
"                     /x/x/x    absolute file address",
"                     ./x/      descent into directory x",
"                     ./x       file x in current directory",
"                     ..        one directory up",
"                     ../       to start directory",
"                     #=# xxxx  additional information",
"                   The first directory address given is absolute and",
"                   sets the start directory for relative addresses.",
"                   By ./x/ and /x/x/x the directory cursor can be",
"                   moved away from the start directory. As long as",
"                   the cursor is in the start directory, addresses",
"                   get prepended a target address which also is",
"                   valid for all relative addresses below.",
"                   The resulting list is usable with -vanish_list_adr",
" -content_list_adr fileaddress",
"                   writes a file list of the backup content. With",
"                   the same rules and formats as -complete_list_adr",
"                   this option lists only those file objects which",
"                   are part of a backup volume. The content lists",
"                   and vanish lists of all levels together contain",
"                   the same information as a single complete list.",
"                   Additionally a content list tells the volume",
"                   numbers by marking volume borders.",
" -conf_dir directoryadress",
"                   sets the location of a backup configuration. It",
"                   contains files which are created and used by",
"                   incremental backups. (see -level)",
"                     specific_arguments : full backup's options list",
"                                          each line is one argument.",
"                     highest_level      : highest valid level number",
"                     ASKME_level_#      : info script of -level #",
"                     level_#/           : level directory, contains",
"                       last_run_date    : time stamp of last run",
"                       number_of_volumes : as it says",
"                       content_file_list.gz : see -content_list_adr",
"                                              and -load_old_tree",
"                       vanished_file_list.gz: see -vanish_list_adr",
"                       VANISHED.sh.gz       : see -vanish_script_adr",
" -delete_filelist  list_adr  directory_prefix",
"                   NOTE: for removing obsolete files during the restore of",
"                         a backup you should better use -delete_targetlist.",
"                   The program ends immediately after this option is",
"                   done. It reads file addresses from list_adr",
"                   together with type, size, checksum properties.",
"                   These informations get compared with the current",
"                   state of the file whose address is composed from",
"                   directory_prefix and the address read from the",
"                   list.",
"                   The file list has to be generated with ",
"                    -vanish_list_adr and must contain type, size",
"                   and checksum informations.",
"                   If the addressed file exists and if the other",
"                   properties match, then the file gets deleted.",
"                   Directories get deleted if they are empty.",
"                   If everything existed and matched, the program",
"                   exits with value 0. Else this value is 1.",
" -delete_targetlist  list_adr  directory_prefix",
"                   like -delete_filelist but processing the target addresses",
"                   of the vanished files. If there have been target=source",
"                   redirections when a backup was made, then this can make",
"                   a big difference.",
" -dir_list_adr fileaddress",
"                   writes a complete directory list of the backup area to",
"                   the given fileaddress. The output format is similar to",
"                   -complete_list_adr but the fields are fixed to",
"                   \"type:chmod:chown:chgrp:date:size:crc:adr\" with",
"                   chmod:chown:chgrp:date defining access permissions,",
"                   ownership and time stamp. If MD5 checksums have been",
"                   enabled by -list_output_format then they are used here",
"                   rather than CRC32 checksums.",
" -dir_script_adr fileaddress",
"                   writes a shell script for restoring the access permissions",
"                   of the directories of the backup area. The script will use",
"                   target addresses relative to the CDs' root directory.",
" -dir_size_cmd text the given text is used as command for determination of",
"                   a directory's size. The command has to expect a single",
"                   argument (the adress of the directory) and give back as",
"                   first word on standard output the size in terms of",
"                   kilobytes (like  du -sk  does).",
"                   Default:  \"-internal\" does not run an external command",
"                   but recursivly sums up the sizes of file objects and the",
"                   length of their addresses. Added is an overhead of 256",
"                   bytes per file. It may be changed by \"-internal+$num\" .",
"                   For example \"-internal+10\" adds 10 extra bytes.",
"                   \"-internal-$num\" disables adding of address length.",
" -disable_safe_leveling  is equivalent to ",
"                   -load_old_tree \"\" \"\"",
"                   -content_list_adr \"\" -complete_list_adr \"\"",
"                   -vanish_script_adr \"\" \"\" -vanish_list_adr \"\" \"\"",
" -exclude_size sizelimit  exclude any single file which contains at least",
"                   sizelimit bytes. A sizelimit of 0 disables this test.",
"                   If sizelimit begins with \"-\" then with symbolic links",
"                   the size of the link target gets tested.",
"                   'k' 'm' and 'g' are recognized like with -max_size .",
" -implicite_ignored text  addresses a problem with mkisofs-1.12b3",
"                   The text tells which characters cause the",
"                   backup command to ignore filenames that are",
"                   not explicitely given as arguments.",
"                   For mkisofs-1.12b3 text should be: \"~#\"",
" -info_script name  set the fileaddress for the separate information",
"                   script. Be aware that two files are created.",
"                   One with the name given and one with name",
"                   extension .pnr (e.g. ASKME and ASKME.pnr).",
"                   Special names are:",
"                    -off     don't generate info script",
"                   other names are used to store the script and",
"                   included in each part of the backup.",
"                   Different from file addresses the separating",
"                   character for the -info_script name is fixed",
"                   \"=\".",
"                   Example:  /ASKME=/home/cd_backup_planer_dir/ASKME",
" -info_exclude     don't include the info script in all backup",
"                   parts automatically despite a file name was set",
"                   with -info_script previously.",
" -filecheck mode   controls the examination of all files.",
"                   Modes can be (default is -off ) :",
"                    -off     no examination (saves much time)",
"                    -on      warn about -implicite_ignored files",
"                             and unreadable files",
"                    -append  like -on. Append -implicite_ignored",
"                             files explicitely to the backup list.",
"                    -append_hexcode  like -append.",
"                             The problem causing characters are",
"                             replaced by +XY+ (XY=ASCII code in hex)",
"                    -exclude like -on. Exclude unreadable files",
"                             explicitely from the backup list.",
"                    -abort   like -on. Abort program if warnings",
"                             have been issued.",
"                    -only    like -on. End program after check.",
"                             (exit 16 = warning issued , 0 = OK)",
"                   useful: -filecheck -exclude -filecheck -append",
" -filter_crc32 recordlist recordname",
"                   the program ends immediately when stdin closes.",
"                   Passes stdin to stdout without changes. Computes",
"                   a CRC32 checksum and counts the bytes. Together",
"                   with the recordname and a time mark they get",
"                   appended as a record to file recordlist.",
"                   The recordname maybe a text without spaces with",
"                   up to 80 characters.",
"                   Such records are used by -compare_checksum .",
"                   If -pre_checksum_size is not 0 there will also",
"                   be written a preview record for -search_crc32 .",
"                   CRC32 got 32 bits and may fail in 1 of 4e9 tests.",
"                   The program will return 0 if any data were read.",
"                   If no bytes were seen at stdin, no record is",
"                   written and exit value is 1.",
" -filter_md5 recordlist recordname",
"                   the program ends immediately when stdin closes.",
"                   Like -filter_crc32 but with \"RSA Data Security,",
"                   MD5 Message-Digest Algorithm\" checksum.",
"                   MD5 got 128 bits and may fail in 1 of 3e38 tests.",
" -filter_pipe shellkommando",
"                   Given before -filter_md5 , -filter_crc32 or",
"                   -filter_through this defines a shell command",
"                   which gets started as soon as the first input",
"                   byte arrives on stdin. The filter's output is",
"                   piped into that command. If no bytes are received",
"                   then the shell command is not started at all.",
" -filter_through   the program ends immediately when stdin closes.",
"                   Passes stdin to stdout without changes. Useful",
"                   together with -filter_pipe .",
" -find directory   prints the directory's name and the names of all",
"                   files and directories below. Like program 'find'",
"                   but without any interpretation of the given name.",
" -find_path_list filename",
"                   reads fileaddresses line by line from file",
"                   filename and applies option -find to each of it.",
" -form_listing     the result list lines are shown in the form",
"                     Type Size Address",
"                   the backup_cmd is not used.",
" -form_path_list_afio listfilename safe_number_of_arguments",
"                   a special version of -form_script_raw. The file",
"                   addresses do not become arguments but are written",
"                   to listfilename (see -form_path_list_mkisofs) and",
"                   this listfilename is handed by option -path-list",
"                   to the writing command. It may then use option",
"                   -find_path_list to get a full file list for afio.",
" -form_path_list_mkisofs listfilename safe_number_of_arguments",
"                   a special version of -form_script suitable for",
"                   versions of mkisofs which know option -path-list",
"                   With this form a very large setting of -max_times",
"                   is permittable (like 100000) since it avoids to",
"                   use them as arguments of shell commands.",
"                   listfilename will be created to store the list of",
"                   items. Therefore write permission is needed for",
"                   its hosting directory.",
"                   safe_number_of_arguments should be small enough",
"                   to comfort your shell (e.g. 100).",
" -form_raw         the result list lines only show the addresses.",
"                   The volume ends are marked by empty lines.",
" -form_script      the result list lines are appended as arguments",
"                   to the command in the form",
"                     /address[/]=address",
"                   this is the default output format.",
" -form_script_here the result list lines are appended as standard",
"                   input lines to the command in the form",
"                     address",
"                   this is done inline by a here-document with",
"                   delimiter \"+\".",
" -form_script_raw  the result list lines are appended as arguments",
"                   to the command in the form",
"                     address",
" -level code       incremental backup with the configuration that",
"                   was selected by -conf_dir. The code determines",
"                   the further action of the backup :",
"                     -create_configuration creates -conf_dir ,",
"                      records the other parameters of the backup",
"                      and finally starts a -level 0 run.",
"                     -update_configuration records new parameters",
"                      in an existing -conf_dir and starts -level 0.",
"                     -off disables the incremental mode.",
"                     -last chooses the level of the most recent run.",
"                     -next chooses the next unused level (-last+1).",
"                     0,1,2,... direct level choice by number.",
"                   By -last, -next or level numbers, many options other",
"                   than -conf_dir and -level are ignored and replaced",
"                   by those from -conf_dir/specific_arguments. A few",
"                   command line options will be able to override the",
"                   recorded ones :",
"                     -backup_cmd -dir_size_cmd -form_* -info_* -max_* -min_*",
"                     -prompt_script -scattering -status_file",
"                     -target_separator -shell_to_use -verify_source_filter",
"                     (I.e. you cannot change the backup area that way)",
"                   If created or updated by this program, specific_arguments",
"                   begins with lines equivalent to these options:",
"                     -working_directory ...address...",
"                     -load_old_tree content_file_list",
"                                    vanished_file_list",
"                     -changetest_options type:size:crc:adr",
"                     -list_output_format type:size:crc:adr",
"                     -content_list_adr content_file_list.gz",
"                     -vanish_script_adr content_file_list",
"                                  /tmp/VANISHED.sh.gz=VANISHED.sh.gz",
"                     -vanish_list_adr content_file_list",
"                                      vanished_file_list.gz",
"                   If you do not want a vanish script in incremental",
"                   backups then you may disable this feature by :",
"                   option -disable_safe_leveling within your own",
"                   options to this program.",
" -list_checksum recordlist searchname old_limit young_limit",
"                   the program ends immediately after listing those",
"                   records from recordlist which match the other",
"                   three arguments. The last one listed would be the",
"                   one used by -compare_checksum (see there).",
"                   The program will return 0 if any records were",
"                   listed. Otherwise its exit value is 1.",
" -list_output_format  code:code:...",
"                   defines which information shall be written to the",
"                   lines resulting from -*_list_adr options.",
"                   Multiple codes are separated by \":\". ",
"                     type  d = directory , l = softlink",
"                           - = datafile  , 0 = unknown",
"                     size  bytecount (-1 = unknown)",
"                     crc   CRC32 content checksum (-1 = unknown)",
"                     md5   MD5 (128 bit) rather than CRC32 (32 bit)",
"                     adr   file address (always on, code ignored)",
"                   The sequence is always type size crc adr . Eg.:",
"                     d 91286 A929B47C /home/thomas/",
" -load_old_tree  old_list_adr vanish_list_adr",
"                   loads a complete file list or a layered set of",
"                   content and vanish lists in order to form a tree",
"                   of file objects. This tree represents an old",
"                   state of the backup area (typically from the time",
"                   when the next lower level's backup was made).",
"                   This tree is used by -younger_than and by",
"                   -vanish_list_adr to determine files which have",
"                   been moved or deleted since then.",
"                   If a -level is given and the addresses contain",
"                   no \"/\" then the files are read from the lower",
"                   level_# directories. Thus no action at level 0.",
"                   If old_list_adr points to a complete list, then",
"                   no vanish lists need to be read and therefore",
"                   vanish_list_adr may be any nonempty text.",
"                   If a list file itself is missing but exists",
"                   with name extension .gz then that file used. If",
"                   the effective filename ends with .gz it gets read",
"                   via a pipe through program gunzip.",
"                   CAUTION: a loaded tree might need much memory.",
"                   NOTE : old_list_adr from this option is identical",
"                   with old_list_adr from -vanish_list_adr. If both",
"                   options are given, the last one will finally set",
"                   this address.",
"                   NOTE: -load_old_tree with nonempty addresses",
"                   sets -changetest_options to \"type:size:crc:adr\"",
" -localdev         (in effect only with NFS on SGI-IRIX systems)",
"                   exclude all remote filesystems.",
"                   Opposite of -alldev. See also -xdev",
" -main_dir text    name of a directory which shall be mapped to the",
"                   root directory of the CD. Especially useful is",
"                   the text  -auto . With this, the common directory",
"                   of the backup list (if any) is used. Text  -off",
"                   disables this mapping (default).",
" -max_items number the maximum number of items in a result list",
"                   of directories and files appended to the",
"                   backup command. (default: 400)",
" -max_size number  the maximum number of bytes per volume.",
"                   Characters 'k', 'm' or 'g' at the end of the",
"                   given number multiply by 1024, 1024*1024 or",
"                   1024*1024*1024. (default: 600m)",
" -max_size_stack size count",
"                   registers a number of volumes (i.e. media) with a",
"                   capacity different from -max_size. This may happen several",
"                   times. The volume group registered last will be used for",
"                   planning first. Watch the size announcement of the",
"                   volume to choose the appropriate media size. Example :",
"                   If you got 25 CDs with 700 MB and the rest stores 650 MB :",
"                     -max_size_stack 685m 25 -max_size 640m",
" -max_size_stack_save -on|off",
"                   It may happen that larger volumes cannot be filled with",
"                   more than -max_size bytes. In that case in default mode",
"                   \"-on\" the -max_size_stack volume is not counted as",
"                   consumed and will be retried with the next backup volume.",
"                   This can cause large media to be required surprisingly",
"                   late. In mode \"-off\" -max_size_stack volumes are",
"                   consumed one by one unconditionally.",
" -merge_trees source target dump options",
"                   The program ends immediately after this option has been",
"                   executed. It moves source and all file objects underneath",
"                   to target and analog addresses underneath. The move may",
"                   cross filesystem borders. target and the existing",
"                   directories underneath get the access permissions of",
"                   source and its subdirectories.",
"                   Files which get overwritten may be saved in a directory",
"                   tree underneath  dump . If dump is a single \"-\" then",
"                   eventually existing files underneath of target get lost",
"                   when overwritten.",
"                   If prior to this option a directory list is given with",
"                   option -load_old_tree then the properties recorded there",
"                   will be copied to the target directories. (Such a list",
"                   may be generated via -dir_list_adr.) Without the list",
"                   directory properties will be copied from source.",
"                   There are the following options :",
"                     overwrite        overwrite existing files",
"                     no_overwrite     do not overwrite such files",
"                                      do not change existing",
"                                      directory permission",
"                     overwrite_older  overwrite only if target file",
"                                      is older than source file",
"                     move             move files (delete source)",
"                     copy             copy files (preserve source)",
"                     owner_plus_w     give w-permission to owner in any case",
"                     chmod_original   copy permissions unchanged",
"                     tree_prefix=<path>  the prefix directory for loading",
"                                      a directory list with -load_old_tree.",
"                                      Only needed if <path> is not source.",
"                     tree_source      with -load_old_tree read source part",
"                                      of addresses.",
"                     tree_target      with -load_old_tree read target part.",
"                   Options may be concatenated by a colon. The empty option",
"                   list \"-\" is the same as :",
"                     \"overwrite:move:chmod_original:tree_target\" .",
" -min_size_percent number  percentage of -max_size which has to be",
"                   used of a volume before a new one may be opened.",
"                   If this percentage isn't reached then a large",
"                   directory may be split to fill the volume.",
"                   (default is 80.0)",
" -no               all following fileaddresses are to be omitted",
"                   from the result lists. This option can cause",
"                   the result lists to fragment since omitting a",
"                   address within a directory requires to put to",
"                   the list all other addresses within this",
"                   directory.",
" -not              same as -no",
"",
" -pacifier mode|rythm  some options concerning incremental backups",
"                   print intermediate messages about their progress",
"                   to stderr. If you are logging stderr this may be",
"                   annoying. -off disables, -on enables with a rythm",
"                   of about 1 message per second. A number sets the",
"                   message rythm to the a value in seconds.",
" -pre_checksum_size number  sets the amount of data from the start",
"                   of the stream which is used to compute a preview",
"                   checksum.If this number is not 0 then -filter_md5",
"                   and -filter_crc32 write two records. The first",
"                   with name appendix \"_pre\" may be used by option",
"                   -search_md5 and -search_crc32 to locate the",
"                   second one which describes the full data stream.",
"                   Default is \"0\"",
" -scattering -on|-off  mode -on allows moving large files to CDs",
"                   different than those which host their directory",
"                   siblings. This helps to fulfill -min_size_percent",
"                   but can lead to scattered directory contents.",
" -search_crc32 recordlist searchname old_limit young_limit",
"                   like -compare_checksum but with automatic volume",
"                   recognition and only for records with CRC32",
"                   checksums. The search method depends on the",
"                   setting of  -pre_checksum_size  . If it is 0 then",
"                   all records are read, sorted by size and compared",
"                   with intermediate checksums of the data stream.",
"                   The stream is read up to its end and more than",
"                   one match may occur if streams with identical",
"                   data were recorded.",
"                   If -pre_checksum_size is >0 then at first a",
"                   matching preview record is searched to determine",
"                   the record with the full stream's checksum.",
"                   The preview checksums are computed by -filter_md5",
"                   and -filter_crc32 if pre_checksum_size was not",
"                   set to 0. The sizes of record and search have",
"                   to match in this case.",
" -search_md5 recordlist searchname old_limit young_limit",
"                   like -search_crc32 but with all records which",
"                   have MD5 checksums.",
" -set_shell_in_scdbackup_scripts shellname filename ... filename",
"                   a very special option which is used only for",
"                   configuration. The given files get tested wether",
"                   they start with \"#!\". If true, the rest of the",
"                   first line gets replaced by shellname. No other",
"                   option may follow since the rest of arguments is",
"                   used as a list of filenames.",
" -shell_to_use progname  the progname is written after \"#!\"",
"                   into the first line of scripts generated by",
"                   this program. Default: /bin/sh",
" -status_file filename  If the filename is not empty and not \"-\"",
"                   then the result script will write the current",
"                   volume number into this file before it calls",
"                   the -backup_cmd . Also the script will be able",
"                   to resume at the start of an interrupted volume",
"                   when given the argument \"AUTO\" rather than",
"                   a volume number.",
" -tmp_dir directory_address",
"                   sets a directory for storing temporary files. Currently",
"                   it is used for list file addresses without '/' which",
"                   would go to the level directories if there was any -level",
"                   set. (see -complete_list_adr)",
" -target_separator text  the first character in text is used as",
"                   separating character between target address",
"                   and file address. (Doesn't apply to -info_script)",
"                   An empty text disables explicit target addresses",
"                   (this is the default setting).",
"                   Example: map /home/office to the / directory of",
"                            the CD:",
"                    -target_separator ':' /:/home/office -target_separator ''",
" -vanish_list_adr  old_list_adr  result_address",
"                   checks the file objects which have been recorded",
"                   some time ago in old_list_adr. Those objects",
"                   which have vanished since then are reported to",
"                   file result_address. If a -level is given and",
"                   result_address contains no \"/\" then the file is",
"                   created in the level_# directory. (See -conf_dir)",
"                   If result_address ends with .gz then the output",
"                   gets piped through program gzip.",
"                   If a target is given in result_address then the",
"                   resulting file will be appended to the last",
"                   backup volume. (e.g. /vanished.gz=vanished.gz)",
"                   old_list_adr may be output of -complete_list_adr",
"                   Alternatively a plain list of file addresses may",
"                   be used.",
"                   If a tree gets loaded by -load_old_tree then it",
"                   is used for comparison. If there is no old tree",
"                   then the old_list_adr gets read directly but it",
"                   has to be a complete list in a single file.",
"                   NOTE : old_list_adr from this option is identical",
"                   with old_list_adr from -load_old_tree. If both",
"                   options are given, the last one will finally set",
"                   the effective address.",
"                   CAUTION: If -load_old_tree adresses are empty or",
"                   -changetest_options is \"off\" then the result",
"                   list may contain old file adresses that have been",
"                   moved but the backup plan will not contain the",
"                   new addresses. Best is to also test for CRCs",
"                   (see -changetest_options)",
" -vanish_script_adr  old_content_list_adr  result_address",
"                   CAUTION: Running the result script is dangerous !",
"                            Especially if generated without CRC test",
"                   Same as -vanish_list_adr but reports shell",
"                   commands to form a script that removes all",
"                   vanished objects. Such a removal is necessary",
"                   before applying an incremental level during",
"                   re-install of a multilevel backup. But this",
"                   script does not check against cruel accidents.",
"                   NOTE: the script contains target addresses only.",
" -verify_filelist  list_adr  directory_prefix",
"                   NOTE: If you want to verify the backuped files on CD or",
"                         during restore then you should better use",
"                         -verify_targetlist.",
"                   The program ends immediately after this option is",
"                   done. It reads file addresses from list_adr",
"                   together with eventual informations about type,",
"                   size and checksum. The informations get compared",
"                   with the current state of the file. Its address",
"                   is composed from directory_prefix and the address",
"                   read from the list.",
"                   If the addressed file exists and if the other",
"                   properties match, then the file's address is",
"                   printed on stdout. Else there will be an error",
"                   message at stderr.",
"                   If everything existed and matched, the program",
"                   exits with value 0. Else this value is 1.",
"                   Currently this is mainly a harmless version of",
"                    -delete_filelist . But that might change.",
" -verify_sources  listfile",
"                   the program ends immediately after this option is",
"                   done. This option is similar to -verify_filelist",
"                   but it expects addresses of form target=source",
"                   and only checks the source parts. No directory",
"                   prefix gets applied.",
" -verify_source_filter  shell_filter",
"                   With output format -form_path_list_mkisofs the",
"                   filelist can be cleaned from undesired addresses",
"                   immediately before the prompt to insert a CD.",
"                   The decision is made by shell_filter which has to",
"                   put out any valid file to stdout and omit any",
"                   invalid file. Use this filter in order to omit",
"                   all files which have been removed since the",
"                   backup was planned :",
"                     ..../cd_backup_planer -verify_sources -",
"                   An empty text as shell_filter disables filtering.",
" -verify_targetlist  list_adr  directory_prefix",
"                   like -verify_filelist but processing the target addresses",
"                   of the listed files. If there have been target=source",
"                   redirections when a backup was made, then this can make",
"                   a big difference.",
" -volume_info      pass additional arguments to the backup command",
"                   (before file names) : -v volume -t total_volumes",
" -working_directory  directory_address",
"                   set the working directory to the given address.",
"                   This option should be used only once and only",
"                   before any fileaddress is given.",
" -xdev             exclude all filesystems mounted below the items",
"                   of the backuplist. Opposite of -alldev .",
" -yes              all following fileaddresses are appended to the",
"                   backuplist (i.e. these files are to be backuped).",
"                   This option ends the range of option \"-no\".",
"                   It is the default mode for fileaddresses.",
"                   If the option -yes is given at least once, then",
"                   the option -no does not apply to the backuplist.",
"                   ( -yes file -not file   : file is in backup",
"                     file -not file        : file is not in backup)",
" -younger_than time_definition  try to backup only files where the",
"                   content or inode changed since the given time.",
"                   Moves of a file within the same file system",
"                   cannot be detected, regrettably.",
"                   Time definitions can have the following forms:",
"                   Absolute timepoints can be defined by the input",
"                   format MMDDhhmm[[CC]YY][.ss]] like with UNIX",
"                   command 'date'. For example:",
"                     020616122000 is 6 Feb 2000 16:12:00 local time.",
"                   Also the format YYMMDD[.hhmm[ss]] is recognized:",
"                     11 May 2001 23:45 = A10511.2345",
"                     1 Mar 2010 = B00301 , 14 Jul 1998 = 980714",
"                   Relative times are preceded by '-' or '+'. They",
"                   are added to the current time. '-' therefore",
"                   points to the past, '+' points to the future. The",
"                   basic unit is the second. It can be modified by",
"                   a letter at the end of the number:",
"                     seconds 1s=1 , hours 1h=3600s , days 1d=24h ,",
"                     weeks 1w=7d , months 1m=31d ,",
"                     years y : Xy=X*365.25d+1d",
"                   Please note : 'm' and 'y' are possibly slightly",
"                   larger than the actual calendar time range.",
"                   Example:",
"                     -3w  is 3 weeks in the past = -1814400s .",
"",
"Examples:",
"",
"Divide the content of the working directory into volumes of",
"625 MB and maximum 200 independend items. Create info script ASKME",
"in working directory and include it in each backup part.",
"",
"  %s \\",
"     -max_items 200 -max_size 625m -info_script ASKME .  >bscript",
"  chmod u+x bscript",
"",
"To actually perfom the backup, execute :",
"  bscript",
"",
"",
"With this option the ASKME script is generated at a separate place",
"and mapped to the root directory :",
"",
"  %s \\",
"    -info_script /ASKME=/home/backup_planer_dir/ASKME",
"",
"",
"Divide the directores \"office\" and \"images\" into volumes of",
"default size. Do not include subdirectories \"office/private\" and",
"\"images/girlfriends\"",
"Replace the omitted \"images/girlfriends\" by \"images/cats\"",
"",
"  %s \\",
"    office images -target_separator = /images/girlfriends=images/cats \\",
"    -not office/private images/girlfriends \\",
"  >bskript",
"",
"",
"Fill DC-6250 tapes with compressed tar archives.",
"",
"  %s \\",
"    -backup_cmd 'read dummy ; tar cvzf /dev/rmt0' \\",
"    -form_script_raw -max_size 245m -max_items 100 \\",
"    -info_script ASKME \\",
"    . \\",
"  >bskript",
"",
"",

"Create a configuration for incremental backups and perform planning",
"of the first -level 0 run. By -level -create_configuration a new",
"directory gets created which will be called $HOME/my_backup_conf .",
"",
"  %s \\",
"    -conf_dir $HOME/my_backup_conf -level -create_configuration \\",
"    ... all arguments defining the full backup ... \\",
"  >bskript",
"",
"The configuration may be used by simply giving this directory and",
"the desired level. All other parameters of the backup will be read",
"from $HOME/my_backup_conf/specific_arguments (1 line = 1 argument).",
"",
"One should use a fixed backup scheme. Let us assume that we made",
"the configuration and the first -level 0 backup at sunday of week 0.",
"",
"A backup scheme which from monday to saturday records only the",
"daily changes and makes a full backup each sunday.",
"  Monday  :  -conf_dir $HOME/my_backup_conf -level -next",
"  ...",
"  Saturday:  -conf_dir $HOME/my_backup_conf -level -next",
"  Sunday  :  -conf_dir $HOME/my_backup_conf -level 0",
"Now the media of monday to saturday may be recycled.",
"",
"A scheme which makes quarterly (13 weeks) full backups, four-weekly",
"and weekly updates (-conf_dir same as above). This scheme lets you",
"retrieve old file versions for quite a long time.",
"  Week  1 :  -level 1",
"  Week  2 :  -level 2",
"  Week  3 :  -level 2",
"  Week  4 :  -level 2",
"  Week  5 :  -level 1",
"  Week  6 :  -level 2",
"  Week  7 :  -level 2",
"  Week  8 :  -level 2",
"  Week  9 :  -level 1",
"  Week 10 :  -level 2",
"  Week 11 :  -level 2",
"  Week 12 :  -level 2",
"  Week 13 :  -level 0",
"For week 14 hop to week 1.",
"The backup of week 12 also covers weeks 11 and 10.",
"The backup of week 9 also covers week 8 to 1.",
"It is advised to keep levels 0 and 1 for a longer time.",
"",
"To combine both schemes shown, simply replace the sunday run of",
"scheme 1 by the weekly run of scheme 2.",
"",
"If you ever need to re-install an incremental backup with multiple",
"levels you will begin with copying level 0 first. Then you need to",
"remove the files which have vanished in the week between the level 0",
"backup and the next level.",
"For that purpose you will find on the last backup volume of level 1",
"a file named /tmp/VANISHED.sh.gz . It does not have x-permission to",
"avoid ugly accidents. You can only trust it if it has been produced",
"reading a content list with CRCs. This is true by default but may be",
"changed with options like -content_list_adr , -list_output_format ,",
" -load_old_tree , -changetest_options , -vanish_script_adr .",
"If you want to be very cautious, you may use option \"md5\" rather",
"than \"crc\" with -list_output_format.",
"",
"You will have to apply  chmod u+x  and  gunzip  after copying or you",
"will have to execute it on-the-fly via a pipe like :",
"   gunzip </cdrom/tmp/VANISHED.sh.gz | /bin/sh",
"Better have a look at its content to decide wether you think the",
"condemned files are really obsolete. Last chance to save old trash.",
"",
"The vanish script of level 1 may be executed at any time after",
"level 0 was installed and before level 2 gets installed. None of",
"the condemned files is contained in the level 1 backup.",
"If disk space is short, better execute it before installing level 1.",
"Otherwise it is more convenient to execute /tmp/VANISHED.sh after",
"it was copied to the disk.",
"When installing the following levels 2,3,4,... do not forget to",
"execute their vanish scripts in time and to remove the last one",
"from the /tmp directory after all work is done.",
"",
"",
"",
"@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@"
 };
 char line[4096];
 int i;

 MELDEN("");
 sprintf(line,"cd_backup_planer version %s",PROG_VERSION);
 MELDEN(line);
 MELDEN("");
 for(i=0;strcmp(hilfetext[i],
                "@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@"
                )!=0;i++){
   sprintf(line,hilfetext[i],progname);
   MELDEN(line);
 }

 return(1);
}


/* ----------------- Backup_plan - end ---------- */


static char cmd[]= {"cd_backup_kommando"};
int max_items= 400;
double max_size= 600*1024*1024;

main(argc,argv)
int argc;
char **argv;
{
 int ret,i;
 char *cpt;
 struct Backup_plaN *plan,plan_data;

 plan= &plan_data;

 ret= Backup_plan_init(plan,max_items,max_size,cmd,0);
 if(ret<=0) return(2);

 ret= Backup_plan_setup(plan,argc,argv,2);
 if(ret<0)
   exit(0); /*    handsome end (e.g. -help) */
 if(ret>0)
   exit(ret); /*  troublesome end */
 if(plan->startliste->anz<=0 && plan->configuration!=2)
   exit(1);

 fprintf(stderr,"\nscdbackup %s : planning backup\n\n",PROG_VERSION);
 
 ret= Backup_plan_conf_main(plan,argc,argv,0);
 if(ret<=0)
   exit(1);

 ret= Backup_plan_list_main_dirs(plan,0);
 if(ret<=0)
   exit(1);

 ret= Backup_plan_generate_lists(plan,0);
 if(ret<=0)
   exit(1-ret);

 if(plan->precheck>0 || plan->precheck_xdev || (plan->precheck_agetest&1)
    || (plan->precheck_sizetest&1)) {
   ret= Backup_plan_precheck(plan,1);
   if(ret&(1<<15)) {
     fprintf(stderr,"--- File check failed. Program aborted.\n");
     exit(16);
   }
 }
 if(plan->precheck==2)
   exit(0);

 Fileliste_destroy(&(plan->old_tree),0);

 if(plan->collision_test==1) 
   if(Backup_plan_collision_test(plan,1)<0) {
     fprintf(stderr,"--- Collision test failed. Program aborted.\n");
     exit(16);
   }

 if(plan->auskunft_name->adressen[0][0]) 
   unlink(plan->auskunft_name->adressen[0]);

 ret= Backup_plan_compose(plan,plan->startliste,2|4|16);
 if(ret<=0)
   exit(1-ret);
 ret= Backup_plan_generate_lists(plan,1);
 if(ret<=0)
   exit(1-ret);
 Backup_plan_compose(plan,NULL,8); /* just print final message */

 if(plan->collision_test==2 )  /* || plan->collision_test==1) */
   if(Backup_plan_collision_test(plan,0)<0) {
     fprintf(stderr,"--- Collision test failed. Program aborted.\n");
     exit(16);
   }

 ret= Backup_plan_skript(plan,0);
 if(ret!=0)
   exit(ret);

 if(plan->configuration==2) {
   ret= Backup_plan_conf_update(plan,0);
   if(ret<=0)
     exit(1);
 }

 exit(0);
}
